#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define asm-formatters
    (append *default-ct-formatters* 
	'((#\e . write-expr)
	  (#\n . write-name)
	  (#\l . write-label))))
(define *relabel-table* #f)

(define-rewriter (asm form)
    (ct-format-rewriter asm-formatters (cons 'ct-format (cons #t (cdr form)))))

(define-rewriter (asm* form)
    (ct-format-rewriter asm-formatters form))

(define (write-multi-expr multi-items port)
    (write-char #\( port)
    (if (pair? multi-items)
	(begin
	    (write-expr (car multi-items) port)
	    (let loop ((i (cdr multi-items)))
		(if (pair? i)
		    (begin
			(write-char #\, port)
			(write-expr (car i) port)
			(loop (cdr i)))))))
    (write-char #\) port))

(define (write-reg ea port)
    (if (< (cadr ea) 10)
	(asm* port "REG~d~n" (cadr ea) (caddr ea))
	(asm* port "REG(~d)~n" (cadr ea) (caddr ea))))

#| Note:  normally, write-name should write any truish
          thing out, but due to a problem with literals
	  containing "*/" (say, for example, the string for
	  this thing itself!), I am temporarily suppressing
	  the printout of all name-comments.
|#	  

(define (write-name item port)
    (if item
	(let ((str (object->string item)))
	    (if (not (or (string-search str "*/")
	    	         (string-search str "/*")))
		(asm* port " /* ~a */" str)))))

(define (write-label item port)
    (let ((new-label (table-lookup *relabel-table* item)))
	(display new-label port)))

(define (write-expr item port)
    (if (pair? item)
	(if (string? (car item))
	    (begin
		(display (car item) port)
		(write-multi-expr (cdr item) port))
	    (case (car item)
		((seq) (write-multi-expr (cdr item) port))
		((ref) (let ((ea (cadr item)))
			 (case (caadr item)
			    ((reg) (write-reg ea port))
			    ((lex-var)
				    (if (< (cadr ea) 10)
					(asm* port "LEXREF~d(~d)~n"
					    (cadr ea)
					    (caddr ea)
					    (cadddr ea))
					(asm* port "LEXREF(~d,~d)~n"
					    (cadr ea)
					    (caddr ea)
					    (cadddr ea))))
			    ((tl-var)
				    (asm* port "TLREF(~d)~n" 
					(cadr ea) 
					(caddr ea)))
			    ((root) (display (cadr ea) port))
			    (else (error/internal "Bad EA in ref")))))
		((literal) 
			(asm* port "LITERAL(~d)~n" 
			    (cadr item) 
			    (caddr item)))
		((if)
			(asm* port "(~e ? ~e : ~e)"
			    (cadr item)
			    (caddr item)
			    (cadddr item)))
		((closure)
			(asm* port "CLOSURE(~d)" (cadr item)))
		(else (error/internal "Bad expr: ~a" item))))
	(display item port)))

(define (asm-set! tab insn)
    (asm "~>" tab)
    (let ((ea (cadr insn))
    	  (expr (caddr insn))
	  (port (current-output-port)))
	(case (car ea)
	    ((reg)  (write-reg ea port)
		    (display " = " port)
		    (write-expr expr port))
	    ((lex-var) (if (< (cadr ea) 10)
			    (asm* port "LEXSET~d(~d~n,~e)"
				(cadr ea)
				(caddr ea)
				(cadddr ea)
				expr)
			    (asm* port "LEXSET(~d,~d~n,~e)"
				(cadr ea)
				(caddr ea)
				(cadddr ea)
				expr)))
	    ((tl-var) (asm* port "TLSET(~d~n,~e)" 
			    (cadr ea)
			    (caddr ea)
			    expr))
	    ((root) (asm* port "~a = ~e" (cadr ea) expr))
	    (else   (error/internal "Bad EA in set!"))))
    (asm ";\n"))
    
(define (asm-jump tab insn)
    (asm "~>JUMP(~d,~l);\n" tab (cadr insn) (caddr insn)))

; should probably just look at the labels and SEE whether or not
; it is a backward jump...

(define (asm-bjump tab insn)
    (asm "~>BJUMP(~d,~l);\n" tab (cadr insn) (caddr insn)))

(define (asm-apply tab insn)
    (asm "~>APPLY(~d,~e);\n" tab (cadr insn) (caddr insn)))

(define (asm-return tab insn)
    (if (eq? (cadr insn) 0)
	(asm "~>RETURN~d();\n" tab (cadr insn))
	(asm "~>RETURN(~d);\n" tab (cadr insn))))

(define (asm-seq tab insn)
    (asm "~>{\n" (sub1 tab))
    (asm-stmts tab (cdr insn))
    (asm "~>}\n" (sub1 tab)))

(define (asm-do tab insn)
    (asm "~>~e;\n" tab (cadr insn)))
    
(define (asm-if tab insn)
    (asm "~>if (~e)\n" tab (cadr insn))
    (asm-stmt (add1 tab) (caddr insn))
    (asm "~>else\n" tab)
    (asm-stmt (add1 tab) (cadddr insn)))

(define (asm-check= tab insn)
    (asm "~>COUNT_ARGS(~d);\n" tab (cadr insn)))
    
(define (asm-check>= tab insn)
    (if (> (cadr insn) 0)
	(asm "~>COUNT_ARGS_AT_LEAST(~d);\n" tab (cadr insn))))
    
(define (asm-collect> tab insn)
    (if (< (cadr insn) 10)
	(asm "~>COLLECT~d();\n" tab (cadr insn))
	(asm "~>COLLECT(~d);\n" tab (cadr insn))))

(define (asm-set-false< tab insn)
    (if (< (cadr insn) 5)
	(asm "~>PAD_WITH_FALSE~d();\n" tab (cadr insn))
	(asm "~>PAD_WITH_FALSE(~d);\n" tab (cadr insn))))

(define (asm-bind tab insn)
    (if (< (length (cdr insn)) 10)
	(asm "~>BEGIN_BIND~d()\n" tab (length (cdr insn)))
	(asm "~>BEGIN_BIND(~d)\n" tab (length (cdr insn))))
    (let loop ((i 0) (e (cdr insn)))
	(if (pair? e)
	    (begin
		(asm "~>BIND_ARG(~d,~e);\n" (add1 tab) i (car e))
		(loop (add1 i) (cdr e)))))
    (asm "~>END_BIND\n" tab))

(define (asm-unbind tab insn)
    (asm "~>POPENVT();\n" tab))

(define (asm-save tab insn)
    (if (< (cadr insn) 10)
	(asm "~>SAVE_CONT~d(~l);\n" tab (cadr insn) (caddr insn))
	(asm "~>SAVE_CONT(~d,~l);\n" tab (cadr insn) (caddr insn))))

(define (asm-restore tab insn)
    (if (< (cadr insn) 10)
	(asm "~>RESTORE_CONT~d();\n" tab (cadr insn))
	(asm "~>RESTORE_CONT(~d);\n" tab (cadr insn))))

#|
(define (asm-stmt tab insn)
    ((case (car insn)
	    ((jump) asm-jump)
	    ((apply) asm-apply)
	    ((return) asm-return)
	    ((seq) asm-seq)
	    ((do) asm-do)
	    ((if) asm-if)
	    ((check=) asm-check=)
	    ((check>=) asm-check>=)
	    ((collect>) asm-collect>)
	    ((set-false<) asm-set-false<)
	    ((bind) asm-bind)
	    ((unbind) asm-unbind)
	    ((save) asm-save)
	    ((restore) asm-restore)
	    ((set!) asm-set!)
	    ((bjump) asm-bjump)
	    (else (error/internal "Bad ASM stmt: ~a" insn)))
	tab
	insn))
|#
(define *asm-stmts* #f)

(define (asm-stmt tab insn)
    (if (not *asm-stmts*)
	(set! *asm-stmts*
	      (list (cons 'jump asm-jump)
		    (cons 'apply asm-apply)
		    (cons 'return asm-return)
		    (cons 'seq asm-seq)
		    (cons 'do asm-do)
		    (cons 'if asm-if)
		    (cons 'bind asm-bind)
		    (cons 'save asm-save)
		    (cons 'restore asm-restore)
		    (cons 'set! asm-set!)
		    (cons 'check= asm-check=)
		    (cons 'unbind asm-unbind)
		    (cons 'check>= asm-check>=)
		    (cons 'collect> asm-collect>)
		    (cons 'set-false< asm-set-false<)
		    (cons 'bjump asm-bjump))))
    ((cdr (assq (car insn) *asm-stmts*)) tab insn))
    
(define (asm-stmts tab insns)
    (for-each 
	(lambda (i) 
	    (asm-stmt tab i)) 
	insns))

(define (write-asm-stmts new-names vinsns)
    (set! *relabel-table* new-names)
    ;(set! *asm-output-port* (current-output-port))	
    (newline)
    (asm-stmts 1 vinsns)
    (set! *relabel-table* #f))

(define (symbol+integer->hash x)
    (if (symbol? x) 
	(symbol->hash x)
	(integer->hash x)))

(define (relabel-monotones base-name monotones label-table)
    (let loop ((i 0) (m monotones) (r '()))
	(if (null? m)
	    (reverse r)
	    (let ((new-label (ct-format #f "~a_~d" base-name i)))
		(table-insert! label-table (caar m) new-label)
		(loop (add1 i)
		      (cdr m)
		      (cons (cons new-label (cdar m)) r))))))
		      
(define (vinsns->monotones* name vinsns)
    (let loop ((accum '()) (i vinsns))
	(if (null? i)
	    (list (cons name (reverse accum)))
	    (if (eq? (caar i) 'label)
		(cons (cons name (reverse accum))
		      (vinsns->monotones* (cadar i) (cdr i)))
		(loop (cons (car i) accum) (cdr i))))))
