#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(prologue {
#include "arith.h"
#include <math.h>
})
(include-h "arith.h")

(define (bad-arg proc args)
    (abort proc "Bad argument in: ~s" args))

(define (_quotient x y)
    (type-check x fixnum? <FixNum>)
    (type-check y fixnum? <FixNum>)
    (quotient x y))

(%define-glue (binary+ x y) literals: ('binary+ (& bad-arg))
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = FX_ADD(x,y);
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( fx2int(x) + extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = make_float( extract_float(x) + fx2int(y) );
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( extract_float(x) + extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else
	goto invoke_generic_fn;
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(%define-glue (binary- x y) literals: ('binary- (& bad-arg))
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = FX_SUB(x,y);
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( fx2int(x) - extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = make_float( extract_float(x) - fx2int(y) );
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( extract_float(x) - extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else
	goto invoke_generic_fn;
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})


(%define-glue (binary* x y) literals: ('binary* (& bad-arg))
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = FX_MUL(x,y);
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( fx2int(x) * extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = make_float( extract_float(x) * fx2int(y) );
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( extract_float(x) * extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else
	goto invoke_generic_fn;
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(%define-glue (binary/ x y) literals: ('binary/ (& bad-arg))
{
double rx, ry;

    if (OBJ_ISA_FIXNUM(x))
	rx = fx2int(x);
    else if (LONGFLOAT_P(x))
	rx = extract_float(x);
    else
	goto invoke_generic_fn;
	
    if (OBJ_ISA_FIXNUM(y))
	ry = fx2int(y);
    else if (LONGFLOAT_P(y))
	ry = extract_float(y);
    else
	goto invoke_generic_fn;

    REG0 = make_float( rx / ry );
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(%define-glue (unary/ x) literals: ('unary/ (& bad-arg))
{
double rx;

    if (OBJ_ISA_FIXNUM(x))
	rx = fx2int(x);
    else if (LONGFLOAT_P(x))
	rx = extract_float(x);
    else
	INVOKE_GENERIC();

    REG0 = make_float( 1.0 / rx );
    RETURN(1);
})

(%define-glue (unary- x) literals: ('unary- (& bad-arg))
{
    if (OBJ_ISA_FIXNUM(x))
	REG0 = int2fx(-fx2int(x));
    else if (LONGFLOAT_P(x))
	REG0 = make_float( -extract_float(x) );
    else
	INVOKE_GENERIC();
    RETURN(1);
})

(%define-glue (floor x) literals: ('floor (& bad-arg))
{
    COUNT_ARGS(1);
    if (OBJ_ISA_FIXNUM(x))
	REG0 = x;
    else if (LONGFLOAT_P(x))
	REG0 = make_float( floor( extract_float(x) ) );
    else
	INVOKE_GENERIC();
    RETURN(1);
})

(%define-glue (ceiling x) literals: ('ceiling (& bad-arg))
{
    COUNT_ARGS(1);
    if (OBJ_ISA_FIXNUM(x))
	REG0 = x;
    else if (LONGFLOAT_P(x))
	REG0 = make_float( ceil( extract_float(x) ) );
    else
	INVOKE_GENERIC();
    RETURN(1);
})

(%define-glue (truncate x) literals: ('truncate (& bad-arg))
{
    COUNT_ARGS(1);
    if (OBJ_ISA_FIXNUM(x))
	REG0 = x;
    else if (LONGFLOAT_P(x))
	REG0 = make_float( (int) extract_float(x) );
    else
	INVOKE_GENERIC();
    RETURN(1);
})

(%define-glue (round x) literals: ('round (& bad-arg))
{
    COUNT_ARGS(1);
    if (OBJ_ISA_FIXNUM(x))
	REG0 = x;
    else if (LONGFLOAT_P(x))
    {
    IEEE_64 a = extract_float(x);
    INT_32 ai;
    
	if (a < 0)
	{
	    ai = (int)(a - 0.5);
	    if (ai >= -536870912)
		REG0 = int2fx( ai );
	    else
		REG0 = x;
	}
	else
	{
	    ai = (int)(a + 0.5);
	    if (ai <= 536870911)
		REG0 = int2fx( ai );
	    else
		REG0 = x;
	}
    }
    else
	INVOKE_GENERIC();
    RETURN(1);
})


(%define-glue (exact->inexact z) literals: ('exact->inexact (& bad-arg))
{
    COUNT_ARGS(1);
    if (OBJ_ISA_FIXNUM(z))
	REG0 = make_float( fx2int(z) );
    else if (LONGFLOAT_P(z))
	REG0 = z;
    else
	INVOKE_GENERIC();
    RETURN(1);
})

(%define-glue (inexact->exact z) literals: ('inexact->exact (& bad-arg))
{
    COUNT_ARGS(1);
    if (OBJ_ISA_FIXNUM(z))
	REG0 = z;
    else if (LONGFLOAT_P(z))
	REG0 = int2fx( (int) extract_float( z ) );
    else
	INVOKE_GENERIC();
    RETURN(1);
})


(%define-glue (binary<? x y) literals: ('binary<? (& bad-arg))
{
ARITHMETIC_ORDER_PREDICATE(<)
})

(%define-glue (binary<=? x y) literals: ('binary<=? (& bad-arg))
{
ARITHMETIC_ORDER_PREDICATE(<=)
})

(%define-glue (binary>? x y) literals: ('binary>? (& bad-arg))
{
ARITHMETIC_ORDER_PREDICATE(>)
})

(%define-glue (binary>=? x y) literals: ('binary>=? (& bad-arg))
{
ARITHMETIC_ORDER_PREDICATE(>=)
})

(%define-glue (binary=? x y) literals: ('binary=? (& bad-arg))
{
ARITHMETIC_ORDER_PREDICATE(==)
})

(export binary+ binary* binary/ binary- unary/ unary-)
(export floor ceiling truncate round)
(export binary<? binary<=? binary>? binary>=? binary=?)

#|
(define (_add1 n)
    (type-check n fixnum? <FixNum>)
    (add1 n))

(define (_sub1 n)
    (type-check n fixnum? <FixNum>)
    (sub1 n))
|#

(define (_modulo n m)
    (type-check n fixnum? <FixNum>)
    (type-check m fixnum? <FixNum>)
    (if (< m 0)
    	(- 0 (modulo (- 0 n) (- 0 m)))
	(modulo n m)))
    
(define (_remainder n m)
    (type-check n fixnum? <FixNum>)
    (type-check m fixnum? <FixNum>)
    (if (< n 0)
	(if (< m 0)
	    (- 0 (remainder (- 0 n) (- 0 m)))
	    (- 0 (remainder (- 0 n) m)))
	(if (< m 0)
	    (remainder n (- 0 m))
	    (remainder n m))))
    
(define (_quotient n m)
    (type-check n fixnum? <FixNum>)
    (type-check m fixnum? <FixNum>)
    (quotient n m))

(export (_add1 add1) (_sub1 sub1)
	(_modulo modulo) (_remainder remainder) (_quotient quotient)
	exact->inexact inexact->exact)

(%define-glue (exp z)
{
    REAL_OPERATOR(exp)
})

(%define-glue (log z)
{
    REAL_OPERATOR(log)
})

(%define-glue (sin z)
{
    REAL_OPERATOR(sin)
})

(%define-glue (cos z)
{
    REAL_OPERATOR(cos)
})

(%define-glue (tan z)
{
    REAL_OPERATOR(tan)
})

(%define-glue (asin z)
{
    REAL_OPERATOR(asin)
})

(%define-glue (acos z)
{
    REAL_OPERATOR(acos)
})

(%define-glue (atan z)
{
    if (arg_count_reg == 2)
    {
    IEEE_64 x, y;
    
	y = get_float( REG0, FUNCTION, "y" );
	x = get_float( REG1, FUNCTION, "x" );
	REG0 = make_float( atan2( y, x ) );
	RETURN(1);
    }
    else
	REAL_OPERATOR(atan)
})

(export exp log sin cos tan asin acos atan)

(%define-glue (sqrt z)
{
    REAL_OPERATOR(sqrt)
})

(%define-glue (float-result-expt #| z1 z1 |#)
{
IEEE_64 z1, z2;

    COUNT_ARGS(2);
    z1 = get_float( REG0, FUNCTION, "z1" );
    z2 = get_float( REG1, FUNCTION, "z2" );
    REG0 = make_float( pow( z1, z2 ) );
    RETURN(1);
})

(export sqrt)
