#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (_%bvec-length item)
    (assert (bvec? item))
    (%bvec-length item))
    
(define (_%bvec-read item index)
    (assert (bvec? item))
    (assert (>= index 0))
    (assert (< index (%bvec-length item)))
    (%bvec-read item index))
    
(define (_%bvec-write item index value)
    (assert (bvec? item))
    (assert (>= index 0))
    (assert (< index (%bvec-length item)))
    (let ((old (%bvec-read item index)))
	(%bvec-write item index value)
	old))

(%define-glue (bvec-write-fixnum bvec i value)
{
INT_32 *d, v = fx2int(value);

    d = (INT_32 *)((char *)PTR_TO_DATAPTR(bvec) + fx2int(i));
    *d = v;
    REG0 = FX_ADD(i,int2fx(sizeof(INT_32)));
    RETURN(1);
})

(%define-glue (bvec-write-16 bvec i value)
{
INT_16 *d, v = fx2int(value);

    d = (INT_16 *)((char *)PTR_TO_DATAPTR(bvec) + fx2int(i));
    *d = v;
    REG0 = FX_ADD(i,int2fx(sizeof(INT_16)));
    RETURN(1);
})

(%define-glue (bvec-write-str bvec i str)
{
UINT_32 len = string_length(str);

    assert( fx2int(i) + string_length(str) <= SIZEOF_PTR(bvec) );
    memcpy( ((char *)PTR_TO_DATAPTR(bvec)) + fx2int(i),
    	    string_text(str),
	    string_length(str) );
    REG0 = FX_ADD( i, int2fx( len ) );
    RETURN(1);
})

;; bug in bvec-write-bvec fixed 94.11.24: sbvec not adj by sindex

(%define-glue (bvec-write-bvec bvec i sbvec sindex slength)
{
    assert( OBJ_ISA_FIXNUM(sindex) && OBJ_ISA_FIXNUM(slength) );
    assert( fx2int(slength) >= 0 );
    assert( fx2int(sindex) + fx2int(slength) <= SIZEOF_PTR(sbvec) );
    memcpy( ((char *)PTR_TO_DATAPTR(bvec)) + fx2int(i),
    	    ((char *)PTR_TO_DATAPTR(sbvec)) + fx2int(sindex),
	    fx2int(slength) );
    REG0 = FX_ADD( i, slength );
    RETURN(1);
})

(%define-glue (bvec-write-float bvec i value)
{
IEEE_64 *d, v = extract_float(value);

    d = (IEEE_64 *)((char *)PTR_TO_DATAPTR(bvec) + fx2int(i));
    *d = v;
    REG0 = FX_ADD(i,int2fx(sizeof(IEEE_64)));
    RETURN(1);
})


(%define-glue (bvec-read-fixnum bvec i)
{
INT_32 *s;

    s = (INT_32 *)((char *)PTR_TO_DATAPTR(bvec) + fx2int(i));
    REG1 = FX_ADD(i,int2fx(sizeof(INT_32)));
    REG0 = int2fx(*s);
    RETURN(2);
})

(%define-glue (bvec-read-16 bvec i)
{
INT_16 *s;

    s = (INT_16 *)((char *)PTR_TO_DATAPTR(bvec) + fx2int(i));
    REG1 = FX_ADD(i,int2fx(sizeof(INT_16)));
    REG0 = int2fx(*s);
    RETURN(2);
})

(%define-glue (bvec-read-str bvec i len)
{
UINT_32 n = fx2int(len);
obj str;

    str = bvec_alloc( n+1, string_class );
    memcpy( PTR_TO_DATAPTR(str), 
	    ((const char *)PTR_TO_DATAPTR(bvec)) + fx2int(i), 
	    n );
    REG1 = FX_ADD( i, len );
    REG0 = str;
    RETURN(2);
})

(%define-glue (bvec-read-float bvec i)
{
IEEE_64 *s;

    s = (IEEE_64 *)((char *)PTR_TO_DATAPTR(bvec) + fx2int(i));
    REG1 = FX_ADD(i,int2fx(sizeof(IEEE_64)));
    REG0 = make_float(*s);
    RETURN(2);
})

(export _%bvec-length _%bvec-read _%bvec-write)
(export bvec-write-fixnum bvec-write-16 bvec-write-str bvec-write-float
	bvec-write-bvec)
(export bvec-read-fixnum bvec-read-16 bvec-read-str bvec-read-float)
