#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (%abort)
    (%call ("abort" ()))
    #f)
    
(define (%exit arg)
    (%call ("exit" (<raw-int>)) arg)
    #f)

(%define-glue (no-handler msg)
{
    fprintf( stderr, "No handler in place for '%s' exception\n",
    	     string_text(msg) );
    RETURN0();
})

(define-primop (%fluid-assq name)
    (%call ("fluid_assq" (<obj>) => <obj>) name))

(define root: "finish_run_tmpl" $finish-run-tmpl 
    (%template (ctcprims 1097) 12))

(define root: "continue_intr_tmpl" $continue-intr-tmpl
    (%template (ctcprims 1097) 13))

;; the exception-vector is used to vector to OS code for
;; non-user-level interrupts (including ^C, but esp. for
;; timeouts, etc.)

(define exception-vector '#(#f #f #f #f))

(define ($init-exceptions)
    (vector-set! exception-vector 0 default-interrupt-handler))

(run-named $init-exceptions)

(define root: "interrupt_dispatch"
	(%interrupt-handler user-break timeout data-ready finalize)

    (if (and finalize (vector-ref exception-vector 1))
	((vector-ref exception-vector 1)))

    (if (and user-break (vector-ref exception-vector 0))
	((vector-ref exception-vector 0)))

    (if (and data-ready (vector-ref exception-vector 2))
	((vector-ref exception-vector 2)))

    ; note: I think the timeout handler needs to be the last of these
    ;       because the presumption is that the other exceptions occur
    ;       and should be serviced soon, wheras any code following the
    ;       servicing of a timeout may not execute for QUITE SOME TIME
    ;       (one time slice of every thread in the system)

    (if (and timeout (vector-ref exception-vector 3))
	((vector-ref exception-vector 3))))

(define root: "exception_handler"
        (%exception-handler type tmpl addr envt descr args)
    ((fluid-ref *exception* default-exception) 
	type tmpl addr envt descr args))

(define (abort abort-place-sym format-string . args)
    (type-check abort-place-sym symbol? <symbol>)
    (type-check format-string string? <string>)
    (apply* abort-place-sym
	    format-string 
	    args 
	    (fluid-ref *abort* default-abort)))

(define (error format-string . args)
    (type-check format-string string? <string>)
    (apply* format-string args (fluid-ref *error* default-error)))

(define (default-error . stuff)
    (no-handler "*error*")
    (%abort))

(define (default-abort . stuff)
    (no-handler "*abort*")
    (%abort))

(define (default-exception . stuff)
    (no-handler "*exception*")
    (%abort))

(define (default-interrupt-handler)
    (no-handler "*interrupt*")
    (%abort))

(define (%with-dynamic-state new thunk)
    (apply-within-dynamic-state thunk new))	;; in apply.scm

(define (%dynamic-state-ref)
    %dynamic-state-reg)

(define (%failed-assert proc)
    (if (pair? proc)
	(abort (car proc) "Assertion failed in: ~a" (cdr proc))
	(abort proc "Assertion failed")))

(export %abort %exit %dynamic-state-ref %dynamic-state-reg
        %with-dynamic-state %fluid-assq)

(export %fluid-unbound
	%failed-assert
	exception-vector
	abort
	error)

