#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
; ** CHARACTERS **

(define (boolean? thing)
    (%call ("OBJ_ISA_BOOLEAN" (<obj>) => <raw-bool>) thing))

(define (unicode-char? thing)
    (%call ("OBJ_ISA_UNICODE_CHAR" (<obj>) => <raw-bool>) thing))

(define-primop (char? thing)
    (%call ("OBJ_ISA_ASCII_CHAR" (<obj>) => <raw-bool>) thing))

(%define-glue (random)
{
UINT_32 os_random( void );

    COUNT_ARGS(0);
    REG0 = OBJ( os_random() & ~PRIMARY_TAG_MASK );
    RETURN(1);
})

(define-macro (integer->char i)
    (%call ("MAKE_ASCII_CHAR" (<raw-int>) => <obj>) i))

(define-macro (char->integer ch)
    (%call ("GET_IMMEDIATE_VALUE" (<obj>) => <raw-int>) ch))

(define (_integer->char int)
    (type-check int fixnum? <FixNum>)
    (if (or (< int 0) (> int 255))
	(abort 'integer->char "Integer ~d is not in 0..255" int))
    (integer->char int))

(define (integer->unicode-char int)
    (type-check int fixnum? <FixNum>)
    (if (or (< int 0) (> int 65535))
	(abort 'integer->unicode-char "Integer ~d is not in 0..65535" int))
    (%call ("MAKE_UNICODE_CHAR" (<raw-int>) => <obj>) int))

(define (_char->integer ch)
    (type-check ch char? <ASCII-Char>)
    (char->integer ch))

(define (unicode-char->integer ch)
    (type-check ch unicode-char? <Unicode-Char>)
    (%call ("GET_IMMEDIATE_VALUE" (<obj>) => <raw-int>) ch))


(define-macro (%char-cmp cmp ch1 ch2)
    (type-check ch1 char? <ASCII-Char>)
    (type-check ch2 char? <ASCII-Char>)
    (cmp (char->integer ch1) (char->integer ch2)))

(define (char=? ch1 ch2)  (%char-cmp eq? ch1 ch2))
(define (char<? ch1 ch2)  (%char-cmp < ch1 ch2))
(define (char<=? ch1 ch2)  (%char-cmp <= ch1 ch2))
(define (char>? ch1 ch2)  (%char-cmp > ch1 ch2))
(define (char>=? ch1 ch2)  (%char-cmp >= ch1 ch2))

(define-macro (%char-ci-cmp cmp ch1 ch2)
    (let-syntax ((as-int (syntax-form (ch) 
			    (let ((i (char->integer ch)))
			      (if (and (>= i 97) (<= i 122))
				  (- i 32)
				  i)))))
      (type-check ch1 char? <ASCII-Char>)
      (type-check ch2 char? <ASCII-Char>)
      (cmp (as-int ch1) (as-int ch2))))

(define (char-ci=? ch1 ch2)  (%char-ci-cmp eq? ch1 ch2))
(define (char-ci<? ch1 ch2)  (%char-ci-cmp < ch1 ch2))
(define (char-ci<=? ch1 ch2)  (%char-ci-cmp <= ch1 ch2))
(define (char-ci>? ch1 ch2)  (%char-ci-cmp > ch1 ch2))
(define (char-ci>=? ch1 ch2)  (%char-ci-cmp >= ch1 ch2))

(define (char-upcase ch)
    (type-check ch char? <ASCII-Char>)
    (let ((i (char->integer ch)))
	(if (and (>= i 97) (<= i 123))
	    (integer->char (- i 32))
	    ch)))

(define (char-downcase ch)
    (type-check ch char? <ASCII-Char>)
    (let ((i (char->integer ch)))
	(if (and (>= i 65) (<= i 91))
	    (integer->char (+ i 32))
	    ch)))

(define (char-alphabetic? ch)
    (type-check ch char? <ASCII-Char>)
    (%call ("isalpha" (<raw-int>) => <raw-bool>) (char->integer ch)))

(define (char-numeric? ch)
    (type-check ch char? <ASCII-Char>)
    (%call ("isdigit" (<raw-int>) => <raw-bool>) (char->integer ch)))

(define (char-whitespace? ch)
    (type-check ch char? <ASCII-Char>)
    (%call ("isspace" (<raw-int>) => <raw-bool>) (char->integer ch)))

(define (char-upper-case? ch)
    (type-check ch char? <ASCII-Char>)
    (%call ("isupper" (<raw-int>) => <raw-bool>) (char->integer ch)))

(define (char-lower-case? ch)
    (type-check ch char? <ASCII-Char>)
    (%call ("islower" (<raw-int>) => <raw-bool>) (char->integer ch)))


(define (bitwise-and i j)
    (type-check i fixnum? <FixNum>)
    (type-check j fixnum? <FixNum>)
    (%call ("FX_AND" (<obj> <obj>) => <obj>) i j))

(define (bitwise-or i j)
    (type-check i fixnum? <FixNum>)
    (type-check j fixnum? <FixNum>)
    (%call ("FX_OR" (<obj> <obj>) => <obj>) i j))

(define (bitwise-xor i j)
    (type-check i fixnum? <FixNum>)
    (type-check j fixnum? <FixNum>)
    (%call ("FX_XOR" (<obj> <obj>) => <obj>) i j))

(define (bitwise-not i)
    (type-check i fixnum? <FixNum>)
    (%call ("FX_NOT" (<obj>) => <obj>) i))

(%define-glue (bit-shift-right value amount)
{
    REG0 = OBJ( ((VAL(value) >> fx2int(amount)) & ~PRIMARY_TAG_MASK) 
    		+ FIXNUM_TAG );
    RETURN(1);
})

(%define-glue (bit-shift-left value amount)
{
    REG0 = OBJ( ((VAL(value) << fx2int(amount)) & ~PRIMARY_TAG_MASK) 
    		+ FIXNUM_TAG );
    RETURN(1);
})

(%define-glue (bit-ref value bit)
{
int i = fx2int(bit);

    if (i < 0 || i >= 30)
	scheme_error( "bit-ref: bit ~d is invalid", 1, bit );

    REG0 = (VAL(value) & (1 << (i+2))) ? TRUE_OBJ : FALSE_OBJ;
    RETURN(1);
})

(%define-glue (bit-set value bit) 
{
int i = fx2int(bit);

    if (i < 0 || i >= 30)
	scheme_error( "bit-set: bit ~d is invalid", 1, bit );

    REG0 = OBJ(VAL(value) | (1 << (i+2)));
    RETURN(1);
})

(export boolean? char? unicode-char?)
(export bitwise-and bitwise-or bitwise-xor bitwise-not)
(export bit-shift-right bit-shift-left bit-ref bit-set)
(export random not)

(export (_integer->char integer->char)
	(_char->integer char->integer)
	unicode-char->integer
	integer->unicode-char)

(export char=? char<? char<=? char>? char>=?)
(export char-ci=? char-ci<? char-ci<=? char-ci>? char-ci>=?)

(export char-alphabetic?
	char-numeric?
	char-whitespace?
	char-upper-case?
	char-lower-case?
	char-upcase
	char-downcase)
