#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#


(define-primop (eq? x y)
    (%call ("EQ" (<obj> <obj>) => <raw-bool>) x y))

(define-primop (symbol? thing)
    (%call ("SYMBOL_P" (<obj>) => <raw-bool>) thing))

(define-primop (vector? thing)
    (%call ("VECTOR_P" (<obj>) => <raw-bool>) thing))

(define-primop (string?  x)
    (%call ("STRING_P" (<obj>) => <raw-bool>) x))


(define-shadow eq? eq?)
(define-shadow symbol? symbol?)
(define-shadow vector? vector?)
(define-shadow string? string?)

(define-primop (double-float? thing)
    (%call ("LONGFLOAT_P" (<obj>) => <raw-bool>) thing))
    
(define-primop (not x)
    (%call ("RB_NOT" (<raw-bool>) => <raw-bool>) x))

(define-shadow not not)

(define-syntax %make
    (syntax-form (c a1)
	(%call ("make1" (<obj> <obj>) => <obj>) 
	       c a1))
    (syntax-form (c a1 a2)
	(%call ("make2" (<obj> <obj> <obj>) => <obj>) 
	       c a1 a2))
    (syntax-form (c a1 a2 a3)
	(%call ("make3" (<obj> <obj> <obj> <obj>) => <obj>) 
	       c a1 a2 a3))
    (syntax-form (c a1 a2 a3 a4)
	(%call ("make4" (<obj> <obj> <obj> <obj> <obj>) => <obj>) 
	       c a1 a2 a3 a4))
    (syntax-form (c . things)
	(_%make c . things)))

(define-shadow %make %make)

(define-syntax vector 
    (syntax-form things
	(%make <Vector> . things))
    (else _vector))

(define-shadow vector vector)

(define-macro (set-car! pair obj)
    (%gvec-write pair 0 obj))
    
(define-macro (set-cdr! pair obj)
    (%gvec-write pair 1 obj))
    
(define-macro (cddr x)
    (cdr (cdr x)))


(define-macro (caddr x)
    (car (cdr (cdr x))))

(define-macro (cdddr x)
    (cdr (cdr (cdr x))))

(define-macro (cadr x)
    (car (cdr x)))

(define-macro (< a b)
    (%call ("FX_LT" (<obj> <obj>) => <raw-bool>) a b))

(define-macro (<= a b)
    (%call ("FX_LE" (<obj> <obj>) => <raw-bool>) a b))

(define-macro (> a b)
    (%call ("FX_GT" (<obj> <obj>) => <raw-bool>) a b))

(define-macro (>= a b)
    (%call ("FX_GE" (<obj> <obj>) => <raw-bool>) a b))

(define-primop (add1 (x <fixnum>))
    (%cast <fixnum> (%call ("ADD1" (<obj>) => <obj>) x)))

(define-primop (sub1 (x <fixnum>))
    (%cast <fixnum> (%call ("SUB1" (<obj>) => <obj>) x)))

(define-macro (mul2 x)
    (%call ("MUL2" (<obj>) => <obj>) x))

(define-macro (div2 x)
    (%call ("DIV2" (<obj>) => <obj>) x))

(define-macro (quotient x y)
    (%call ("FX_DIV" (<obj> <obj>) => <obj>) x y))
    
(define-macro (remainder x y)
    (%call ("REMDR" (<raw-int> <raw-int>) => <raw-int>) x y))
    
(define-macro (modulo x y)
    (%call ("MOD" (<raw-int> <raw-int>) => <raw-int>) x y))



(define-macro (%bvec-length thing)
    (%call ("SIZEOF_PTR" (<obj>) => <raw-int>) thing))

(define-macro (%bvec-read thing index)
    (%call ("bvec_read" (<obj> <raw-int>) => <raw-int>) thing index))

(define-macro (%bvec-write thing index value)
    (%call ("bvec_write" (<obj> <raw-int> <raw-int>)) thing index value))



(define-macro (string-length str)
    (%call ("string_length" (<obj>) => <raw-int>) str))
	
(define-macro (string-ref  str i)
    (%call ("string_ref" (<obj> <raw-int>) => <obj>) str i))

(define-macro (string-set!  str i ch)
    (%call ("string_set" (<obj> <raw-int> <obj>)) str i ch))



(define-primop (vector-length (vec <Vector>))
    (%gvec-length vec))

(define-primop (vector-ref (vec <Vector>) 
			   (i <FixNum>
			      (and (>= i 0) (< i (vector-length vec)))))
    (%gvec-read vec i))

(define-primop (vector-set! (vec <Vector>) (i <FixNum>) x)
    (%gvec-write vec i x)
    '#undef)

(define-shadow vector-length vector-length)
(define-shadow vector-ref vector-ref)
(define-shadow vector-set! vector-set!)

(define-macro (+ x y)
    (%call ("FX_ADD" (<obj> <obj>) => <obj>) x y))

(define-macro (- x y)
    (%call ("FX_SUB" (<obj> <obj>) => <obj>) x y))


(export %make)
(export %bvec-length %bvec-read %bvec-write)

(export eq?
	(symbol? %symbol?)
	(not %not))
	

