#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-syntax cons*
    (syntax-form (item1 item2 . more-items) 
    			(cons item1 (cons* item2 . more-items)))
    (syntax-form (last-item) last-item))


(define-rewriter (define-macro form)
    (let ((name (car (car (cdr form))))
    	  (args (cdr (car (cdr form))))
	  (body (cdr (cdr form))))
	(list 'define-syntax
	      name
	      (cons 'syntax-form
	            (cons args body)))))

(define-syntax (assert expr)
    (if (not expr)
	(scheme-error "assertion failed in ~s: ~s" 
		      (*FUNCTION*) 
		      (mquote expr))))

(define-syntax type-check 
    (syntax-form (arg predicate class)
	(if (not (predicate arg))
	    (%call ("failed_type_check" (<obj> <obj> <obj> <obj>))
		    (*FUNCTION*) 
		    (mquote arg)
		    arg
		    (cons class '()))))
    (syntax-form (arg class)
	(if (not (instance? arg class))
	    (%call ("failed_type_check" (<obj> <obj> <obj> <obj>))
		    (*FUNCTION*) 
		    (mquote arg)
		    arg
		    (cons class '())))))

(define root: ("dynamic_state_reg" register 0) %dynamic-state-reg)

(define-syntax fluid-let
    (syntax-form (bindings . body)
	(letrec-syntax ((extend-fluid
			    (syntax-form ((var init) . more-bdgs) 
			(cons (cons (mquote var) init)
					(extend-fluid . more-bdgs)))
			    (syntax-form () %dynamic-state-reg)))
	    (%with-dynamic-state
	        (extend-fluid . bindings)
		(lambda ()
		    (begin . body))))))

(define-primop (%fluid-unbound name)
    (scheme-error "Fluid variable `~a' is unbound" name))

(define-syntax fluid-ref 
    (syntax-form (name)	; unforgiving form
	(let ((x (%fluid-assq (mquote name))))
	    (if x
		(cdr x)
		(%fluid-unbound (mquote name)))))
    (syntax-form (name default-expr)  ; forgiving form
	(let ((x (%fluid-assq (mquote name))))
	    (if x
		(cdr x)
		default-expr))))

(define-syntax fluid-set!
    (syntax-form (name val)
	(let ((x (%fluid-assq (mquote name))))
	    (if x
		(let ((old (cdr x)))
		    (set-cdr! x val)
		    old)
		(%fluid-unbound (mquote name))))))

(export type-check assert)
(export cond cons* or and case let* do quasiquote)
(export define-macro fluid-let fluid-ref delay)
