#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define-macro (car* p)
    (let ((p p))
	(type-check p pair? <Pair>)
	(car p)))

(define-macro (set-car!* p v)
    (let ((p p))
	(type-check p pair? <Pair>)
	(begin
	    (set-car! p v)
	    #f)))


(define-macro (cdr* p)
    (let ((p p))
	(type-check p pair? <Pair>)
	(cdr p)))

(define-macro (set-cdr!* p v)
    (let ((p p))
	(type-check p pair? <Pair>)
	(begin
	    (set-cdr! p v)
	    #f)))

#| obsolete in v2.0.5g

(define (_cons a b)
    (cons a b))

(define (_pair? thing)
    (pair? thing))

(define (_null? thing)
    (null? thing))
|#

(define (_list . things)
    things)

;(define (_car x)
;    (car* x))

(define (_cdr x)
    (cdr* x))

(define (_set-car! x a)
    (set-car!* x a))

(define (_set-cdr! x a)
    (set-cdr!* x a))

    
(define (delq! item list)
    (let loop ((prev #f) (l list))
	(if (pair? l)
	    (if (eq? (car l) item)
		(if prev
		    (begin
			(set-cdr! prev (cdr l))
			(loop prev (cdr l)))
		    (begin
			(set! list (cdr l))
			(loop #f (cdr l))))
		(loop l (cdr l)))
	    (if prev
		list
		'()))))

; Some fancy functions...

(define (reverse* from to)
    (if (pair? from)
        (reverse* (cdr from) (cons (car from) to))
	to))

(define (reverse l)
    (reverse* l '()))

(define (append1 list1 list2)
    (if (pair? list1)
	(cons (car list1)
	      (append1 (cdr list1) list2))
	list2))

(define (appendn lists)
    (if (null? lists)
	'()
	(if (null? (cdr lists))
	    (car lists)
	    (append1 (car lists) (appendn (cdr lists))))))
	    
(define (append . lists)
    (appendn lists))

; List Access Functions



; c[ad]^2r

(define (_caar x) (car* (car* x)))
(define (_cadr x) (car* (cdr* x)))
(define (_cddr x) (cdr* (cdr* x)))
(define (_cdar x) (cdr* (car* x)))

; c[ad]^3r

(define (_caaar x) (car* (car* (car* x))))
(define (_caadr x) (car* (car* (cdr* x))))
(define (_caddr x) (car* (cdr* (cdr* x))))
(define (_cadar x) (car* (cdr* (car* x))))

(define (_cdaar x) (cdr* (car* (car* x))))
(define (_cdadr x) (cdr* (car* (cdr* x))))
(define (_cdddr x) (cdr* (cdr* (cdr* x))))
(define (_cddar x) (cdr* (cdr* (car* x))))

; c[ad]^4r

(define (_caaaar x) (car* (car* (car* (car* x)))))
(define (_caaadr x) (car* (car* (car* (cdr* x)))))
(define (_caaddr x) (car* (car* (cdr* (cdr* x)))))
(define (_caadar x) (car* (car* (cdr* (car* x)))))

(define (_cadaar x) (car* (cdr* (car* (car* x)))))
(define (_cadadr x) (car* (cdr* (car* (cdr* x)))))
(define (_cadddr x) (car* (cdr* (cdr* (cdr* x)))))
(define (_caddar x) (car* (cdr* (cdr* (car* x)))))

(define (_cdaaar x) (cdr* (car* (car* (car* x)))))
(define (_cdaadr x) (cdr* (car* (car* (cdr* x)))))
(define (_cdaddr x) (cdr* (car* (cdr* (cdr* x)))))
(define (_cdadar x) (cdr* (car* (cdr* (car* x)))))

(define (_cddaar x) (cdr* (cdr* (car* (car* x)))))
(define (_cddadr x) (cdr* (cdr* (car* (cdr* x)))))
(define (_cddddr x) (cdr* (cdr* (cdr* (cdr* x)))))
(define (_cdddar x) (cdr* (cdr* (cdr* (car* x)))))

; other stuff

#|  breaks on infinite loops
(define (list? x)
    (let loop ((l x))
	(if (pair? l)
	    (loop (cdr l))
	    (null? l))))
|#

(define (ok-so-far seen item)
  (and (pair? item)
       (not (memq item seen))))

(define (list? x)
  (let loop ((l x) (a '()))
    (if (ok-so-far a l)
	(let ((l (cdr l)))
	  (if (ok-so-far a l)
	      (let ((l (cdr l)))
		(if (ok-so-far a l)
		    (let ((l (cdr l)))
		      (if (ok-so-far a l)
			  (loop (cdr l) (cons l a))
			  (null? l)))
		    (null? l)))
	      (null? l)))
	(null? l))))

(define (last-pair x)
    (let loop ((p #f) (l x))
	(if (pair? l)
	    (loop l (cdr l))
	    (if (null? l)
		(if p
		    p
		    (abort 'last-pair "Empty list"))
		(abort 'last-pair "Not a list at: ~s" l)))))

(define (last l)
  (car (last-pair l)))

(define (length l)
  (let loop ((l l) ((i <fixnum>) 0))
    (if (pair? l)
	(loop (cdr l) (add1 i))
	(if (null? l)
	    i
	    (abort 'length "Not a list at # ~d: ~s" i l)))))

(define (list-tail x n)
    (if (not (fixnum? n))
	(abort 'list-tail "Argument `n' is ~s, not a <FixNum>" n)
	(let loop ((l x) ((i <fixnum>) n))
	    (if (eq? i 0)
		l
		(if (pair? l)
		    (loop (cdr l) (sub1 i))
		    (abort 'list-tail "List too short: ~d missing at ~s"
		    	   i l))))))

(define (list-ref list k) 
    (car* (list-tail list k)))

(define (list-set! list k value) 
    (set-car!* (list-tail list k) value))

(export car ; (_car car)
	(_cdr cdr)
	(_set-car! set-car!)
	(_set-cdr! set-cdr!)
	delq!
	reverse
	append
	length
	list?
	last last-pair
	list-ref
	list-tail
	list-set!
	(_list list)
	(_caar caar) (_cadr cadr) (_cdar cdar) (_cddr cddr)
	(_caaar caaar) (_caadr caadr) (_cadar cadar) (_caddr caddr)
	(_cdaar cdaar) (_cdadr cdadr) (_cddar cddar) (_cdddr cdddr)

	(_caaaar caaaar) (_caaadr caaadr) (_caadar caadar) (_caaddr caaddr)
	(_cadaar cadaar) (_cadadr cadadr) (_caddar caddar) (_cadddr cadddr)
	
	(_cdaaar cdaaar) (_cdaadr cdaadr) (_cdadar cdadar) (_cdaddr cdaddr)
	(_cddaar cddaar) (_cddadr cddadr) (_cdddar cdddar) (_cddddr cddddr))

#| obsolete in 2.0.5g	
(export	(_cons cons)
	(_pair? pair?)
	(_null? null?))

 (define-shadow _cons cons)
 (define-shadow _pair? pair?)
 (define-shadow _null? null?)
|#

(define-shadow _car car)
(define-shadow _cdr cdr)
(define-shadow _cadr cadr)
(define-shadow append append)

