#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;  Define some useful macros...

(define-syntax cond
    (syntax-form (('else . action))
	(begin . action))
    (syntax-form ((guard) . more-clauses)
	(let ((temp guard)) 
	   (if temp 
	       temp 
	       (cond . more-clauses))))
    (syntax-form ((guard '=> proc) . more-clauses)
	(let ((temp guard)) 
	   (if temp 
	       (proc temp)
	       (cond . more-clauses))))
    (syntax-form ((guard . actions) . more-clauses)
	(if guard 
            (begin . actions) 
            (cond . more-clauses)))
    (syntax-form ()
	(begin)))

(define-syntax or
    (syntax-form (clause)
	clause)
    (syntax-form (clause . more-clauses)
	(let ((temp clause))
		(if temp
		    temp
		    (or . more-clauses))))
    (syntax-form () #f))
 
(define-syntax and
    (syntax-form (clause1 clause2 . more-clauses)
	(if clause1 (and clause2 . more-clauses) #f))
    (syntax-form () #t)
    (syntax-form (clause) clause))

(define-syntax case
  (syntax-form (expr . branches)
	       (letrec-syntax
		   ((test-cases
			(syntax-form (var ((g0 . g1) . actions) . more-cases)
			    (if (test-guards var g0 . g1)
				(begin . actions)
				(test-cases var . more-cases)))
			(syntax-form (var ('else . actions))
			    (begin . actions))
			(syntax-form (var)
			    (begin)))
		    (test-guards
			(syntax-form (var item)
			    (eq? var (mquote item)))
			(syntax-form (var item . more-items)
			    (if (eq? var (mquote item))
				#t
				(test-guards var . more-items)))
			(syntax-form (var)
			    #f)))
		    (let ((temp expr))
			(test-cases temp . branches)))))



(define-shadow let*-rewriter let*-rewriter)
(define-shadow do-rewriter do-rewriter)

(define-rewriter (let* form)
    (let*-rewriter form))

(define-rewriter (do form)
    (do-rewriter form))
    
(export do-rewriter let*-rewriter)

(define-rewriter (quasiquote form)
    (%expand-quasiquote (car (cdr form))))

(define-syntax delay 
    (syntax-form (expr)
	(%make-promise (lambda () expr))))

