/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "runtim.h"
#include "scmtime.h"

struct scheme_time extract_schemetime( obj ptr )
{
    return *(struct scheme_time *)PTR_TO_DATAPTR(ptr);
}

obj make_time( struct scheme_time t, obj t_class )
{
obj ptr = alloc( sizeof(struct scheme_time), t_class );

    *(struct scheme_time *)PTR_TO_DATAPTR(ptr) = t;
    return ptr;
}

obj make_time_sec( int t, obj t_class )
{
struct scheme_time t;

    t.usec = 0;
    t.sec = t;
    return make_time( t, t_class );
}

int timecmp( obj t1, obj t2 )
{
struct scheme_time raw_t1 = extract_schemetime(t1);
struct scheme_time raw_t2 = extract_schemetime(t2);

    if (raw_t1.sec == raw_t2.sec)
    {
        if (raw_t1.usec < raw_t2.usec)
	    return -1;
	else if (raw_t1.usec > raw_t2.usec)
	    return 1;
	return 0;
    }
    else if (raw_t1.sec < raw_t2.sec)
	return -1;
    else if (raw_t1.sec > raw_t2.sec)
	return 1;
    return 0;
}

struct scheme_time timeadd( obj t, obj dt )
{
struct scheme_time raw_t = extract_schemetime(t);
struct scheme_time raw_dt = extract_schemetime(dt);

    raw_t.usec += raw_dt.usec;
    if (raw_t.usec > 1000000)
    {
	raw_t.usec -= 1000000;
	raw_t.sec++;
    }
    else if (raw_t.usec < 0)
    {
	raw_t.usec += 1000000;
	raw_t.sec--;
    }
    raw_t.sec += raw_dt.sec;
    return raw_t;
}

struct scheme_time timesub( obj t1, obj t2 )
{
struct scheme_time raw_t1 = extract_schemetime(t1);
struct scheme_time raw_t2 = extract_schemetime(t2);
struct scheme_time raw_dt;

    raw_dt.usec = raw_t1.usec - raw_t2.usec;
    raw_dt.sec = raw_t1.sec - raw_t2.sec;
    if (raw_dt.usec < 0)
    {
	raw_dt.usec += 1000000;
	raw_dt.sec--;
    }
    return raw_dt;
}

struct scheme_time intervalsub( obj t1, obj t2 )
{
struct scheme_time raw_t1 = extract_schemetime(t1);
struct scheme_time raw_t2 = extract_schemetime(t2);
struct scheme_time raw_dt;

    raw_dt.usec = raw_t1.usec - raw_t2.usec;
    raw_dt.sec = raw_t1.sec - raw_t2.sec;
    if (raw_dt.usec < 0)
    {
	raw_dt.usec += 1000000;
	raw_dt.sec--;
    }
    else if (raw_dt.usec > 1000000)
    {
	raw_dt.usec -= 1000000;
	raw_dt.sec++;
    }
    return raw_dt;
}
