#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(%define-glue (make-string num ch)
{
char c;
UINT_32 n = fx2int(num);

    if (arg_count_reg == 1)
	c = ' ';
    else
	c = GET_IMMEDIATE_VALUE(ch);

    REG0 = bvec_alloc( fx2int(num)+1, string_class );
    memset( PTR_TO_DATAPTR(REG0), c, n );
    RETURN(1);
})

(%define-glue (string-fill! str fill)
{
UINT_32 i;
char src, *dst;

    src = GET_IMMEDIATE_VALUE(fill);
    dst = (char *)PTR_TO_DATAPTR(str);
    for (i=string_length(str); i>0; i--)
	*dst++ = src;
    RETURN(1);
})

(%define-glue (string-copy str)
{ 
UINT_32 len = SIZEOF_PTR(str);
obj new_str;

    new_str = bvec_alloc( len, string_class );
    memcpy( PTR_TO_DATAPTR(new_str), PTR_TO_DATAPTR(str), len );
    REG0 = new_str;
    RETURN(1);
})

(%define-glue (substring str start end)
{
INT_32 i, j;
UINT_32 len, new_len;
obj new_str;

    len = SIZEOF_PTR(str) - 1;
    i = fx2int(start);
    if (arg_count_reg == 2)
        j = len;
    else
	j = fx2int(end);
    
    if (i < 0 || i > len)
	scheme_error( "start ~d out of range 0..~d", 
			2, start, int2fx(len) );
    if (j < i || j > len)
	scheme_error( "end ~d out of range ~d..~d\n", 
			3, end, start, int2fx(len) );
    
    new_len = j - i;
    new_str = bvec_alloc( new_len + 1, string_class );
    
    /* Note that we don't have to copy or put in a terminal NUL,
       because bvec_alloc guarantees that the whole last
       word is ZERO, so if you alloc n+1 BYTES, then the last
       BYTE will surely be zero (but the next-to-the-last byte
       might not be, because it may fall on the preceding word).
       This is even true if n=0. */
       
    memcpy( PTR_TO_DATAPTR(new_str), 
    	    string_text(str)+i,
	    new_len );
    REG0 = new_str;
    RETURN(1);
})

(%define-glue (string-search str search_crit)
{
char *s0, *s;

    s0 = (char *)string_text(str);
    if (OBJ_ISA_ASCII_CHAR(search_crit))
	s = strchr( s0, (char)GET_IMMEDIATE_VALUE(search_crit) );
    else
	s = strstr( s0, string_text(search_crit) );
    if (s)
	REG0 = int2fx( s - s0 );
    else
	REG0 = FALSE_OBJ;
    RETURN(1);
})

(%define-glue (string)
{
char *p;
int i;
obj str, ch;

    str = bvec_alloc( arg_count_reg+1, string_class );
    p = (char *)PTR_TO_DATAPTR(str);
    for (i=0; i<arg_count_reg; i++)
    {
	ch = reg_ref(i);
	*p++ = GET_IMMEDIATE_VALUE(ch);
    }
    REG0 = str;
    RETURN(1);
})

(%define-glue (list->string)
{
obj i, new_str;
UINT_32 n;
char *p;

    n = 0;
    for (i=REG0; PAIR_P(i); i=pair_cdr(i))
    {
	if (!OBJ_ISA_ASCII_CHAR(pair_car(i)))
	    scheme_error( "not char: ~a", 1, pair_car(i) );
	n++;
    }
    new_str = bvec_alloc( n+1, string_class );
    p = (char *)PTR_TO_DATAPTR(new_str);
    for (i=REG0; PAIR_P(i); i=pair_cdr(i))
    {
	*p++ = GET_IMMEDIATE_VALUE(pair_car(i));
    }
    REG0 = new_str;
    RETURN(1);
})

(%define-glue (string->list string)
{
obj result;
const char *str;
unsigned i;

    result = NIL_OBJ;
    i = string_length(string);
    str = string_text(string) + i;
    while (i > 0)
    {
        result = cons( MAKE_ASCII_CHAR(*--str), result );
	i--;
    }
    REG0 = result;
    RETURN(1);
})


(define (string=? str1 str2)
    (type-check str1 string? <String>)
    (type-check str2 string? <String>)
    (%call ("string_eq" (<obj> <obj>) => <raw-bool>) str1 str2))

(define (string-ci=? str1 str2)
    (type-check str1 string? <String>)
    (type-check str2 string? <String>)
    (%call ("string_ci_eq" (<obj> <obj>) => <raw-bool>) str1 str2))

(define-macro (string-cmp pred str1 str2)
    (type-check str1 string? <String>)
    (type-check str2 string? <String>)
    (pred (%call ("string_cmp" (<obj> <obj>) => <raw-int>) str1 str2) 0))

(define (string<? s1 s2) (string-cmp < s1 s2))
(define (string>? s1 s2) (string-cmp > s1 s2))
(define (string<=? s1 s2) (string-cmp <= s1 s2))
(define (string>=? s1 s2) (string-cmp >= s1 s2))

(define-macro (string-ci-cmp pred str1 str2)
    (type-check str1 string? <String>)
    (type-check str2 string? <String>)
    (pred (%call ("string_ci_cmp" (<obj> <obj>) => <raw-int>) str1 str2) 0))

(define (string-ci<? s1 s2) (string-ci-cmp < s1 s2))
(define (string-ci>? s1 s2) (string-ci-cmp > s1 s2))
(define (string-ci<=? s1 s2) (string-ci-cmp <= s1 s2))
(define (string-ci>=? s1 s2) (string-ci-cmp >= s1 s2))

(define (symbol->string sym)
    (type-check sym symbol? <Symbol>)
    (%gvec-read sym 0))

(define (string->symbol str)
    (type-check str string? <String>)
    (%call ("intern" (<obj>) => <obj>) str))

(define *temp-symbol-ctr* -1)

(define ($gensym)
    (set! *temp-symbol-ctr* (add1 *temp-symbol-ctr*))
    (string->symbol 
	(string-append "G#" 
			(number->string *temp-symbol-ctr*))))

(define-shadow $gensym $gensym)

(%define-glue (string-append)
{
unsigned i, len, n;
char *p;
obj new_str, str;

    len = 0;
    for (i=0; i<arg_count_reg; i++)
    {
	str = reg_ref(i);
	if (!STRING_P(str))
	    scheme_error( "string-append: Not a string ~a", 1, str );
	len += string_length(str);
    }
    new_str = bvec_alloc( len+1, string_class );
    p = (char *)PTR_TO_DATAPTR(new_str);
    for (i=0; i<arg_count_reg; i++)
    {
	str = reg_ref(i);
        n = string_length( str );
	memcpy( p, string_text(str), n );
	p += n;
    }
    REG0 = new_str;
    RETURN(1);
})

(define (_string-length str)
    (type-check str string? <String>)
    (string-length str))
    
(define (_string-ref str i)
    (type-check str string? <String>)
    (type-check i fixnum? <FixNum>)
    (string-ref str i))
    
(define (_string-set! str i c)
    (type-check str string? <String>)
    (type-check i fixnum? <FixNum>)
    (type-check c char? <ASCII-Char>)
    (%call ("string_set" (<obj> <raw-int> <obj>)) str i c)
    c)

(export string make-string string-fill! string-copy string-append 
	substring string-search)
(export string->list list->string
	symbol->string
	string->symbol $gensym
	string?
	symbol?
	(_string-length string-length)
	(_string-ref string-ref)
	(_string-set! string-set!))
	
(export string=? string<? string>? string<=? string>=?)
(export string-ci=? string-ci<? string-ci>? string-ci<=? string-ci>=?)

(define (c-text? thing)
    (isa-ptr? thing directly: <C-Text>))
    
(define (string->c-text str)
    (type-check str string? <String>)
    (%make <C-Text> str))

(define (c-text->string ctext)
    (type-check ctext c-text? <C-Text>)
    (%gvec-read ctext 0))

(export c-text? string->c-text c-text->string)

