#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(%define-glue (make-vector num_ fill_) 
    literals: ("make-vector: argument `num' is not a fixnum: ~s")
{
UINT_32 i, len;
obj fill = FALSE_OBJ;

    if (arg_count_reg != 1)
    {
	if (arg_count_reg == 2)
	    fill = REG1;
	else
	    wrong_num_args_range( FUNCTION, 1, 2 );
    }

    if (!OBJ_ISA_FIXNUM(REG0))
	scheme_error( string_text(LITERAL(0)), 1, REG0 );

    len = FXWORDS_TO_RIBYTES(REG0);
    
    REG0 = alloc( len, vector_class );
    if (OBJ_ISA_PTR(fill))
    {
	for (i=0; i<len; i+=SLOT(1))
	    gvec_write_fresh_ptr( REG0, i, fill );
    }
    else
    {
	for (i=0; i<len; i+=SLOT(1))
	    gvec_write_fresh_non_ptr( REG0, i, fill );
    }
    RETURN(1);
})

(%define-glue (list->vector list)
{
UINT_32 i, len;
obj vector;
obj lp;

    len = 0;
    for (lp=list; PAIR_P(lp); lp=pair_cdr(lp))
	len += sizeof(obj);
    if (!NULL_P(lp))
	scheme_error( "list->vector: ~a not a proper list at: ~a",
			2, list, lp );

    vector = alloc( len, vector_class );

    i = 0;
    while (!NULL_P(list))
    {
        gvec_write_fresh( vector, i, pair_car(list) );
	list = pair_cdr( list );
	i += sizeof(obj);
    }
    REG0 = vector;
    RETURN(1);
})    

(%define-glue (vector-fill! vector fill)
{
UINT_32 i, len;

    COUNT_ARGS(2);
    len = SIZEOF_PTR(vector);
    
    if (OBJ_ISA_PTR(fill))
    {
	for (i=0; i<len; i+=SLOT(1))
	    gvec_write_ptr( vector, i, fill );
    }
    else
    {
	for (i=0; i<len; i+=SLOT(1))
	    gvec_write_non_ptr( vector, i, fill );
    }
    RETURN(1);
})

(%define-glue (_vector)
{
unsigned i;
obj old0;

    old0 = REG0;
    REG0 = alloc( SLOT(arg_count_reg), vector_class );
    switch (arg_count_reg)
    {
	default:
		for (i=10; i<arg_count_reg; i++)
		    gvec_write_fresh( REG0, SLOT(i), REG(i) );
	case 10: gvec_write_fresh( REG0, SLOT(9), REG9 );
	case 9: gvec_write_fresh( REG0, SLOT(8), REG8 );
	case 8: gvec_write_fresh( REG0, SLOT(7), REG7 );
	case 7: gvec_write_fresh( REG0, SLOT(6), REG6 );
	case 6: gvec_write_fresh( REG0, SLOT(5), REG5 );
	case 5: gvec_write_fresh( REG0, SLOT(4), REG4 );
	case 4: gvec_write_fresh( REG0, SLOT(3), REG3 );
	case 3: gvec_write_fresh( REG0, SLOT(2), REG2 );
	case 2: gvec_write_fresh( REG0, SLOT(1), REG1 );
	case 1: gvec_write_fresh( REG0, SLOT(0), old0 );
	case 0: /* the empty vector */;
    }
    RETURN(1);
})


(define (vector->list (v <vector>))
    (let loop ((i (vector-length v)) (r '()))
	(if (eq? i 0)
	    r
	    (loop (sub1 i) (cons (vector-ref v (sub1 i)) r)))))

(export list->vector
	vector->list
	make-vector
	vector-fill!
	vector
	vector?
	vector-length
	vector-ref
	vector-set!)

(define-shadow vector->list vector->list)
(define-shadow list->vector list->vector)
(define-shadow make-vector make-vector)
(define-shadow _vector vector)
