#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define ($std-class-find-method method-list args)
  (let ((key-class (class-of (car args))))
    (let loop ((i method-list))
      (if (pair? i)
	  (if (subclass? key-class (method-key-class (car i)))
	      (car i)
	      (loop (cdr i)))
	  #f))))

;; Given a list of <Method> objects
;; and a <Method>, destructively modifies the given
;; list to install the new method.  It places it in
;; the right place for <<Class>> inheritance (ie,
;; so that the first match when doing a find-method
;; is the right one)
;; Returns the new list (which will be eq? the original
;; list unless the new method went at the head of the list)

(define ($std-class-install-method! methods method)
  (let-syntax ((insert-before (syntax-form (prev rest)
				   (if prev
				       (begin
					 (set-cdr! prev (cons method rest))
					 methods)
				       (cons method rest)))))
    (let loop ((i methods)
    	       (prev #f))
	(cond ((null? i) 
			(insert-before prev '()))
	      ((eq? (method-key-class method) (method-key-class (car i)))
			; this is a REPLACEMENT of an existing method
			(set-car! i method)
			methods)
	      ((subclass? (method-key-class method) 
			  (method-key-class (car i)))
			(insert-before prev i))
	      (else (loop (cdr i) i))))))
