#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(%define-glue $std-gf-dispatch literals: ((& $std-gf-cache-miss))
{
obj c, m, impl;

    if (arg_count_reg < 1)
    {
	scheme_error( "~a: generic function called with no arguments",
		      1,
		      LEXREF0(2) );
    }
    c = class_of(REG0);
    for (m=LEXREF0(0); !NULL_P(m); m=pair_cdr(m))
    {
        assert( PAIR_P(m) );
	impl = pair_car(m);
	if (EQ(pair_car(impl),c))
            APPLY( arg_count_reg, pair_cdr(impl) );
    }
    COLLECT0();
    REG1 = envt_reg;
    APPLY( 2, TLREF(0) );
})

;; this function gets called by the standard dispatch routine
;; whenever there is a cache miss; that is, there is not an
;; exact association between the class of the "receiver" and
;; an implementation in the method cache

(define ($std-gf-cache-miss args gfe)
    (let ((m ($std-gf-find-method gfe args)))
	(if m
	    (let ((impl (method-proc m)))
		(std-gfe-set-method-cache!
			gfe
			(cons (cons (class-of (car args)) impl)
			      (std-gfe-method-cache gfe)))
		(apply* args impl))
	    (does-not-understand (std-gfe->gf gfe) args))))


(define-macro (std-gfe-clear-method-cache! gfe)
    (std-gfe-set-method-cache! gfe '()))

(define ($std-gf-find-method gfe args)
    ($std-class-find-method (std-gfe-methods gfe) args))
    
(define ($std-gf-add-method! gfe method)
    (std-gfe-clear-method-cache! gfe)
    (std-gfe-set-methods!
	gfe
	($std-class-install-method! (std-gfe-methods gfe) method))
    method)

(define ($std-gf-remove-method! gfe method)
    (std-gfe-clear-method-cache! gfe)
    (std-gfe-set-methods!
	gfe
	(delq! method (std-gfe-methods gfe)))
    method)

(define ($std-gf-remove-all-methods! gfe)
    (std-gfe-clear-method-cache! gfe)
    (std-gfe-set-methods! gfe '())
    gfe)

(define $std-gf-meta-methods
    (%make-gf-envt
	    '()
	    $std-gf-find-method
	    $std-gf-add-method!
	    $std-gf-remove-method!
	    $std-gf-remove-all-methods!))

(define (make-generic-fn name)
    (let* ((gfe (%make-gf-envt $std-gf-meta-methods
			    '()		;; LEXREF0(0): implementation cache
			    '()		;; LEXREF0(1): methods
			    name	;; LEXREF0(2): name
			    #f))
	   (gf (%make-closure gfe $std-gf-dispatch)))
	(%gvec-write gfe 4 gf)
	gf))

(define-rewriter (define-generic-function form)
    (let ((name (car (cdr form))))
	(list 'define
	      name
	      (list 'quote
		    (make-generic-fn name)))))

(define-rewriter (define-generic-fn form)
    (cons 'define-generic-function (cdr form)))

(export define-generic-fn make-generic-fn define-generic-function)

(define-shadow %define-method-rewriter %define-method-rewriter)
(define-shadow subclass? subclass?)

(define-macro (%define-method generic-fn specializers code)
    (add-method! generic-fn 
    		 (make-method
			specializers
			(%make-closure '() code))))

(define-rewriter (define-method form)
    (%define-method-rewriter form))
    
(define (add-method! gf method)
    (let ((gfe (gf->gfe gf)))
	((mme-add-method (gfe->mme gfe)) gfe method)))

(define (find-method gf args)
    (let ((gfe (gf->gfe gf)))
	((mme-find-method (gfe->mme gfe)) gfe args)))

(define (find-implementation gf args)
    (method-proc (find-method gf args)))

(define (generic-function-name gf)
    (gfe-name (gf->gfe gf)))

(define (generic-function? thing)
    (if (procedure? thing)
	(let ((tmpl (%gvec-read thing 1)))
	    (or (eq? tmpl $std-gf-dispatch)
		(eq? tmpl $writer-dispatch)
		(eq? tmpl $init-dispatch)))
	#f))

(export %define-method define-method
	%define-method-rewriter
	add-method! find-method find-implementation
	generic-function-name)
	
(export $std-gf-dispatch generic-function?)

(define (class-name-of x)
    (class-name (class-of x)))

(define (method-signature-name m)
    (class-name (method-key-class m)))

(define (does-not-understand gf args)
    (let ((gfe (gf->gfe gf)))
      (abort (gfe-name gfe)
	     "Do not understand: ~s (only understand ~s)"
	     (map class-name-of args)
	     (map method-signature-name (gfe-methods gfe)))))


		