#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
*******NOTICE********
Do NOT implement this feature
until the new system is running
within itself.
(it will make it a lot easier, anyway)

;; special code to handle the head-compilation of a
;; <TopLevelVar> when the value is (a) constant
;; and (b) a generic function

;; processing amounts to seeing if there is a type
;; restriction on the first argument such that we
;; can statically select a method -- which is to say,
;; there is an APPLICABLE <InlinableMethod>
;; of course, this involves compiling the argument
;; forms, which is redundant in most cases (even if
;; we expand it as a combo, we will wind up re-compiling
;; them in compile/combo)

(define (compile-head-gf gf form lex-envt dyn-envt mode)
    (let ((arg-types (static-arg-types (cdr form) lex-envt dyn-envt)))
      (if (null? arg-types)
	  (begin
	    (warning "Call to generic function ~a has no arguments"
		     (car form))
	    (compile/combo form lex-envt dyn-envt mode))
	  ;; note that this will work with multi-dispatch as well
	  (let ((method (find-method-from-types gf arg-types)))
	    (if (isa? f <InlinableMethod>)
		(compile-head (method-macro f)
			      form
			      lex-envt
			      dyn-envt
			      mode)
		;; otherwise, we have to resort to a full gf call
		(compile/combo form lex-envt dyn-envt mode))))))
		    
      
(define (static-arg-types arg-exprs lex-envt dyn-envt)
    (map first-return-type (compile/multi arg-exprs lex-envt dyn-envt 'value)))

;; create a macro to handle a getter

(define (make-getter-macro name slot lex-envt)
    (make <Macro> name: name
	          envt: lex-envt
		  forms: (list (make <MacroForm>
				     args: '(self)
				     body: (list (list '%gvec-read
						       'self
						       slot))))))

(define (make-setter-macro name slot type lex-envt)
    (define (setter-macro-body)
      (if (eq? type <Object>)
	  (list (list '%gvec-write 'self slot value))
	  (list `(let ((x value)) 
		   (if (isa? x ',type)
		       (%gvec-write self ,slot x)
		       (abort ',name
			      "value ~s is not acceptable"
			      x))))))
		 
    (make <Macro> name: name
	          envt: lex-envt
		  forms: (list (make <MacroForm>
				     args: '(self value)
				     body: (setter-macro-body)))))

 
(begin
  (define <Point> (make-class ...))
  (add-method! x (make-inlinable-method 
		  (list <Point>) 
		  (make-getter 'x 0)
		  (macro (self)
		    (%gvec-read self 0))))
  (add-method! y (make-inlinable-method
		  (list <Point>)
		  (make-getter 'y 1)
		  (macro (self)
		    (%gvec-read self 1))))
  (add-method! set-x! (make-inlinable-method
		       (list <Point>)
		       (make-setter 'x 0 <Number>)
		       (macro (self value)
			 (%gvec-write self 0 value))))

