#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; (define-class <Point> (<Object>) x y)


(define (%define-class-rewriter form eval)
    (let ((options (list (car form)))
    	  (f (cdr form)))
	(if (eq? (car f) 'root:)
	    (begin
		(set! options (append options (list (car f) (cadr f))))
		(set! f (cddr f))))
	(let* ((name (car f))
	       (info (%form-class-inits name 
					(cadr f) 
					(cddr f)
					eval))
	       (definition (append (cons 'define (cdr options))
				   (list name
					 (cdr info)))))
	  (if (car info)
	      (list 'begin
		    definition
		    (list '%define-getters name)
		    (list '%define-setters name))
	      definition))))

#|
    (define <Point> (<Object>) x y)
  ==>
    (begin
      (define <Foo> (%make-class* '<Foo> 0 1 (%list <Object>) ...))
      (add-method! x (make-method (%list <Foo>)
				  (%make-getter 'x 0)))
      (add-method! y (make-method (%list <Foo>)
				  (%make-getter 'y 1)))
      (add-method! set-x! (make-method (%list <Foo> <Object>)
				       (%make-unchecked-setter 'set-x! 0)))
      (add-method! set-y! (make-method (%list <Foo> <Object>)
				       (%make-unchecked-setter 'set-y! 1))))
|#

(define (%make-class-rewriter form eval)
    (cdr (%form-class-inits (cadr form) (caddr form) (cdddr form) eval)))


(define (%form-class-inits name parents flags eval)
    (let* ((base-size (if (null? parents)
    			  0
			  (%class-instance-size (eval (car parents)))))
    	   (v (%collect-class-options flags base-size)))
	(list
	    (vector-ref v 3)
	    '%make-class*
	    (list 'quote name)		;; [0] name
	    (vector-ref v 0)		;; [1] heap type
	    (vector-ref v 1)		;; [2] image mode
	    (cons '%list parents)	;; [3] parents
	    (vector-ref v 2))))		;; [4] slot descriptors

(define-syntax %make-class* 
    (syntax-form (name heap-type image-mode parents slots)
       (%make <<Class>> name			;; [0] name
			heap-type		;; [1] heap type
			image-mode		;; [2] image mode
			parents			;; [3] parents
			slots			;; [4] slot descriptors
			$write-trap		;; [5] write method impl
			$display-trap		;; [6] display method impl
			$initialize-trap	;; [7] initialize method impl
			#f #f #f #f #f #f #f #f
			#f #f #f #f #f #f #f #f
			#f #f #f #f)))		;; [8...27]  future expansion

(define (%collect-class-options options base-size)
    (let ((heap-type 0)
    	  (image-mode 0)
	  (slots '())
	  (size base-size)
	  (make-ivar-gfs #t))
	(let loop ((i options))
	    (if (pair? i)
		(case (car i)
		    ((:gvec) 
			(set! heap-type 0)
			(loop (cdr i)))
		    ((:bvec) 
			(set! heap-type 1)
			(loop (cdr i)))
		    ((:immob) 
			(set! heap-type 2)
			(loop (cdr i)))
		    ((:weak1)
			(set! heap-type 3)
			(loop (cdr i)))
		    
		    ((:no-ivar-gfs) 
			(set! make-ivar-gfs #f)
			(loop (cdr i)))
		    
		    ((image-mode:) 
			(set! image-mode (cadr i))
			(loop (cddr i)))
		    (else 
			(set! slots 
			      (cons (%slot-constructor (car i) size) 
				    slots))
			(set! size (add1 size))
			(loop (cdr i))))))
	(vector heap-type
		image-mode
		(cons '%list (reverse slots))
		(if (null? slots)
		    #f
		    make-ivar-gfs))))

(define (%slot-constructor slot-descr position)
    (let ((name slot-descr)
	  (init #f)
	  (type '<Object>))
	(if (pair? slot-descr)
	    (begin
		(set! name (car slot-descr))
		(set! init (cadr slot-descr))
		(if (pair? (cddr slot-descr))
		    (set! type (caddr slot-descr)))))
	(list '%make 
	      '<Slot-Descriptor>
	      (list 'quote name)
	      init
	      type
	      position
	      (list 'quote 
		    (string->symbol 
			(string-append 
			    (symbol->string name) 
			    ":"))))))



(define (%make-rewriter form eval)
    (let* ((the-class (eval (cadr form)))
	   (opts (parse-options the-class))
	   (exprs (collate-initializers! (cdr opts) the-class)))
	(for-each
	    (lambda (init)
		(if (car init)
		    (abort '%make-rewriter 
		    	   "argument `~a ~s' not used"
		    	   (car init)
			   (cdr init))))
	    (cdr opts))
	(list '%initialize
	    (if (car opts)
		(append '(%make*) 
			(list (cadr form)) 
			exprs 
			(list (car opts)))
		(append '(%make) 
			(list (cadr form)) 
			exprs))))


    ; this function takes a list of initializers
    ; (which is an assoc.list mapping keywords to expressions)
    ; and returns a list of expressions for initializing the
    ; instance (in layout order).  As it does so, the initializers
    ; list is side effected so that each used keyword is set to #f.
    
    (define (collate-initializers! initializers the-class)
	(let ((the-slots (all-slots the-class)))
	    (map (lambda (x) (cdr x))
	    	;^-- why this?  So that it is portable between
		;               old and new, and generates FAST
		;		code in the new system (car, in
		;		builtins, is a type-unsafe macro)
		 (sort
		    (map
			(lambda (slot keyword)
			    (let ((a (assq keyword initializers)))
				(cons (%slot-position slot)
				      (if a
					  (begin
					    (set-car! a #f)
					    (cdr a))
					  (list 'quote 
						(%slot-default-value slot))))))
			the-slots
			(map %slot-keyword the-slots))
		    (lambda (x y)
			(binary<? (car x) (car y)))))))

    (define (all-slots the-class)
	(apply append 
	       (map %class-slots (%linear-ancestors the-class))))

    ; returns a list,
    ;	the first being the "rest" expr or #f if none
    ; the remainder being pairs:
    ;	with car => the initialization keyword
    ;   with cdr => the initialization expression
    
    (define (parse-options the-class)
	(let loop ((i (cddr form)) (r '()))
	    (cond ((null? i) 
			(cons #f r))
	    	  ((eq? (car i) '#rest)
			(cons (cadr i) r))
		  ((keyword? (car i))
			(loop (cddr i)
			      (cons (cons (car i) (cadr i)) r)))
		  (else
			(abort 'make/parse-options
			    "Expecting keyword at: ~s" 
			    i))))))

(define (%class-supers class)
    (%gvec-read class 3))

(define (%class-slots class)
    (%gvec-read class 4))

(define (%slot-keyword slot)
    (%gvec-read slot 4))
    
(define (%slot-position slot)
    (%gvec-read slot 3))

(define (%slot-default-value slot)
    (%gvec-read slot 1))

(define (%slot-name slot)
    (%gvec-read slot 0))

(define (%linear-ancestors class)
    (cons class
    	  (if (null? (%class-supers class))
	      '()
	      (%linear-ancestors (car (%class-supers class))))))

(define (%class-all-slots class)
  (apply append (map %class-slots (%linear-ancestors class))))

(define (%class-instance-size class)
    (add1 (apply max 
		 (cons -1 
		       (map %slot-position (%class-all-slots class))))))

(define (keyword? x)
    (and (symbol? x)
    	 (let ((s (symbol->string x)))
	    (eq? (string-ref s (sub1 (string-length s)))
	    	 #\:))))

(define (keyword->symbol x)
    (let ((s (symbol->string x)))
	(string->symbol (substring s 0 (sub1 (string-length s))))))

(define (%define-method-rewriter form)
    (let ((name (cadr form))
    	  (specializers (map (lambda (x)
	  			(if (pair? x)
				    (cadr x)
				    '<Object>))
			     (caddr form)))
	  (body (cdddr form)))
	(list '%define-method
	      name
	      (cons 'list specializers) 
	      (cons '%code
		    (cons (list name (car specializers))
			  (cons (caddr form) body))))))

#|
(define (%define-method-rewriter form)
    (let ((name (cadr form))
    	  (specializers (map (lambda (x)
	  			(if (pair? x)
				    (cadr x)
				    '<Object>))
			     (caddr form)))
	  (vars (map (lambda (x)
	  			(if (pair? x)
				    (car x)
				    x))
			     (caddr form)))
	  (body (cdddr form)))
	(list '%define-method
	      name
	      (cons 'list specializers) 
	      (cons '%code
		    (cons (list name (car specializers))
			(cons vars
				body))))))
|#
(define (ensure-gf envt name expr)
    (let ((b (envt 'lookup name)))
	(if b
	    expr
	    (list 'begin
	    	  (list 'define-generic-fn name)
		  expr))))

(define (%form-getter envt class-name slot-descr)
    (ensure-gf 
	envt 
	(%slot-name slot-descr)
	(list 'add-method!
		(%slot-name slot-descr)
		(list 'make-method
		    (list '%list class-name)
		    (list '%make-getter 
			    (list 'quote (%slot-name slot-descr))
			    (%slot-position slot-descr))))))

(define (%form-setter envt class-name slot-descr)
    (let ((strnm (string->symbol (string-append "set-"
						(symbol->string
						  (%slot-name slot-descr))
						"!"))))
      (ensure-gf 
	    envt 
	    strnm
	    (list 'add-method!
		strnm
		(list 'make-method
			(list '%list class-name '<Object>)
			(list '%make-unchecked-setter
			    (list 'quote strnm)
			    (%slot-position slot-descr)))))))

(define (%define-getters-rewriter form eval envt)
    (let* ((name (cadr form))
	   (class (eval name)))
      (let loop ((s (%class-slots class))
		 (r '()))
	(if (null? s)
	    (cons 'begin r)
	    (loop (cdr s)
		  (cons (%form-getter envt name (car s))
			r))))))

(define (%define-setters-rewriter form eval envt)
    (let* ((name (cadr form))
	   (class (eval name)))
      (let loop ((s (%class-slots class))
		 (r '()))
	(if (null? s)
	    (cons 'begin r)
	    (loop (cdr s)
		  (cons (%form-setter envt name (car s))
			r))))))
