#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-macro (class-of thing)
    (%call ("class_of" (<obj>) => <obj>) thing))

(define-macro (class-of-ptr thing)
    (%call ("CLASSOF_PTR" (<obj>) => <obj>) thing))

(define-macro (ptr? thing)
    (%call ("OBJ_ISA_PTR" (<obj>) => <raw-bool>) thing))

(define-syntax isa-ptr? 
    (syntax-form (thing 'directly: class)
	(if (ptr? thing)
	    (eq? (class-of-ptr thing) class)
	    #f))
    (syntax-form (thing 'indirectly: class)
	(if (ptr? thing)
	    (subclass? (class-of-ptr thing) class)
	    #f)))


(define-macro (subclass? class1 class2)
    (%call ("subclass_p" (<obj> <obj>) => <raw-bool>) class1 class2))

(define-shadow subclass? new-subclass?)

(define (_class-of thing)
    (class-of thing))

(define (_subclass? class1 class2)
    (subclass? class1 class2))

(define-shadow class-of class-of)


(define (class? thing)
    (isa-ptr? thing indirectly: <<Class>>))


(define (instance? object class)
    (subclass? (class-of object) class))
    
(define isa? instance?)
    
(define (class-name x)
    (%gvec-read x 0))

(define-primop (class-heap-type class)
    (%gvec-read class 1))

(define-primop (class-gvec? class)
    (eq? (class-heap-type class) 0))
    
(define-primop (class-bvec? class)
    (eq? (class-heap-type class) 1))
    
(define-primop (class-immob? class)
    (eq? (class-heap-type class) 2))
    
(define (gvec? thing)
    (and (ptr? thing)
    	 (class-gvec? (class-of-ptr thing))))

(define (bvec? thing)
    (and (ptr? thing)
    	 (class-bvec? (class-of-ptr thing))))

(define-primop (fixnum?  thing)
    (%call ("OBJ_ISA_FIXNUM" (<obj>) => <raw-bool>) thing))

(define-primop (immob? thing)
    (%call ("OBJ_ISA_IMMOB" (<obj>) => <raw-bool>) thing))

(export (_class-of class-of)
	(_subclass? subclass?))

(export class-name
	class?
	instance? isa?
	(isa-ptr? %isa-ptr?))

(export class-gvec? class-bvec? class-immob?
	gvec?       bvec?       immob?       fixnum?)
