#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; we have to manually install these methods
;; because the classes were defined in the builtins
;; module

(define-method name ((self <Binding>))
    (%gvec-read self 0))
    
(define-method value ((self <TopLevelVar>))
    (%gvec-read self 1))
    
(define-method write-prot ((self <TopLevelVar>))
    (%gvec-read self 2))
    
(define-method set-value! ((self <TopLevelVar>) x)
    (%gvec-write self 1 x)
    0)
    
(define-method set-write-prot! ((self <TopLevelVar>) x)
    (%gvec-write self 2 x)
    0)


;; Recognizers

(define (macro? x)
    (eq? (class-of x) <Macro>))

(define (substitution? x)
    (eq? (class-of x) <Substitution>))
    

;; Implementation

(define-method write ((self <TopLevelVar>) port)
    (format port "#[tlv ~a]" (name self)))

(define-method compile-head ((self <Rewriter>) form lex-envt dyn-envt mode)
  (compile ((compiler self) form) lex-envt dyn-envt mode))

(define-method compile-head ((self <ExtendedRewriter>) 
			     form 
			     lex-envt 
			     dyn-envt 
			     mode)
  (compile ((compiler self) form 
			    (lambda (expr)
			      (eval expr lex-envt)))
	   lex-envt
	   dyn-envt
	   mode))

(define-method compile-head ((self <SpecialForm>) form lex-envt dyn-envt mode)
    (if (and (syntax-checker self)
	     (not ((syntax-checker self) form lex-envt)))
	(error/syntax "~a form doesn't pass syntax-checker" (name self))
	((compiler self) self form lex-envt dyn-envt mode)))

(define-method compile-ref ((self <SpecialForm>) lex-envt dyn-envt mode)
    (error/syntax "Special form name `~a' referenced as variable" (name self)))

(define-method compile-set ((self <SpecialForm>) 
			    value-expr
			    lex-envt
			    dyn-envt
			    mode)
    (error/syntax "Special form name `~a' used in variable position in set!" 
    		  (name self)))


(define-method compile-head ((self <TopLevelVar>) form lex-envt dyn-envt mode)
    (compile/combo form lex-envt dyn-envt mode))

(define-method compile-ref ((self <TopLevelVar>) lex-envt dyn-envt mode)
    (if (write-prot self)
	(make-const (value self) mode)
	(make <ic-tl-ref> var: self mode: mode)))

(define-method compile-set ((self <TopLevelVar>) 
			    value-expr lex-envt dyn-envt mode)
    (if (write-prot self)
	(error/syntax "Top-level `~a' in set! is write protected" 
    		  (name self))
	(make <ic-tl-set> var: self
			  rhs: (compile value-expr lex-envt dyn-envt 'value)
			  mode: mode)))


(define-method compile-head ((self <LexicalVar>) form lex-envt dyn-envt mode)
    (compile/combo form lex-envt dyn-envt mode))

(define-method compile-ref ((self <LexicalVar>) lex-envt dyn-envt mode)
    (make <ic-lex-ref> var: self 
		       mode: mode
		       return-types: (list (type self))))

(define-method compile-set ((self <LexicalVar>) 
			    value-expr lex-envt dyn-envt mode)
    (make <ic-lex-set> var: self
    		      rhs: (compile value-expr lex-envt dyn-envt 'value)
		      mode: mode))

; Note that compiling a macro use (or a substitution, for that matter)
; preserves the mode (ie, even if it's top).  Hence, you can have
; top-level forms INSIDE of macro bodies that are properly interpreted
; at the top level.
; (THIS IS A FEATURE, and indeed is almost the entire purpose to
;  the mode parameter (certainly to the 'top value of mode)!)

(define-method compile-head ((self <Macro>) form lex-envt dyn-envt mode)
    (if (and *compile-seeing-shadows* (shadow self))
	;; compile as a combo to our shadow, instead!
	(compile/combo (cons (list 'quote (shadow self))
		             (cdr form))
			lex-envt 
			dyn-envt 
			mode)
	(let ((specific (find-match self (cdr form) lex-envt)))
	    (if specific
		(fluid-let ((*place* (name self)))
		    (compile/begin self
				    (cons 'begin (body (car specific)))
				    (cdr specific)
				    dyn-envt
				    mode))
		(if (else-bdg self)
		    (compile-head (else-bdg self) form lex-envt dyn-envt mode)
		    (error/syntax "Use of macro ~a does not match defn" 
		    		(car form)))))))


(define-method compile-ref ((self <Macro>) lex-envt dyn-envt mode)
    (if (else-bdg self)
	(compile-ref (else-bdg self) lex-envt dyn-envt mode)
	(error/syntax "Macro name `~a' referenced as variable" (name self))))

(define-method compile-set ((self <Macro>) 
			    value-expr
			    lex-envt
			    dyn-envt
			    mode)
    (error/syntax "Macro name `~a' used in variable position in set!" 
    		  (name self)))

(define-method compile-ref ((self <Substitution>) lex-envt dyn-envt mode)
    (compile (expr self) (envt self) dyn-envt mode))

(define-method compile-head ((self <Substitution>) form lex-envt dyn-envt mode)
    ; if our expr is a symbol, that symbol may be ct-bound to 
    ; a substitution rather than a variable...
    (let loop ((e (expr self)) (lxe (envt self)))
	(if (symbol? e)
	    (let ((bdg (lookup lxe e)))
		
		; random thoughts from space:
		;  if we think of macros as syntactic CLOSURES, is
		;  what we're doing here (threading through substitutions,
		;  etc.) analagous to syntactic CONTINUATIONS?
		
		(if (substitution? bdg)
		    (loop (expr bdg) (envt bdg))
		    (if (macro? bdg)
			(compile-head bdg form lex-envt dyn-envt mode)
			(make <ic-call> 
			    function: (compile e lxe dyn-envt 'value)
		    	    args: (compile/multi (cdr form) 
			    			 lex-envt 
						 dyn-envt 
						 'value)
		    	    mode: mode))))
	    (make <ic-call> function: (compile e lxe dyn-envt 'value)
			    args: (compile/multi (cdr form) 
					      lex-envt 
					      dyn-envt 
					      'value)
			    mode: mode))))


(define-method compile-set ((self <Substitution>) 
			    value-expr
			    lex-envt
			    dyn-envt
			    mode)
    ; if our expr is a symbol, that symbol may be ct-bound to 
    ; a substitution rather than a variable...
    (let loop ((e (expr self)) (lxe (envt self)))
	(if (symbol? e)
	    ; if this subst. is bound to a name
	    ; then look up that name and transfer responsibility
	    ; to it (note that we can't just recurse on substitutions,
	    ; because we have to remember the lex. envt of the
	    ; value-expr)
	    (let ((bdg (lookup lxe e)))
		(if (substitution? bdg)
		    (loop (expr bdg) (envt bdg))
		    (compile-set bdg value-expr lex-envt dyn-envt mode)))
	    (error/syntax 
		"Cannot set! subst. `~a' (replace with `~a')" 
			(name self)
			(expr self)))))

;============================= Root Variables =============================

(define-method compile-ref ((self <RootVar>) lex-envt dyn-envt mode)
    (if (write-prot self)
	(make-const (value self) mode)
	(make <ic-root-ref> var: self mode: mode)))

(define-method compile-set ((self <RootVar>) 
			    value-expr lex-envt dyn-envt mode)
    (if (write-prot self)
	(error/syntax "Root `~a' in set! is write protected" 
    		  (name self))
	(make <ic-root-set> var: self
    		      rhs: (compile value-expr lex-envt dyn-envt 'value)
		      mode: mode)))

;============================= Loop Variables =============================

(define-method compile-head ((self <loop-var>) form lex-envt dyn-envt mode)
    (if (eq? mode 'tail)
	(make <ic-jump> 
	    args: (compile/multi (cdr form) lex-envt dyn-envt 'value)
	    loop-var: self)
	(begin
	    (format #t 
		"Compiling loop `~s' with full semantics (non-tail)\n" 
		(name self))
	    ((compile-with-full-semantics self)))))

;; if a loop variable is ever referenced or set (or called from a non-tail
;; position)
;; then we jump back to compile it with full semantics

(define-method compile-ref ((self <loop-var>) lex-envt dyn-envt mode)
    (format #t "Compiling loop `~s' with full semantics (refd)\n" (name self))
    ((compile-with-full-semantics self)))

(define-method compile-set ((self <loop-var>) 
			    value-expr lex-envt dyn-envt mode)
    (format #t "Compiling loop `~s' with full semantics (set!)\n" (name self))
    ((compile-with-full-semantics self)))


