#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; we have to manually install this methods
;; because the classes were defined in the builtins
;; module

(define-method table ((self <TopLevelEnvt>))
    (%gvec-read self 0))

(define-method set-table! ((self <TopLevelEnvt>) table)
    (type-check table table? <table>)
    (%gvec-write self 0 table)
    table)


(define (lookup-type name envt)
    (let ((bdg (lookup envt name)))
	(if bdg
	    (if (class? (value bdg))
		(value bdg)
		(error/syntax "~a is not a type name" name))
	    (error/syntax "~a is not bound" name))))

(define (make-top-level-envt)
    (make <TopLevelEnvt>
	table: (make-table eq? symbol->hash)))

(define (make-lexical-envt bindings lex-enclosing dyn-enclosing)
    (make <LexicalEnvt>
	name->bindings: (map (lambda (bdg)
				(cons (name bdg) bdg))
			     bindings)
	bindings: bindings
	lexical-enclosing: lex-enclosing
	dynamic-enclosing: dyn-enclosing))

;; proc is a procedure of one argument (the new environment)
;; which is expected to return some bindings for the environment
;; (this interface supercedes the old "set-bindings!" kluge,
;; wherin the CALLER was responsible for breaking the construction
;; of a lex envt into two parts)

(define (make-lexrec-envt proc lex-enclosing dyn-enclosing)
    (let* ((new-envt (make <LexicalEnvt> lexical-enclosing: lex-enclosing
					  dynamic-enclosing: dyn-enclosing))
	   (new-bdgs (proc new-envt)))
	(set-bindings! new-envt new-bdgs)
	(set-name->bindings! new-envt (map (lambda (bdg)
						(cons (name bdg) bdg))
					   new-bdgs))
	new-envt))

(define (num-bindings envt)
    (length (bindings envt)))

(define (shallow-lookup tl-envt name)
    (table-lookup (table tl-envt) name))

(define-method lookup ((self <LexicalEnvt>) name)
    (let ((b (assq name (name->bindings self))))
	(if b
	    (cdr b)
	    (lookup (lexical-enclosing self) name))))

(define-method lookup ((self <TopLevelEnvt>) name)
    (table-lookup (table self) name))

(define-method bind! ((self <TopLevelEnvt>) bdg)
    (table-insert! (table self) (name bdg) bdg))

;; avoid dynamic dispatch...

(define (tle-bind! (self <TopLevelEnvt>) bdg)
    (table-insert! (%gvec-read self 0) (name bdg) bdg))

;; later on we can optimize this to look up the bind!
;; method for envt and just call it repeatedly, saving
;; a GF dispatch per item

(define (multi-bind! envt stuff)
    (for-each
	(lambda (bdg)
	    (bind! envt bdg))
     stuff))

(define (the-top-level envt)
    (if (instance? envt <TopLevelEnvt>)
	envt
	(the-top-level (lexical-enclosing envt))))
