#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (process-formals$ formals-list nil-thunk last-proc middle-proc)
    (let loop ((p formals-list))
	(cond ((null? p) (nil-thunk))
	      ((symbol? p) (last-proc p))
	      ((pair? p) 
	       (if (eq? (car p) '#rest)
		   (if (pair? (cdr p))
		       (last-proc (cadr p))
		       (error/syntax "Illegal #rest clause in formals: ~s" p))
		   (middle-proc (car p) (loop (cdr p)))))
	      (else (error/syntax "Badly formatted formals: ~s" 
	      		          formals-list)))))

#|
(define (formals-form->formal-names formals-list)
    (process-formals$ 
	formals-list
	(lambda () '())
	(lambda (last) (list last))
	(lambda (item rest) (cons item rest))))
|#

(define (compute-num-args formals-list)
    (process-formals$ 
	formals-list
	(lambda () 0)
	(lambda (last) 1)
	(lambda (item rest) (add1 rest))))

(define (compute-has-rest formals-list)
    (process-formals$ 
	formals-list
	(lambda () #f)
	(lambda (last) #t)
	(lambda (item rest) rest)))


;; convert a "specifier" into a binding
;; of the given class.  The lex-envt is
;; used to look up type names, if given
;; a specifier is either "name" or "(name type)", for now

;; (this needs to be fixed to handle intentional capturing
;;  of variables.  In particular, if a name is given in the
;;  variable position, and the name is bound to a substitution,
;;  then we descend into the substitution, looking to bottom
;;  out in a name that is not so bound.  Then, an "alias" binding
;;  needs to be inserted into the dynamic chain to record that
;;  a particular name in a particular (captured) lexical environment
;;  really means something completely different.  THEN, when
;;  we do a lookup, after we find the lexical var, we need to
;;  look up the dynamic chain for an aliasing.)

(define (specifier->lex-var spec lex-envt)
    (if (pair? spec)
	(make <LexicalVar> name: (car spec) 
			   type: (lookup-static-type lex-envt (cadr spec)))
	(if (symbol? spec)
	    (make <LexicalVar> name: spec type: <Object>)
	    (error/syntax "~s: Illegal specifier" spec))))

;; return the class for a type specification

(define (lookup-static-type envt type-expr)
    (let ((c (lookup envt type-expr)))
	(if c
	    (if (class? (value c))
		(value c)
		(error/syntax 
		    "~s: binding of ~a is ~a, not a class" 
		    type-expr
		    c
		    (value c)))
	    (error/syntax "Type name ~a not bound" type-expr))))

;; formals is a list of formal variables in CANONICAL format
;;   formals ::= ( specifier ... )
;;               ( specifier ... #rest specifier )
;;               ( specifier ... . name )
;;   specifier ::= name
;;                 | ( name type )

(define (make-lex-vars formals lex-envt)
  (process-formals$
   formals
   (lambda () '())
   (lambda (last) 
     (list (specifier->lex-var last lex-envt)))
   (lambda (item rest) 
     (cons
      (specifier->lex-var item lex-envt)
      rest))))
