#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (find-lex-addr ct-envt var)
    (let loop ((e ct-envt) (i 0))
	(let ((x (memq var (car e))))
	    (if x
		(cons i (- (length (car e)) (length x)))
		(loop (cdr e) (add1 i))))))


; Generate NATIVE closure-threaded code

(define %if (%template (ctcprims 1097) 0))
(define %multi (%template (ctcprims 1097) 1))
(define %combo (%template (ctcprims 1097) 2))
(define %literal (%template (ctcprims 1097) 3))

; imported from builtins:
;	(define root: "procedure_template" 
;		%procedure (%template (ctcprims 1097) 4))

(define %bind (%template (ctcprims 1097) 5))
(define %seq (%template (ctcprims 1097) 6))
(define %lex-set! (%template (ctcprims 1097) 7))
(define %tl-set! (%template (ctcprims 1097) 8))
(define %lambda (%template (ctcprims 1097) 9))
(define %root-ref (%template (ctcprims 1097) 10))
(define %root-set! (%template (ctcprims 1097) 11))

(define %ctc-eval (make-closure '() (%template (ctcprims 1097) 14)))
;(snarf %ctc-eval)
#|
(define %ctc-eval
    (%make <Closure>
	'()
	(%make <Template>
	    (dynamic-load-lookup 
		(open-dynamic-load "runtim/foo.o") 
		"_ctc_eval_22_00")
	    0)))
|#

(define (nctc-wrap template . things)
    (make-closure (list->binding-envt '() things)
	          template))

(define-method gen-nctc ((self <ic-lex-ref>) ct-envt)
    (let ((lex-addr (find-lex-addr ct-envt (var self))))
	(bitwise-or (bit-shift-left 1 29)
		    (bitwise-or (bit-shift-left (car lex-addr) 14)
		    		(cdr lex-addr)))))

(define-method gen-nctc ((self <ic-lex-set>) ct-envt)
    (let ((lex-addr (find-lex-addr ct-envt (var self)))
    	  (val-expr (gen-nctc (rhs self) ct-envt)))
	(nctc-wrap %lex-set! 
	    (car lex-addr) 
	    (cdr lex-addr) 
	    val-expr)))

(define-method gen-nctc ((self <ic-root-ref>) ct-envt)
    (let ((root-var (var self)))
	(ct-format #t "gen-nctc <ic-root-ref>: ~a is ~a.~d\n"
		      (name (var self))
		      (module root-var)
		      (slot root-var))
	(nctc-wrap %root-ref
	    (symbol->string (module root-var))
	    (slot root-var))))
	    
(define-method gen-nctc ((self <ic-root-set>) ct-envt)
    (let ((root-var (var self)))
	(nctc-wrap %root-set! 
	    (symbol->string (module root-var)) 
	    (slot root-var) 
	    (gen-nctc (rhs self) ct-envt))))


(define-method gen-nctc ((self <ic-tl-ref>) ct-envt)
    ; a TLV by itself is an immediate expr which evaluates
    ; to the current value of the TLV
    (var self))

(define-method gen-nctc ((self <ic-tl-set>) ct-envt)
    (let ((tl-var (var self))
    	  (val-expr (gen-nctc (rhs self) ct-envt)))
	(nctc-wrap %tl-set!
	    tl-var
	    val-expr)))

(define-method gen-nctc ((self <ic-const>) ct-envt)
    (let ((val (value self)))
	
	; most constants are evaluated by means of an
	; immediate expression of the same value.  However,
	; some objects have a different meaning when used
	; as an immediate expr (eg, some integers indicate
	; a lexical lookup, and a TLV indicates a tl-ref)
	
	(if (or (procedure? val)
		(and (integer? val)
		     (eq? (bit-shift-right val 28) 2))
		(eq? (class-of val) <TopLevelVar>))
	    (nctc-wrap %literal val)
	    val)))


(define-method gen-nctc ((self <ic-multi>) ct-envt)
    (let ((things (map (lambda (x) (gen-nctc x ct-envt)) (arg-list self))))
	(nctc-wrap %multi things)))

(define-method gen-nctc ((self <ic-seq>) ct-envt)
    (let ((things (map (lambda (x) (gen-nctc x ct-envt)) (stmt-list self))))
	(nctc-wrap %seq things)))

(define-method gen-nctc ((self <ic-call>) ct-envt)
    (let ((fn (gen-nctc (function self) ct-envt))
	  (args (gen-nctc (args self) ct-envt)))
	(nctc-wrap %combo fn args)))

(define-method gen-nctc ((self <ic-call-prim>) ct-envt)
    (warning "Could not generate xctc code for call to ~a" (function self))
    (nctc-wrap %literal #f))
	    
(define-method gen-nctc ((self <ic-prim-expr>) ct-envt)
    (warning "Could not generate xctc code for `~a'" (expr self))
    (nctc-wrap %literal #f))

(define-method gen-nctc ((self <ic-if>) ct-envt)
    (let ((test (gen-nctc (condition self) ct-envt))
    	  (true-br (gen-nctc (if-true self) ct-envt))
    	  (false-br (gen-nctc (if-false self) ct-envt)))
	(nctc-wrap %if test true-br false-br)))
	
;; NOTE:  Because we have multiple-value return now, the
;; initializer expr for a %bind need not necessarily be
;; a %multi.

(define-method gen-nctc ((self <ic-bind>) ct-envt)
    (nctc-wrap %bind
	(num-args self)
	(rest? self)
	'%bind
	(gen-nctc (inits self) ct-envt)
	(gen-nctc (body self) (cons (bindings (envt self)) ct-envt))))

;; this returns a really funny thing, which is not
;; a piece of ctc code, but rather a binding environment
;; which represents the procedure.  This BE can be
;; put in a %lambda to make a (lambda ...), or it
;; can be shoved in a %procedure (with an extra BE
;; for the RT-ENVT) to make an executable ctc procedure.

#|
(define-method gen-nctc ((self <ic-procedure>) ct-envt)
    (let ((arg-binder (body self)))
	(list->binding-envt '() 
	    (list
		(num-args arg-binder)
		(rest? arg-binder)
		(name self)
		(gen-nctc (body arg-binder) 
			  (cons (bindings (envt arg-binder))
				ct-envt))))))
|#

(define-method gen-nctc ((self <ic-procedure>) ct-envt)
    (let ((arg-binder (body self)))
	(make-ctc-template
	    (list->binding-envt '() 
		(list
		    (num-args arg-binder)
		    (rest? arg-binder)
		    (name self)
		    (gen-nctc 
			(body arg-binder) 
			(cons (bindings (envt arg-binder))
			      ct-envt)))))))

(define-method gen-nctc ((self <ic-lambda>) ct-envt)
    (nctc-wrap %lambda (gen-nctc (proc self) ct-envt)))

;; take a procedure (icode) at the top level and return
;; a xctc closure that effects it

;; use only on <ic-procedure>'s
;; (ie, as returned from compile/procedure)

#|
(define (xwrap-procedure icode)
    (let ((proc (gen-nctc icode '())))
	(make-closure (list->binding-envt proc '(()))
		      %procedure)))
|#

(define (xwrap-procedure icode)
    (make-closure '() (gen-nctc icode '())))

(define (compile/%code sf form lex-envt dyn-envt mode)
    (let ((name (cadr form))
	  (formals (caddr form))
	  (body (cdddr form)))
	(make-const (gen-nctc (compile/procedure 
				    name 
				    formals
				    body
				    lex-envt
				    dyn-envt 
				    #f) 
			      '())
		    mode)))


;; evaluate a piece of CTC in a given RT environment

#|
(%define-glue (%ctc-eval envt ctc)
{
extern bool immediate_expr( obj expr );
extern obj eval_immediate_expr( obj expr );

    if (immediate_expr(ctc))
    { 
	REG0 = eval_immediate_expr(ctc); 
	RETURN(1); 
    }
    else 
	APPLY( 1, ctc );
})
|#

(define (eval expr envt)
    (let ((ic (compile expr envt envt 'top)))
	(if (symbol? ic)
	    ;; it was a DEFINE-type form
	    ic
	    (if (icode? ic)
		(%ctc-eval '() (gen-nctc ic '()))
		(values)))))
