#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (icode? thing)
    (instance? thing <icode>))

(define-method initialize ((self <icode>))
    (warning "Did not get initialized: ~s" self)
    self)
    

(define-method first-return-type ((self <expr-icode>))
    (if (null? (return-types self))
	(class-of '#none)
	(car (return-types self))))

(define-method initialize ((self <ic-call>))
    (set-does-lambda?! self (or (does-lambda? (function self))
		                (does-lambda? (args self))))
    (set-does-apply?! self #t)
    self)

(define-method initialize ((self <ic-call-prim>))
    (set-does-lambda?! self (does-lambda? (args self)))
    (set-does-apply?! self (does-apply? (args self)))
    self)


(define-method initialize ((self <ic-prim-expr>))
    (set-does-lambda?! self #f)
    (set-does-apply?! self #f)
    self)



(define-method initialize ((self <ic-bind>))
    (set-does-lambda?! self (or (does-lambda? (body self))
    				(and (inits self) 
				     (does-lambda? (inits self)))))
    (set-does-apply?! self (or (does-apply? (body self))
    			       (and (inits self) 
			            (does-apply? (inits self)))))
    (set-num-args! self (- (num-bindings (envt self)) 
    			   (if (rest? self) 1 0)))
    (set-return-types! self (return-types (body self)))
    self)

    
(define-method initialize ((self <ic-tl-ref>))
    self)

(define-method initialize ((self <ic-lex-ref>))
    self)
(define-method initialize ((self <ic-root-ref>))
    self)



(define-method initialize ((self <ic-tl-set>))
    (set-does-lambda?! self (does-lambda? (rhs self)))
    (set-does-apply?! self (does-apply? (rhs self)))
    self)

(define-method initialize ((self <ic-lex-set>))
    (set-does-lambda?! self (does-lambda? (rhs self)))
    (set-does-apply?! self (does-apply? (rhs self)))
    self)

(define-method initialize ((self <ic-root-set>))
    (set-does-lambda?! self (does-lambda? (rhs self)))
    (set-does-apply?! self (does-apply? (rhs self)))
    self)



(define-method initialize ((self <ic-if>))
    (set-does-lambda?! self (or (does-lambda? (condition self)) 
    				(does-lambda? (if-true self)) 
				(does-lambda? (if-false self))))
    (set-does-apply?! self (or (does-apply? (condition self)) 
			       (does-apply? (if-true self)) 
			       (does-apply? (if-false self))))
    self)


(define-method initialize ((self <ic-const>))
    self)

(define (const-type value)
    (cond
	((integer? value) <Integer*>)
	((string? value) <String*>)
	((boolean? value) <Boolean*>)
	(else <Object*>)))
	
(define (make-const value mode)
    (make <ic-const>
	value: value
	mode: mode
	return-types: (list (const-type value))))
	
(define-method initialize ((self <ic-procedure>))
    (set-does-lambda?! self (does-lambda? (body self)))
    (set-does-apply?! self (does-apply? (body self)))
    self)


(define-method initialize ((self <ic-lambda>))
    (set-does-lambda?! self #t)
    (set-does-apply?! self #f)
    (set-return-types! self (list <Closure*>))
    self)


(define-method num-args ((self <ic-multi>))
    (length (arg-list self)))
    
(define-method initialize ((self <ic-multi>))
    (set-does-lambda?! self (any? does-lambda? (arg-list self)))
    (set-does-apply?! self (any? does-apply? (arg-list self)))
    (set-return-types! self (map first-return-type (arg-list self)))
    self)


(define-method initialize ((self <ic-seq>))
    (set-does-lambda?! self (any? does-lambda? (stmt-list self)))
    (set-does-apply?! self (any? does-apply? (stmt-list self)))
    self)


(define-method write ((self <ic-call>) port)
    (format port "[call ~s ~s]" (function self) (args self)))

(define-method write ((self <ic-call-prim>) port)
    (format port "[call-prim ~a ~s]" (function self) (args self)))

(define-method write ((self <ic-prim-expr>) port)
    (format port "{~a}" (expr self)))

(define-method write ((self <ic-const>) port)
    (format port "'~s" (value self)))

(define-method write ((self <ic-multi>) port)
    (display "{multi" port)
    (let loop ((x (arg-list self)))
	(if (pair? x)
	    (begin
		(write-char #\space port)
		(write (car x) port)
		(loop (cdr x)))))
    (write-char #\} port))

(define-method write ((self <ic-bind>) port)
    (format port "{bind ~s=~s; ~s}" 
    	    (map name (bindings (envt self)))
	    (inits self)
	    (body self)))

(define-method write ((self <ic-seq>) port)
    (display "{seq" port)
    (for-each
	(lambda (i)
	    (write-char #\space port)
	    (write i port))
	(stmt-list self))
    (write-char #\} port))

(define-method write ((self <ic-tl-ref>) port)
    (format port "[tlref ~a]" (name (var self))))

(define-method write ((self <ic-tl-set>) port)
    (format port "[tlset ~a=~s]" (name (var self)) (rhs self)))

(define-method write ((self <ic-lex-ref>) port)
    (format port "[lexref ~a]" (name (var self))))

(define-method write ((self <ic-lex-set>) port)
    (format port "[lexset ~a=~s]" (name (var self)) (rhs self)))

(define (any? predicate list)
    (let loop ((x list))
	(if (pair? x)
	    (if (predicate (car x))
		#t
		(loop (cdr x)))
	    #f)))

(define (all? predicate list)
    (let loop ((x list))
	(if (pair? x)
	    (if (predicate (car x))
		(loop (cdr x))
		#f)
	    #t)))



(define-method num-args ((self <ic-loop>))
    (length (bindings (envt self))))

(define-method rest? ((self <ic-loop>))
    #f)

(define-method initialize ((self <ic-loop>))
    (set-does-lambda?! self (or (does-lambda? (body self))
    				(does-lambda? (inits self))))
    (set-does-apply?! self (or (does-apply? (body self))
    			       (does-apply? (inits self))))
    self)
    
(define-method initialize ((self <ic-jump>))
    (set-does-lambda?! self (does-lambda? (args self)))
    (set-does-apply?! self (does-apply? (args self)))
    self)
    
