#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define (binding-isa-subst binding)
    (and binding
	(instance? binding <Substitution>)))

(define (match-context-cancel context)
    ((vector-ref context 0) #f))
    

(define (match-context-add-subst context symbol expr envt)
    (vector-set! context 1
	(cons
	    (make <Substitution> name: symbol
	    			 expr: expr
				 envt: envt)
	    (vector-ref context 1))))

;; NOTICE:  This algorithm does not work
;;          if you expect the type of an argument to
;;          change because of intentional capturing
;;  (define-syntax foo
;;	(syntax-form (a (b isa: <pair>))
;;	    (let (((a <pair>) (cons 1 1)))
;;		b)))
;;  (foo x x) ==> no match, because x is not obviously
;;                a <pair> at this point, even though
;;		  the actual binding that's important for the
;;		  second x would be a <pair>

(define (match-type? actual-expr actual-envt formal-type)
    (format #t "matching type of ~s to ~s...\n" actual-expr formal-type)
    (let ((r (return-types (compile actual-expr 
				    actual-envt 
				    actual-envt 
				    'value))))
	(format #t "actual types are: ~s\n" r)
        (if (pair? r)
	    (subclass? (car r) formal-type)
	    #f)))

(define (match formal actual actual-envt context)
    (cond
	((symbol? formal)
	    (match-context-add-subst context formal actual actual-envt))
	((and (pair? formal)
	      (pair? (cdr formal))
	      (eq? (cadr formal) 'isa:))
	    (if (not (match-type? actual 
	    			  actual-envt 
				  (lookup-static-type (vector-ref context 2)
				  		      (caddr formal))))
		(match-context-cancel context))
	    (match-context-add-subst context (car formal) actual actual-envt))
	((and (pair? formal) (eq? (car formal) 'quote))
	    (if (not (eq? (cadr formal) actual))
		(match-context-cancel context)))
	((pair? formal)
	    (if (pair? actual)
		(begin
		    (match (car formal) (car actual) actual-envt context)
		    (match (cdr formal) (cdr actual) actual-envt context))
		;;
		;; the last chance for this match is if actual is
		;; a symbol which is bound to a substitution
		;;
		(if (symbol? actual)
		    (let ((bdg-info (lookup actual-envt actual)))
			(if (binding-isa-subst bdg-info)
			    (match formal
				   (expr bdg-info)
				   (envt bdg-info)
				   context)
			    (match-context-cancel context)))
		    (match-context-cancel context))))
	((null? formal)
	    (if (null? actual)
		#t
		;; there is the possibility that actual is a
		;; substitution name that is bound to the empty list...
		(if (symbol? actual)
		    (let ((bdg-info (lookup actual-envt actual)))
			(if (binding-isa-subst bdg-info)
			    (match formal
				   (expr bdg-info)
				   (envt bdg-info)
				   context)
			    (match-context-cancel context)))
		    (match-context-cancel context))))
	(#t (error/internal "Bad macro formal" formal))))

(define (match-args formals actuals actual-envt defn-envt)
    (call-with-current-continuation
	(lambda (exit)
	    (let ((ctx (vector exit '() defn-envt)))
		(match formals actuals actual-envt ctx)
		(make-lexical-envt (vector-ref ctx 1) defn-envt #f)))))

(define (compile-syntax-form form)
    (make <MacroForm> args: (cadr form)
		      body: (cddr form)))
    
(define (compile-macro macro-name definition lexical-envt)
    (if (pair? macro-name)
    	;; special-case the form "(define-syntax (foo x) ...)"
	(make <Macro> name: (car macro-name)
		      forms: (list (make <MacroForm> args: (cdr macro-name)
		      				     body: definition))
		      envt: lexical-envt)
	(let ((else-bdg #f))
	    (let loop ((s definition) (fs '()))
		(if (null? s)
		    (make <Macro> name: macro-name
				    forms: (reverse fs)
				    envt: lexical-envt
				    else-bdg: else-bdg)
		    (if (pair? s)
			(let ((d (car s)))
			    (if (eq? (car d) 'else)
				(if else-bdg
				    (error/syntax 
					"Multiple else parts in macro ~s's defn" 
					macro-name)
				    (begin
					(set! else-bdg 
						(make <Substitution> 
						    name: macro-name
						    expr: (cadr d)
						    envt: lexical-envt))
					(loop (cdr s) fs)))
				(loop (cdr s)
					(cons (compile-syntax-form d) fs))))
			(error/syntax "Bad syntax-form (for macro ~s) at: ~s"
					macro-name
					s)))))))

		    

(define (find-match macro-binding actuals lexical-envt)
    (call-with-current-continuation
	(lambda (found-it)
	    (for-each
		(lambda (specific-form)
		    (let ((match-result (match-args
				(args specific-form) 
				actuals 
				lexical-envt 
				(envt macro-binding))))
			(if match-result
			    (found-it (cons specific-form match-result)))))
		(forms macro-binding))
	    #f)))
