#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

;; Standard Special Forms
;; ----------------------

(define (compile/values sf form lex-envt dyn-envt mode)
    (compile/multi (cdr form) lex-envt dyn-envt mode))

;; BEGIN is a little strange because the Scheme standard
;; defines that (begin (define ...)) is the same as (define ...)
;; so we have to process the contents of a top-level begin
;; at the top level
;;
;; Other compound constructs, when compiled at the top-level,
;; regard their tail part as being in 'tail position (using the untop fn)
;; (although they continue to regard themselves as being at top)

(define (compile/begin sf form lex-envt dyn-envt mode)
    (if (eq? mode 'top)
	(let loop ((stmts (cdr form)) (r '()))
	    (if (null? stmts)
		(cond ((null? r) (make-const '#undef mode))
		      ((null? (cdr r)) (car r))
		      (else (make <ic-seq> return-types: (return-types (car r))
		      		           stmt-list: (reverse r)
					   mode: mode)))
		(if (symbol? stmts)
		    ;; check for substituion...
		    (let ((bdg (lookup lex-envt stmts)))
			(if (substitution? bdg)
			    (begin
				(set! lex-envt (envt bdg))
				(loop (expr bdg) r))
			    (error/syntax
				"Rest symbol ~a not a substitution" stmts)))
		    (let ((ic (compile (car stmts) lex-envt dyn-envt 'top)))
			(loop (cdr stmts) (if (icode? ic) (cons ic r) r))))))
	(compile/body (cdr form) lex-envt dyn-envt mode)))
	
(define (compile/let sf form lex-envt dyn-envt mode)
    (if (symbol? (cadr form))
	(compile/named-let sf form lex-envt dyn-envt mode)
	(compile/let-bdg sf form lex-envt dyn-envt mode)))

(define (compile/named-let* form lex-envt dyn-envt mode)
    (let ((loop-vars (map car (caddr form)))
    	  (loop-inits (map cadr (caddr form)))
	  (loop-body (cdddr form))
	  (loop-name (cadr form))
	  (loop-binding (specifier->lex-var (cadr form)
					    ; could be:
					    ;  (list (cadr form) <Closure*>)
					    ; if nobody side-effects `loop'
					    lex-envt)))
	(let ((new-envt (make-lexical-envt  (list loop-binding)
					    lex-envt
					    dyn-envt)))
	    (make-letrec new-envt
	    		 (list (make <ic-lambda> mode: 'value
			 			 proc: (compile/procedure
						 	   loop-name
							   loop-vars
							   loop-body
							   new-envt
							   new-envt)))
			 (make <ic-call> 
				function: (make <ic-lex-ref> 
						var: loop-binding 
						mode: 'value)
			 	args: (compile/multi loop-inits 
						     lex-envt 
						     dyn-envt 
						     'value)
				mode: mode)))))


(define *compile-optimized-loops* #f)	

;; note that we don't generate <ic-loop> code when
;; compiling to closure-threaded code

(define (compile/named-let sf form lex-envt dyn-envt mode)
    (if (and *compile-optimized-loops*
    	     (not *compile-seeing-shadows*))
	(let ((ic (call-with-current-continuation
		    (lambda (exit)
			(let* ((loop-var (make <loop-var> 
					    name: (cadr form)
					    compile-with-full-semantics: 
						(lambda () 
						    (exit #f))))
				(loopv-envt (make-lexical-envt (list loop-var)
								lex-envt
								dyn-envt))
				(loop-args (caddr form))
				(arg-envt (make-lexical-envt 
						(make-lex-vars 
						    (map car loop-args) 
						    loopv-envt) 
						loopv-envt 
						loopv-envt))
				(ic (make <ic-loop>
					mode: mode
					envt: arg-envt
					loop-var: loop-var
					inits: (compile/multi 
						(map cadr loop-args) 
						loopv-envt
						loopv-envt
						'value)
					body: (compile/body 
						(cdddr form) 
						arg-envt 
						arg-envt 
						(untop mode)))))
			    (set-loop-icode! loop-var ic)
			    ic)))))
	    ; NOTE:  The loop code generation is buggy,
	    ;        so this is disabled.  Hence, all loops
	    ;        will be compiled with full semantics (-dmk 94.03.05)
	    (if ic
		;; it was successful as an optimized loop
		ic
		;; it failed as an optimized loop.. compile it
		;; as a loop with full semantics
		(compile/named-let* form lex-envt dyn-envt mode)))
	(compile/named-let* form lex-envt dyn-envt mode)))
	
(define (make-letrec new-envt inits body)

    (make <ic-bind>
	mode: mode
	envt: new-envt
	inits: (make <ic-multi> 
			arg-list: (map (lambda (x)
					(make-const '#uninit 'value))
					inits)
			mode: 'value)
	body: (make <ic-seq>
		stmt-list: 
		    (list (make <ic-seq> stmt-list:(map make-set
							(bindings new-envt)
							inits))
			    body)))

    (define (make-set var val)
	(make <ic-lex-set> var: var
			    rhs: val
			    mode: 'effect)))

(define (compile/letrec sf form lex-envt dyn-envt mode)
    (let ((new-envt (make-lexical-envt 
    				(make-lex-vars (map car (cadr form)) lex-envt) 
				lex-envt 
				dyn-envt)))
	(make-letrec new-envt
		     (map (lambda (bdg-info)
			    (compile (cadr bdg-info) new-envt new-envt 'value))
			  (cadr form))
		     (compile/body (cddr form) 
				    new-envt 
				    new-envt 
				    (untop mode)))))

(define (compile/let-bdg sf form lex-envt dyn-envt mode)
    (let ((new-envt (make-lexical-envt 
    				(make-lex-vars (map car (cadr form)) lex-envt) 
				lex-envt 
				dyn-envt)))
	(make <ic-bind>
	    mode: mode
	    envt: new-envt
	    inits: (compile/multi (map cadr (cadr form)) 
				    lex-envt
				    dyn-envt
				    'value)
	    body: (compile/body (cddr form) 
	    			new-envt 
				new-envt 
				(untop mode)))))

;; Dylan's version of "let" (but with the additional ability to
;; have a #rest and arbitrary function for the initializer)
;;
;; usage something like:
;;
;;     (bind ((a b c (values 1 2 3))
;;            (x y #rest z (list->values *a-bunch*)))
;;        ...)

(define (all-but-last l)
  (if (pair? l)
      (if (pair? (cdr l))
	  (cons (car l) (all-but-last (cdr l)))
	  '())
      l))
 
(define (compile/bind sf form lex-envt dyn-envt mode)

    (define (compile-binds binds body lex-envt dyn-envt)
	(if (pair? binds)
	    ;; "binds" is a list of binding forms
	    ;; each binding form is like:
	    ;;    (a b #rest c init-expr)
	    (let* ((bdg-form (all-but-last (car binds)))
		   (init-expr (last (car binds)))
		   (new-envt (make-envt bdg-form lex-envt dyn-envt)))
		(compile-one-bind init-expr
				  (compute-has-rest bdg-form)
				  lex-envt
				  dyn-envt
				  new-envt
				  (compile-binds (cdr binds)
				  		 body 
						 new-envt 
						 new-envt)))
	    (compile/body body lex-envt dyn-envt (untop mode))))
    
    (define (make-envt formals lex-envt dyn-envt)
	(make-lexical-envt (make-lex-vars formals lex-envt) 
			    lex-envt 
			    dyn-envt))
    
    (define (compile-one-bind init-expr rest? lex-envt dyn-envt new-envt body)
	(make <ic-bind>
	    mode: mode
	    envt: new-envt
	    inits: (compile init-expr lex-envt dyn-envt 'bind)
	    body: body
	    rest?: rest?))
    
    (compile-binds (cadr form) (cddr form) lex-envt dyn-envt))

(define (compile/let-syntax sf form lex-envt dyn-envt mode)
    (let ((new-lexical-envt
    		(make-lexical-envt
		    (map 
			(lambda (x) 
			    (compile-macro (car x) (cdr x) lex-envt))
			(cadr form))
		    lex-envt
		    #f)))
	(compile/body 
	    (cddr form) 
	    new-lexical-envt 
	    dyn-envt
	    (untop mode))))

(define (compile/letrec-syntax sf form lex-envt dyn-envt mode)
    (let ((new-lexical-envt
    		(make-lexrec-envt
		    (lambda (new-envt)
			(map 
			    (lambda (x) 
				(compile-macro (car x) (cdr x) new-envt))
			    (cadr form)))
		    lex-envt
		    #f)))
	(compile/body
	    (cddr form) 
	    new-lexical-envt 
	    dyn-envt
	    (untop mode))))


(define (compile/set! sf form lex-envt dyn-envt mode)
    (let ((bdg (find lex-envt (cadr form))))
	(compile-set bdg (caddr form) lex-envt dyn-envt mode)))
	
(define (compile/quote sf form lex-envt dyn-envt mode)
    (if (eq? (length form) 2)
	(make-const (cadr form) mode)
	(error/syntax "Badly constructed: ~a" form)))

(define (compile-FUNCTION sf form lex-envt dyn-envt mode)
    (make-const (current-place-name) mode))

(define (compile/mquote sf form lex-envt dyn-envt mode)

    (define (compile-mquoted mquoted-expr lex-envt)
	(if (symbol? mquoted-expr)
	    (let ((bdg (lookup lex-envt mquoted-expr)))
		(if (substitution? bdg)
		    (compile-mquoted
			(expr bdg)
			(envt bdg))
		    (make-const mquoted-expr mode)))
	    (make-const mquoted-expr mode)))
    
    (if (eq? (length form) 2)
	(compile-mquoted (cadr form) lex-envt)
	(error/syntax "Badly constructed: ~a" form)))

(define (compile/lambda sf form lex-envt dyn-envt mode)
    (if (and (pair? (cadr form))
    	     (eq? (caadr form) 'quote))
	(make <ic-lambda>
	    mode: mode
	    proc: (compile/procedure 
			    (cadadr form) 
			    (caddr form) 
			    (cdddr form) 
			    lex-envt 
			    dyn-envt))
	(make <ic-lambda>
	    mode: mode
	    proc: (compile/procedure 
			    #f 
			    (cadr form) 
			    (cddr form) 
			    lex-envt 
			    dyn-envt))))

(define (compile/if sf form lex-envt dyn-envt mode)
    (make <ic-if>
	mode: mode
	condition: (compile (cadr form) lex-envt dyn-envt 'value)
	if-true: (compile (caddr form) lex-envt dyn-envt mode)
	if-false: (if (null? (cdddr form))
		      (make-const '#undef mode)
		      (compile (cadddr form) lex-envt dyn-envt mode))))

(define (compile/tlv sf form lex-envt dyn-envt mode)
    (let ((ic (compile (cadr form) lex-envt dyn-envt 'value)))
	(if (instance? ic <ic-tl-ref>)
	    (make <ic-const> value: (var ic))
	    (error/semantic "~s: ~s does not compile to a tl-ref"
	    		    form
			    (cadr form)))))


(define (make-special-forms)
    (list
	(make <SpecialForm> name: 'begin compiler: compile/begin)
	(make <SpecialForm> name: 'quote compiler: compile/quote)
	(make <SpecialForm> name: 'mquote compiler: compile/mquote)

	(make <SpecialForm> name: 'if compiler: compile/if)
	(make <SpecialForm> name: 'let compiler: compile/let)
	(make <SpecialForm> name: 'letrec compiler: compile/letrec)
	(make <SpecialForm> name: 'bind compiler: compile/bind)
	(make <SpecialForm> name: 'let-syntax 
    				    compiler: compile/let-syntax)
	(make <SpecialForm> name: 'letrec-syntax 
    				    compiler: compile/letrec-syntax)
	(make <SpecialForm> name: 'set! compiler: compile/set!)
	(make <SpecialForm> name: 'values compiler: compile/values)
	(make <SpecialForm> name: '*FUNCTION* 
    				    compiler: compile-FUNCTION)
	(make <SpecialForm> name: 'lambda compiler: compile/lambda)
	(make <SpecialForm> name: '%top-level-var compiler: compile/tlv)))
