#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-syntax (unsafe-all? predicate list)
    (let loop ((x list))
	(if (null? x)
	    #t
	    (if (predicate (car x))
		(loop (cdr x))
		#f))))

(define (map1 proc list)
    (if (pair? list)
	(cons (proc (car list)) (map1 proc (cdr list)))
	(if (null? list)
	    '()
	    (abort 'map "Not a list at: ~s" list))))

(define (testing x)
  (let loop ((i x))
    (if (pair? i)
	(let ((x (cdr i)))
	  (loop x))
	#t)))

(define-syntax (unsafe-map1 proc list)
    (if (null? list)
	'()
	(let ((first (cons (proc (car list)) '())))
	  (let loop ((i (cdr list))
		     (prev first))
	    (if (null? i)
		first
		(let ((latest (cons (proc (car i)) '())))
		  (set-cdr! prev latest)
		  (loop (cdr i) latest)))))))

(define (unsafe-map-car list)
  (unsafe-map1 car list))

(define (unsafe-map-cdr list)
  (unsafe-map1 cdr list))

(define (unsafe-car x)
    (car x))

(define (unsafe-cdr x)
    (cdr x))

(define (mapn proc lists)
    (if (unsafe-all? pair? lists)
	(cons
	    (apply proc (unsafe-map-car lists))
	    (mapn proc (unsafe-map-cdr lists)))
	'()))

(define (map proc list1 . more-lists)
    (if (null? more-lists)
	(map1 proc list1)
	(mapn proc (cons list1 more-lists))))

(define (for-each1 proc list)
    (if (pair? list)
	(if (null? (cdr list))
	    (proc (car list))
	    (begin
		(proc (car list))
		(for-each1 proc (cdr list))))
	(values)))

(define (for-eachn proc lists)
    (if (null? (car lists))
	(values)
	(if (null? (cdr (car lists)))
	    (apply proc (map1 unsafe-car lists))
	    (begin
		(apply proc (map1 unsafe-car lists))
		(for-eachn proc (map1 unsafe-cdr lists))))))

(define (for-each proc list1 . more-lists)
    (if (null? more-lists)
	(for-each1 proc list1)
	(for-eachn proc (cons list1 more-lists))))

(define (vector-map proc vec)
    (let ((v (make-vector (vector-length vec))))
	(let loop ((i (vector-length vec)))
    	    (if (eq? i 0)
		v
		(begin
		    (vector-set! v (sub1 i) (proc (vector-ref vec (sub1 i))))
		    (loop (sub1 i)))))))

(define (vector-for-each proc vec)
    (if (eq? (vector-length vec) 0)
	(values)
	(let ((max (sub1 (vector-length vec))))
	    (let loop ((i 0))
		(if (eq? i max)
		    (proc (vector-ref vec i))
		    (begin
			(proc (vector-ref vec i))
			(loop (add1 i))))))))

(define (any? predicate list)
    (let loop ((x list))
	(if (pair? x)
	    (if (predicate (car x))
		#t
		(loop (cdr x)))
	    #f)))

(define (all? predicate list)
    (let loop ((x list))
	(if (pair? x)
	    (if (predicate (car x))
		(loop (cdr x))
		#f)
	    #t)))

(define-shadow map map)
(define-shadow for-each for-each)

(export for-each map vector-map vector-for-each any? all?)
