#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;(%use hacker)(load "/tmp/rdln.scm")

(define (string-prefix=? str prefix)
    (and (>= (string-length str) (string-length prefix))
	 (string=? (substring str 0 (string-length prefix))
	 	   prefix)))

(define (seperator? str)
    (string-prefix=? str "----------"))

(define (title-header? str)
    (string-prefix=? str "Title: "))

(define (for-each-section proc input)
    (let sections-loop ()
	(let ((sect (make-seq)))
	    (let lines-loop ()
		(let ((line (read-line input)))
		    (cond
			((eof-object? line)
			    (proc (seq->list sect)))
			((seperator? line)
			    (proc (seq->list sect))
			    (sections-loop))
			(else
			    (seq-add! sect line)
			    (lines-loop))))))))

(define (collect-bugs file)
    (let ((bugs (make-seq)))
	(call-with-input-file
	    file
	    (lambda (f)
		(for-each-section
		    (lambda (s)
			(if (and (pair? s)
				    (title-header? (car s)))
			    (seq-add! bugs s)))
		    f)))
	(seq->list bugs)))

(define (parse-bugs bugs)
    (map
	(lambda (bug)
	    (let ((kwds (make-seq)))
		(let loop ((i bug))
		    (cond
			((null? i)
			    (cons (seq->list kwds) '()))
			((eq? (string-length (car i)) 0)
			    (cons (seq->list kwds) (cdr i)))
			(else
			    (seq-add! kwds (parse-header-line (car i)))
			    (loop (cdr i)))))))
	bugs))
			    

(define (skip-spaces ch)
    (if (and (pair? ch) (char-whitespace? (car ch)))
	(skip-spaces (cdr ch))
	ch))
    
(define (parse-header-line str)
    (let* ((chars (string->list str))
	   (split (memq #\space chars))
	   (colon (memq #\: chars)))
	(if (eq? (cdr colon) split)
	    (begin
		(set-cdr! colon '())
		(cons (list->string chars) 
		      (list->string (skip-spaces (cdr split)))))
	    (abort 'parse-header-line "invalid: ~s" str))))

(define (summarize-bugs bugs)
    (for-each
	(lambda (status)
	    (format #t "--- Status: ~a ---\n" status)
	    (for-each
		(lambda (bug)
		    (let ((info (car bug)))
			(format #t "~a: ~a\n"
				(cdr (assoc "Id:" info))
				(cdr (assoc "Title:" info)))))
		(select (lambda (b)
			    (string=? (cdr (assoc "Status:" (car b)))
				      status))
			bugs))
	    (newline))
	*valid-status*)
    (let ((any-invalid? #f))
	(for-each
	    (lambda (bug)
		(if (not (member (cdr (assoc "Status:" (car bug))) 
				 *valid-status*))
		    (begin
			(if (not any-invalid?)
			    (format #t "--- Invalid Status ---\n"))
			(format #t "~a: ~a (~a)\n"
				(cdr (assoc "Id:" (car bug)))
				(cdr (assoc "Title:" (car bug)))
				(cdr (assoc "Status:" (car bug))))
			(set! any-invalid? #t))))
	    bugs)))
	    
(define *valid-status* 
    '("Scheduled"
      "Deferred"
      "Open"
      "Closed"))

(define *bugs* (parse-bugs (collect-bugs "bugs.txt")))

(with-output-to-file
    "bugs.sum"
    (lambda ()
	(summarize-bugs *bugs*)))

;(%exit 0)

#|
mail -s "EasyScheme Bug Report (summary)" wilson@cs < bugs.sum
mail -s "EasyScheme Bug Report Appendix" wilson@cs < bugsappx.txt
mail -s "EasyScheme Bug Report" wilson@cs < bugs.txt
|#
