#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;============================= module.scm =============================

(define-class <Module> (<Object>)
    (name 'user)
    (files '())
    top-level-envt
    link-name
    source-dir
    dest-dir
    (num-roots 0)
			;; <JoinFile>'s...
    root-definitions
    link-table
    public-interface
    private-interface
    init-procs
    class-table
    makefile-parts
    exports-table
    imports-table
    module-descr)	;; a <PartDescr> that will refer to our module_descr

(define-generic-fn add-file!)
(define-generic-fn update)

(define *modules-table* '())	; a global assoc list (name -> module)
(define *defined-by* 
    (make-table
	eq?
	(lambda (bdg) 	;; keys are Bindings
	    (symbol->hash (name bdg)))))
;(define *compile-seeing-shadows* #t)	; most of the time, use the shadows

;; *shadows* is an association list (assq) mapping
;; arbitrary objects to underlying-system closures.
;; shadows are used in the following way:  Just before
;; closure-threaded code does an apply, it checks for
;; a shadow (ie, it sees if there is a mapping from
;; the value that is to be the "function" in the apply)
;; If there is one, the RHS of the mapping is used instead.
;; In this way, you can execute code in the underlying system
;; and closure-threaded code in the right way.
;;
;; <Macro> bindings are handled slightly differently -- they
;; store their shadow directly with them, and at compile time
;; use the shadow in a COMBO rather than expanding as a macro.

(define *shadows* '())

(define *private-interface* #f)
(define *current-module* #f)
(define *parts* #f)
(define *public-interface* #f)

;============================= sourcefile.scm =============================

(define-class <SourceFile> (<Object>) 
    pathname
    module
    (num-parts 0)
    (timestamp #f)
    (pervasive-impact? #f)
    (templates '())
    (part-tag #f))	;; if set, fixes the tag of the first part

(define *templates* '())

(define-generic-fn init)
(define-generic-fn dirty?)
(define-generic-fn set-clean!)
(define-generic-fn modify-time)

(define-class <Part> (<Object>)
    source-file
    dest-pathname
    link-name		;; the name that we use as a hook to link to it
    tag
    part-descr		;; the part descriptor used for linking
    (link-table '())
    (total-size 0))

(define *current-part* #f)
(define *current-part-port* (current-output-port))
(define *max-part-size* 300)
(define *root-defns* '())
(define *classes* '())
(define *generic-functions* '())
(define *source-load-commands* '())
(define *make-commands* '())
(define *imports* '())		; the names of imported modules
(define *exports* '())		; the exportations

(define-rewriter (push! form)
    (let ((var (cadr form))
    	  (expr (caddr form)))
	`(set! ,var (cons ,expr ,var))))
	

;============================= joinfile.scm =============================

(define-class <JoinFile> (<Object>)
    module
    (dirty? #t)
    (contributions '())
    (equal-proc equal?)
    make-proc
    filename)

;============================= preload.scm =============================

(define *current-source-file* #f)
(define *pre-load* #t)



;============================= definers.scm =============================

(define *use-soft-links* #f)
