#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-class <t-expr> (<Object>)
    type
    expr)

(define-class <t-const> (<t-expr>)
    value)

(define (t-const? e)
    (eq? (class-of e) <t-const>))
    
(define (make-t-expr t e)
    (make <t-expr> type: t expr: e))

(define (make-t-const t e v)
    (make <t-const> type: t expr: e value: v))

;; types:

(define (type->hash t)
    (symbol->hash (class-name t)))	; types are classes
    
;; ALWAYS-ISA? returns #t if an instance of type1 can always be
;; converted into an (equivalent) instance of type2.
;; (eg, an <Integer*> is always a <Float*>,
;;      but a <Float*> is NOT always an <Integer*>)

(define (always-isa? type1 type2)
)

;; NEVER-ISA? returns #t if an instance of type1 can never be
;; converted into an (equivalent) instance of type2.
;; (eg, an <Object*> is sometimes a <Integer*>,
;;      but a <Float*> is never an <Integer*>)

(define (never-isa? type1 type2)
)

;; DEVISE-TYPE-TEST returns an (icode) expression that tests
;; the given expression to make sure it is of the given type

(define (devise-type-test expr type)
    (error/internal "devise-type-test not implemented"))
    
(define yes-const (make-t-const <raw-bool> "YES" #t))
(define no-const (make-t-const <raw-bool> "NO" #f))


(define (coercion->hash c)
    (* (type->hash (car c)) 
       (type->hash (cdr c))))

(define coercion-table 
    (make-table 
	(lambda (a b) 
	    (and (eq? (car a) (car b)) 
		 (eq? (cdr a) (cdr b))))
	coercion->hash))

(define (obj-expr t-expr)
    (expr (coerce t-expr <Object*>)))

(define (raw-bool-expr t-expr)
    (expr (coerce t-expr <raw-bool>)))
    
(define (coerce t-expr to-type)
    (if (eq? to-type (type t-expr))
	t-expr
	(let ((fn (table-lookup coercion-table 
					(cons (type t-expr) to-type))))
	    (if fn
		(fn t-expr to-type)
		;; special case the (trivial) default coercion of
		;; an expression of any normal class to a superclass
		(if (and (not (subclass? to-type <raw-type>))
		         (subclass? (type t-expr) to-type))
		    t-expr
		    (fluid-let ((t-expr t-expr)
		    	        (to-type to-type))
			(error/internal "Impossible coercion needed: ~a -> ~a"
			    (type t-expr) 
			    to-type)))))))



(table-insert! coercion-table (cons <raw-bool> <Object*>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <Object*> 
	    		  (if (value t-expr) "TRUE_OBJ" "FALSE_OBJ")
			  (value t-expr))
	    (make-t-expr <Object*> (list "rb_to_bo" (expr t-expr))))))
    
(table-insert! coercion-table (cons <Object*> <raw-bool>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <raw-bool> 
	    		  (if (value t-expr) "YES" "NO")
			  (value t-expr))
	    (make-t-expr <raw-bool> (list "truish" (expr t-expr))))))

(table-insert! coercion-table (cons <raw-int> <Object*>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <Object*> 
	    		  (list "int2fx" (expr t-expr))
			  (value t-expr))
	    (make-t-expr <Object*> (list "int2fx" (expr t-expr))))))

(table-insert! coercion-table (cons <raw-int> <FixNum>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <FixNum> 
	    		  (list "int2fx" (expr t-expr))
			  (value t-expr))
	    (make-t-expr <FixNum> (list "int2fx" (expr t-expr))))))

(table-insert! coercion-table (cons <Object*> <raw-int>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <raw-int> 
	    		  (number->string (value t-expr))
			  (value t-expr))
	    (make-t-expr <raw-int> (list "fx2int" (expr t-expr))))))

(table-insert! coercion-table (cons <Object*> <raw-int-bytes>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <raw-int-bytes> 
	    		  (list "SLOT" (number->string (value t-expr)))
			  (value t-expr))
	    (make-t-expr <raw-int-bytes> 
	    	 	 (list "FXWORDS_TO_RIBYTES" (expr t-expr))))))

(table-insert! coercion-table (cons <raw-int-bytes> <Object*>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <Object*> 
	    		  (number->string (value t-expr))
			  (value t-expr))
	    (make-t-expr <Object*> 
	    	 	 (list "RIBYTES_TO_FXWORDS" (expr t-expr))))))

(table-insert! coercion-table (cons <raw-int> <raw-int-bytes>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <raw-int-bytes> 
	    		  (list "SLOT" (number->string (value t-expr)))
			  (value t-expr))
	    (make-t-expr <raw-int-bytes> 
	    	 	 (list "4*" (expr t-expr))))))

(table-insert! coercion-table (cons <raw-float> <Object*>)
    (lambda (t-expr to-type)
	(if (t-const? t-expr)
	    (make-t-const <Object*>
	    		  (list 'literal 
				(alloc-literal (value t-expr))
				(number->string (value t-expr)))
	    		  (value t-expr))
	    (make-t-expr <Object*> (list "make_float" (expr t-expr))))))

(define (trivial-coercion t-expr to-type)
    (make-t-expr to-type (expr t-expr)))

(table-insert! coercion-table (cons <Closure*> <Object*>)
    trivial-coercion)

(table-insert! coercion-table (cons <raw-str> <Object*>)
    (lambda (t-expr to-type)
	(make-t-expr <Object*> (list "make_string" (expr t-expr)))))

(table-insert! coercion-table (cons <Object*> <raw-str>)
    (lambda (t-expr to-type)
	(make-t-expr <raw-str> (list "string_text" (expr t-expr)))))

#|
(define (install-types! envt)
    (let-syntax ((bind-type! 
    			(syntax-form (name)
			    (bind! envt
			           (make <TopLevelVar>
				         name: (mquote name)
					 value: name
					 write-prot: #t))))
		 (bind-type-as! 
    			(syntax-form (name type)
			    (bind! envt
			           (make <TopLevelVar>
				         name: (mquote name)
					 value: type
					 write-prot: #t)))))
	(bind-type-as! <obj> <Object>)
	(bind-type! <raw-bool>)
	(bind-type! <raw-float>)
	(bind-type! <raw-str>)
	(bind-type! <raw-int>)
	(bind-type! <raw-int-bytes>)))
|#
