#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (make-runtime-bind*2 N mode)

    (define (make-shift-regs)
	(let loop ((num (if (caddr mode)
			    (add1 (cadr mode))
			    (cadr mode)))
		    (src N)
		    (dst (cadddr mode))
		    (result '()))
	    (if (eq? num 0)
		result
		(loop (sub1 num)
			(add1 src)
			(add1 dst)
			(cons `(set! (reg ,dst #f) (ref (reg ,src #f)))
			    result)))))

    (define (make-heap-bind)
	(let loop ((num (if (caddr mode)
			    (add1 (cadr mode))
			    (cadr mode)))
		    (src N)
		    (result '()))
	    (if (eq? num 0)
		(list (cons 'bind (reverse result)))
		(loop (sub1 num)
			(add1 src)
			(cons `(ref (reg ,src #f))
			    result)))))

    (if (eq? (car mode) 'reg-bind)
	    (make-shift-regs)
	    (make-heap-bind)))

;; code-gen-bind*0 is used for loops that bind no vars
; -- untested as of 94.05.07

(define (code-gen-bind*0 self ct-envt in-regs mode)
    (let ((body (code-gen (body self) ct-envt in-regs mode)))
	(if (eq? mode 'value)
	    (let ((save-reg `(reg ,(length in-regs) #f)))
		(m (append (s body)
			   `((set! ,save-reg ,(obj-expr (e body)))))
		   (make-t-expr <Object*> `(ref ,save-reg))))
	    body)))

(define (code-gen-bind*2 self ct-envt in-regs mode)
    (let* ((reg-alloc? (not (or (does-apply? (body self)) 
    				(does-lambda? (body self)))))
    	   (dbl-bind? (and reg-alloc? 
	   		   (pair? mode) 
			   (eq? (car mode) 'reg-bind)))
	   (bindings (bindings (envt self)))
	   (body (code-gen (body self)
			    (if reg-alloc?
			    	ct-envt
				(cons bindings ct-envt))
			    (if reg-alloc?
			    	(append (reverse bindings) in-regs)
				in-regs)
			    mode))
	   (bind-mode   (if reg-alloc?
			    (list 'reg-bind 
				(num-args self) 
				(rest? self) 
				(if dbl-bind?
				    (+ (cadddr mode) (length in-regs))
				    (length in-regs)))
			    (list 'bind (num-args self) (rest? self))))
	   (binder  (make-runtime-bind*2 (length in-regs) bind-mode)))
	(if (eq? mode 'value)
	    (let ((save-reg `(reg ,(length in-regs) #f)))
		(m (append binder 
			   (s body)
			   `((set! ,save-reg ,(obj-expr (e body))))
			   (if reg-alloc? '() '((unbind))))
		   (make-t-expr <Object*> `(ref ,save-reg))))
	    (if (or (eq? mode 'tail) reg-alloc?)
		(append binder body)
		(append binder body '((unbind)))))))

(define (code-gen-loop self ct-envt in-regs mode)
    (let ((new-label (alloc-label))
    	  (n (num-args (inits self))))
	(set-first-arg-reg! self (length in-regs))
	(set-ct-envt! self ct-envt)
	(set-label! (loop-var self) new-label)
	(if (> n 0)
	    (let ((args-code (code-gen (inits self) 
					ct-envt
					in-regs
					`(reg-bind ,n #f ,(length in-regs)))))
		(set-inits! self #f)
		(append args-code
			`((jump ,(+ n (length in-regs)) ,new-label)
			    (label ,new-label))
			(code-gen-bind*2 self ct-envt in-regs mode)))
	    (begin
		(set-inits! self #f)
		(append `((jump ,n ,new-label)
			  (label ,new-label))
			(code-gen-bind*0 self ct-envt in-regs mode))))))

(define (code-gen-jump self cte in-regs mode)
    (let* ((n (num-args (args self)))
	   (args-code (code-gen (args self) 
			        cte
			        in-regs
			        `(reg-bind ,n #f ,(length in-regs)))))
	(append args-code
		(pop-envts cte
			   (ct-envt (loop-icode (loop-var self))))
		(shift-regs n 
			    (length in-regs) 
			    (first-arg-reg (loop-icode (loop-var self))))
		`((bjump ,n ,(label (loop-var self)))))))

(define-method return-types ((self <ic-jump>))
    '())
    