#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-method add-file! ((m <Module>) f)
    (set-files! m (append (files m) (list f))))

(define-method write ((thing <Module>) port)
    (format port "#[Module ~s]" (name thing)))

(define (dest-aux-file module suffix ext)
    (make <file-name> directory: (dest-dir module)
		      filename: (if suffix
		      		    (string-append 
					    (symbol->string (name module))
					    "_"
					    suffix)
				    (symbol->string (name module)))
		      extension: ext))


(define-method initialize ((self <Module>))

    (set-link-name! self (symbol->string (name self)))
    (set-module-descr! self (make <PartDescr> 
    				module-name: (symbol->string (name self))
				part-tag: 0))
    (set-root-definitions! self
		(make <JoinFile> make-proc: make-roots
				 filename: (dest-aux-file self "r" "h")
				 module: self))
    (set-makefile-parts! self
		(make <JoinFile> make-proc: make-makefile
				 filename: (make <file-name>
						    directory: (dest-dir self)
						    filename: "Makefile")
				 module: self))
    (set-link-table! self
		(make <JoinFile> make-proc: make-link-file
				 filename: (dest-aux-file self "l" "c")
				 module: self))
    (let ((f (file-within-dir (filename (link-table self)))))
	(set-contribution! (makefile-parts self)
			    self
			    (list (cons 'c-file f))))

    (set-exports-table! self
    		(make <JoinFile> module: self))
    (set-imports-table! self
    		(make <JoinFile> module: self))

    (set-public-interface! self
		(make <JoinFile> make-proc: make-public-interface
				 filename: (dest-aux-file self #f "h")
				 module: self))
    (set-private-interface! self
		(make <JoinFile> make-proc: make-private-interface
				 filename: (dest-aux-file self "p" "h")
				 module: self))
    (set-init-procs! self
		(make <JoinFile> module: self
		                 equal-proc: (lambda (a b) #f)))
    (set-class-table! self
		(make <JoinFile> module: self
				 equal-proc: (lambda (a b) #f)))

    ;(install-types! (top-level-envt self))
    (multi-bind! (top-level-envt self) (make-special-forms))
    (install-ezc-forms! (top-level-envt self))
    
    ; install us into the global module table
    
    (let ((i (assq (name self) *modules-table*)))
	(if i
	    (set-cdr! i self)
	    (push! *modules-table* (cons (name self) self))))
    self)




(define-method update ((self <Module>))
    (set! *current-module* self)
    (let ((out-of-date (remake-files (files self))))
	(for-each
	    (lambda (x)
		(ct-format #t "Need to make: ~a\n" x))
	    out-of-date)
	(for-each
	    update
	    out-of-date)
	(make-if-needed (root-definitions self))
	(make-if-needed (link-table self))
	(make-if-needed (public-interface self))
	(make-if-needed (private-interface self))
	(make-if-needed (makefile-parts self))
	(warn-duplicate-exports self)
	(warn-unbound-vars self)))

(define (with-output-to-header pathname thunk)
    (with-output-to-file (pathname->string pathname)
	(lambda ()
	    (let ((guard (string-upcase (filename pathname))))
		(display-disclaimer)
		(ct-format #t "\n#ifndef _H_~a\n" guard)
		(ct-format #t "#define _H_~a\n\n" guard)
		(thunk)
		(ct-format #t "#endif /* _H_~a */\n" guard)))))
    
(define (string-upcase str)
    (list->string (map char-upcase (string->list str))))

;; MAKE-ROOTS is the make proc for the `module_r.h' file

(define (make-roots jf)
    (ct-format #t "Making roots (~a)\n" (filename jf))
    (with-output-to-header (filename jf)
	(lambda ()
	    (ct-format #t 
		"/* Roots for the ~a module */\n\n"
		(name (module jf)))
	    (ct-format #t "extern obj ~a_root_vector[];\n" 
	    		  (link-name (module jf)))
	    
	    ; loop over each source file's request for roots
	    (for-each
		(lambda (c)
		    
		    (ct-format #t "/* roots registered in ~a */\n" 
			          (file-within-dir (pathname (car c))))
		    
		    ; loop over the roots requested by this source file
		    (for-each
			(lambda (r)
			    (ct-format #t 
				"#define ~a ~a_root_vector[~d]  /* ~a */\n" 
				(caddr r)
				(link-name (module jf))
				(cadr r)
				(car r)))
			(cdr c))
		    (newline))
		(contributions jf)))))

;; load in the makefile postamble.  It should be in this directory...

(define *makefile-postamble* 
    (file->string 
	(pathname->string 
	    (append-path 
		(fluid-ref *current-dir*) 
		(string->file "makefile.pos")))))

;; MAKE-MAKEFILE is the make proc for the `Makefile' file

(define (make-makefile jf)
    (ct-format #t "Making Makefile\n")

    (with-output-to-file
	(pathname->string
	    (make <file-name> filename: "depends"
	    		      directory: (directory (filename jf))))
	newline)

    (with-output-to-file (pathname->string (filename jf))
	(lambda ()
	    (let ((m (module jf))
	    	  (get-parts 
			(let ((u (unified-contributions jf)))
			    (lambda (key)
				(let loop ((r '()) (i u))
				    (if (null? i)
					(reverse r)
					(if (eq? (caar i) key)
					    (loop (cons (cdar i) r)
					    	  (cdr i))
					    (loop r (cdr i)))))))))
		
		(ct-format #t "CFILES=")
		(for-each
		    (lambda (x)
			(ct-format #t " ~a" x))
		    (get-parts 'c-file))

		(newline)
		(ct-format #t "PRODUCT=~a_m.o\n" (link-name m))
		(display *makefile-postamble*)))))

;; MAKE-LINK-FILE is the make proc for the `module.c' file

(define (make-link-file jf)
    (ct-format #t "Making link file (~a)\n" (filename jf))
    (with-output-to-file (pathname->string (filename jf))
	(lambda ()
	    (let ((m (module jf))
	    	  (unified (unified-contributions jf)))
		(ct-format #t "#include \"~a\"\n" 
			    (file-within-dir (filename (private-interface m))))
		(ct-format #t
		    "\n/* Link file for the ~a module */\n\n"
		    (name m))
		(ct-format #t "#define NUM_ROOTS (~d)\n"
			   (if (eq? (num-roots m) 0)
			       1
			       (num-roots m)))
		(ct-format #t "obj ~a_root_vector[NUM_ROOTS];\n\n"
			    (link-name m))

		(ct-format #t "#ifdef PROVIDE_ROOT_NAMES\n")
		(ct-format #t "static struct root_info roots[] = {\n")
		(for-each
		    (lambda (root-def)
			(ct-format #t "    { ~d, \"~a\" },\n"
				   (cadr root-def)
				   (car root-def)))
		    (unified-contributions (root-definitions m)))
		(ct-format #t "    { 0, NULL } };\n")
		(ct-format #t "#else /* PROVIDE_ROOT_NAMES */\n")
		(ct-format #t "#define roots ((struct root_info *)0)\n")
		(ct-format #t "#endif /* PROVIDE_ROOT_NAMES */\n")
		
		; loop over each part (flattening out where they're from)
		(for-each
		    (lambda (c)
			(ct-format #t 
			    "extern struct part_descr ~a_part_~a;\n" 
			    (link-name m)
			    c))
		    unified)
		(newline)
		(display "static struct part_descr *(tab[]) = {\n")
		(for-each
		    (lambda (x)
			(ct-format #t "    &~a_part_~a,\n" (link-name m) x))
		    unified)
		(display "    NULL };\n")
		(ct-format #t 
			   "struct module_descr module_~a = { \"~a\", tab,\n"
			   (link-name m)
			   (name m))
		(ct-format #t "\tNUM_ROOTS, /* num_roots */\n")
		(ct-format #t "\t~a_root_vector,\n" (link-name m))
		(ct-format #t "\troots};\n\n")))))

;; MAKE-PRIVATE-INTERFACE is the make proc for the `module_p.h' file

(define (make-private-interface jf)
    (ct-format #t "Making private interface (~a)\n" (filename jf))
    (with-output-to-header (filename jf)
	(lambda ()
	    (let ((m (module jf)))
		(ct-format #t "#include \"~a\"\n" 
			    (file-within-dir 
				(filename (public-interface m))))
		(ct-format #t "#include \"~a\"\n" 
			    (file-within-dir 
				(filename (root-definitions m))))
		(display "#include \"vinsns.h\"\n")
		(for-each
		    display
		    (unified-contributions jf)) ))))

;; MAKE-PUBLIC-INTERFACE is the make proc for the `module.h' file

(define (make-public-interface jf)
    (ct-format #t "Making public interface (~a)\n" (filename jf))
    (with-output-to-header (filename jf)
	(lambda ()
	    (let ((m (module jf)))
		(display "#include \"obj.h\"\n")
		(display "#include \"linktype.h\"\n")

		(ct-format #t "extern struct module_descr module_~a;\n"
	    			(link-name m))))))


; REVERSE-TOPOLOGICAL-SORT sorts the list of classes so that no
; class appears in the list before its superclass appears in the list
; works by flipping around the superclass links and then doing a
; depth-first search from each "local eff. root", which is a class
; with no superclass in the list.

; Note that the situation where A <-- B <-- C <-- D
; (A superclass of B, B superclass of C, etc.)
; where B does not appear in the list but A and C do should
; be impossible, since that would imply that B is defined
; elsewhere, which it could not have been since THIS constitutes
; the definition of A.

(define (reverse-topological-sort classes)

    (define (zilch classes)
	(map (lambda (x) (cons x '())) classes))
    
    (define (depth-first ci)
	(cons (car ci) 
	      (apply append (map depth-first (cdr ci)))))

    (let ((flip (zilch classes)))
	(for-each
	    (lambda (ci)
		(let ((c (car ci)))
		    (for-each
			(lambda (super)
			    (let ((a (assq super flip)))
				(if a
				    (set-cdr! a (cons ci (cdr a))))))
			(%gvec-read c 3))))
	    flip)
	(apply append
	       (map
		 (lambda (ci)
		    (if (not (any? (lambda (super)
					(memq super classes))
				   (%gvec-read (car ci) 3)))
			(begin
			    (ct-format #t "Local effective root ~a\n" (car ci))
			    (depth-first ci))
			'()))
		 flip))))
	    

;; Takes a <TopLevelEnvt> and mutates it so
;; that instead of a hash table, it has a vector
;; representing the contents of the table.
;; calls select on each value to determine whether
;; or not to include the entry (for example, to
;; filter out stuff)
;; Returns the original hash table, so it can be put back later

(define (flatten-tle! tle proc)
  (let ((table (table tle))
	(l '()))
    (table-for-each table
       (lambda (h k v)
	 (if (proc v)
	     (set! l (cons* h k v l)))))
    (set-table! tle (list->vector l))
    table))

;; check an exports listing for duplicates

(define (warn-duplicate-exports module)
    (let ((t (make-table eq? symbol->hash)))
	(let loop ((sources (contributions (exports-table module))))
	    (if (pair? sources)
		(let ((source-file (caar sources)))
		    (for-each
			(lambda (x)
			    (let ((also (table-insert! t 
						       (cadr x) 
						       source-file)))
			        (if also
				    (warning "~a is in ~a and ~a"
				    	(cadr x)
					source-file
					also))))
			(cdar sources))
		    (loop (cdr sources)))))))
			    
;; an export looks like:  (bdg remote-name shared? read-only?)

(define (export->import export)
    (let ((old-bdg (car export))
    	  (new-name (cadr export))
	  (shared? (caddr export))
	  (write-prot? (cadddr export)))
	(if shared?
	    old-bdg	; bind the *same* TLV
	    (let ((x (clone (car export))))
		(%gvec-write x 0 new-name)	;; aka set-name!
		(if write-prot?
		    (set-write-prot! x write-prot?))
		x))))


(define (warn-unbound-vars module)
    (define (tlv? x)
	(eq? (class-of x) <TopLevelVar>))
	
    (for-each 
	(lambda (x) 
	    (if (and (tlv? x) (eq? (value x) '#unbound))
		(ct-format #t "warning: ~s is unbound\n" (name x)))) 
	(table-values->list (table (top-level-envt module)))))

