#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define (c-compile/load sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "load not at top-level: ~a" form))
    (if *pre-load*
	(push! *load-next* 
	    (list (directory (pathname *current-source-file*))
		  (cadr form)
		  (module *current-source-file*))))
    #f)	;; return #f ==> cause eval to ignore this form

(define *load-next* '())

(define (pre-load in-dir file module)
    (let ((p&f (locate in-dir '() '("scm") file)))
	(if p&f
	    (let ((port (car p&f))
		  (old *current-source-file*)
		  (file-obj (make <SourceFile> pathname: (cdr p&f) 
					       module: module)))
		(close-input-port port)
		(add-file! module file-obj)
		(set! *load-next* '())
		(update file-obj)
		(for-each
		    (lambda (load-cmd)
			(apply pre-load load-cmd))
		    (reverse *load-next*)))
   	   (abort 'pre-load "File not found: ~a from dir ~a" file in-dir))))

(define (build-module name source-dir dest-dir)
    (let ((m (make <Module> name: name
			    source-dir: source-dir
			    dest-dir: dest-dir
			    top-level-envt: (make-default-envt))))

	(set! *current-source-file* #f)
	(set! *pre-load* #t)
	(set! *current-module* m)
	(pre-load source-dir (symbol->string name) m)
	(set! *pre-load* #f)
	m))

(define (make-boot-initial)
    (let ((envt (make-top-level-envt)))
    	; these are from the underlying REPL
	(multi-bind! envt (make-special-forms))
	(multi-bind! envt (make-definers))
	envt))
	
(define (build-boot-module name source-dir dest-dir)
    (let ((m (make <Module> name: name
			    source-dir: source-dir
			    dest-dir: dest-dir
			    top-level-envt: (make-boot-initial))))
	(set! *current-source-file* #f)
	(set! *pre-load* #t)
	(set! *current-module* m)
	(pre-load source-dir (symbol->string name) m)
	(set! *pre-load* #f)
	m))

(define (testx) 
    (load-module "../testx"))

(define (load-bi)
    (load-module "../builtins"))

(define (mkdir dir)
    (let* ((p (pathname->string dir))
    	   (p0 (substring p 0 (sub1 (string-length p)))))
	(if (not (file-exists? p))
	    (let ((rc (system (string-append "mkdir " p0))))
		(if (not (eq? rc 0))
		    (abort 'mkdir "Error in `mkdir ~a'" p0))))))

(define (load-module path)
    (let* ((s (string->dir path))
    	   (d (append-dirs s
			  (string->dir 
				(string-append 
				    "../"
				    (last-directory s)
				    "_m")))))
	(mkdir d)
	(build-module (string->symbol (last-directory s))
		      s
		      d)))

(define (load-module-in path dest-dir-path)
    (let ((d (string->dir dest-dir-path)))
	(mkdir d)
	(build-module (string->symbol (last-directory (string->dir path)))
			(string->dir path)
			d)))

(define (load-boot-module-in path dest-dir-path)
    (let ((d (string->dir dest-dir-path)))
	(mkdir d)
	(build-boot-module (string->symbol (last-directory (string->dir path)))
			(string->dir path)
			d)))
