#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (form-triv-args n)
  (let loop ((i n) (r '()))
    (if (eq? i 0)
        r
        (loop (sub1 i) 
              (cons (string->symbol (string-append "arg-"
                                                   (number->string i)))
                    r)))))

;; takes as input a sequence of primop declarators of the form:
;;	(<args> <body>...)
;;  where <args> is (<arg>...), and <arg> is:
;;	<arg> ::= <formal>
;;		 | (<formal> <type>)
;;  if the first form has any type checks, then there is also a 
;;  ':trust-me' form added (at the end) which suppresses those
;;  type checks.

(define (reform-as-syntax-forms forms)
  (let ((result (map reform-1-as-syntax forms)))
    (if (caar result)
	(append (map cdr result)
		(list (reform-1-as-trustme (car forms))))
	(map cdr result))))

(define (reform-1-as-trustme form)
    `(syntax-form (':trust-me ,@(map strip-type (car form)))
    		  ,@(cdr form)))

(define (reform-1-as-syntax form)
    (let* ((any-type-checks? #f)
	   (result (cons 'syntax-form 
			(cons
			    (map (lambda (a)
				    (if (pair? a)
					(begin
					    (set! any-type-checks? #t)
					    (list (car a) 'isa: (cadr a)))
					a))
				  (car form))
			    (cdr form)))))
	(cons any-type-checks?
		result)))

(define (strip-type x)
    (if (pair? x)
        (car x)
        x))

(define (strip-assert x)
    (if (and (pair? x) (pair? (cdr x)) (pair? (cddr x)))
	(list (car x) (cadr x))
	x))
	
;; note: this use of primops relies on the compiler
;;       generating safe code for:
;;		(define (car (x <pair>))
;;	            (_unsafe_car x))
;;	which means I need Marc Vossman's static type checkers

(define (define-primop-rewriter form)
    (if (pair? (cadr form))
        `(define-primop ,(caadr form) (,(cdadr form) ,@(cddr form)))
        (let* ((name (cadr form))
               (else-args (car (last (cddr form))))
               (else-name (string->symbol 
                           (string-append "_"
                                          (symbol->string name)))))
          `(begin
             (define-syntax ,name
               ,@(reform-as-syntax-forms (cddr form))
               (else ,else-name))
             (define (,else-name ,@(map strip-assert else-args))
               ,@(map primop-type-check else-args)
	       (,name ,@(map strip-type else-args)))))))
#|
(define test-1 '(define-primop (char? thing)
                  (%call ("ASCII_CHAR_P" (<obj>) => <raw-bool>) thing)))

(define test-2 '(define-primop (car (x <pair>))
                  (%call ("pair_car" (<obj>) => <obj>) x)))

|#
(define (primop-type-check arg)
    (if (pair? arg)
	(if (and (pair? (cdr arg))
		 (pair? (cddr arg)))
	    ;; an extra assertion
	    `(begin
		(type-check ,(car arg)
			    ,(class-name->predicate (cadr arg))
			    ,(cadr arg))
		(assert ,(caddr arg)))
	    ;; a plain type check
	    `(type-check ,(car arg)
	    		 ,(class-name->predicate (cadr arg))
			 ,(cadr arg)))
	;; no checks
	'(values)))

;; a hack!!!
;; leveraging off the convention that for class <foo>, the predicate is foo?

(define (class-name->predicate name)
    (let ((str (symbol->string name)))
	(string->symbol
	    (string-append
		(substring str
			   1
			   (- (string-length str) 1))
		"?"))))
    
;; other rewriters

(define (scheme-error-rewriter form)
  (cons 'abort (cons ''scheme (cdr form))))
#|
    (let ((msg (car (cdr form)))
          (args (cdr (cdr form))))
	(if (string? msg)
	    (list 'begin
		    (cons '%call
		          (cons
				(list "scheme_error"
				    (cons '<raw-str>
					    (cons '<raw-int>
						(map (lambda (x) 
							    '<obj>) 
							args))))
			        (cons msg 
				      (cons (length args) 
					    args))))
		    '(values))
	    (syntax-error "scheme-error: format arg not a string")))
|#

(define (bind-tlvs-rewriter form)
  (let ((envt-var (cadr form))
	(bdgs (cddr form)))
    `(tle-multi-bind! ,envt-var
		      ,@(apply append 
			       (map (lambda (breq)
				      (if (pair? breq)
					  (list (list 'quote (cadr breq))
						(car breq))
					  (list (list 'quote breq)
						breq)))
				    bdgs)))))
