#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-method modify-time ((file <SourceFile>))
    (vector-ref (stat (pathname->string (pathname file))) 2))

(define-method dirty? ((file <SourceFile>))
    (or (not (timestamp file))
	(time>? (modify-time file) (timestamp file))))

(define-method set-clean! ((file <SourceFile>))
    (set-timestamp! file (modify-time file)))

(define (remake-files file-list)
    (if (pair? file-list)
	(let ((f (car file-list)))
	    (if (dirty? f)
		(if (pervasive-impact? f)
		    file-list
		    (cons f (remake-files (cdr file-list))))
		(remake-files (cdr file-list))))
	'()))

(define-method write ((thing <SourceFile>) port)
    (format port "#[<SourceFile> ~a]" (pathname thing)))

(define-method display ((thing <SourceFile>) port)
    (display (pathname thing) port))

#|
(define f1 (make <SourceFile> pathname: "test1.scm"))
(define f2 (make <SourceFile> pathname: "test2.scm"))
(define f3 (make <SourceFile> pathname: "test3.scm"))

(set-pervasive-impact?! f2 #t)

(define files (list f1 f2 f3))
|#

(define-method update ((self <SourceFile>))
    (if *pre-load*
	(ct-format #t "Pre-loading: ~a\n" self)
	(ct-format #t "Updating: ~a\n" self))
    (do ((i 0 (add1 i)))
    	((>= i (num-parts self)))
      (ct-format #t "    `Deleting' part ~d\n" i))
    (set-num-parts! self 0)
    (let ((old *current-source-file*))
	(set! *current-source-file* self)
	(set! *root-defns* '())
	(set! *private-interface* '())
	(set! *public-interface* '())
	(set! *parts* '())
	(set! *classes* '())
	(set! *generic-functions* '())
	(set! *make-commands* '())
	(set! *imports* '())
	(set! *exports* '())
	(set! *templates* '())
	(set! *source-load-commands* '())
	(set! *current-part* #f)
	(set! *current-part-port* (current-output-port))

	(set-compile-seeing-shadows! #t) ;; top-level exprs
					 ;; should use shadows
	(let ((first #t)
	      (envt (top-level-envt (module self))))
	    (with-objects-from-file 
		(pathname->string (pathname self))
		(if $allow-form-restart
		    (lambda (form)
			(if (and first
				 (pair? form)
				 (eq? (car form) 'part-tag))
			    (set-part-tag! self (cadr form))
			    (call-with-current-continuation
				(lambda (exit)
				    (fluid-let ((*next-top-level-form* exit))
					(eval form envt)))))
			(set! first #f))
		    (lambda (form)
			(if (and first
				 (pair? form)
				 (eq? (car form) 'part-tag))
			    (set-part-tag! self (cadr form))
			    (eval form envt))
			(set! first #f)))))
	(flush-part!)
	(if (> (num-parts self) 1)
	    (with-output-to-file
		(pathname->string 
		    (make <file-name> directory: (dest-dir (module self))
				      filename: (string-append
						    (filename (pathname self))
						    "x")
				      extension: "c"))
		(lambda ()
		    (display-disclaimer)
		    (let ((nm (filename (pathname self))))
			(ct-format #t "#include \"~a.c\"\n" nm)
			(let loop ((i 1))
			    (if (< i (num-parts self))
				(begin
				    (ct-format #t "#include \"~a~d.c\"\n" nm i)
				    (loop (+ i 1)))))))))
	(set-templates! self *templates*)
	(set-contribution! (root-definitions (module self))
			   self
			   *root-defns*)
	(set-contribution! (link-table (module self))
			   self
			   *parts*)
	(set-contribution! (makefile-parts (module self))
			   self
			   *make-commands*)
	(set-contribution! (init-procs (module self))
			    self
			    (reverse *source-load-commands*))
	(set-contribution! (class-table (module self))
			    self
			    *classes*)
	(set-contribution! (private-interface (module self))
			    self
			    *private-interface*)
	(set-contribution! (public-interface (module self))
			    self
			    *public-interface*)
	(set-contribution! (imports-table (module self))
			   self
			   (reverse *imports*))
	(set-contribution! (exports-table (module self))
			   self
			   *exports*)
	(set! *current-source-file* old))
    (if (not *pre-load*)
	(set-clean! self)))

(define (flush-part!)
    (if *current-part*
	(begin
	    (with-output-to-port *current-part-port*
		(lambda ()
		    (display "\n/*** Part Link Table ***/\n\n")
		    (display "static struct function_descr ")
		    (format #t "*(part_~a_tab[]) = {\n"
				(link-name *current-part*))
		    (for-each 
		    	(lambda (x)
			    (ct-format #t "    &~a_descr,\n" x))
			(reverse (link-table *current-part*)))
		    (display "    NULL };\n")
		    (ct-format #t 
			"struct part_descr ~a_part_~a = {\n"
			(link-name (module *current-source-file*))
			(link-name *current-part*))
		    (ct-format #t 
			"    ~d,\n    &module_~a,\n"
			(tag *current-part*)
			(link-name (module *current-source-file*)))
		    (ct-format #t 
			"    part_~a_tab,\n    \"~a\",\n    0 };\n"
			(link-name *current-part*)
			(link-name *current-part*))
		    (ct-format #t 
			"#undef _SUBPROJ_~a\n" 
			(string-upcase 
			    (symbol->string 
				(name (module *current-source-file*)))))
		    (ct-format #t 
			"#undef _SCM_~a\n" 
			(string-upcase 
			    (filename 
				(pathname *current-source-file*))))
		    (ct-format #t 
			"#undef _C_~a\n" 
			(string-upcase (link-name *current-part*)))))
	    (close-output-port *current-part-port*)
	    (ct-format #t "Done with part ~a\n" 
	    		(dest-pathname *current-part*))
	    (set! *current-part* #f)
	    (set! *current-part-port* #f))))

(define (new-part-name)
    (let* ((sf *current-source-file*)
    	   (n (num-parts sf)))
	(make <file-name> directory: (dest-dir (module sf))
			  filename: (if (eq? n 0)
					(filename (pathname sf))
					(string-append
					    (filename (pathname sf))
					    (number->string n)))
			  extension: "c")))

;; part tag numbers smaller than 10,000 are reserved
;; for internal use (for now)

(define (make-part-tag)
    (if (and (part-tag *current-source-file*)
	     (eq? (num-parts *current-source-file*) 0))
	(part-tag *current-source-file*)
	(let loop ((i (random)))
	    (if (< i 10000)
		(loop (random))
		i))))

(define (new-part!)
    (flush-part!)
    (let ((dest (new-part-name))
    	  (sf *current-source-file*))
	(ct-format #t "Creating new part in: ~a\n" dest)
	(set! *current-part*
	    (make <Part> source-file: sf
			 dest-pathname: dest
			 link-name: (filename dest)
			 tag: (make-part-tag)))
	(set-part-descr! *current-part*
	    (make <PartDescr> module-name: (symbol->string (name (module sf)))
	    		      part-tag: (tag *current-part*)))

	(set! *current-part-port* (open-output-path dest))
	(set-num-parts! sf (add1 (num-parts sf)))
	(push! *parts* (link-name *current-part*) *parts*)
	(push! *make-commands* (cons 'c-file (file-within-dir dest)))
	(with-output-to-port *current-part-port*
	    (lambda ()
		(display-disclaimer)
		(ct-format #t 
		    "#define _SUBPROJ_~a\n" 
		    (string-upcase (symbol->string (name (module sf)))))
		(ct-format #t 
		    "#define _SCM_~a\n" 
		    (string-upcase (filename (pathname sf))))
		(ct-format #t 
		    "#define _C_~a\n" 
		    (string-upcase (link-name *current-part*)))
		(ct-format #t "#include \"~a\"\n" 
			(file-within-dir 
			    (filename 
				(private-interface 
				    (module *current-source-file*)))))
		(ct-format #t "extern struct part_descr ~a_part_~a;\n"
			(link-name *current-module*)
			(link-name *current-part*))
		(newline)))))

(define (display-disclaimer)
    (display "/**********************************************\n")
    (display "THIS FILE WAS AUTOMATICALLY GENERATED, AND MAY\n")
    (display "BE AUTOMATICALLY RE-GENERATED WHEN THE COMPILER\n")
    (display "CHANGES.  DO NOT MODIFY THIS FILE BY HAND!\n")
    (display "**********************************************/\n\n"))

; WITHIN-PART executes the given procedure (passing it the C name
; to use for the function) in the context of a "part".  It returns
; a pair whose car is the <PartDescr> for the part in which the
; function was written, and whose cdr is the index of the function
; within the part.

(define (within-part name size proc)
    (if (not *current-part*)
	(new-part!))
	
    (if (> (+ (total-size *current-part*) size)
    	   *max-part-size*)
	(begin
	    (if (> size *max-part-size*)
		(warning "In ~a: object ~a has size ~d > maximum size ~d"
			 (pathname *current-source-file*)
			 name
			 size
			 *max-part-size*))
	    ; create a new part
	    ; unless the current one is actually EMPTY
	    ; (in which case, the above message should have
	    ;  been generated)
	    (if (not (eq? (total-size *current-part*) 0))
		(new-part!))))

    ; now, the current part should be able to hold the monotones
    ; (if anything can!)
    
    (let ((c-name (string-append (link-name *current-part*)
    			         "_"
				 (c-ify (object->string name))
				 "_"
				 (number->string 
					(length (link-table 
							*current-part*))))))
    
	(set-link-table! *current-part*
	    (cons c-name
		  (link-table *current-part*)))
	(set-total-size! *current-part*
	    (+ size (total-size *current-part*)))
		  
	(with-output-to-port 
	    *current-part-port*
	    (lambda ()
		(proc c-name))))
    (cons (part-descr *current-part*)
          (sub1 (length (link-table *current-part*)))))

; a cover for within-part that returns a template
; which, when linked in, will refer to the emitted
; part, and holds all the literals besides

(define (template-within-part name size proc literals)
    (let* ((info (within-part name size proc))
	   (tmpl (make-template (cdr info) (car info) literals)))
	(set! *templates* (cons tmpl *templates*))
	tmpl))

(define (c-ify str)
    (let loop ((i (string->list str)) (r '()))
	(if (null? i)
	    (if (null? r)
		"anon"
		(list->string (reverse r)))
	    (if (char-alphabetic? (car i))
		(loop (cdr i) (cons (car i) r))
		(if (eq? (car i) #\-)
		    (loop (cdr i) (cons #\_ r))
		    (loop (cdr i) r))))))
		    
(define (size-of-monotones monotones)
    (length monotones))	; cheesy metric for now

(define *virt-asm-profile* #f)

(define (display-some-monotones name c-name monotones splat-proc)
    (ct-format #t "static char ~a_name[] = \"~a\";\n" c-name name)
    (ct-format #t "#define FUNCTION ~a_name\n\n" c-name)
    (ct-format #t "PROLOGUE(~a)\n\n" c-name)
    (ct-format #t "BEGIN_FWD(~a)\n" c-name)
    (for-each
	(lambda (m)
	    (ct-format #t "  FWD_MONOTONE(~a)\n" (car m)))
	monotones)
    (ct-format #t "END_FWD(~a)\n\n" c-name)
    (for-each
	(lambda (m)
	    (ct-format #t "MONOTONE(~a)\n{" (car m))
	    (splat-proc (cdr m))
	    (display "}\n\n"))
	monotones)
    (ct-format #t "EPILOGUE(~a)\n\n" c-name)

    (ct-format #t "BEGIN_BACK(~a)\n" c-name)
    (for-each
	(lambda (m)
	    (ct-format #t "  BACK_MONOTONE(~a)\n" (car m)))
	monotones)
    (ct-format #t "END_BACK(~a)\n\n" c-name)

    (ct-format #t "static struct function_descr ~a_descr = {\n" 
		    c-name)
    (ct-format #t "\tJUMP_TABLE( ~a ),\n" c-name)
    (ct-format #t "\t~a_name,\n" c-name)
    (ct-format #t "\t&~a_part_~a };\n" 
	          (link-name *current-module*)
		  (link-name *current-part*))
    (ct-format #t "#undef FUNCTION\n\n"))


(define (alloc-root-slot c-name name)
    (let* ((m (module *current-source-file*))
	   (uc (unified-contributions (root-definitions m)))
    	   (a (assq name uc))
	   (slot (if a
		     (cadr a)
		     (let ((i (num-roots m)))
			(set-num-roots! m (add1 i))
			i))))
	(push! *root-defns* (list name slot c-name))
	(ct-format #t "    Assigning root ~a to slot ~d\n" name slot)
	slot))


