#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define *system-defs* '())

(define (add-system-def name thing) 
    (set! *system-defs* 
	(cons (cons (if (string? name) 
			name
			(symbol->string name))
		    thing) 
	    *system-defs*)))

(define (add-system-class class)
    (add-system-def (%gvec-read class 0) class))

(add-system-def "<<Class>>" <<Class>>)
(add-system-def "<String>" <String>)
(add-system-def "<Vector>" <Vector>)
(add-system-def "<Symbol>" <Symbol>)
(add-system-def "<Pair>" <Pair>)
(add-system-def "<Closure>" <Closure>)
(add-system-def "<Template>" <Template>)
(add-system-def "<TopLevelVar>" <TopLevelVar>)
(add-system-def "<TopLevelEnvt>" <TopLevelEnvt>)
(add-system-def "<BindingEnvt>" <BindingEnvt>)
;(add-system-def "<LongFloat>" <LongFloat>)
(add-system-def "<LongFloat>" <double-float>)

#|  We need these for compiling "normal" code
(add-system-def '%if %if)
(add-system-def '%multi %multi)
(add-system-def '%combo %combo)
(add-system-def '%literal %literal)
(add-system-def '%bind %bind)
(add-system-def '%seq %seq)
(add-system-def '%lex-set! %lex-set!)
(add-system-def '%tl-set! %tl-set!)
(add-system-def '%lambda %lambda)
(add-system-def '%root-ref %root-ref)
(add-system-def '%root-set! %root-set!)
|#

(define (save-all)
  (apply save-system-image (map cdr *modules-table*)))

(define (save-loadable-image path module)
    (%save-image
	path
    	(vector
	    (find-version-defn (list module))
	    (unified-contributions (init-procs module))
	    (map
		(lambda (i)
		    (cons (cadr i)
		    	  (value (car i))))
		(unified-contributions (exports-table module))))
	(list->vector *system-defs*)))

(define (uniquify-on-cdr l)
    (let ((t (make-object-table)))
	(for-each
	    (lambda (i)
		(table-insert! t (cdr i) i))
	    l)
	(table-values->list t)))

(define (preview-defs v)
    (print v)
    v)
    
(define (save-system-image path . modules)
    (let ((inits (list->vector
		  (apply append
			 (map (lambda (m)
			         (unified-contributions (init-procs m)))
			      modules)))))
	;; save the initialization proc vector in the
	;; top-level variable *initialization-thunks*
	;; of the FIRST module
	(let ((bdg (lookup (top-level-envt (car modules))
			   '*initialization-thunks*)))
	    (if (and bdg (instance? bdg <TopLevelVar>))
		(set-value! bdg inits)))
	;; now, actually save the image
	(%save-image path
		(vector
		    ;; [0] name of module
		    (find-version-defn modules)
		    ;; [1] initialization procedures
		    inits
		    ;; [2] the top-level executive procedure
		    (find-start-binding modules))
		(preview-defs 
		    (list->vector (uniquify-on-cdr *system-defs*))))))

	
    
(define (renew-system-image . modules)
    (%save-image "../system-2.img"
           (vector
	        ;; [0] name of module
	        (find-version-defn modules)
		;; [1] initialization procedures
		(list->vector
		 (apply append
			(map (lambda (m)
			       (unified-contributions
				(init-procs m)))
			     modules)))
		;; [2] the top-level executive procedure
		(find-start-binding modules))
	   (list->vector *system-defs*)))

#|
	    (vector-map
		(lambda (cmd)
		    (if (vector? cmd)
			(vector (part-addr->part-descr (vector-ref cmd 0))
				(vector-ref cmd 1)
				(vector-ref cmd 2))
			cmd))
		(vector-ref (STUB) 1))
|#

(define (find-start-binding modules)
    (let loop ((m modules))
      (if (null? m)
	  (error/semantic "No $start binding")
	  (let ((bdg (lookup (top-level-envt (car m)) '$start)))
	    (if bdg
		(value bdg)
		(loop (cdr m)))))))

(define *default-version-line* "RScheme")

(define (find-version-defn modules)
    (let loop ((m modules))
      (if (null? m)
	  *default-version-line*
	  (let ((bdg (lookup (top-level-envt (car m)) '$version)))
	    (if bdg
		(value bdg)
		(loop (cdr m)))))))
	    

;; important system bootstrapping definitions:
;;	$std-gf:meta-method-envt
;;	$write-gf
;;	$display-gf
;;	$initialize-gf

(define (go . files)
    (let ((x (map load-module files)))
	(for-each update x)
	(save-all)
	x))

