/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "hashmain.h"
#include "runtim.h"
#include "htstruct.h"
#include "tables_r.h"

bool HASHTABLE_P( obj thing )
{
    return (OBJ_ISA_PTR(thing) 
    	    && obj_isa_instance_of(thing,table_class))
    	   ? YES : NO;
}

obj make_bucket( int bucket_bits )
{
obj x;

    x = gvec_alloc( (3 * BUCKET_CAPACITY) + 2, table_bucket_class );
    gvec_write_fresh_non_ptr( x, SLOT(0), int2fx( bucket_bits ) );
    return x;
}

/* obj make_primitive_table( void ) */
obj make_table( obj of_table_class, unsigned size_bits )
{
    if (EQ(of_table_class, eq_table_class))
	return make4( of_table_class,
		      make_empty_vector(1<<size_bits),
		      int2fx(size_bits),
		      ZERO,		/* table-size */
		      FALSE_OBJ );  	/* table-hash-function */
    else
	return make3( of_table_class,
		      make_empty_vector(1<<size_bits),
		      int2fx(size_bits),
		      ZERO );  		/* table-size */
}

/*  Low-level operation:  Install a (key,value) pair into
    a table, having already computed the hash value.
*/

void hashtable_install( obj table, obj hash, obj key, obj value )
{
obj vec, bucket;
UINT_32 i;

    inserting_one( table );
    vec = gvec_read( table, HASHTABLE_DIRECTORY );
    
    for (bucket = read_dir(vec,hash);
         !EQ(bucket,FALSE_OBJ);
	 bucket = gvec_read( bucket, BUCKET_OVERFLOW ))
    {
        for (i=SLOT(2); i<SLOT(2+BUCKET_CAPACITY); i+=sizeof(obj))
	{
	    if (EQ(read_bucket_hash(bucket,i),FALSE_OBJ))
	    {
		write_bucket_hash( bucket, i, hash );
		write_bucket_key( bucket, i, key );
		write_bucket_value( bucket, i, value );
		return;
	    }
	}
    }

    /* grow things... */
    
    split_bucket( table, read_dir( vec, hash ), hash, key, value );
}

void hashtable_foreach( obj table, void *info, 
			void (*fn)( void *info, obj h, obj key, obj value ) )
{
obj vec, bucket;
UINT_32 mask, i, j, n;

    vec = gvec_read( table, HASHTABLE_DIRECTORY );
    n = SIZEOF_PTR(vec);
    for (j=0; j<n; j+=SLOT(1))
    {
        bucket = gvec_read( vec, j );
	if (!EQ(bucket,FALSE_OBJ))
	{
	    mask = (~(0UL)-3) << fx2int( gvec_read( bucket, BUCKET_BITS ) );
	    
	    if (!(j & mask))
	    {
		do {
		    for (i=SLOT(2); i<SLOT(2+BUCKET_CAPACITY); i+=sizeof(obj))
		    {
			if (!EQ(read_bucket_hash(bucket,i),FALSE_OBJ))
			{
			    fn( info, read_bucket_hash( bucket, i ),
			    	      read_bucket_key( bucket, i ),
				      read_bucket_value( bucket, i ) );
			}
		    }
		    bucket = gvec_read( bucket, BUCKET_OVERFLOW );
		} while (!EQ(bucket,FALSE_OBJ));
	    }
	}
    }
}

obj hashtable_chains( obj table )
{
obj chain_list, vec, bucket;
UINT_32 mask, j, n;

    chain_list = NIL_OBJ;
    vec = gvec_read( table, HASHTABLE_DIRECTORY );
    n = SIZEOF_PTR(vec);
    for (j=0; j<n; j+=SLOT(1))
    {
        bucket = gvec_read( vec, j );
	if (!EQ(bucket,FALSE_OBJ))
	{
	    mask = (~(0UL)-3) << fx2int( gvec_read( bucket, BUCKET_BITS ) );
	    
	    if (!(j & mask))
	    {
		chain_list = cons( bucket, chain_list );
	    }
	}
    }
    return chain_list;
}

static void accum_key( void *info, obj h, obj k, obj v )
{
    *(obj *)info = cons( k, *(obj *)info );
}

static void accum_value( void *info, obj h, obj k, obj v )
{
    *(obj *)info = cons( v, *(obj *)info );
}

obj hashtable_keys_to_list( obj table )
{
obj list = NIL_OBJ;

    hashtable_foreach( table, &list, accum_key );
    return list;
}

obj hashtable_values_to_list( obj table )
{
obj list = NIL_OBJ;

    hashtable_foreach( table, &list, accum_value );
    return list;
}


#if 0
/**  Probe for a specific hash value
     (if YES, their is a good chance that the key with that
      hash value is not present; if NO, guarantees that the
      key with that hash value is absent)

    P.S. I was going to use this for lookup_symbol(),
         but it turns out it wouldn't do any good!
*/

bool hashtable_probe( obj table, obj h )
{
UINT_32 i;
obj vec, bucket;

    vec = gvec_read( table, HASHTABLE_DIRECTORY );
    
    bucket = read_dir( vec, h );
    
    while (!EQ(bucket,FALSE_OBJ))
    {
        for (i=SLOT(2); i<SLOT(2+BUCKET_CAPACITY); i+=sizeof(obj))
	{
	    if (EQ(read_bucket_hash(bucket,i),h))
		return YES;
	}
	bucket = gvec_read( bucket, BUCKET_OVERFLOW );
    }
    return NO;
}
#endif /* 0 */

#ifdef INLINES
#define eq_fn_inline inline
#else
#define eq_fn_inline /* nothing */
#endif

/********************* string=?-hash-table *********************/

#define SPECIAL_TABLE_LOOKUP	stringtable_lookup
#define SPECIAL_TABLE_INSERT    stringtable_insert
#define SPECIAL_TABLE_REMOVE    stringtable_remove
#define SPECIAL_CMP_FN		string_equal

static eq_fn_inline bool string_equal( obj str1, obj str2 )
{
    return string_cmp( str1, str2 ) == 0 ? YES : NO;
}

#include "special.ci"

#undef SPECIAL_TABLE_LOOKUP
#undef SPECIAL_TABLE_INSERT
#undef SPECIAL_TABLE_REMOVE
#undef SPECIAL_CMP_FN

/********************* string-ci=?-hash-table *********************/

#define SPECIAL_TABLE_LOOKUP	cistringtable_lookup
#define SPECIAL_TABLE_INSERT    cistringtable_insert
#define SPECIAL_TABLE_REMOVE    cistringtable_remove
#define SPECIAL_CMP_FN		string_ci_equal

static eq_fn_inline bool string_ci_equal( obj str1, obj str2 )
{
    return string_ci_cmp( str1, str2 ) == 0 ? YES : NO;
}

#include "special.ci"

#undef SPECIAL_TABLE_LOOKUP
#undef SPECIAL_TABLE_INSERT
#undef SPECIAL_TABLE_REMOVE
#undef SPECIAL_CMP_FN

/********************* eq?-hash-table *********************/

#define SPECIAL_TABLE_LOOKUP	objecttable_lookup
#define SPECIAL_TABLE_INSERT    objecttable_insert
#define SPECIAL_TABLE_REMOVE    objecttable_remove
#define SPECIAL_CMP_FN		object_cmp

static eq_fn_inline bool object_cmp( obj sym1, obj sym2 )
{
    return (bool)EQ(sym1,sym2);
}

#include "special.ci"

#undef SPECIAL_TABLE_LOOKUP
#undef SPECIAL_TABLE_INSERT
#undef SPECIAL_TABLE_REMOVE
#undef SPECIAL_CMP_FN

