/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "heapi.h"
#include "imagfile.h"
#include "runtim.h"
#include "hashmain.h"
#include "hashfn.h"
#include "mapf.h"
#include <string.h>
#include "smemory.h"

void string_lower_case( obj str );

#define IGNORE_FAULTS

/* #define LOADIMAGE_TRACE */
/* #define PAUSE_IN_MODE2 */

#ifdef LOADIMAGE_TRACE
/* #define SKIP_TEMPLATE_SWIZZLING */
FILE *debug_file;
FILE *debug_binary_file;
char info_buffer[1024*1024], *info_buf_ptr;
obj target_comments;	/* a hash table of comments */
char *comment_str;

obj unswiz_data[1024*60], *unswiz_data_ptr;

static void write_trace_info( obj item, int mode )
{
int n;

    fwrite( &item, sizeof(obj), 1, debug_binary_file );
    fwrite( &mode, sizeof(int), 1, debug_binary_file );
    n = unswiz_data_ptr - unswiz_data;
    fwrite( &n, sizeof(int), 1, debug_binary_file );
    fwrite( unswiz_data, sizeof(obj), n, debug_binary_file );
    n = info_buf_ptr - info_buffer;
    fwrite( &n, sizeof(int), 1, debug_binary_file );
    fwrite( info_buffer, 1, n, debug_binary_file );
}

static void reset_trace_info( void )
{
    info_buf_ptr = info_buffer;
    unswiz_data_ptr = unswiz_data;
}

static void tri_printf( const char *msg, ... )
{
va_list a;
unsigned n;

    va_start(a,msg);
    vsprintf( info_buf_ptr, msg, a );
    va_end(a);
    n = strlen(info_buf_ptr);
    fwrite( info_buf_ptr, 1, n, debug_file );
    info_buf_ptr += n;
}

static void tri_swizzled( int offset, obj from, obj to )
{
obj comment = FALSE_OBJ;

    if (OBJ_ISA_PTR(from))
    {
    obj k = OBJ( VAL(from) - POINTER_TAG );
    
	comment = objecttable_lookup( target_comments, k, k );
    }
    if (offset == -1)
	tri_printf( "  class" );
    else if (offset == -2)
	tri_printf( "   self" );
    else
	tri_printf( "    [%d]", offset/4 );
	
    tri_printf( " %#x => %#x", VAL(from), VAL(to) );
    
    if (OBJ_ISA_PTR(to))
    {
	*unswiz_data_ptr++ = to;
        *info_buf_ptr++ = ' ';
	*info_buf_ptr++ = '*';
    }

    if (!EQ(comment,FALSE_OBJ))
    {
        *info_buf_ptr++ = ' ';
	strcpy( info_buf_ptr, string_text(comment) );
	info_buf_ptr = strchr( info_buf_ptr, 0 );
    }
    *info_buf_ptr++ = '\n';
}


#endif

#define NUM_USEFUL_LOAD2_MODES (4)

/*  This is a special flag that disables asserting that
    the class for an allocation is ZERO or a CLASS_P
    -- used only when assert's are used (ie, no NDEBUG)
       or when counting allocations (because the COUNT_ALLOCS
       mechanism needs to be turned off during loading, too)
*/

#if !defined(NDEBUG) || defined(COUNT_ALLOCS)
extern bool loading_image;
#endif

/* #ifdef NDEBUG
   #undef NDEBUG
   static bool loading_image;
   #endif
*/

static bool remap_constants; /* do certain constants (like #none) need
				to get switched around? (ie, for loading
				version 1 images) */

typedef void swizzler_fn( obj item );

static void swizzle_class_only( obj item );
static void swizzle_gvec( obj item );
static void swizzle_template( obj item );
static void swizzle_partcont( obj item );

static struct {
    enum load2_mode mode;
    swizzler_fn		*swizzler;
} useful_load2_modes[NUM_USEFUL_LOAD2_MODES] =
    { { LOAD2_CLASS_ONLY, swizzle_class_only },
      { LOAD2_GVEC, swizzle_gvec },
      { LOAD2_TEMPLATE, swizzle_template },
      { LOAD2_PARTCONT, swizzle_partcont }, };


static obj load_ref( void *src, UINT_32 len, obj preamble );
static obj load_symbol( void *src, UINT_32 len, obj preamble );
static obj load_part( void *src, UINT_32 len, obj preamble );
static obj load_array_8( void *src, UINT_32 len, obj preamble );
static obj load_array_16( void *src, UINT_32 len, obj preamble );
static obj load_array_32( void *src, UINT_32 len, obj preamble );
static obj load_array_64( void *src, UINT_32 len, obj preamble );

static void *find_a_part( const char *module, int tag );

static obj *obj_table;
static obj reference_table;

static volatile void load_image_error( const char *msg )
{
    fprintf( stderr, "Load error: %s\n", msg );
    abort();
}


static inline obj swizzle( obj item )
{
    if (OBJ_ISA_PTR(item))
    {
    obj result;
    
	result = *(obj *)(((char *)obj_table) + VAL(item) - POINTER_TAG);
#ifdef LOADIMAGE_TRACE
	fprintf( debug_file, "\t%#x ==> %#x\n", VAL(item), VAL(result) );
#endif
	return result;
    }
    else
    {
        if (remap_constants && OBJ_ISA_IMMOB(item))
	{
	    switch ((int)SECONDARY_TAG(item))
	    {
	      case 4: /* old #undef */
		item = UNDEFINED_OBJ;
		break;
	      case 5: /* old #uninit */
		item = UNINITIALIZED_OBJ;
		break;
	      case 6: /* old #none */
		item = NOVALUE_OBJ;
		break;
	      case 7: /* old #unbound */
		item = UNBOUND_OBJ;
		break;
	    }
	}
	return item;
    }
}


static void swizzle_class_only( obj item )
{
POBHeader *h = PTR_TO_HDRPTR(item);
obj n;

    n = swizzle( h->pob_class );
#ifdef LOADIMAGE_TRACE
    tri_swizzled( -1, h->pob_class, n );
#endif
    h->pob_class = n;
}

static void swizzle_gvec( obj item )
{
POBHeader *h = PTR_TO_HDRPTR(item);
obj n, *d = (obj *)PTR_TO_DATAPTR(item);
UINT_32 i, len = SIZEOF_PTR(item);

    n = swizzle( h->pob_class );
#ifdef LOADIMAGE_TRACE
    tri_swizzled( -1, h->pob_class, n );
#endif
    h->pob_class = n;

    for (i=0; i<len; i+=SLOT(1))
    {
	n = swizzle( *d );
#ifdef LOADIMAGE_TRACE
	tri_swizzled( i, *d, n );
#endif /* LOADIMAGE_TRACE */
        gvec_write_fresh( item, i, n );
	d++;
    }
}

static void swizzle_template( obj item )
{
struct part_descr *part_ptr;
UINT_32 ix;
jump_addr entry_point;
struct function_descr *fn;

    swizzle_gvec( item );
    if (SIZEOF_PTR(item) < 8)
    {
	fprintf( stderr, 
	    "Warning: %#x (TEMPLATE) is too small! (code_ptr = %#x)\n", 
	    VAL(item), VAL(gvec_read(item,0)) );
	return;
    }
#ifndef SKIP_TEMPLATE_SWIZZLING
    part_ptr = (struct part_descr *)
    			OBJ_TO_RAW_PTR( gvec_read( item, SLOT(1) ) );
    if (!part_ptr)
    {
        fprintf( stderr, "Error: %#x (TEMPLATE) has no part ptr\n",
		 VAL(item) );
#ifndef IGNORE_FAULTS
        abort();
#else
        gvec_write_fresh_non_ptr( item, SLOT(0), ZERO );
        return;  
#endif /* IGNORE_FAULTS */
    }
    ix = fx2int( gvec_read( item, SLOT(0) ) );
    fn = part_ptr->functions[ix];
    entry_point = fn->monotones[0];
    gvec_write_fresh_non_ptr( item, SLOT(0), JUMP_ADDR_TO_OBJ(entry_point) );
    gvec_write_fresh_non_ptr( item, SLOT(1), RAW_PTR_TO_OBJ(fn) );
#endif /* SKIP_TEMPLATE_SWIZZLING */
}

static void swizzle_partcont( obj item )
{
    swizzle_gvec( item );
}

obj load_image_file( const char *path, obj the_reference_table )
{
struct file_header *hdr;
UINT_32 i, *input_ptr, *input_buffer;
obj result;

#ifdef LOADIMAGE_TRACE
    debug_file = fopen( "/tmp/loadimg.trc", "w" );
    debug_binary_file = fopen( "/tmp/loadimg.bin", "wb" );
    target_comments = make_primitive_table();
    info_buf_ptr = NULL;
#endif /* LOADIMAGE_TRACE */
    if (!mapf_open( path ))
	return FALSE_OBJ;
	
    reference_table = the_reference_table;
    mapf_seek( FILE_HDR_OFFSET );
    hdr = (struct file_header *)mapf_read( sizeof( struct file_header ) );
#ifndef PLATFORM_IS_BIG_ENDIAN
    hdr->magic = BIG_ENDIAN_TO_HOST_32(hdr->magic);
    hdr->version = BIG_ENDIAN_TO_HOST_32(hdr->version);
    for (i=0; i<NUM_LOAD2_MODES; i++)
    {
	hdr->load2_offset[i] = BIG_ENDIAN_TO_HOST_32(hdr->load2_offset[i]);
	hdr->load2_count[i] = BIG_ENDIAN_TO_HOST_32(hdr->load2_count[i]);
    }
    hdr->data_area_length = BIG_ENDIAN_TO_HOST_32(hdr->data_area_length);
    hdr->num_objects = BIG_ENDIAN_TO_HOST_32(hdr->num_objects);
    hdr->entry_object = BIG_ENDIAN_TO_HOST_OBJ(hdr->entry_object);
#endif

    if (hdr->magic != IMAGE_MAGIC_NUMBER)
    {
	load_image_error( "Not an image file" );
    }
    
    remap_constants = NO;
    switch (hdr->version)
    {
	case IMAGE_VERSION_NUMBER:
		break;
	case 1:
		/* we can handle version 1 images by remapping
		    some small constants */
	    	remap_constants = YES;
		break;
	default:
	    load_image_error( "Version number is incorrect" );
    }
    
    obj_table = (obj *)malloc( sizeof(obj) * hdr->num_objects );
    if (!obj_table)
    {
	load_image_error( "Could not malloc space for object table" );
    }

    /*  Pass I. Allocate space for and 
    		load the data for the objects
    		(and execute LOAD1 commands, involving
		 table lookups (REF and SYMBOL), and convert
		 endianess) */

    input_buffer = (UINT_32 *)mapf_read( hdr->data_area_length );
    input_ptr = input_buffer;

#if !defined(NDEBUG) || defined(COUNT_ALLOCS)
    loading_image = YES;
#endif
    
    for (i=0; i<hdr->num_objects; i++)
    {
    UINT_32 len;
    enum load1_mode mode1;
    obj preamble, item;
    void *p;
    
	len = BIG_ENDIAN_TO_HOST_32( input_ptr[0] );
	mode1 = (enum load1_mode) BIG_ENDIAN_TO_HOST_32( input_ptr[1] );
	preamble = BIG_ENDIAN_TO_HOST_OBJ( OBJ(input_ptr[2]) );
	p = &input_ptr[3];
#ifdef LOADIMAGE_TRACE
	fprintf( debug_file, 
		"Loading %#x := mode(I) %d (%u bytes) preamble %#x",
		(i*4+3), mode1, len, VAL(preamble) );
	comment_str = NULL;
#endif /* LOADIMAGE_TRACE */
	switch (mode1)
	{
	    case LOAD1_REF:	item = load_ref( p, len, preamble );
	    			break;
	    case LOAD1_SYMBOL:	item = load_symbol( p, len, preamble );
	    			break;
	    case LOAD1_PART:	item = load_part( p, len, preamble );
	    			break;
	    case LOAD1_ARRAY8:	item = load_array_8( p, len, preamble );
	    			break;
	    case LOAD1_ARRAY16:	item = load_array_16( p, len, preamble );
	    			break;
	    case LOAD1_ARRAY32:	item = load_array_32( p, len, preamble );
	    			break;
	    case LOAD1_ARRAY64:	item = load_array_64( p, len, preamble );
	    			break;
	    default:		item = ZERO;
	    			assert( !EQ(item,ZERO) );
				break;
	}
#ifdef LOADIMAGE_TRACE
	if (comment_str)
	{
	obj key = int2fx(i);
	
	    hashtable_install( target_comments, 
	    			key,
				key,
	    			make_string(comment_str) );
	    fprintf( debug_file, " %s", comment_str );
	}
        fprintf( debug_file, " := %#x\n", VAL(item) );
#endif /* LOADIMAGE_TRACE */
	obj_table[i] = item;
	input_ptr = (UINT_32 *)(((char *)input_ptr) + ((len + 3 + 12) & ~3));
    }
#ifdef LOADIMAGE_TRACE
    fflush( debug_file );
#endif /* LOADIMAGE_TRACE */

    /*  Pass II.  Go through the LOAD2 commands, executing
    		  them in order
    */
    
    for (i=0; i<NUM_USEFUL_LOAD2_MODES; i++)
    {
    enum load2_mode mode2 = useful_load2_modes[i].mode;
    swizzler_fn *proc = useful_load2_modes[i].swizzler;
    obj *table;
    UINT_32 j, count;
    
	count = hdr->load2_count[mode2];
#ifdef LOADIMAGE_TRACE
        fprintf( debug_file, "** LOAD2 (mode2 = %d): %u\n",
		             mode2,
		             count );
#endif
        if (count == 0)
	  continue;

	table = (obj *)mapf_read( sizeof(obj) * count );

#if defined(LOADIMAGE_TRACE) && defined(PAUSE_IN_MODE2)
	if (i > 0)
	{ 
	char temp[10];
	UINT_32 n = count;
	int getpid( void );

	    if (n > 20)
		n = 20;
		
	    boot_list = alloc( SLOT(n), ZERO );
	    for (j=0; j<n; j++)
		gvec_write( boot_list, SLOT(j), swizzle(table[j]) );
	    printf( "-m %d %d -- about to load %u in mode(II) %d?", 
			    getpid(), (int)master_table, count,  mode2 );
	    gets(temp);
	}
#endif /* LOADIMAGE_TRACE */

	for (j=0; j<count; j++)
	{
	obj item = swizzle(BIG_ENDIAN_TO_HOST_OBJ( table[j] ));

#ifdef LOADIMAGE_TRACE
	    reset_trace_info();
	    tri_printf( "Loading mode(II) %d %#x\n", mode2, VAL(item) );
	    tri_swizzled( -2, BIG_ENDIAN_TO_HOST_OBJ( table[j] ), item );
#endif /* LOADIMAGE_TRACE */
	    proc( item );
#ifdef LOADIMAGE_TRACE
	    tri_printf( "done\n" );
	    write_trace_info( item, mode2 );
#endif /* LOADIMAGE_TRACE */
	}
    }
    result = swizzle( hdr->entry_object );
#ifdef LOADIMAGE_TRACE
    fclose( debug_file );
    fclose( debug_binary_file );
#endif /* LOADIMAGE_TRACE */
    mapf_close();
    free( obj_table );

#if !defined(NDEBUG) || defined(COUNT_ALLOCS)
    loading_image = NO;
#endif
    return result;
}

obj load_ref( void *src, UINT_32 len, obj preamble )
{
obj str = load_array_8( src, len, string_class );
obj item;

#ifdef LOADIMAGE_TRACE
    sprintf( info_buffer, "(REF '%s')", string_text(str) );
    comment_str = info_buffer;
#endif /* LOADIMAGE_TRACE */

    /* in the future, this will not be computed here... */
    preamble = raw_bytes_hash( src, len );
    
    /* note that you can't bind a REF to #f, because we
       would think that it's not bound! */
  
    item = stringtable_lookup( reference_table, preamble, str );
    if (EQ(item,FALSE_OBJ))
    {
	fprintf( stderr, "REF `%s' not bound\n", (char *)src );
    }
    return item;
}

obj load_symbol( void *src, UINT_32 len, obj preamble )
{
obj str = load_array_8( src, len, string_class );

#ifdef LOADIMAGE_TRACE
    sprintf( info_buffer, " (SYMBOL '%s')", string_text(str) );
    comment_str = info_buffer;
#endif /* LOADIMAGE_TRACE */
#ifndef CASE_SENSITIVE
    string_lower_case( str );
#endif
    return intern( str );
}

obj load_part( void *src, UINT_32 len, obj preamble )
{
obj objp;
void *p;

    /* preamble is the part's tag,
       and the string is the name of the module.
       Returns a pointer to the part_descr (which is NOT
       a heap object, note) */
    assert( OBJ_ISA_FIXNUM(preamble) );
    p = find_a_part( (const char *)src, fx2int(preamble) );
    objp = RAW_PTR_TO_OBJ(p);
    assert( !OBJ_ISA_PTR(objp) );
#ifdef LOADIMAGE_TRACE
    if (p)
    {
	if (EQ(preamble,ZERO))
	    sprintf( info_buffer, 
		    " (PART %s)", ((struct module_descr *)p)->name );
	else
	    sprintf( info_buffer, 
		    " (PART %s, %s)", 
		    ((struct part_descr *)p)->in_module->name, 
		    ((struct part_descr *)p)->name );
    }
    else
	sprintf( info_buffer, " (NO PART)");
    comment_str = info_buffer;
#endif /* LOADIMAGE_TRACE */
    return objp;
}

obj load_array_8( void *src, UINT_32 len, obj preamble )
{
obj item = bvec_alloc( len, preamble );

    memcpy( PTR_TO_DATAPTR(item), src, len );
#ifdef LOADIMAGE_TRACE
    if (info_buf_ptr) /* might not be in mode II yet */
    {
    unsigned char *p = (unsigned char *)src;
    UINT_32 i, n;
    
	if (len > 64)
	    n = 64;
	else
	    n = len;
	tri_printf( "array8[%u] \"", len );
	for (i=0; i<len; i++)
	{
	unsigned char ch = *p++;
	
	    if ((ch >= 32) && (ch < 127))
		*info_buf_ptr++ = ch;
	    else
	    {
		*info_buf_ptr++ = '.';
	    }
	}
	if (len > 64)
	    tri_printf( "..." );
	*info_buf_ptr++ = '\"';
	*info_buf_ptr++ = '\n';
    }
#endif /* LOADIMAGE_TRACE */
    return item;
}

obj load_array_16( void *src, UINT_32 len, obj preamble )
{
obj item = bvec_alloc( len, preamble );
UINT_16 *s, *d, *limit;

    assert( (len & (sizeof(UINT_16)-1)) == 0 );
    s = (UINT_16 *)src;
    limit = (UINT_16 *)(len + (char *)s);
    
    d = (UINT_16 *)PTR_TO_DATAPTR(item);

    while (s < limit)
	*d++ = BIG_ENDIAN_TO_HOST_16(*s++);

    return item;
}

obj load_array_32( void *src, UINT_32 len, obj preamble )
{
obj item = alloc( len, preamble );
UINT_32 *s, *d, *limit;

    assert( (len & (sizeof(UINT_32)-1)) == 0 );
    s = (UINT_32 *)src;
    limit = (UINT_32 *)(len + (char *)s);
    d = (UINT_32 *)PTR_TO_DATAPTR(item);

    while (s < limit)
	*d++ = BIG_ENDIAN_TO_HOST_32(*s++);

    return item;
}

obj load_array_64( void *src, UINT_32 len, obj preamble )
{
obj item = alloc( len, preamble );
UINT_32 *d32, *s32, *limit32;

    assert( (len & (sizeof(IEEE_64)-1)) == 0 );
    s32 = (UINT_32 *)src;
    d32 = (UINT_32 *)PTR_TO_DATAPTR(item);
    limit32 = (UINT_32 *)(len + (char *)s32);

    while (s32 < limit32)
    {
	*d32++ = *s32++;
	*d32++ = *s32++;
    }
#ifndef PLATFORM_IS_BIG_ENDIAN
    {
    IEEE_64 *d, *limit;
	
	d = (IEEE_64 *)PTR_TO_DATAPTR(item);
	limit = (IEEE_64 *)(len + (char *)d);
    
	while (d < limit)
	{
	    *d = BIG_ENDIAN_TO_HOST_IEEE_64(*d);
	    d++;
	}
    }
#endif
    return item;
}

static void *find_a_part( const char *module, int tag )
{
struct module_descr *m = find_module( module );

    if (tag)
	return (void *)find_part( m, tag );
    else
	return (void *)m;
}
