/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "mapf.h"
#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#ifdef __cplusplus
}
#endif

static caddr_t image_mapping_addr;
static size_t image_mapping_size;
static void *image_mapping_ptr;
static int image_file;

bool mapf_open( const char *path )
{
  image_file = open( path, O_RDONLY, 0 );
  if (image_file < 0)
    {
      perror( path );
      return NO;
    }
  image_mapping_size = lseek( image_file, 0L, SEEK_END );
  
  image_mapping_addr = mmap( (caddr_t)0, image_mapping_size, PROT_READ,
	   MAP_VARIABLE, image_file, (off_t)0 );
  if (!image_mapping_addr)
    {
      fprintf( stderr, "could not map file: %s\n", path );
      close( image_file );
      return NO;
    }

  image_mapping_ptr = (void *)image_mapping_addr;
  return YES;
}

void mapf_seek( UINT_32 offset )
{
    image_mapping_ptr = (void *)((char *)image_mapping_addr + offset);
}

void *mapf_read( UINT_32 bytes )
{
void *p = image_mapping_ptr;

    image_mapping_ptr = (void *)((char *)image_mapping_ptr + bytes);
    return p;
}

void mapf_close( void )
{
    if (image_file >= 0)
    {
        munmap( image_mapping_addr, image_mapping_size );
	close( image_file );
    }
}

