/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include <stdio.h>
#include <stdlib.h>
#include "saveq.h"

/* the -4 is because in struct SaveQueue_item is char data[4] */

#define EXTRA_SIZE	(sizeof( struct SaveQueue_item ) - 4)

void init_queue( SaveQueue *q, unsigned item_size, ncanonicalize_fn *fn )
{
    q->front = NULL;
    q->tail = NULL;
    q->item_size = item_size + EXTRA_SIZE;
    q->netcanon_fn = fn;
}

void *enqueue_item( SaveQueue *q )
{
struct SaveQueue_item *x;

    x = (struct SaveQueue_item *)malloc( q->item_size );
    if (q->tail)
	q->tail->next = x;
    else
	q->front = x;
    q->tail = x;
    x->next = NULL;
    return x->data;
}


void init_queue_state( SaveQueueIterator *state, SaveQueue *q )
{
    state->q = q;
    state->p = NULL;
}

const void *dequeue_item( SaveQueueIterator *state )
{
    if (!state->p)
    {
	state->p = state->q->front;
    }
    else
    {
	state->p = state->p->next;
    }
    return (state->p) ? state->p->data : NULL;
}

unsigned fwrite_queue_contents( SaveQueue *q, FILE *f )
{
unsigned n, per_item;
struct SaveQueue_item *i;
#ifndef PLATFORM_IS_BIG_ENDIAN
ncanonicalize_fn *canonicalize = q->netcanon_fn;
#endif

    n = 0;
    per_item = q->item_size - EXTRA_SIZE;
    for (i=q->front; i; i=i->next)
    {
#ifdef PLATFORM_IS_BIG_ENDIAN
	fwrite( i->data, 1, per_item, f );
#else
	fwrite( canonicalize( i->data ), 1, per_item, f );
#endif
	n++;
    }
    return n;    
}

void free_queue( SaveQueue *q )
{
struct SaveQueue_item *i, *j;

    for (i=q->front; i; i=j)
    {
	j = i->next;
	free( i );
    }
}

#if 0
void test( const char *teststr )
{
SaveQueue q;
SaveQueueIterator i;
int *p;

    init_queue( &q, sizeof(int) );
    init_queue_state( &i, &q );
    
    printf( "From '%s': ", teststr );
    while (*teststr)
    {
        if (*teststr == '.')
	{
	    p = dequeue_item(&i);
	    if (p)
		printf( " (%d)", *p );
	    else
		printf( " ()" );
	}
	else
	{
	    p = enqueue_item( &q );
	    *p = *teststr - '0';
	    printf( " %d", *p );
	}
	teststr++;
    }
    printf( "\nAgain: " );
    init_queue_state( &i, &q );
    while (p = dequeue_item(&i))
    {
	printf( " %d", *p );
    }
    printf( "\n" );
}

int main( int argc, const char **argv )
{
int i;

    for (i=1; i<argc; i++)
	test( argv[i] );
    return 0;
}
#endif
