#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;  OS Library / (High Level) Dynamic Linking Interface

(prologue {
#include "schemedl.h"
})

(include-c "dynlink.c")
(include-c "schemedl.c")
(include-h "schemedl.h")

(%define-glue %c-apply
{
union {
    jump_addr p;
    obj (*fn0)( void );
    obj (*fn1)( obj );
    obj (*fn2)( obj, obj );
    obj (*fn3)( obj, obj, obj );
    obj (*fn4)( obj, obj, obj, obj );
    obj (*fn5)( obj, obj, obj, obj, obj );
} fn;

    fn.p = OBJ_TO_JUMP_ADDR(LEXREF0(0));
    if (arg_count_reg != fx2int(LEXREF0(1)))
    {
	scheme_error( "~a: Wrong number args (~d) to C function; expected ~d", 
			3, 
			LEXREF0(2), 
			int2fx(arg_count_reg), 
			LEXREF0(1) );
    }
    assert( arg_count_reg <= 5 );
    switch (arg_count_reg)
    {
	case 0:	REG0 = fn.fn0();
		break;
	case 1:	REG0 = fn.fn1(REG0);
		break;
	case 2:	REG0 = fn.fn2(REG0,REG1);
		break;
	case 3:	REG0 = fn.fn3(REG0,REG1,REG2);
		break;
	case 4:	REG0 = fn.fn4(REG0,REG1,REG2,REG3);
		break;
	case 5:	REG0 = fn.fn5(REG0,REG1,REG2,REG3,REG4);
		break;
    }
    RETURN(1);
})

(%define-glue (call-gate gate_name)
{
const char *n = string_text(gate_name);
struct call_gate_descr *i;

    for (i=call_gates; i->name; i++)
    {
	if (strcmp(n,i->name) == 0)
	{
	    REG0 = JUMP_ADDR_TO_OBJ(i->proc);
	    RETURN(1);
	}
    }
    REG0 = FALSE_OBJ;
    RETURN(1);
})

(define (dynamic-load-lookup dl-file name)
    (%call ("scheme_dl_lookup" (<obj> <obj>) => <obj>) dl-file name))

(define (open-dynamic-load path)
    (%call ("scheme_dl_load" (<obj>) => <obj>) path))

(define (close-dynamic-load dlf)
    (%call ("scheme_dl_done_loading" (<obj>)) dlf)
    (values))


(%define-glue (install-module module)
{
    install_module( (struct module_descr *)OBJ_TO_RAW_PTR(module) );
    RETURN0();
})

(%define-glue (name->module module_name)
{
    REG0 = RAW_PTR_TO_OBJ( find_module( string_text(module_name) ) );
    RETURN(1);
})

(%define-glue (module&tag->part module tag)
{
    REG0 = RAW_PTR_TO_OBJ( 
    		find_part( 
		    (struct module_descr *)OBJ_TO_RAW_PTR(module), 
		    fx2int(tag) ) );
    RETURN(1);
})


(%define-glue (part&index->function the_part_descr the_fn_index)
{
struct part_descr *pd;
struct function_descr *fn;

    pd = (struct part_descr *)OBJ_TO_RAW_PTR(the_part_descr);
    fn = pd->functions[ fx2int(the_fn_index) ];
    REG0 = JUMP_ADDR_TO_OBJ(fn->monotones[0]);
    REG1 = RAW_PTR_TO_OBJ(fn);
    RETURN(2);
})

(define (raw-fn-ptr->procedure raw-ptr name num-args)
  (%make <Closure> 
	 (%make <BindingEnvt> '() raw-ptr num-args name)
	 %c-apply))

(define (make-c-ffi dl-file name num-args)
    (let ((f (dynamic-load-lookup dl-file name)))
	(if f
	    (raw-fn-ptr->procedure f name num-args)
	    (abort 'make-c-ffi "symbol `~a' not linked" name))))

(define (gate->procedure name num-args)
  (let ((f (call-gate name)))
    (if f
	(raw-fn-ptr->procedure f name num-args)
	(abort 'gate->procedure "call gate ~s not defined" name))))

;; use DYNAMIC-LOAD-CODE when the image is already
;; in memory, but needs to be fixed up.  This would be
;; the case if we spit out a .c file and compile it, never
;; shutting down our heap.

(define (dynamic-load-code path image name)
    (let ((dlf (open-dynamic-load path)))
	(if dlf
	    (let ((md (dynamic-load-lookup 
			    dlf 
			    (string-append "_module_" 
					   (symbol->string name)))))
		(install-module md)
		(let ((templates-to-fix-up (vector-ref image 1)))
		    (for-each
			(lambda (tmpl)
			    (fixup-part-descr (%gvec-read tmpl 1)))
			templates-to-fix-up)
		    (for-each
			fixup-template
			templates-to-fix-up)
		    (close-dynamic-load dlf)
		    (vector-ref image 0)))
	    (abort 'dynamic-load "could not load ~s" path))))

(define (fixup-part-descr part-descr)
    (if (not (%gvec-read part-descr 2))
	(%gvec-write 
	    part-descr 
	    2 
	    (module&tag->part 
		(name->module (%gvec-read part-descr 0)) 
		(%gvec-read part-descr 1))))
    part-descr)

(define (fixup-template tmpl)
    (let ((fn-ix (%gvec-read tmpl 0))
    	  (pd (%gvec-read (%gvec-read tmpl 1) 2)))
	(bind ((code fn-descr (part&index->function pd fn-ix)))
	    (%gvec-write tmpl 0 code)
	    (%gvec-write tmpl 1 fn-descr)
	    tmpl)))

(export call-gate %c-apply 
	gate->procedure raw-fn-ptr->procedure
	dynamic-load-code
	open-dynamic-load
	 dynamic-load-lookup 
	close-dynamic-load
	install-module name->module module&tag->part part&index->function
	make-c-ffi 
	fixup-part-descr 
	fixup-template)
