#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(prologue {
#include "socks.h"
})

(include-c "socks.c")
(include-h "socks.h")

(%define-glue (%write-str/SOCK sock str)
{
    sock_write( sock, string_text(str), string_length(str) );
    RETURN0();
})

(%define-glue (%write-bvec/SOCK sock bvec use_len)
{
    sock_write( sock, (const char *)PTR_TO_DATAPTR(bvec), fx2int(use_len) );
    RETURN0();
})

(%define-glue (%write-binary/SOCK sock val)
{
INT_32 i;

    assert (OBJ_ISA_FIXNUM(val));
    i = fx2int(val);
    sock_write( sock, (const char *)&i, sizeof(INT_32) );
    RETURN0();
})

(%define-glue (%read-str/SOCK sock)
{
char *s, temp[1000];
UINT_32 n;

    n = sock_read( sock, temp, 1000 );
    if (n == 0)
	REG0 = FALSE_OBJ;
    else
    {
	REG0 = bvec_alloc( n+1, string_class );
	s = (char *)string_text(REG0);
	memcpy( s, temp, n );
	/* s[n] = 0 because bvec_alloc() zero's out the last word
			    AND we allocated n+1 bytes */
    }
    RETURN(1);
})

(%define-glue (%read-bvec/SOCK sock bvec)
{
char *s = (char *)PTR_TO_DATAPTR(bvec);
UINT_32 n;

    n = sock_read( sock, s, SIZEOF_PTR(bvec) );
    REG0 = int2fx(n);
    RETURN(1);
})

(export %write-str/SOCK %write-bvec/SOCK %write-binary/SOCK 
	%read-str/SOCK %read-bvec/SOCK)

#| constructors |#

(define (open-named-socket name async?)
    (%call ("sock_open_named_socket" (<obj> <raw-bool>) => <obj>) name async?))

(define (close-named-socket sock)
    (%call ("sock_close_named_socket" (<obj>)) sock)
    (values))

(define (connect-to-name host-id name async?)
    (%call ("sock_connect_to_name" (<obj> <obj> <raw-bool>) => <obj>) 
    		host-id name async?))

(define (accept-connection named-sock async?)
    (%call ("sock_accept_connection" (<obj> <raw-bool>) => <obj>)
    	   named-sock async?))

(define (close-connection sock)
    (%call ("sock_close_connection" (<obj>)) sock)
    (values))

(define (collect-socket-data items)
    (%call ("sock_collect_data" (<obj>) => <obj>) items))

(export open-named-socket close-named-socket connect-to-name)
(export accept-connection close-connection collect-socket-data)
  
(define (hostname->host-id hostname)
    (let ((v (%call ("sock_hostname_to_hostid" (<obj>) => <obj>) hostname)))
	(if v
	    (values (vector-ref (%cast <vector> v) 0)
		    (vector-ref (%cast <vector> v) 1))
	    #f)))

(define (host-id->hostname host-id)
    (let ((v (%call ("sock_hostid_to_hostname" (<obj>) => <obj>) host-id)))
	(if v
	    (values (vector-ref v 0)
		    (vector-ref v 1))
	    #f)))

(define (host-id-eq? host1 host2)
    (%call ("sock_hostid_eq" (<obj> <obj>) => <raw-bool>) host1 host2))

(export host-id-eq? hostname->host-id host-id->hostname)
