#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define (getenv key)
    (%call ("os_getenv" (<raw-str>) => <raw-str>) key))

(define (mkdir path)
  ;; actually returns a <boolean>
  (%call ("os_make_directory" (<raw-str>) => <obj>) path))

(define (os-file-open path read? write? create? append? dontblock?)
  (%call ("os_open_file" (<raw-str>
			  <raw-bool>
			  <raw-bool>
			  <raw-bool>
			  <raw-bool>
			  <raw-bool>) => <obj>)
	 path read? write? create? append? dontblock?))

(define (os-file-seek file position)
  (%call ("os_file_seek" (<obj> <raw-int>) => <obj>) file position))
	 
(define (os-file-read-str file length)
  (%call ("os_file_read" (<obj> <obj> <raw-int>) => <obj>)
	 file
	 <string>
	 length))

(define (os-file-read-bvec file length)
  (%call ("os_file_read" (<obj> <obj> <raw-int>) => <obj>)
	 file
	 <bytevector>
	 length))

(define (os-file-write file thing)
  (%call ("os_file_write" (<obj> <obj>) => <obj>) file thing))

(define (getpid)
    (%call ("os_getpid" () => <obj>)))

#|
(define (sleep sec)
    ;(type-check number? <number> sec)
    (if (integer? sec) 
	(%call ("os_sleep" (<raw-int>)) (* sec 1000000))
	(%call ("os_sleep" (<raw-int>)) (round (* sec 1000000))))
    (values))

|#
(define (sleep sec)
    ;(type-check real? <real> sec)
    (if (> sec 500)
        (abort 'sleep "Max sleep time (500 sec) exceeded: ~s" sec)
        (begin
            (%call ("os_sleep" (<raw-int>)) 
                   (if (fixnum? sec)
                       (* sec 1000000)
                       (inexact->exact (round (* sec 1000000)))))
            (values))))

(define (time-now)
    (%call ("os_time_now" (<obj>) => <obj>) <Time>))

(define (system cmd)
    (%call ("os_invoke" (<obj>) => <obj>) cmd))

(define (fork)
    (%call ("os_fork_wait" () => <obj>)))

(define (fork-no-wait)
    (%call ("os_fork" () => <obj>)))

(define (wait child)
    (%call ("os_wait" (<obj>) => <raw-int>) child))

;; get-user-info
;; returns a vector of at least 2 elements
;   [0] user name (string)
;   [1] user id
;  if user is #f, info returned is for the effective user

(define (get-user-info user)
    (assert (or (eq? user #f) (string? user) (fixnum? user)))
    (%call ("os_get_user_info" (<obj>) => <obj>) user))

;; returns a vector of at least 2 intervals
;; (the contents of the vector beyond the first 2 elements
;;  is OS-specific)
;; (vector-ref (getrusage) 0)  ==> user time
;; (vector-ref (getrusage) 1)  ==> system time

(define (getrusage)
    (%call ("os_get_resource_usage" (<obj>) => <obj>) <Interval>))

(define (errorstr code)
    (%call ("os_error_string" (<raw-int>) => <obj>) code))

;; time-thunk calls thunk with zero arguments
;; and returns it's (first) value as the first result
;; the second return value is a vector containing:
;;	[0] user time used
;;	[1] real time used (wall clock)
;;	[2] system time used
;; the user & system time are taken from (getrusage)
;; the real time is taken from (time-now)

(define (time-thunk thunk)
    (let ((rt0 (time-now))
	  (ut0 (getrusage))
    	  (result (thunk))
	  (rt1 (time-now))
	  (ut1 (getrusage)))
	(values result
		(vector (interval- (vector-ref ut1 0) (vector-ref ut0 0))
			(time- rt1 rt0)
			(interval- (vector-ref ut1 1) (vector-ref ut0 1))))))

(export getenv getpid sleep time-now time-thunk
	system fork fork-no-wait wait getrusage errorstr
	get-user-info)

(export mkdir 
	os-file-open os-file-seek 
	os-file-read-str os-file-read-bvec 
	os-file-write)

