#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define-class <Time> (<Object>) :bvec image-mode: 7)
(define-class <Interval> (<Object>) :bvec image-mode: 7)

(define (time? thing)
    (isa? thing <Time>))

(define (interval? thing)
    (isa? thing <Interval>))

(define (time->string t . fmt)
    (if (null? fmt)
	(%call ("time_to_string" (<obj> <obj>) => <obj>) t #f)
	(%call ("time_to_string" (<obj> <obj>) => <obj>) t (car fmt))))

(define-method write ((self <Time>) port)
    (format port "#[<Time> ~a]" (time->string self)))

(define-method display ((self <Time>) port)
    (display (time->string self) port))


(define-method write ((self <Interval>) port)
    (format port "#[<Interval> ~a]" (interval->string self)))

(define-method display ((self <Interval>) port)
    (display (interval->string self) port))

(%define-glue (%interval->seconds dt)
{
struct scheme_time raw_dt = extract_schemetime(dt);

    if (raw_dt.usec == 0 && (raw_dt.sec < (1<<29)))
	REG0 = int2fx(raw_dt.sec);
    else
	REG0 = make_float( raw_dt.usec * 1.0e-06 + raw_dt.sec );
    RETURN(1);
})

(define (interval->seconds dt)
    (type-check dt interval? <Interval>)
    (%interval->seconds dt))

(%define-glue (seconds->interval t) literals: ((& <Interval>))
{
struct scheme_time raw_dt;
IEEE_64 s;

    if (OBJ_ISA_FIXNUM(t))
    {
        raw_dt.sec = fx2int(t);
	raw_dt.usec = 0;
    }
    else
    {
	s = extract_float(t);
	raw_dt.sec = (int) s;
	raw_dt.usec = (int) ((s * 1.0e6) - (raw_dt.sec * 1.0e6));
    }
    REG0 = make_time( raw_dt, TLREF(0) );
    RETURN(1);
})


(%define-glue (time+ base_time interval) literals: ((& <Time>))
{
struct scheme_time t = timeadd( base_time, interval );

    REG0 = make_time( t, TLREF(0) );
    RETURN(1);
})

(%define-glue (time- base_time target_time) literals: ((& <Interval>))
{
struct scheme_time t = timesub( base_time, target_time );

    REG0 = make_time( t, TLREF(0) );
    RETURN(1);
})

(%define-glue (interval- iv1 iv2) literals: ((& <Interval>))
{
struct scheme_time t = intervalsub( iv1, iv2 );

    REG0 = make_time( t, TLREF(0) );
    RETURN(1);
})

(%define-glue (interval->string interval)
{
struct scheme_time t = extract_schemetime(interval);
char *p, temp[50];
double sec;

    p = temp;
    if (t.sec < 0)
    {
        *p++ = '-';
	if (t.usec == 0)
	    t.sec = -t.sec;
	else
	{
	    t.sec = -(t.sec + 1);
	    t.usec = 1000000 - t.usec;
	}
    }

    if (t.sec == 0)
    {
	if (t.usec < 1000)
	{
	    if (t.usec == 0)
	    {
		p[0] = '0';
		p[1] = 0;
	    }
	    else
		sprintf( p, "%d us", t.usec );
	}
	else
	    sprintf( p, "%.3g ms", t.usec / 1000.0 );
    }
    else
    {
	sec = t.sec + t.usec * 1.0e-6;
	if (sec < 60.0)
	{
	    sprintf( p, "%.3g s", sec );
	}
	else if (sec < 3600.0)
	{
	    sprintf( p, "%.3g m", sec / 60.0 );
	}
	else if (sec < 24.0 * 3600.0)
	{
	    sprintf( p, "%.3g h", sec / 3600.0 );
	}
	else
	{
	    sprintf( p, "%.3g d", sec / (24.0 * 3600.0) );
	}
    }
    REG0 = make_string( temp );
    RETURN(1);
})

(%define-glue (time<? t1 t2)
{
    REG0 = rb_to_bo( timecmp(t1,t2) < 0 );
    RETURN(1);
})

(%define-glue (time<=? t1 t2)
{
    REG0 = rb_to_bo( timecmp(t1,t2) <= 0 );
    RETURN(1);
})

(%define-glue (time>? t1 t2)
{
    REG0 = rb_to_bo( timecmp(t1,t2) > 0 );
    RETURN(1);
})

(%define-glue (time>=? t1 t2)
{
    REG0 = rb_to_bo( timecmp(t1,t2) >= 0 );
    RETURN(1);
})

(%define-glue (time=? t1 t2)
{
    REG0 = rb_to_bo( timecmp(t1,t2) == 0 );
    RETURN(1);
})

(export time=? time<? time>? time<=? time>=?)
(export time->string interval->string)

(export interval- time+ time-)
(export seconds->interval interval->seconds)
(export <Interval> <Time>)

(export interval? time?)

