/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#ifndef _H_OSGLUE
#define _H_OSGLUE

#include "obj.h"
#include <stdio.h>

void init_os( void );

/* return a 31-bit random number */

UINT_32 os_random( void );

/*  open a file
    (do pathname localization, so I can say
     os_fopen("foo/bar") and on a Mac get fopen("foo:bar")
*/

FILE *os_fopen( const char *path, const char *mode );

const char *os_getenv( const char *key );
obj os_scan_directory( const char *path );

obj os_getwd( void );
void os_setwd( const char *path );

/*  return the pathname for a "user configuration" file
    (e.g., "~/.<basename>rc"
         or "\config\<basename>.ini")
*/

obj os_user_config_file( obj basename );

/*
    return a list of directories to be searched for
    system library files
*/

obj os_library_path( void );

/* note: process identifiers can be arbitrary objects, but they
         must exhibit object identity (ie, two PID's for the
	 same process must be eq?)
*/

obj os_getpid( void );

/*  send a command to the system shell. 
    Returns the integer return code of the command
    or #f if the system shell could not be invoked.
*/

obj os_invoke( obj command );

/*
    returns a string describing an OS error (errno)
    or #f if one cannot be ascertained
*/

obj os_error_string( int rc );

obj os_get_resource_usage( obj i_class );
obj os_file_status( const char *path, obj t_class );
obj os_get_user_info( obj usr );

/*  return YES or NO accordingly as whether the named
    file exists
*/

bool os_file_exists_p( const char *path );
obj os_fork_wait( void );  	/* return FALSE_OBJ to child, 
				     exit code to parent */
obj os_fork( void );		/* return FALSE_OBJ to child, 
				     child's PID to parent */
int os_wait( obj child );	/* return exit code */

void os_sleep( UINT_32 usec );

obj os_time_now( obj t_class );

/*
   arrange for some interrupts...
*/

void os_set_timer( UINT_32 msec );
UINT_32 os_get_time_remaining( void );
UINT_32 os_halt_timer( void );

/*
   low-level file manipulation
*/

obj os_open_file(const char *path,
		 bool read_p, bool write_p, bool create_p, 
		 bool append_p, bool nonblock_p );
obj os_file_write( obj file, obj data );
obj os_file_read( obj file, obj target_class, UINT_32 len );
obj os_file_seek( obj file, UINT_32 place );
obj os_make_directory( const char *path );

#endif /* _H_OSGLUE */
