/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#define template _their_template
#include "socks.h"
#include "runtim.h"
#include <string.h>
#ifdef __cplusplus
extern "C" {
#endif
#ifdef PLATFORM_NEXT
# include <bsd/netdb.h>
# include <libc.h>
#else
# include <netdb.h>
# include <arpa/inet.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <netinet/in.h>
#ifdef __cplusplus
#ifdef PLATFORM_NEXT
extern unsigned long inet_addr( char *cp );
extern char *inet_ntoa ( struct in_addr in );
#endif
}
#endif
#ifdef PLATFORM_LINUX
#include <termios.h>
#endif
#ifdef SUBPLATFORM_SOLARIS
#include <sys/file.h>
#endif

#include "intrs.h"

#define MAKE_VECTOR_2(a,b) make2( vector_class, a, b )

#ifdef PLATFORM_NEXT /* somehow, this is getting lost... */
#undef bzero
#define bzero(b,len) memset(b,0,len)
#endif

#ifdef PLATFORM_AIX
/* #include <sys/socketvar.h> */
#include <sys/select.h>
#undef bzero
#define bzero(b,len) memset(b,0,len)
#ifdef __cplusplus
extern "C" {
#endif
extern int socket( int AF, int TYPE, int PROTOCOL );
extern int bind( int fd, struct sockaddr *addr, int addr_len );
extern int listen( int fd, int len );
extern int accept( int fd, struct sockaddr *addr, int *addr_len );
extern int connect( int fd, struct sockaddr *addr, int addr_len );
#ifdef __cplusplus
}
#endif
#endif

#ifndef TRUE
#define TRUE (1)
#define FALSE (0)
#endif


static void register_listen_to( int fd );


obj sock_open_named_socket( obj name, bool async )
{
int s, n, rc;
struct sockaddr_in addr;

    s = socket( PF_INET, SOCK_STREAM, 0 );
    if (s < 0)
    {
	perror( "socket" );
	return FALSE_OBJ;
    }
    addr.sin_family = PF_INET;
    addr.sin_port = fx2int(name);
    addr.sin_addr.s_addr = INADDR_ANY;
    memset( addr.sin_zero, 0, 8 );
    
    n = sizeof addr;
    rc = bind( s, (struct sockaddr *)&addr, n );
    if (rc < 0)
    {
	perror( "bind" );
	close(s);
	return FALSE_OBJ;
    }
    rc = listen( s, 3 );
    if (rc < 0)
    {
	perror( "listen" );
	close(s);
	return FALSE_OBJ;
    }
    if (async)
	register_listen_to( s );
	
    return int2fx(s);
}

void sock_close_named_socket( obj named_socket )
{
    close( fx2int(named_socket) );
}


obj sock_connect_to_name( obj host_id, obj name, bool async )
{
struct sockaddr_in addr;
int fd, n, rc;

    fd = socket( PF_INET, SOCK_STREAM, 0 );

    addr.sin_family = PF_INET;
    addr.sin_port = fx2int(name);
    addr.sin_addr.s_addr = inet_addr( (char *)string_text(host_id) );
    memset( addr.sin_zero, 0, 8 );
    
    n = sizeof addr;
    rc = connect( fd, (struct sockaddr *)&addr, n );
    if (rc != 0)
	return FALSE_OBJ;
    if (async)
	register_listen_to( fd );
    return MAKE_VECTOR_2( int2fx(fd), host_id );
}

obj sock_accept_connection( obj named_socket, bool async )
{
int n, fd;
struct sockaddr_in addr;

    n = sizeof addr;
    fd = accept( fx2int(named_socket), (struct sockaddr *)&addr, &n );
    if (fd < 0)
	return FALSE_OBJ;
    if (async)
	register_listen_to( fd );
    return MAKE_VECTOR_2( int2fx(fd), 
		          make_string( inet_ntoa( addr.sin_addr ) ) );
}

static obj collect_aliases( struct hostent *h )
{
obj names, prev_name;
char **alias;

    if (!h)
	return FALSE_OBJ;
	
    names = cons( make_string( h->h_name ), NIL_OBJ );
    prev_name = names;
    for (alias=h->h_aliases; *alias; alias++)
    {
    obj n;
    
	n = cons( make_string( *alias ), NIL_OBJ );
	gvec_write_fresh_ptr( prev_name, SLOT(1), n );
	prev_name = n;
    }
    return MAKE_VECTOR_2( 
	    make_string( inet_ntoa( *(struct in_addr *)*h->h_addr_list )),
	    names );
}

obj sock_hostname_to_hostid( obj host_name )
{
struct hostent *h;

    h = gethostbyname( (char *)string_text(host_name) );
    return collect_aliases( h );
}

obj sock_hostid_to_hostname( obj hostid )
{
struct in_addr addr;
struct hostent *h;

    addr.s_addr = inet_addr( (char *)string_text(hostid) );
    h = gethostbyaddr( (char *)&addr, 4, AF_INET );
    return collect_aliases( h );
}

bool sock_hostid_eq( obj hostid1, obj hostid2 )
{
    return string_eq( hostid1, hostid2 );
}



void sock_write( obj connection, const char *text, UINT_32 length )
{
int fd = fx2int(gvec_read(connection,SLOT(0)));

    write( fd, text, length );
}

UINT_32 sock_read( obj connection, char *buffer, UINT_32 max_len )
{
int fd = fx2int(gvec_read(connection,SLOT(0)));
int n;

    n = read( fd, buffer, max_len );
    if (n < 0)
	scheme_error( "sock_read: return code ~d", 1, int2fx(n) );
    return n;
}

void sock_close_connection( obj connection )
{
int fd = fx2int(gvec_read(connection,SLOT(0)));

    close( fd );
}

#if defined(PLATFORM_SUNOS) && !defined(SUBPLATFORM_SOLARIS)
static void notice_input( int sig, int code, struct sigcontext *s, char *a )
#else
static void notice_input( int sig )
#endif
{
    /* printf( "\nNoticed input!\n" ); */
    interrupt_pending = YES;
    interrupt_flag[ INTR_FLAG_DATA_READY ] = TRUE_OBJ;
}

static void register_listen_to( int fd )
{
int rc;
static int setup_signal_q = FALSE;

    if (!setup_signal_q)
    {
	setup_signal_q = TRUE;
	signal( SIGIO, notice_input );
    }
#ifdef PLATFORM_LINUX
    rc = fcntl( fd, F_SETFL, FIOASYNC );
#else
    rc = fcntl( fd, F_SETFL, FASYNC );
#endif
    if (rc != -1)
    {
	rc = fcntl( fd, F_SETOWN, getpid() );
    }
}


/* items is a list of objects returned from
   sock_open_named_socket, sock_connect_to_name,
   or sock_accept_connection with async = YES
*/

#define INPUT_SET_SIZE 128

obj sock_collect_data( obj items )
{
fd_set inp;
obj item, result;
int i, n, num_fds, fds[INPUT_SET_SIZE];
obj itemvec[INPUT_SET_SIZE];
struct timeval t = { 0, 0 };

    num_fds = 0;
    FD_ZERO( &inp );
    
    while (!EQ(items,NIL_OBJ))
    {
	item = pair_car( items );
	if (OBJ_ISA_FIXNUM(item))
	    i = fx2int(item);
	else
	    i = fx2int( gvec_read( item, SLOT(0) ) );
	    
	FD_SET( i, &inp );
	itemvec[num_fds] = item;
	fds[num_fds++] = i;
	items = pair_cdr( items );
    }
    n = select( INPUT_SET_SIZE, &inp, NULL, NULL, &t);
    if (n <= 0)
	return NIL_OBJ;

    result = NIL_OBJ;
    while (num_fds)
    {
	i = fds[--num_fds];
	if (FD_ISSET(i,&inp))
	{
	    result = cons( itemvec[num_fds], result );
	}
    }
    return result;
}
