/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "mapf.h"
#include <stdio.h>
#include <assert.h>

struct image_buffer {
    void *block;
    struct image_buffer *next;
};

static struct image_buffer *image_buffers;
static FILE *image_file;

bool mapf_open( const char *path )
{
    image_file = fopen( path, "rb" );
    if (!image_file)
    {
	perror( path );
	return NO;
    }
    image_buffers = NULL;
    return YES;
}

void mapf_seek( UINT_32 offset )
{
int rc;

    rc = fseek( image_file, offset, SEEK_SET );
    assert( rc == 0 );
}

void *mapf_read( UINT_32 bytes )
{
struct image_buffer *b = (struct image_buffer *)malloc( sizeof( struct image_buffer ) );
int n;

    b->next = image_buffers;
    image_buffers = b;
    b->block = malloc( bytes );
    assert( b->block );
    n = fread( b->block, 1, bytes, image_file );
    assert( n == bytes );
    return b->block;
}

void mapf_close( void )
{
struct image_buffer *b, *n;

    for (b=image_buffers; b; b=n)
    {
	n = b->next;
	free( b->block );
	free( b );
    }
}

