/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include <string.h>
#include <signal.h>
#include <time.h>
#include "runtim.h"
#include "scmtime.h"

static void oslib_error( const char *fn, int rc )
{
    scheme_error( "OS library error in ~s: rc ~d",
    		  2, make_string(fn), int2fx(rc) );
}

static void oslib_ni( const char *fn )
{
    scheme_error( "OS library function not implemented: ~s",
		 1,
		 make_string(fn) );
}

static obj os_time( clock_t c, obj t_class )
{
struct scheme_time t;

    t.sec = c / CLOCKS_PER_SEC;
    t.usec = (c % CLOCKS_PER_SEC) * (1000000 / CLOCKS_PER_SEC);
    return make_time( t, t_class );
}

obj os_time_now( obj t_class )
{
struct scheme_time st;

    st.sec = time(NULL);
    st.usec = 0;
    return make_time( st, t_class );
}

const char *os_getenv( const char *key )
{
  return getenv(key);
}

obj os_scan_directory( const char *path )
{
  oslib_ni( "os_scan_directory" );
  return FALSE_OBJ;
}

obj os_getwd( void )
{
  oslib_ni( "os_getwd" );
  return FALSE_OBJ;
}

obj os_getpid( void )
{
  oslib_ni( "os_getpid" );
  return FALSE_OBJ;
}

obj os_invoke( obj command )
{
int rc;

    OUT_CALL( rc = system( string_text( command ) ); );
    return int2fx( rc );
}

obj os_error_string( int rc )
{
  return FALSE_OBJ;
}

obj os_get_resource_usage( obj i_class )
{
  oslib_ni( "os_get_resource_usage" );
  return FALSE_OBJ;
}

obj os_make_directory( const char *path )
{
  oslib_ni( "os_make_directory" );
  return FALSE_OBJ;
}

obj os_file_seek( obj file, UINT_32 offset )
{
  oslib_ni( "os_file_seek" );
  return FALSE_OBJ;
}

obj os_file_read( obj file, obj bvec_class, UINT_32 bytes )
{
  oslib_ni( "os_file_read" );
  return FALSE_OBJ;
}

obj os_file_write( obj file, obj data )
{
  oslib_ni( "os_file_write" );
  return FALSE_OBJ;
}

obj os_open_file( const char *path,
		  bool read_p, bool write_p, bool create_p, 
		  bool append_p, bool nonblock_p )
{
  oslib_ni( "os_getpid" );
  return FALSE_OBJ;
}

bool os_file_exists_p( const char *path )
{
FILE *f = fopen( path, "r" );

    if (f)
      {
	fclose(f);
	return YES;
      }
    else
      return NO;
}

obj os_file_status( const char *path, obj t_class )
{
  oslib_ni( "os_file_status" );
  return FALSE_OBJ;
}

void os_setwd( const char *path )
{
  oslib_ni( "os_setwd" );
}

obj os_fork( void )
{
  oslib_ni( "os_fork" );
  return FALSE_OBJ;
}

int os_wait( obj child )
{
  oslib_ni( "os_wait" );
  return FALSE_OBJ;
}

obj os_fork_wait( void )
{
  oslib_ni( "os_fork_wait" );
  return FALSE_OBJ;
}

obj os_get_user_info( obj usr )
{
  oslib_ni( "os_get_user_info" );
  return FALSE_OBJ;
}

/* convert a unix-standard pathname to local OS format */

obj os_path( obj path )
{
    return path;
}

/* open a file using unix-std pathnames */

FILE *os_fopen( const char *path, const char *mode )
{
  return fopen( path, mode );
}

/* TIMER_IS_MONOTONE_COUNTER is required for ANSI */
extern UINT_32 system_timeout;

void os_set_timer( UINT_32 msec )
{
    system_timeout = msec;
}

UINT_32 os_get_time_remaining( void )
{
    return system_timeout;
}

UINT_32 os_halt_timer( void )
{
UINT_32 t = system_timeout;

    system_timeout = 0;
    return t;
}


void init_os( void )
{
}

obj os_user_config_file( obj basename )
{
char temp[1024];

    sprintf( temp, "~/.%src", string_text(basename) );
    return make_string(temp);
}

obj os_library_path( void )
{
    return cons( make_string( "~/lib/es" ),
    		 cons( make_string( "/usr/local/lib/es" ),
			NIL_OBJ ) );
}

/*  
    certain functions rely on the range of random
    numbers being fairly large (they will spin
    indefinitely otherwise).
    We'd use random() if it were standard...
*/

UINT_32 os_random( void )
{
    return (rand() << 16) ^ rand();
}

void os_sleep( UINT_32 usec )
{
    sleep( usec / 1000000 );
}


