/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#ifndef _H_PLATFORM
#define _H_PLATFORM

/* platform-independence layer */

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <alloca.h>
#include <netinet/in.h>

/*  Linux (Intel)  */

#ifdef INLINES
#define CI_DECL static inline
#define CIH_DECL static inline
#else
#define CI_DECL /* nothing */
#define CIH_DECL /* nothing */
#endif

#define PLATFORM_LINUX

/* typedef unsigned long long UINT_64; */
typedef unsigned UINT_32;
typedef unsigned short UINT_16;
typedef unsigned char UINT_8;

/* typedef long long INT_64; */
typedef int INT_32;
typedef short INT_16;
typedef signed char INT_8;

typedef double IEEE_64;
typedef float IEEE_32;

typedef enum { NO = 0, YES = !NO } bool;

#define PLATFORM_IS_LITTLE_ENDIAN	/* define this to get htonf, etc. */
IEEE_32 htonf( IEEE_32 f );
IEEE_32 ntohf( IEEE_32 f );
IEEE_64 htond( IEEE_64 d );
IEEE_64 ntohd( IEEE_64 d );
#define HOST_TO_BIG_ENDIAN_32(x) htonl(x)
#define HOST_TO_BIG_ENDIAN_16(x) htons(x)
#define HOST_TO_BIG_ENDIAN_IEEE_32(x) htonf(x)
#define HOST_TO_BIG_ENDIAN_IEEE_64(x) htond(x)

#define BIG_ENDIAN_TO_HOST_32(x) ntohl(x)
#define BIG_ENDIAN_TO_HOST_16(x) ntohs(x)
#define BIG_ENDIAN_TO_HOST_IEEE_32(x) ntohf(x)
#define BIG_ENDIAN_TO_HOST_IEEE_64(x) ntohd(x)

/*
    The sign convention for the C operator `%' is
    implementation dependent, so REMDR insulates us from that, and
    promises to use the convention defined by Scheme,
    which apparently is that the remainder always has the sign
    of the dividend (the first argument)
*/

/*
    "Remainder and modulo differ on negative arguments ---
     the remainder is either zero or has the sign of the dividend,
     while the modulo always has the sign of the divisor:"
    
				    -- Revised^4 Scheme, p.22
*/

#define UMOD(x,y) ((int)((unsigned)x % (unsigned)y))
#define XFER_SIGN(x,y) ((y < 0) ? -x : x)

#define REMDR(x,y) XFER_SIGN(UMOD(x,y),x)
#define MOD(x,y) XFER_SIGN(UMOD(x,y),y)

#endif


