/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "easyscm.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/*
    Rob Browning. Fall, 1993.
*/

/*
This seems like a good short-term solution, except (and I hate to say this)
we should use a static buffer to assuage the possibility of a memory leak.
And we should call it unix_path_to_native_path to be more consistent with
naming convention and style.

Also, don't forget to handle the unix "..", which turns into something like
"::".   For example (and this is something that will happen):

unix_path_to_native_path( "../foo/bar.scm" ) ==> "::foo:bar.scm"
unix_path_to_native_path( "quux/../foo/bar.scm" ) ==> ":quux::foo:bar.scm"
unix_path_to_native_path( "/tmp/temp.scm" ) ==> "tmp:temp.scm"

NOTE:  a reference to a DIRECTORY will always end in "/"
       (ie, via (chdir ...) and possibly in the future (opendir ...))
       a root directory ("/") is never used

../   =>  ::
../../ => :::
../foo/./bar  => ::foo:bar
../foo/../bar ==> ::foo::bar
../foo/../bar/ ==> ::foo::bar:
/foo/bar/../baz ==> foo:bar::baz
../foo ==> ::foo
foo ==> foo       (":foo" might be OK, but not preferable)
foo/bar ==> :foo:bar
foo/bar/./baz/../bleen ==> :foo:bar:baz::bleen

*/

void unix_path_to_native_path(char* path) {
        char* orig_str = (char *) malloc(strlen(path)+1);
        int i=0,j=0,in;

        strcpy(orig_str,path);

        /* Check to see if just a file name */
        if(!strpbrk(orig_str,"/")) return;

        if(orig_str[i]!='/') {
                path[j++]=':';
                while(1) {
                        if(orig_str[i]=='/') i++;
                        if(strncmp(&(orig_str[i]),"./",2)==0) {
                                i+=2;
                        }
                        else if(strncmp(&(orig_str[i]),"../",3)==0) {
                                path[j++]=':';
                                i+=3;
                        }
                        else break;
                }
        }
        else i++;

        while(orig_str[i]!='\0') {
                if(orig_str[i]=='/') {
                        path[j++]=':';
                        i++;
                        while(1) {
                                if(orig_str[i]=='/') i++;
                                if(strncmp(&(orig_str[i]),"./",2)==0) {
                                        i+=2;
                                }
                                else if(strncmp(&(orig_str[i]),"../",3)==0) {
                                        path[j++]=':';
                                        i+=3;
                                }
                                else break;
                        }
                }
                else path[j++]=orig_str[i++];
        }
        path[j]='\0';

        free(orig_str);
}

#if 0
void unix_path_to_native_path(char* path) {
        char* orig_str = (char *) malloc(strlen(path)+1);
        int i,j;

        strcpy(orig_str,path);

        i=0; j=0;

        if(orig_str[i]=='/') i++; /* This is an absolute path */
        else if(strncmp(orig_str,"../",3)==0) {
                path[j++] = ':';
                path[j++] = ':';
                i+=3;
        }
        else if(strncmp(orig_str,"./",2)==0) {
                path[j++] = ':';
                i+=2;
        }
        else {
                path[j++] = ':';
        }

        while(orig_str[i]!='\0') {
                if(strncmp(&(orig_str[i]),"/../",4)==0) {
                        /* Move up a level */
                        path[j++]=':';
                        path[j++]=':';
                        i+=4; /* Skip the ../ */
                }
                else if(strncmp(&(orig_str[i]),"/./",3)==0) {
                        path[j++]=':';
                        i+=3;
                }
                else if(orig_str[i]=='/') {
                        path[j++]=':';
                        i++;
                }
                else path[j++]=orig_str[i++];
        }
        path[j]='\0';

        free(orig_str);
}
#endif

FILE* fopen_unix_path(const char* filename,const char* mode) {
        FILE* opened_file;
        char* translated_name = (char *) malloc(strlen(filename)+12);
        /* The 12 is just a safety margin; I think that 2 would be sufficient */
        /* The only time a mac path could be larger is when the path is
relative. */
        /* The mac adds a : on the front then. */

        if(!translated_name) {
                fprintf(stderr, "Easy Scheme: Allocation failed in
fopen_unix_path.\n");
                exit(-1);
        }
        strcpy(translated_name,filename);
        unix_path_to_native_path(translated_name);
        opened_file = fopen(translated_name,mode);
        free(translated_name);
        return opened_file;
}

#if 0

void main() {
        int i;

        char paths[][2][64] = {
                {"../foo/bar.scm",                      "::foo:bar.scm"},
                {"quux/../foo/bar.scm",         ":quux::foo:bar.scm"},
                {"/tmp/temp.scm",                       "tmp:temp.scm"},
                {"../",                                         "::"},
                {"../../",                                      ":::"},
                {"../foo/./bar",                        "::foo:bar"},
                {"../foo/../bar",                       "::foo::bar"},
                {"../foo/../bar/",                      "::foo::bar:"},
                {"/foo/bar/../baz",                     "foo:bar::baz"},
                {"../foo",                                      "::foo"},
                {"foo",                                         "foo"},
/*(":foo" might be OK, but not preferable)*/
                {"foo/bar",                                     ":foo:bar"},
                {"foo/bar/./baz/../bleen"       ":foo:bar:baz::bleen"},
                {NULL,                                          NULL}};

        i=0;
        for(i=0;paths[i][1][0]!='\0';i++) {
                unix_path_to_native_path(paths[i][0]);
                printf("%s\n%s\n\n",paths[i][1],paths[i][0]);
        }
}

#endif
