/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "mapf.h"
#ifdef __cplusplus
extern "C" {
#endif
#include <mach/mach.h>
/* #include <bsd/libc.h> */
#include <sys/file.h>

int open( const char *path, int flags, int mode );
int close( int fd );
long lseek( int, long, int );

char *getsectdata( const char *segname, const char *sectname, int *size );
extern kern_return_t map_fd(int fd, vm_offset_t offset, 
	vm_offset_t *addr, boolean_t find_space, vm_size_t numbytes);

#ifdef __cplusplus
}
#endif
#include <string.h>

static vm_address_t image_mapping_addr;
static vm_size_t image_mapping_size;
static void *image_mapping_ptr;
static int image_file;

bool mapf_open( const char *path )
{
kern_return_t rc;

    if (strncmp(path,"sect:",5) == 0)
    {
    char *s, name[1000];
    
	strcpy( name, path+5 );
	s = strchr( name, ':' );
	*s++ = 0;
	
	image_file = -1;
	image_mapping_addr = 
	    (vm_address_t)getsectdata( name, s, (int *)&image_mapping_size );
	if (!image_mapping_addr)
	{
	    fprintf( stderr, "could not map segment %s, section %s\n",
	    			name, s );
	    return NO;
	}
    }
    else
    {
	image_file = open( path, O_RDONLY, 0 );
	if (image_file < 0)
	{
	    perror( path );
	    return NO;
	}
	image_mapping_size = lseek( image_file, 0L, SEEK_END );
	
	rc = map_fd( image_file, 
		    (vm_offset_t)0, 
		    &image_mapping_addr, 
		    /* find_space */ TRUE,
		    image_mapping_size );
	if (rc != KERN_SUCCESS)
	{
	    fprintf( stderr, "could not map file: %s\n", path );
	    close( image_file );
	    return NO;
	}
    }
    image_mapping_ptr = (void *)image_mapping_addr;
    return YES;
}

void mapf_seek( UINT_32 offset )
{
    image_mapping_ptr = (void *)(image_mapping_addr + offset);
}

void *mapf_read( UINT_32 bytes )
{
void *p = image_mapping_ptr;

    image_mapping_ptr += bytes;
    return p;
}

void mapf_close( void )
{
    if (image_file >= 0)
    {
	vm_deallocate( task_self(), image_mapping_addr, image_mapping_size );
	close( image_file );
    }
}

