/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <streams/streams.h>
#include <mach-o/rld.h>
#ifdef __cplusplus
}
#endif
#include "schemedl.h"

struct resolved {
    struct resolved *next;
    const char *sym;
};

struct dlfile {
    struct resolved *lookups;
    const char *path;
};

static const char *strdup( const char *str )
{
char *p = (char *)malloc( strlen(str)+1 );

    strcpy( p, str );
    return p;
}

static NXStream *errors = NULL;

void *resolve_link_symbol( void *info, const char *sym )
{
unsigned long val, rc;
struct resolved *i;
struct dlfile *f = (struct dlfile *)info;

    if (info)
    {
	i = (struct resolved *)malloc( sizeof( struct resolved ) );
	i->sym = strdup( sym );
	i->next = f->lookups;
	f->lookups = i;
    }

    if (!errors)
	errors = NXOpenFile( 2, NX_WRITEONLY );

    rc = rld_lookup( errors, sym, &val );
    NXFlush(errors);
    return (rc == 1) ? (void *)val : NULL;
}

void *dynamic_link_file( const char *path )
{
const char *(files[2]) = { path, NULL };
unsigned long rc;

    if (!errors)
	errors = NXOpenFile( 2, NX_WRITEONLY );

    rc = rld_load( errors, NULL, files, NULL );
    NXFlush(errors);
    if (rc == 1)
    {
    struct dlfile *f = (struct dlfile *)malloc( sizeof( struct dlfile ) );
    
	f->lookups = NULL;
        f->path = strdup( path );
	return f;
    }
    return NULL;
}

void done_resolving( void *info )
{
struct dlfile *f = (struct dlfile *)info;
struct resolved *i, *j;

    NXPrintf( errors, "Done with: %s\n", f->path );
    for (i=f->lookups; i; i=j)
    {
	NXPrintf( errors, "    forgetting %s\n", i->sym );
	rld_forget_symbol( errors, i->sym );
	j = i->next;
	free( (void *)i->sym );
	free( (void *)i );
    }
    rld_unload_all( errors, 0 );
    NXFlush( errors );
    free( (void *)f->path );
    free( (void *)f );
}

void init_dynamic_link( const char *argv0 )
{
}
