/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
/*
    read-line

    Scheme interface to the GNU readline library
*/

#include "platform.h"
#ifdef __cplusplus
extern "C" {
#endif
#include <readline/readline.h>
#include <readline/history.h>
#ifdef __cplusplus
}
#endif

#ifdef RETURN	/* this gets defined in readline/chardefs.h */
#undef RETURN
#endif

#include "runtim.h"
#include "vinsns.h"
#include <string.h>

static const char *scheme_generator( char *text, int state );

static obj the_completions;

/* Strip whitespace from the start and end of STRING. */

static void stripwhite( char *string )
{
int i = 0;

  while (whitespace (string[i]))
    i++;

  if (i)
    strcpy (string, string + i);

  i = strlen (string) - 1;

  while (i > 0 && whitespace (string[i]))
    i--;

  string[++i] = '\0';
}

/* **************************************************************** */
/*                                                                  */
/*                  Interface to Readline Completion                */
/*                                                                  */
/* **************************************************************** */

/* Tell the GNU Readline library how to complete.  We want to try to complete
   on command names if this is the first word in the line, or on filenames
   if not. */
   
static void init_readline( void )
{
  /* Allow conditional parsing of the ~/.inputrc file. */
    rl_readline_name = "easyscm";
    rl_basic_word_break_characters = " \t\n\"\\()[]{},";

  /* Tell the completer that we want a crack first. */
  
    rl_completion_entry_function = (Function *)scheme_generator;
}

static const char *scheme_generator( char *text, int state )
{
static obj current;
static int len;
obj item;
const char *name;

    if (state == 0)	/* restarting generation */
    {
	current = the_completions;
	len = strlen( text );
    }
    
    while (!EQ( current, NIL_OBJ ))
    {
        assert( PAIR_P(current) );
	item = pair_car( current );
	current = pair_cdr( current );

	if (STRING_P(item))
	{
	    name = string_text(item);
	}
	else
	{
	    assert( SYMBOL_P(item) );
	    name = symbol_text(item);
	}

	if (strncmp( name, text, len ) == 0)
	{
	char *name2;
	
	    name2 = (char *)malloc( strlen( name ) + 1 );
	    strcpy( name2, name );
	    return name2;
	}
    }
    return NULL;
}

const char *read_console_line( obj completions, const char *prompt )
{
static char *buffer = NULL;
static bool init_yet = NO;

    if (!init_yet)
	init_readline();
	
    if (buffer)
	free( buffer );

    the_completions = completions;

    buffer = readline( prompt );
    if (!buffer)
	return NULL;

    /* put it in the history buffer */

    stripwhite( buffer );
    if (*buffer)
	add_history( buffer );
    return buffer;
}
