#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
#|
    <string_matcher> ::= <sequence>
    		       | <choice>
		       | <kleene-star>
		       | <kleene-plus>
		       | <zero-or-one>
		       | <collector>
		       | <range>
		       | <literal>
    <sequence> := ( <string_matcher> ... )
    <choice> := ( or <string_matcher> ... )
    <kleene-star> := ( * <string_matcher> )
    <kleene-plus> := ( + <string_matcher> )
    <zero-or-one> := ( ? <string_matcher> )
    <collector> := ( <identifier> = <string_matcher> )
    <range> := ( <CHAR> <DOTDOTDOT> <CHAR> )
    <literal> := <STRING> | <CHAR>


  (* ( #\[ (foo = (* ( or #\a #\b ))) #\]))
    when matching
	"[aba][ba][bbb]"
    returns
    	((foo . ("aba" "ba" "bbb")))
    because the lexical structure of "foo" is that it is inside
    a <kleen-star>.  Its value, if present, will always be a list,
    possibly empty (or is it possible to be empty? -- then it wouldn't
    be present at all!).
  
  (or (#\[ (foo = (* (or #\a #\b))) #\])
      (#\{ (bar = (* (or #\a #\b))) #\}))

    then:	"[aab]"		=>	((foo . "aab"))
		"{aab}"		=>	((bar . "aab"))

    however:
       (* (or (foo = (or #\+ #\-))
       	      (bar = (or #\x #\y))))
    can get wierd:
    		"+x-"		=>	((foo . ("+" "-")) (bar . ("x")))

    and:
    	(* (or (foo = (or #\+ #\-))
	       (#\[ (* (foo = (or #\a #\b))) #\])))
    can get non-uniform:
    		"+[aba]-"		((foo . ("+" ("a" "b" "a") "-")))
|#

(define-syntax (debug-out msg . args))
;(define-syntax (debug-out msg . args) (format #t msg . args))

(define (find-best-disjunct patterns chars)
    (let ((best #f)
    	  (best-length 0))
	(debug-out "|^^^disjunct at ~s\n" chars)
	(for-each
	    (lambda (pat)
		(debug-out "|---disjunct ~s at ~s\n" pat chars)
		(let (((result <vector>) (pat chars)))
		    (if result
			(let ((len (length (vector-ref result 1))))
			    ;; len is amount remaining, so length being
			    ;; smaller is a better match
			    (if (or (< len best-length)
			            (not best))
				(begin
				    (set! best result)
				    (set! best-length len)))))))
	    patterns)
	(debug-out "|___best disjunct ~s\n" best)
	best))

(define (kleene-merge-reverse bdgss)
    (let ((t '()))
	(for-each
	    (lambda (bdgs)
		(for-each
		    (lambda (bdg)
			(let ((key (car bdg))
			      (value (cdr bdg)))
			    (let ((e (assq key t)))
				(if e
				    (set-cdr! e (cons value (cdr e)))
				    (set! t (cons (cons key (cons value '())) t))))))
		    bdgs))
	    bdgss)
	t))
    
;; takes a substring in list form and
;; converts it to a string

(define (sub-str from to)
    (let loop ((p from) (prev #f))
	(if (eq? p to)
	    (if prev
		(begin
		    (set-cdr! prev '())
		    (let ((x (list->string from)))
			(set-cdr! prev to)
			x))
		"")
	    (loop (cdr p) p))))


(define (compile-slow-match pattern)
    (cond
	((char? pattern)
	    (lambda 'match-char (chars)
		(if (null? chars)
		    #f
		    (if (eq? (car chars) pattern)
			(vector '() (cdr chars))
			#f))))
	((string? pattern)
	    (let ((n (string-length pattern)))
		(lambda 'match-string (chars)
		    (let loop ((i 0) (remdr chars))
			(if (eq? i n)
			    (vector '() remdr)
			    (if (null? remdr)
				#f
				(if (eq? (car remdr) (string-ref pattern i))
				    (loop (+ i 1) (cdr remdr))
				    #f)))))))
	((null? pattern)
	    (lambda 'match-empty (chars)
		(vector '() chars)))
	((not (pair? pattern))
	    (abort 'slow-match "Illegal pattern: ~s" pattern))
	((eq? (car pattern) 'or)
	    (let ((subpats (map compile-slow-match (cdr pattern))))
		(lambda 'match-or (chars)
		    (find-best-disjunct subpats chars))))
	((eq? (car pattern) '*)
	    ;; have a <kleene-star> of a pattern that matches
	    ;; the empty string is not yet implemented...
	    (if (match-empty? (cadr pattern))
		(abort 'compile-slow-match 
			"Subpattern of <kleene-star> matches empty: ~s" 
			(cadr pattern)))
	    (let ((subpat (compile-slow-match (cadr pattern))))
		(lambda 'match-* (chars)
		    (debug-out "^^^~s matching: ~s\n" pattern chars)
		    (let loop ((remdr chars)
			       (bdgs '()))
			(debug-out "~s matching: ~s\n" pattern remdr)
			(let (((m <vector>) (subpat remdr)))
			    (if m
				(loop (vector-ref m 1)
				      (cons (vector-ref m 0) bdgs))
				;; this works even if there were no matches
				;; (kleene-* can match 0 times)
				;; because (kleen-merge-reverse '()) ==> ()
				(begin
				(debug-out "___~s matched: ~s\n" pattern bdgs)
				(vector (kleene-merge-reverse bdgs)
					remdr))))))))
	((eq? (car pattern) '+)
	    (let ((subpat (compile-slow-match (cadr pattern))))
		(lambda 'match-+ (chars)
		    (let loop ((remdr chars)
			       (bdgs '()))
			(let (((m <vector>) (subpat remdr)))
			    (if m
				(loop (vector-ref m 1)
				      (cons (vector-ref m 0) bdgs))
				(if (null? bdgs)
				    #f
				    ; there was at least one, so its ok
				    (vector (kleene-merge-reverse bdgs)
					    remdr))))))))
	((eq? (car pattern) '?)
	    (let ((subpat (compile-slow-match (cadr pattern))))
		(lambda 'match-? (chars)
		    (let (((r <vector>) (subpat chars)))
			(if r
			    r
			    (vector '() chars))))))
	((symbol? (car pattern))
	    (let ((subpat (compile-slow-match (caddr pattern)))
		  (bind-name (car pattern)))
		(if (possibly-binds? bind-name (caddr pattern))
		    (abort 'compile-slow-match
			   "subpattern of ~a illegally rebinds ~a"
			   bind-name
			   bind-name))
		(lambda 'match&bind (chars)
		    (let (((result <vector>) (subpat chars)))
			(if result
			    (let ((bdgs (vector-ref result 0))
				  (remdr (vector-ref result 1)))
				(vector (cons (cons bind-name
						    (sub-str chars remdr))
						bdgs)
					remdr))
			    #f)))))
	((eq? (cadr pattern) '...)
	    (let ((from (car pattern))
	    	  (to (caddr pattern)))
		(lambda 'match-range (chars)
		    (if (null? chars)
			#f
			(if (and (char>=? (car chars) from)
				 (char<=? (car chars) to))
			    (vector '() (cdr chars))
			    #f)))))
	(else  ;; concatenation
	    ;; note: we don't support (or detect) multiple
	    ;; bindings of the same name from different subregexps
	    (let ((subpats (map compile-slow-match pattern)))
		(lambda 'match-seq (chars)
		    (let loop ((i subpats) (bdgs '()) (remdr chars))
			(if (null? i)
			    (vector (apply append bdgs) remdr)
			    (let (((r <vector>) ((car i) remdr)))
				(if r
				    (loop (cdr i)
					  (cons (vector-ref r 0) bdgs)
					  (vector-ref r 1))
				    #f)))))))))

(define (possibly-binds? name pattern)
    (cond
	((char? pattern)
	    #f)
	((string? pattern)
	    #f)
	((null? pattern)
	    #f)
	((not (pair? pattern))
	    (abort 'possibly-binds? "Illegal pattern: ~s" pattern))
	((eq? (car pattern) 'or)
	    (any? (lambda (x)
		    (possibly-binds? name x))
		  (cdr pattern)))
	((eq? (car pattern) '*)
	    (possibly-binds? name (cadr pattern)))
	((eq? (car pattern) '+)
	    (possibly-binds? name (cadr pattern)))
	((eq? (car pattern) '?)
	    (possibly-binds? name (cadr pattern)))
	((symbol? (car pattern))
	    (if (eq? name (car pattern))
		#t
		(possibly-binds? name (caddr pattern))))
	((eq? (cadr pattern) '...)
	    #f)
	(else
	    (any? (lambda (x)
		    (possibly-binds? name x))
		  pattern))))
		  
(define (match-empty? pattern)
    (cond
	((char? pattern)
	    #f)
	((string? pattern)
	    (string=? pattern ""))
	((null? pattern)
	    #t)
	((not (pair? pattern))
	    (abort 'match-empty? "Illegal pattern: ~s" pattern))
	((eq? (car pattern) 'or)
	    (any? match-empty? (cdr pattern)))
	((eq? (car pattern) '*)
	    #t)
	((eq? (car pattern) '+)
	    (match-empty? (cadr pattern)))
	((eq? (car pattern) '?)
	    #t)
	((symbol? (car pattern))
	    (match-empty? (caddr pattern)))
	((eq? (cadr pattern) '...)
	    #f)
	(else
	    (all? match-empty? pattern))))

(define (first-match-char pattern)
    (let ((v (make-vector 256)))
	(first-match-char* v pattern)
	(let loop ((i 255) (r '()))
	    (if (eq? i -1)
		r
		(loop (- i 1)
		      (if (vector-ref v i)
		          (cons (integer->char i) r)
			  r))))))

(define (first-match-char* vec pattern)
    (cond
	((char? pattern)
	    (vector-set! vec (char->integer pattern) #t))
	((string? pattern)
	    (if (> (string-length pattern) 0)
		(vector-set! vec (char->integer (string-ref pattern 0)) #t)))
	((null? pattern)
	    #f)
	((not (pair? pattern))
	    (abort 'first-match-char* "Illegal pattern: ~s" pattern))
	((eq? (car pattern) 'or)
	    (for-each
		(lambda (subp)
		    (first-match-char* vec subp))
		(cdr pattern)))
	((eq? (car pattern) '*)
	    (first-match-char* vec (cadr pattern)))
	((eq? (car pattern) '+)
	    (first-match-char* vec (cadr pattern)))
	((eq? (car pattern) '?)
	    (first-match-char* vec (cadr pattern)))
	((symbol? (car pattern))
	    (first-match-char* vec (caddr pattern)))
	((eq? (cadr pattern) '...)
	    (let ((from (char->integer (car pattern)))
	    	  (to (char->integer (caddr pattern))))
		(let loop ((i from))
		    (if (< i to)
			(begin
			    (vector-set! vec i #t)
			    (loop (+ i 1)))))))
	(else
	;; need to collect these as long as the pred. matches the empty string
	    (let loop ((i pattern))
		(if (not (null? i))
		    (begin
			(first-match-char* vec (car i))
			(if (match-empty? (car i))
			    (loop (cdr i)))))))))

(define (regex->prefix-matcher pattern)
    (let ((p (compile-slow-match pattern)))
	(lambda 'prefix-matcher (str)
	    (let ((r (p (string->list str))))
		(if r
		    (vector (vector-ref r 0)
		    	    (- (string-length str)
			       (length (vector-ref r 1))))
		    #f)))))

(define (regex->exact-matcher pattern)
    (let ((p (compile-slow-match pattern)))
	(lambda 'exact-matcher (str)
	    (let ((r (p (string->list str))))
		(if (and r
		         (null? (vector-ref r 1)))
		    (vector-ref r 0)
		    #f)))))
	
(define (regex->any-matcher pattern)
    (let ((p (compile-slow-match pattern))
    	  (h (first-match-char pattern)))
	(if (null? h)
	    ; if there are no first characters, then the only possible
	    ; match is the empty string, which we can readily find in
	    ; all strings
	    (lambda 'any-matcher-null (str)
		'#(() 0 0))
	    (if (null? (cdr h))
		; there is only one first character, so we can find it fast
		; (after this is compiled, it will look like a "memq" that
		;  keeps on searching and keeps track of how far its gone)
		; (except (effective v0.3.1) it uses string-search to get started)
		(let ((first-char (car h)))
		    (lambda 'any-matcher-fast (str)
			(let ((firstposn (string-search str first-char)))
			    (if firstposn
				(let loop ((startposn firstposn)
					   (strl (list-tail (string->list str) firstposn)))
				    (if (null? strl)
					#f ;; never found any matches
					(if (eq? (car strl) first-char)
					    ;; this is a first char, so try matching from here
					    (let ((r (p strl)))
						(if r
						    (vector (vector-ref r 0)
							    startposn
							    (- (string-length str)
								(length (vector-ref r 1))))
						    (loop (+ startposn 1) (cdr strl))))
					    ;; this is not a first char, so move along
					    (loop (+ startposn 1) (cdr strl)))))
				#f))))
		; there are several possible first characters
		; it's a little more work, then
		(lambda 'any-matcher-slow (str)
		    (let loop ((startposn 0) (strl (string->list str)))
			(if (null? strl)
			    #f ;; never found any of them
			    (if (memq (car strl) h)
				;; this is a first char, so try matching from here
				(let ((r (p strl)))
				    (if r
					(vector (vector-ref r 0)
						startposn
						(- (string-length str)
						    (length (vector-ref r 1))))
					(loop (+ startposn 1) (cdr strl))))
				;; this is not a first char, so move along
				(loop (+ startposn 1) (cdr strl))))))))))

(export regex->prefix-matcher
	regex->exact-matcher
	regex->any-matcher)
