#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-class <dequeue> (<object>)
    head
    tail)
    
(define (make-dequeue)
    (let ((i (cons 0 '())))
	(make <dequeue> head: i tail: i)))

(define-method write ((self <dequeue>) port)
    (format port "#[<dequeue> ~s]" (dequeue->list self)))

(define (dequeue->list dq)
    (cdr (head dq)))
    
(define (push dq item)
    (let ((i (cons item (cdr (head dq)))))
	(set-cdr! (head dq) i)
	(if (eq? (tail dq) (head dq))
	    (set-tail! dq i))
    	(values)))

(define (pop dq)
    (let ((front (cdr (head dq))))
	(set-cdr! (head dq) (cdr front))
	(car front)))

(define (put dq item)
    (let ((back (cons item '())))
	(set-cdr! (tail dq) back)
	(set-tail! dq back)
	(values)))

(define (empty? dq)
    (null? (cdr (head dq))))
    
#|
    <string_matcher> ::= <sequence>
    		       | <choice>
		       | <kleene-star>
		       | <kleene-plus>
		       | <zero-or-one>
		       | <collector>
		       | <range>
		       | <literal>
    <sequence> := ( <string_matcher> ... )
    <choice> := ( or <string_matcher> ... )
    <kleene-star> := ( * <string_matcher> )
    <kleene-plus> := ( + <string_matcher> )
    <zero-or-one> := ( ? <string_matcher> )
    <collector> := ( <identifier> = <string_matcher> )
    <range> := ( <CHAR> <DOTDOTDOT> <CHAR> )
    <literal> := <STRING> | <CHAR>


  (* ( #\[ (foo = (* ( or #\a #\b ))) #\]))
    when matching
	"[aba][ba][bbb]"
    returns
    	((foo . ("aba" "ba" "bbb")))
    because the lexical structure of "foo" is that it is inside
    a <kleen-star>.  Its value, if present, will always be a list,
    possibly empty.
  
  (or (#\[ (foo = (* (or #\a #\b))) #\])
      (#\{ (bar = (* (or #\a #\b))) #\}))

    then:	"[aab]"		=>	((foo . "aab"))
		"{aab}"		=>	((bar . "aab"))

    however:
       (* (or (foo = (or #\+ #\-))
       	      (bar = (or #\x #\y))))
    can get wierd:
    		"+x-"		=>	((foo . ("+" "-")) (bar . ("x")))

    and:
    	(* (or (foo = (or #\+ #\-))
	       (#\[ (* (foo = (or #\a #\b))) #\])))
    can get non-uniform:
    		"+[aba]-"		((foo . ("+" ("a" "b" "a") "-")))
|#

(define (try-all-disjuncts patterns chars)
    (select pair?
	    (map (lambda (pat)
		(let ((result (slow-match pat chars)))
		    (if result
			(cons (length 
				(vector-ref result 1))
				result)
			#f)))
		patterns)))

(define (regex->prefix-matcher pattern)
    (lambda (str)
	(let ((chars (string->list str)))
	    (let ((r (slow-match pattern chars)))
		(if r
		    (vector (vector-ref r 0)
		    	    (- (string-length str)
			       (length (vector-ref r 1))))
		    #f)))))

(define (regex->exact-matcher pattern)
    (lambda (str)
	(let ((chars (string->list str)))
	    (let ((r (slow-match pattern chars)))
		(if (and r
		         (null? (vector-ref r 1)))
		    (vector-ref r 0)
		    #f)))))

(define (slow-match pattern chars)
    (cond
	((char? pattern)
	    (if (pair? chars)
		(if (eq? (car chars) pattern)
		    (vector '() (cdr chars))
		    #f)
		#f))
	((string? pattern)
	    (let ((n (string-length pattern)))
		(let loop ((i 0) (remdr chars))
		    (if (< i n)
			(if (null? remdr)
			    #f
			    (if (eq? (car remdr) (string-ref pattern i))
				(loop (+ i 1) (cdr remdr))
				#f))
			(vector '() remdr)))))
	((not (pair? pattern))
	    (abort 'slow-match "Illegal pattern: ~s" pattern))
	((eq? (car pattern) 'or)
	    (let ((summary (try-all-disjuncts (cdr pattern) chars)))
		;(format #t "disjuncts: ~s\n" summary)
		(if (null? summary)
		    #f
		    (cdar (sort summary 
				(lambda (a b)
				    (< (car a) (car b))))))))
	((eq? (car pattern) '*)
	    ;; have a <kleene-star> of a pattern that matches
	    ;; the empty string is not yet implemented...
	    (assert (not (match-empty? (cadr pattern))))
	    (let loop ((remdr chars)
			(bdgs '()))
		(let ((m (slow-match (cadr pattern) remdr)))
		    (if m
			(loop (vector-ref m 1)
				(cons (vector-ref m 0) bdgs))
			(if (null? bdgs)
			    ;; there were no matches
			    #f
			    (vector (kleene-merge bdgs)
				    remdr))))))
	((symbol? (car pattern))
	    (let ((result (slow-match (caddr pattern) chars)))
		(if result
		    (let ((bdgs (vector-ref result 0))
			    (remdr (vector-ref result 1)))
			(assert (not (assq (car pattern) bdgs)))
			(vector (cons (cons (car pattern)
					    (sub-str chars remdr))
					bdgs)
				remdr))
		    #f)))
	((eq? (cadr pattern) '...)
	    (if (null? chars)
		#f
		(if (and (char>=? (car chars) (car pattern))
		         (char<=? (car chars) (caddr pattern)))
		    (vector '() (cdr chars))
		    #f)))
	(else
	    (let loop ((i pattern) (bdgs '()) (remdr chars))
		(if (null? i)
		    (vector bdgs remdr)
		    (let ((r (slow-match (car i) remdr)))
			(if r
			    (loop (cdr i)
				    (append bdgs (vector-ref r 0))
				    (vector-ref r 1))
			    #f)))))))


(define (kleene-merge bdgss)
    (let ((t '()))
	(for-each
	    (lambda (bdgs)
		(for-each
		    (lambda (bdg)
			(let ((key (car bdg))
			      (value (cdr bdg)))
			    (let ((e (assq key t)))
				(if e
				    (if (seq? (cdr e))
					(seq-add! (cdr e) value)
					(let ((s (make-seq)))
					    (seq-add! s (cdr e))
					    (seq-add! s value)
					    (set-cdr! e s)))
				    (set! t (cons (cons key value) t))))))
		    bdgs))
	    bdgss)
	(map (lambda (b)
		(if (seq? (cdr b))
		    (cons (car b) (seq->list (cdr b)))
		    b))
	     t)))
    
(define (match-empty? x)
    #f)	;; stub
    

;; takes a substring in list form and
;; converts it to a string

(define (sub-str from to)
    (let loop ((p from) (prev #f))
	(if (eq? p to)
	    (if prev
		(begin
		    (set-cdr! prev '())
		    (let ((x (list->string from)))
			(set-cdr! prev to)
			x))
		"")
	    (loop (cdr p) p))))

#|
    compile-string-matcher returns a procedure of one argument (a string)
    which returns #f for no match, or an assoc list of collected stuff
    (may be empty, of course; good thing '() and #f are different)
|#

(define (compile-string-matcher regex)
)

(define (simulate-regex-machine machine input exit)
    (let ((dequeue (make-dequeue))
    	  (chars (string->list input)))
	  
	(define (add use state)
	    (if (vector? state)
		(if (< (vector-ref state 4) 2)
		    (begin 
			(use dequeue state)
			(vector-set! state 4 (+ (vector-ref state 4) 1))))
		(use dequeue state)))

	(put dequeue #f)
	(let loop ((state machine))
	    (format #t "/^^char = ~s state = " (or (null? chars) (car chars)))
	    (if (vector? state)
		(format #t "#(~a ...)" (vector-ref state 0))
		(format #t "~s" state))
	    (format #t "  queue = ~s\n" (map (lambda (qe)
	    					   (if (vector? qe)
							(vector-ref qe 0)
							qe))
						(dequeue->list dequeue)))
	    (if (eq? state 'done)
		(exit chars)
		(if (empty? dequeue)
		    (exit #f)))
	    (if state
		(let ((match (vector-ref state 1)))
		    (vector-set! state 4 (- (vector-ref state 4) 1))
		    (if match
			;; match is not #f, so it's a char
			(if (null? chars)
			    (exit #f)
			    (if (eq? match (car chars))
				(add put (vector-ref state 2))))
			;; match is #f, so its a null state
			(begin
			    (add push (vector-ref state 2))
			    (if (vector-ref state 3)
				(add push (vector-ref state 3))))))
		; state is #f, so eat the character
		(begin
		    (if (null? chars)
			(exit #f))
		    (set! chars (cdr chars))
		    (put dequeue #f)))
	    (format #t "\\__char = ~s state = " (or (null? chars) (car chars)))
	    (if (vector? state)
		(format #t "#(~a ...)" (vector-ref state 0))
		(format #t "~s" state))
	    (format #t "  queue = ~s\n" (map (lambda (qe)
	    					   (if (vector? qe)
							(vector-ref qe 0)
							qe))
						(dequeue->list dequeue)))
	    (loop (pop dequeue)))))

(define *test-machine*
    (let ((states '#(#(0 #f    5   #f   0)
    		     #(1 #\a   2   #f   0)
		     #(2 #f    3    1   0)
		     #(3 #\b   4   #f   0)
		     #(4 #f    8   #f   0)
		     #(5 #f    6    2   0)
		     #(6 #\a   7   #f   0)
		     #(7 #\c   8   #f   0)
		     #(8 #\d   9   #f   0)
		     #(9 #f done   #f   0))))
	(define (fixup v i)
	    (if (exact? (vector-ref v i))
		(vector-set! v i (vector-ref states (vector-ref v i)))))
	(set! states 
	      (vector-map (lambda (v) (vector-map (lambda (i) i) v)) states))
	(vector-for-each
	    (lambda (e)
		(fixup e 2)
		(fixup e 3))
	    states)
	(vector-ref states 0)))

(define (test str)
    (call-with-current-continuation
	(lambda (exit)
	    (simulate-regex-machine *test-machine* str exit))))
    
