#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; continuation chain backtrace
;; $Id: ccbt.scm,v 1.2 94/11/01 16:43:42 donovan Exp $

;; partial continuation stuff

(define (partial-continuation? thing)
  (instance? thing <PartialContinuation>))

(define (pc-envt-reg pc)
  (%gvec-read pc 0))

(define (pc-template-reg pc)
  (%gvec-read pc 1))

(define (pc-dynamic-state-reg pc)
  (%gvec-read pc 2))

(define (pc-continuation-reg pc)
  (%gvec-read pc 3))

(define (pc-jump-addr pc)
  (%gvec-read pc 4))

(define (pc-regs pc)
  (let loop ((i (%gvec-length pc)) (r '()))
    (if (> i 5)
	(loop (- i 1) (cons (%gvec-read pc (- i 1)) r))
	r)))

;; fc is a <closure> whose binding envt
;; contains a partial cont

(define (full->partial fc)
  (%gvec-read (%gvec-read fc 0) 1))

(define (print-bdg-envt frame-number frame)
  (let ((num-slots (- (%gvec-length frame) 1)))
    (let loop ((slot 0))
      (if (< slot num-slots)
	  (begin
	    (format #t "    ~d:~d => ~s\n" 
		    frame-number 
		    slot 
		    (%gvec-read frame (+ slot 1)))
	    (loop (+ slot 1)))))))

(define (print-bdg-envt-chain frame-number frame)
  (if (instance? frame <BindingEnvt>)
      (begin
	(if (not (eq? frame-number 0))
	    (newline))
	(print-bdg-envt frame-number frame)
	(print-bdg-envt-chain (+ frame-number 1) (%gvec-read frame 0)))))

(define (ccbt)
  (call-with-current-continuation
   (lambda (full-continuation)
     (let loop ((pc (full->partial full-continuation)))
       (if (partial-continuation? pc)
	   (let ((port (current-output-port)))
	     (format port "will continue at:\n")
	     (format port "    Program counter: 0x~r" (pc-jump-addr pc))
	     (if (instance? (pc-template-reg pc) <Template>)
		 (format port 
			 " (~a)" 
			 (addr->name (pc-template-reg pc)
			             (pc-jump-addr pc))))
	     (format port "\n    Template: ~s\n" (pc-template-reg pc))
	     (let loop ((i 0) (r (pc-regs pc)))
	       (if (pair? r)
		   (begin
		     (if (< i 10)
			 (format port "    REG~d => ~s\n" i (car r))
			 (format port "    REG(~d) => ~s\n" i (car r)))
		     (loop (+ i 1) (cdr r)))))
	     (format port "    In binding environment:\n")
	     (print-bdg-envt-chain 0 (pc-envt-reg pc))
	     (loop (pc-continuation-reg pc)))
	   (values))))))
