#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; $Id: cmdloop.scm,v 1.3 1994/12/11 00:23:47 donovan Exp $

(%define-glue (set-debug! flag)
{
#ifdef STEP_DUMP
extern int do_step_dump;

    if (arg_count_reg > 0)
    {
	if (EQ(flag,FALSE_OBJ))
	    do_step_dump = 0;
	else
	    do_step_dump = 1;
    }
    REG0 = TRUE_OBJ;
#else
    REG0 = FALSE_OBJ;
#endif
    RETURN(1);
})

(%define-glue (set-backtrace! flag)
{
#ifdef RECORD_CALL_CHAIN
extern bool do_record_call_chain;

    if (arg_count_reg > 0)
    {
	if (EQ(flag,FALSE_OBJ))
	    do_record_call_chain = NO;
	else
	    do_record_call_chain = YES;
    }
    REG0 = TRUE_OBJ;
#else
    REG0 = FALSE_OBJ;
#endif
    RETURN(1);
})

(define (current-display-limit)
    (let ((repl (fluid-ref *REPL* #f)))
	(if repl
	    (display-limit repl)
	    64)))

(define (display-values v)
    (let ((line-len (current-display-limit)))
	(if (and (pair? v) (null? (cdr v)))
	    (begin
		(format #t "value := " (car v))
		(display (object->bounded-string 
				(if line-len 
				    (+ line-len 3) 
				    #f) 
				(car v)))
		(newline))
	    (let loop ((i 0) (v v))
		(if (not (null? v))
		    (begin
		    (format #t "value[~d] := " i)
		    (display (object->bounded-string line-len (car v)))
		    (newline)
		    (loop (add1 i) (cdr v))))))))

(define (repl-ensure-tlv repl name)
    (let* ((envt (top-level-envt repl))
    	   (bdg (lookup envt name)))
	(if bdg
	    bdg
	    (let ((bdg (make <TopLevelVar> name: name)))
		(bind! envt bdg)
		bdg))))

(define (go-up! this-repl)
    (let ((outer (outer-repl this-repl)))
	(if outer
	    ((restart outer))
	    (begin
		(format #t "This REPL has no outer-repl\n")
		(values)))))

(define (go-top! this-repl)
    (let loop ((a-repl this-repl))
	(if (outer-repl a-repl)
	    (loop (outer-repl a-repl))
	    ((restart a-repl)))))

(define (update-completions! repl)
    (if (completions-are-dirty repl)
	(begin
	    (set-input-port-completions! 
		(input-port repl)
		(append *special-completions*
			(table-keys->list 
			    (table (top-level-envt repl)))))
	    (set-completions-are-dirty! repl #f))))

(define (read-eval-print envt label)
    (let ((i 0)
	  (eval-print #f)
	  (using-state #f)
	  (shift-in-value #f)
	  (on-console? (console-input-port? (current-input-port)))
	  (repl (make <REPL> input-port: (current-input-port)
    			     output-port: (current-output-port)
			     top-level-envt: envt
			     prompt-label: label
			     outer-repl: (fluid-ref *REPL* #f))))
	
	(if (not envt)
	    (let ((e (if (outer-repl repl)
			 (top-level-envt (outer-repl repl))
			 (make-default-envt))))
		(set-top-level-envt! repl e)
		(set! envt e)))

	(fluid-let ((*abort* repl-abort)
		    (*error* repl-error)
		    (*interrupt* repl-interrupt)
		    (*exception* repl-exception)
		    (*REPL* repl))

	    (set! shift-in-value
		(let ((var-%1 (repl-ensure-tlv repl '%))
		      (var-%2 (repl-ensure-tlv repl '%%)))
		    (lambda (results)
			(if (not (null? results))
			    (begin
				(set-value! var-%2 (value var-%1))
				(set-value! var-%1 (car results)))))))
    
	    (set! eval-print 
		(lambda (datum)
		    (set-input-port-prompt! (input-port repl) "?")
		    (set-backtrace! (foreground-backtrace-flag repl))
		    (set-debug! (foreground-debug-flag repl))
		    (bind ((#rest vals (eval datum envt)))
		      (set-debug! (background-debug-flag repl))
		      (set-backtrace! (background-backtrace-flag repl))
		      (display-values vals)
		      (shift-in-value vals))))
	    (call-with-current-continuation
		(lambda (exit)
		    (set-restart! repl exit)))
    
	    (let loop ()
		(set-input-port-prompt! (input-port repl)
					(string-append label 
						       "[" 
						       (number->string i)
						       "]=>"))
		(if on-console?
		    (update-completions! repl))
		(let ((datum (read)))
		    (if (eof-object? datum)
			(begin
			    (newline)
			    ;(values)
			    ;changed to EXIT instead of RESUME
			    ; (unless this is "top", in which case exit)
			    (if (outer-repl repl)
				((restart (outer-repl repl)))
				(values)))
			(let ((s (special-cmd datum)))
			  (if (and s (eq? s 'resume))
			      (list->values
			       (map (lambda (expr) 
				      (eval expr envt)) 
				    (cdadr datum)))
			      (begin
				(if s
				    (s repl)
				    (eval-print datum))
				(set! i (add1 i))
				(loop))))))))))

(define (in-triv-context envt thunk)
    (let ((repl (make <REPL> input-port: (current-input-port)
    			     output-port: (current-output-port)
			     top-level-envt: envt
			     prompt-label: label
			     outer-repl: (fluid-ref *REPL* #f))))
	
	(if (not envt)
	    (let ((e (if (outer-repl repl)
			 (top-level-envt (outer-repl repl))
			 (make-default-envt))))
		(set-top-level-envt! repl e)
		(set! envt e)))
	(set-input-port-prompt! (input-port repl) "?")
	(fluid-let ((*abort* repl-abort)
		    (*error* repl-error)
		    (*interrupt* repl-interrupt)
		    (*exception* repl-exception)
		    (*REPL* repl))
	    (thunk))))


(define $gensym #f)

(define ($init-gensym)
    (set! $gensym (make-gensym)))

(run-named $init-gensym)

