#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
; so-called "colon commands"
; (now should be called comma commands)
;$Id: colonc.scm,v 1.4 1994/12/11 00:26:08 donovan Exp $

(define *special-completions*
    '(debug no-debug
      bg-debug no-bg-debug
      fg-abt no-fg-abt
      bg-abt no-bg-abt
      bt abt
      up top exit))


(define (if-debugging repl flag cmd setter)
    (if (set-debug!)
	(setter repl flag)
	(format #t ",~s -- not compiled with STEP_DUMP\n" cmd)))

(define (special-cmd-debug repl)
    (if-debugging repl #t 'debug set-foreground-debug-flag!))

(define (special-cmd-no-debug repl)
    (if-debugging repl #f 'no-debug set-foreground-debug-flag!))

(define (special-cmd-bg-debug repl)
    (if-debugging repl #t 'bg-debug set-background-debug-flag!))

(define (special-cmd-no-bg-debug repl)
    (if-debugging repl #f 'no-bg-debug set-background-debug-flag!))


(define (if-backtracing repl flag cmd setter)
    (if (set-backtrace!)
	(setter repl flag)
	(format #t ",~s -- not compiled with RECORD_CALL_CHAIN\n" cmd)))

(define (special-cmd-fg-abt repl)
    (if-backtracing repl #t 'fg-abt set-foreground-backtrace-flag!))

(define (special-cmd-no-fg-abt repl)
    (if-backtracing repl #f 'no-fg-abt set-foreground-backtrace-flag!))

(define (special-cmd-bg-abt repl)
    (if-backtracing repl #t 'bg-abt set-background-backtrace-flag!))

(define (special-cmd-no-bg-abt repl)
    (if-backtracing repl #f 'no-bg-abt set-background-backtrace-flag!))

(define (special-cmd-abt repl)
    (if-backtracing repl 
		    #f 
		    'bt
		    (lambda (x y)
		      (bt))))

(define (special-cmd form)
  (if (and (pair? form)
	   (eq? (car form) 'unquote)
	   (pair? (cdr form)))
      (let ((comma-cmd (cadr form)))
	(if (symbol? comma-cmd)
	    (simple-comma-cmd comma-cmd)
	    (if (and (list? comma-cmd)
		     (not (null? comma-cmd))
		     (eq? (car comma-cmd) 'resume))
		'resume
		(lambda (repl)
		  (display "misformed `,'-command -- use `,help' for help\n")
		  (values)))))
      #f))

(define (simple-comma-cmd name)
  (case name
    ((abt) special-cmd-abt)
    ((ds) (lambda (repl)
	    (bt)))
    ((bt) (lambda (repl)
	    (ccbt)))
    ((up) go-up!)
    ((top) go-top!)
    ((debug) special-cmd-debug)
    ((no-debug) special-cmd-no-debug)
    ((bg-debug) special-cmd-bg-debug)
    ((no-bg-debug) special-cmd-no-bg-debug)
    ((fg-abt)  special-cmd-fg-abt)
    ((no-fg-abt)  special-cmd-no-fg-abt)
    ((bg-abt)  special-cmd-bg-abt)
    ((no-bg-abt)  special-cmd-no-bg-abt)
    ((warranty) comma-cmd-warranty)
    ((exit)
     (lambda (repl)
       (%exit 1)))
    ((help h ?)
     (lambda (repl)
       (display "Help for `,' commands:\n")
       (display-help-table
	'((",bt" "show back trace")
	  (",abt" "show APPLY back trace (if available)")
	  (",ds" "show dynamic state")
	  (",fg-abt" "enable back tracing of normal code")
	  (",bg-abt" "enable back tracing of the compiler's code")
	  (",debug"
	   "enable (very slow) monotone tracing of normal code")
	  (",bg-debug"
	   "enable monotone tracing of the compiler's code")
	  (",up" "exit to outer REPL")
	  (",top" "exit to outer-most REPL")
	  (",exit" "exit process")
	  (",warranty" "show (lack of) warranty")
	  (",h ,? ,help" "print this help info")))
       (display "(Note: a `no-' prefix disables flags, ")
       (display "e.g. `,no-fg-abt')\n")
       (display "(Note: monotone tracing info ")
       (display "gets put into ./stepdump.tra)\n")))
    (else
     (lambda (repl)
       (display "unrecognized `,'-command -- use `,help' for help\n")
       (values)))))

(define (comma-cmd-warranty repl)
    (display
" 		                NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.\n"))

