#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; note that these definers return a symbol instead
;; of #f to indicate that the TL should ignore them

;; DEFINERs are special kinds of SPECIAL FORMs, with the
;; property that they create new bindings for top-level
;; variables (or other kinds of objects, like macros)
;; A given binding can only have one definer for it in
;; all of input.  If a binding is not marked write-protected,
;; then it can have it value changed, of course...

(define (car-most form)
    (if (pair? form)
	(car-most (car form))
	form))

(define (compile/define sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "define not at top-level: ~a" form))
    (let ((name (car-most (cdr form))))
	(if (not (symbol? name))
	    (error/syntax "define target not a symbol: ~s" name))
	(let ((bdg (lookup (the-top-level lex-envt) name)))
	    (if (not bdg)
		(begin
		    (set! bdg (make <TopLevelVar> name: name))
		    (bind! (the-top-level lex-envt) bdg)))
	    (if (pair? (cadr form))
		(compile/define-fn bdg (cdr form) lex-envt dyn-envt)
		(compile/define-var bdg (cdr form) lex-envt dyn-envt))
	    name)))

; note that you can say something like
;   (define ((foo bar) x y) ...)
; to make a function of two arguments with a compound
; name.  `bdg' has already been bound to the tlv foo
; but we will compute name as '(foo bar)
;
; (at this point, form starts with the name-arg list)

(define *compile-verbose* #f)

(define (set-compile-verbose! flag)
  (set! *compile-verbose* flag))

(define (compile/define-fn bdg form lex-envt dyn-envt)
    (let ((name (caar form))
	  (formals (cdar form))
	  (body (cdr form)))
        (if *compile-verbose*
	    (format #t "compiling function: ~a\n" name))
	(set-value! bdg
		    (xwrap-procedure
			(compile/procedure name 
					    formals
					    body
					    lex-envt
					    dyn-envt 
					    #f)))))

(define (compile/define-var bdg form lex-envt dyn-envt)
    (if *compile-verbose*
        (format #t "compiling declaration: ~a\n" (car form)))
    (let ((val (eval (cadr form) lex-envt)))
	(set-value! bdg val)
    	0))


(define (compile/define-syntax sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "define-syntax not at top-level: ~a" form))

    (let ((name (cadr form)))
        (if *compile-verbose*
	    (format #t "compiling syntax: ~a\n" name))
	(bind! (the-top-level lex-envt)
	       (compile-macro name (cddr form) lex-envt))
	name))


(define (compile/define-alias sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "define-alias not at top-level: ~a" form))

    (let ((name (cadr form))
	  (expr (caddr form)))
        (if *compile-verbose*
	    (format #t "compiling alias: ~a\n" name))
	(bind! (the-top-level lex-envt)
	       (make <Substitution> name: name
	       			    expr: expr
				    envt: lex-envt))
	name))


(define (compile/define-rewriter sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "define-rewriter not at top-level: ~a" form))

    (let* ((decl (cadr form))
           (name (car decl))
    	   (formals (cdr decl))
	   (envt (the-top-level lex-envt))
	   (proc (xwrap-procedure 
			    (compile/procedure name 
						formals
						(cddr form)
						envt
						envt
						#t))))
	(bind! envt (make-rewriter name proc (length formals)))
	name))

(define (make-definers)
  (list
    (make <SpecialForm> name: 'define-syntax 
			compiler: compile/define-syntax)
    (make <SpecialForm> name: 'define-alias 
			compiler: compile/define-alias)
    (make <SpecialForm> name: 'define-rewriter 
			compiler: compile/define-rewriter)
    (make <SpecialForm> name: 'define 
			compiler: compile/define)))
