#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;$Id: load.scm,v 1.2 94/11/01 16:45:25 donovan Exp $
;;
;; script processing mode causes the first
;; lines of the file to be ignored as long as they start with a '#'

(define (repl-load-1* file script?)
    (let* ((dir (fluid-ref *current-dir* (current-directory)))
	   (noisy (not (fluid-ref *load-quiet* #f)))
    	   (p&f (locate dir
	  	        (fluid-ref *search-path* '())
		        '("scm")
		        file)))
	(if p&f
	    (let ((envt (fluid-ref *top-level-envt*)))
	        (if noisy
		    (format #t ">> loading ~a\n" (cdr p&f)))
		(fluid-let ((*current-dir* (directory (cdr p&f))))
		    (if script?
			(let loop ()
			    (if (eq? (peek-char (car p&f)) #\#)
				(begin
				    (read-line (car p&f))
				    (loop)))))
		    (with-objects-from-port
			(car p&f)
			(lambda (item)
			    (eval item envt))))
		(close-input-port (car p&f))
		(if noisy
		    (format #t "<< done loading ~a\n" (cdr p&f)))
		(values))
	    (abort 'load "Could not find file ~s in ~s" file dir))))

(define (repl-load-1 file)
    (repl-load-1* file #f))
    
(define (repl-load . files)
    (for-each repl-load-1 files)
    (let ((repl (fluid-ref *REPL* #f)))
	(if repl
	    (set-completions-are-dirty! repl #t)))
    (values))

(define (dload envt file)
    (for-each
	(lambda (xp)
	    (bind! envt
		    (make <TopLevelVar> name: (car xp)
					value: (cdr xp))))
	(vector-ref (dynamic-load file) 2)))

