#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;$Id: mktlvs.scm,v 1.9 1994/12/11 00:23:09 donovan Exp $
#|
(define-syntax bind-tlvs!
    (syntax-form (envt (var name) . more-vars)
	(let-syntax ((%initialize (syntax-form (x) x)))
	    (tle-bind! envt (make <TopLevelVar> name: (mquote name)
				      value: var))
	    (bind-tlvs! envt . more-vars)))
    (syntax-form (envt var . more-vars)
	(let-syntax ((%initialize (syntax-form (x) x)))
	    (tle-bind! envt (make <TopLevelVar> name: (mquote var)
				      value: var))
	    (bind-tlvs! envt . more-vars)))
    (syntax-form (envt) '()))
|#
(snarf bind-tlvs-rewriter)
(define-rewriter (bind-tlvs! form)
  (bind-tlvs-rewriter form))

(define (tle-multi-bind! envt . bdgs)
    (let loop ((i bdgs))
      (if (null? i)
	  #f
	  (let ((r (cdr i)))
	    (tle-bind! envt 
		       (make <TopLevelVar>
			     name: (car i) 
			     value: (car r)))
	    (loop (cdr r))))))

(define (bind-hacker-functions-1! envt)
    (bind-tlvs! envt
	(_%gvec-length %gvec-length)
	(_%gvec-read %gvec-read)
	(_%gvec-write %gvec-write)
	
	(_%bvec-length %bvec-length)
	(_%bvec-read %bvec-read)
	(_%bvec-write %bvec-write)
	%for-each-output-file %for-each-input-file 
	%close-all-input-files %close-all-output-files 
	%close-all-files
	%abort %exit
	print-bvec print-gvec
	%save-image %load-image
	clone gvec? bvec? clear-user-hash parse-format-string
	garbage-collect
	allocation-stats reset-allocation-stats
	with-bounded-string-port with-bounded-string-port*
	*initialization-thunks*))

(define (bind-hacker-functions-2! envt)
    (bind-tlvs! envt
	%format *default-formatters* 
	ct-format-rewriter *default-ct-formatters*
	(_%write %write) (_%display %display) %write-int %write-str
	%write-binary %write-bvec
	%char-ready? %peek-char %read-char %scan-token
	<Port> <InputPort> <OutputPort> *eof-object*
	<Console-InputPort> <Token>
	<User-InputPort> <User-OutputPort>
	write-hex write-bin write-oct write-tab4
	write-radix write-hex-w/pad
	%gvec-alloc %bvec-alloc %console-input-port
	%console-output-port
	bvec-write-fixnum bvec-write-16 bvec-write-str bvec-write-float
	bvec-write-bvec
	bvec-read-fixnum bvec-read-16 bvec-read-str bvec-read-float))	

(define (bind-hacker-functions! envt)
    (bind-hacker-functions-1! envt)
    (bind-hacker-functions-2! envt))
    
(define (bind-socket-functions! envt)
    (bind-tlvs! envt
	%write-str/SOCK %write-bvec/SOCK %write-binary/SOCK 
	%read-str/SOCK %read-bvec/SOCK
	open-named-socket close-named-socket connect-to-name
	accept-connection close-connection collect-socket-data
	host-id-eq? hostname->host-id host-id->hostname))

(define (bind-os-functions! envt)
    (bind-tlvs! envt
	getenv system errorstr getrusage get-user-info
	fork fork-no-wait wait sleep getpid
	scandir getwd setwd!
	stat stat-directory? stat-access? file-exists?
	time=? time<? time>? time<=? time>=?
	time->string interval->string
	time-thunk
	interval- time+ time-
	time-now seconds->interval interval->seconds
	<Interval> <Time>
	set-timer! halt-timer! get-time-remaining
	interval? time?
	mkdir os-file-open os-file-seek os-file-read-str os-file-read-bvec
	os-file-write))

(define (bind-extended-functions-1! envt)
  (bind-tlvs! envt
	(envt *self*)
	STUB Gui dload
	set-compile-verbose!
	; benchmark
	warn-unbound-envt-vars))

(define (bind-extended-functions-2! envt)
  (bind-tlvs! envt
	<Binding> <TopLevelVar>
	<Closure> <Template> <BindingEnvt> <TopLevelEnvt>
	<table>
	<Environment> <Method>
	<Slot-Descriptor>
	<PartDescr>
	<PartialContinuation>))


(define (bind-scheme-functions-1! envt)
    (bind-tlvs! envt
	close-input-port close-output-port
	%failed-assert 	
	format print
	open-output-string open-output-file flush-output-port
	open-buffered-input
	file-output-port? output-port? open-output-file
	input-port? file-input-port? console-input-port?
	newline write-char object->string object->bounded-string
	open-input-file open-input-file* read-char peek-char char-ready?
	read-line
	console-input-port console-output-port
	open-console-input-port
	input-port-prompt input-port-completions
	set-input-port-prompt! set-input-port-completions!
	read eof-object?))
	
(define (bind-scheme-functions-12! envt)
    (bind-tlvs! envt
	with-output-to-port with-input-from-port
	call-with-output-file with-output-to-file
	call-with-input-file with-input-from-file with-input-from-string
	file->string
	with-objects-from-port with-objects-from-file
	current-input-port current-output-port
	open-input-string))


(define (bind-scheme-functions-2! envt)
    (bind-tlvs! envt
	%dynamic-state-ref
	%fluid-assq %fluid-unbound
	%with-dynamic-state
	%expand-quasiquote
	%make-promise
	* + - / < <= = >= >
	binary<? binary<=? binary>? binary>=? binary=?
	<Object>
	<number> 
        <complex> <real> <float> <rational> <integer> <ratio> 
	<fixnum> <double-float>))
	
(define (bind-scheme-functions-11! envt)
    (bind-tlvs! envt
 	<<Class>>
	<Symbol> <String> <Vector> <Pair> <List>
	<ByteVector> <C-Text> c-text? string->c-text c-text->string
	<Boolean>
	<Null>
	<ASCII-Char>
	<Unicode-Char>
	<UniqueObject>
	%make* (_%make %make)
	error abort abs
	any? all?))

(define (bind-scheme-functions-3! envt)
    (bind-tlvs! envt
	bit-ref bit-set bit-shift-left bit-shift-right
	bitwise-and bitwise-not bitwise-or bitwise-xor
	boolean?
	char=? char<? char<=? char>? char>=?
	char-ci=? char-ci<? char-ci<=? char-ci>? char-ci>=?
	char-alphabetic?
	char-numeric?
	char-whitespace?
	char-upper-case?
	char-lower-case?
	char-upcase
	char-downcase
	random))

(define (bind-scheme-functions-4! envt)
    (bind-tlvs! envt 
	cons car cdr
	caaaar caaadr caaar caadar caaddr caadr
	caar cadaar cadadr cadar caddar cadddr caddr
	cadr cdaaar cdaadr cdaar cdadar cdaddr cdadr
	cdar cddaar cddadr cddar cdddar cddddr cdddr
	cddr add1 append apply
	apply* assoc assq assv
	<seq> seq? make-seq seq-add! seq->list 
	select range))

(define (bind-scheme-functions-5! envt)
    (bind-tlvs! envt 
	call-with-current-continuation ceiling
	char->integer char? class-name class-of class?
	complex? exception-vector delq!  display eq?
	equal? eqv? even? exact? floor for-each force
	gcd))

(define (bind-scheme-functions-6! envt)
    (bind-tlvs! envt
	table-insert! table-keys->list
	table-lookup table-for-each table-remove!
	table-values->list table? inexact?
	initialize integer->char integer->hash identity
	integer->unicode-char integer? instance? lcm length
	list list->string list->values list->vector
	list-ref list-set!  list-tail list? last last-pair))


(define (bind-scheme-functions-7! envt)
    (bind-tlvs! envt 
	table-join make-table make-object-table make-generic-fn
	make-string
	make-vector map max
	member memq memv min modulo remainder negative? not null?
	number->string number? odd? pair? positive?
	procedure? generic-function?
	top-level-var? quotient rational? real? reverse
	round set-car!  set-cdr!))
	
(define (bind-scheme-functions-8! envt)
    (bind-tlvs! envt 
	sort string string->hash
	string->list string->number string->symbol
	string-append string-ci->hash string-ci=?
	string-copy string-fill!  string-length string-ref
	string-set!  string=? string?))

(define (bind-scheme-functions-9! envt)
    (bind-tlvs! envt 
	sub1 subclass?
	substring string-search symbol->hash immob->hash symbol->string symbol?
	truncate unary- unary/ unicode-char->integer
	unicode-char? vector vector->list vector-length
	vector-ref vector-set!  vector? write zero?
	exp log sin cos tan asin acos atan
	sqrt expt exact->inexact inexact->exact))

(define (bind-scheme-functions-10! envt)
    (bind-tlvs! envt 
	%define-class-rewriter
	%make-class-rewriter
	%slot-constructor
	%make-rewriter
	%define-setters-rewriter
	%define-getters-rewriter
	find-method find-implementation add-method!
	$getter $unchecked-setter $checked-setter))

(define (bind-scheme-functions-13! envt)
    (bind-tlvs! envt 
	vector-fill! vector-for-each vector-map
	string<? string>? string<=? string>=?
	string-ci<? string-ci>? string-ci<=? string-ci>=?))
	
(define (bind-scheme-functions-14! envt)
    (bind-tlvs! envt 
	<weak-cell> weak-cell contents set-contents! 
	<weak-pair> weak-cons weak-car weak-cdr
	
	regex->prefix-matcher
	regex->exact-matcher
	regex->any-matcher))
	
(define (bind-ctc-prims! envt)
    (bind-tlvs! envt
	%if %multi %combo %literal %bind %seq %lex-set! %tl-set!
	%lambda %root-ref %root-set! %procedure))


(define (bind-compiler-1! envt)
    (bind-tlvs! envt
	compile current-place-name current-procedure-name
	compile/procedure
	<icode>
	    <expr-icode>
	    <ic-call>
	    <ic-call-prim>
	    <ic-prim-expr>
	    <ic-bind>
	    <ic-tl-ref>
	    <ic-lex-ref>
	    <ic-root-ref>
	    <ic-tl-set>
	    <ic-lex-set>
	    <ic-root-set>
	    <ic-if>
	    <ic-const>
	    <ic-procedure>
	    <ic-lambda>
	    <ic-multi>
	    <ic-seq>
	    <ic-loop>
	    <ic-jump>
	xwrap-procedure
	error/syntax error/semantic error/internal warning))

(define (bind-compiler-2! bind-into-envt)
  (bind-tlvs! bind-into-envt
	gen-nctc compile <RootVar> <TopLevelVar> <SpecialForm>
	<Rewriter>  <ExtendedRewriter> <Macro> <MacroForm> <Substitution>
	<loop-var>
	
	return-types does-lambda? does-apply? mode
	envt ct-envt inits body num-args rest? var rhs
	inits loop-var first-arg-reg proc arg-list stmt-list
	condition if-true if-false function args arg-types
	table bindings write-dec make-closure make-template
	nctc-wrap %ctc-eval set-shadow! shadow slot make-const
	make-special-forms make-definers the-top-level name
	value set-value! 
	set-compile-seeing-shadows!
	set-compile-optimized-loops!
	%define-method-rewriter c-name set-c-name! module set-module!
	set-first-arg-reg! find-lex-addr set-ct-envt!
	set-label! set-inits! loop-icode label set-slot!
	set-write-prot! set-table!
	compile/multi lookup-type set-return-types!))

(define (bind-compiler-3! envt)
    (bind-tlvs! envt
	write-asm-stmts vinsns->monotones*
	relabel-monotones symbol+integer->hash))

(define (bind-compiler! envt)
    (bind-compiler-1! envt)
    (bind-compiler-2! envt)
    (bind-compiler-3! envt))

(define (bind-eval! envt)
  (bind-tlvs! envt
	eval
	make-default-envt
	make-top-level-envt
	make-repl-envt
	bind! tle-bind!
	lookup
	multi-bind!
	read-eval-print
	repl-top-level-envt))

(define (bind-pathname! envt)
    (bind-tlvs! envt
	<file-name> <directory-name> <root-dir>
	pathname->string
	open-output-path* open-output-path 
	open-input-path* open-input-path
	append-path filename extension directory path=?
	file-within-dir append-dirs string->file string->dir
	last-directory locate dir-from-to current-directory))
	

(define (bind-dl-stuff! envt)
    (bind-tlvs! envt
	%c-apply call-gate
	gate->procedure raw-fn-ptr->procedure
	dynamic-load-lookup open-dynamic-load close-dynamic-load
	install-module name->module module&tag->part part&index->function
	make-c-ffi fixup-part-descr fixup-template
	dynamic-load-code
	*dload-defs* dynamic-load dynamic-load-module))

(define (bind-threads! envt)
    (bind-tlvs! envt
	<Thread>
	thread-state thread-return-values thread-name
	set-thread-state! set-thread-return-values! set-thread-name!
	get-current-thread get-time-slice set-time-slice!
	%thread-yield %timeout
	%threads-did-unblock
	make-thread with-interrupts-disabled get-ready-threads
	thread-done?
	$init-threads
	$run-threads show-threads
	<Queue>
	enqueue!
	dequeue!
	queue-empty?
	%call-with-lightweight-continuation
	%jump-with-timeout
	%suspend-current-thread
	%disable-interrupts %reenable-interrupts
	%mark-thread-ready 
	%dispatch-next-thread
	thread?
	resume-thread suspend-thread thread-sleep))

(define (bind-vector-ops! envt)
    (bind-tlvs! envt
		vector-or
		vector-and!
		vector-andn!
		vector-bool-op))

;; extra table classes & functions

(define (bind-table-vars! envt)
    (bind-tlvs! envt
	<symbol-table>
	<string-table>
	<string-ci-table>
	<object-table>
	<eq-table>
	<generic-table>
	<table-bucket>
	<integer-table>
	<hash-integer-table>
	table-key-present?
	table-size
	table-hash-function table-equal-function
	table-stats
	get-system-symbol-table))
	
(define (bind-extended-functions! envt)
  (bind-extended-functions-1! envt)
  (bind-extended-functions-2! envt))
		
(define (bind-scheme-functions! envt)
    (bind-scheme-functions-1! envt)
    (bind-scheme-functions-2! envt)
    (bind-scheme-functions-3! envt)
    (bind-scheme-functions-4! envt)
    (bind-scheme-functions-5! envt)
    (bind-scheme-functions-6! envt)
    (bind-scheme-functions-7! envt)
    (bind-scheme-functions-8! envt)
    (bind-scheme-functions-9! envt)
    (bind-scheme-functions-10! envt)
    (bind-scheme-functions-11! envt)
    (bind-scheme-functions-12! envt)
    (bind-scheme-functions-13! envt)
    (bind-scheme-functions-14! envt)
    (bind-extended-functions! envt)
    (bind-user-envt-introspection! envt))

(define (bind-inside! envt)
    (multi-bind! envt (repl-top-level-envt)))

(define (bind-all! envt)
  (for-each
   (lambda (m)
     (if (not (eq? (cdr m) bind-all!))
	 ((cdr m) envt)))
   (get-modules)))

(define *modules* #f)

(define (get-modules)
    (if (not *modules*)
	(set! *modules*
	      (list (cons 'compiler bind-compiler!)
		    (cons 'ctc-prims bind-ctc-prims!)
		    (cons 'dynamic-linking bind-dl-stuff!)
		    (cons 'pathname bind-pathname!)
		    (cons 'os bind-os-functions!)
		    (cons 'threads bind-threads!)
		    (cons 'eval bind-eval!)
		    (cons 'hacker bind-hacker-functions!)
		    (cons 'sockets bind-socket-functions!)
		    (cons 'all bind-all!)
		    (cons 'vector-ops bind-vector-ops!)
		    (cons 'tables bind-table-vars!)
		    (cons 'inside bind-inside!))))
    *modules*)

(define (use-compiler sf form lxe dyne mode)
  (if (null? (cdr form))
      (format #t "choose one of: ~s\n" (map car (get-modules)))
      (for-each
       (lambda (module)
	 (let ((proc (assq module (get-modules))))
	   (if proc
	       ((cdr proc) (the-top-level lxe))
	       (error/syntax "module ~s not available" module))))
	(cdr form)))
  (let ((repl (fluid-ref *REPL* #f)))
    (if repl
	(set-completions-are-dirty! repl #t)))
  (values))
