#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; $Id: replexc.scm,v 1.2 94/11/01 16:47:09 donovan Exp $

;; set the TLV *error* in the current REPL
;; to some info about the error

(define (set-error! value)
    (let ((r (fluid-ref *REPL* #f)))
	(if r
	    (set-value! (repl-ensure-tlv r '*error*) value))))

(define (at-depth label)
    (let loop ((i 0) (a-repl (fluid-ref *REPL* #f)))
	(if a-repl
	    (loop (add1 i) (outer-repl a-repl))
	    (if (eq? i 1)
		label
		(string-append label "^" (number->string i))))))


(define (with-io-for-repl thunk)
    (let ((repl (fluid-ref *REPL* #f)))
	(with-input-from-port
	    (if repl
		(input-port repl)
		(console-input-port))
	    (lambda ()
		(with-output-to-port
		    (if repl
			(output-port repl)
			(console-output-port))
		    thunk)))))

(define (repl-error msg . args)
    (with-io-for-repl 
	(lambda ()
	    (format #t "*** ERROR ***\n")
	    (apply* #t msg args format)
	    (newline)
	    (set-error! (vector msg args))
	    (read-eval-print #f (at-depth "break")))))

(define (repl-abort place msg . args)
    (with-io-for-repl 
	(lambda ()
	    (format #t "*** ERROR IN ~a ***\n" place)
	    (apply* #t msg args format)
	    (newline)
	    (set-error! (vector place msg args))
	    (read-eval-print #f (at-depth "break")))))

(define (repl-interrupt)
    (with-io-for-repl 
	(lambda ()
	    (display "*** INTERRUPT ***\n")
	    (read-eval-print #f (at-depth "break")))))

(define (repl-exception type tmpl addr envt descr args)
    (with-io-for-repl
	(lambda ()
	    (display "*** EXCEPTION ***\n")
	    (apply* #t descr args format)
	    (newline)
	    (set-error! (vector type tmpl envt descr args))
	    (read-eval-print #f (at-depth "break")))))
