#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;  Define some useful macros...

(define (bind-std-scheme-macros! envt)
  (letrec-syntax ((syntax-form (syntax-form (args . body)
				 (make <MacroForm> args: (mquote args)
				       body: (mquote body))))
		  (%initialize (syntax-form (x) x))
		  (bind-syntax! (syntax-form (name . body)
				  (bind! envt
					 (make <Macro>
					       name: (mquote name)
					       envt: envt
					       forms: (list . body))))))
    (bind-syntax! cond
      (syntax-form (('else . action))
	(begin . action))
      (syntax-form ((guard) . more-clauses)
	(let ((temp guard)) 
	  (if temp 
	      temp 
	      (cond . more-clauses))))
      (syntax-form ((guard '=> proc) . more-clauses)
	(let ((temp guard)) 
	  (if temp 
	      (proc temp)
	      (cond . more-clauses))))
      (syntax-form ((guard . actions) . more-clauses)
	(if guard 
	    (begin . actions) 
	    (cond . more-clauses)))
      (syntax-form ()
	(begin)))
    
    (bind-syntax! or
      (syntax-form (clause)
	clause)
      (syntax-form (clause . more-clauses)
	(let ((temp clause))
	  (if temp
	      temp
	      (or . more-clauses))))
      (syntax-form () #f))
    
    (bind-syntax! and
      (syntax-form (clause1 clause2 . more-clauses)
	(if clause1 (and clause2 . more-clauses) #f))
      (syntax-form () #t)
      (syntax-form (clause) clause))
    
    (bind-syntax! case
      (syntax-form (expr . branches)
	(letrec-syntax
	    ((test-cases
	      (syntax-form (var ((g0 . g1) . actions) . more-cases)
		(if (test-guards var g0 . g1)
		    (begin . actions)
		    (test-cases var . more-cases)))
	      (syntax-form (var ('else . actions))
		(begin . actions))
	      (syntax-form (var)
		(begin)))
	     (test-guards
	      (syntax-form (var item)
		(eq? var (mquote item)))
	      (syntax-form (var item . more-items)
		(if (eq? var (mquote item))
		    #t
		    (test-guards var . more-items)))
	      (syntax-form (var)
		#f)))
	  (let ((temp expr))
	    (test-cases temp . branches)))))

    (bind-syntax! delay
      (syntax-form (expr)
	(%make-promise (lambda () expr))))))

(define (bind-es-scheme-macros! envt)
  (letrec-syntax ((syntax-form (syntax-form (args . body)
				 (make <MacroForm> args: (mquote args)
				       body: (mquote body))))
		  (%initialize (syntax-form (x) x))
		  (bind-syntax! (syntax-form (name . body)
				  (bind! envt
					 (make <Macro>
					       name: (mquote name)
					       envt: envt
					       forms: (list . body))))))
    
    (bind-syntax! fluid-let
      (syntax-form (bindings . body)
	(letrec-syntax ((extend-fluid
			 (syntax-form ((var init) . more-bdgs) 
			   (cons (cons (mquote var) init)
				 (extend-fluid . more-bdgs)))
			 (syntax-form () (%dynamic-state-ref))))
	  (%with-dynamic-state
	   (extend-fluid . bindings)
	   (lambda ()
	     (begin . body))))))
    (bind-syntax! fluid-set!
      (syntax-form (name val)
	(let ((x (%fluid-assq (mquote name))))
	  (if x
	      (let ((old (cdr x)))
		(set-cdr! x val)
		old)
	      (%fluid-unbound (mquote name))))))
    (bind-syntax! fluid-ref
      (syntax-form (name)	; unforgiving form
	(let ((x (%fluid-assq (mquote name))))
	  (if x
	      (cdr x)
	      (%fluid-unbound (mquote name)))))
      (syntax-form (name default-expr)  ; forgiving form
	(let ((x (%fluid-assq (mquote name))))
	  (if x
	      (cdr x)
	      default-expr))))
    
    (bind-syntax! cons*
      (syntax-form (item1 item2 . more-items)
	(cons item1 (cons* item2 . more-items)))
      (syntax-form (last-item) last-item))
    
    (bind-syntax! type-check
      (syntax-form (arg predicate class)
	(if (not (predicate arg))
	    (abort (last (*FUNCTION*))
		   "argument ~s (~s) is not a ~s" 
		   (mquote arg)
		   arg 
		   (class-name class))))
      (syntax-form (arg class)
	(if (not (instance? arg class))
	    (abort (last (*FUNCTION*))
		   "argument ~s (~s) is not a ~s" 
		   (mquote arg)
		   arg 
		   (class-name class)))))
    
    (bind-syntax! assert
      (syntax-form (expr)
	(if (not expr)
	    (%failed-assert
	     (*FUNCTION*)))))))

(define (bind-scheme-macros! envt)
  (bind-std-scheme-macros! envt)
  (bind-es-scheme-macros! envt)
  (values))

