#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; $Id: start.scm,v 1.4 1994/12/11 00:24:52 donovan Exp $

(define (user-config-file name)
    (make <file-name> directory: (fluid-ref *HOME*) 
		      filename: "" 
		      extension: name))

(define (process-args envt args)
    (if (and (pair? args)
    	     (string=? (car args) "-norc"))
	(set! args (cdr args))
	(let* ((f (pathname->string (user-config-file "R.scm")))
	       (p (open-input-file* f)))
	    (if p
		(begin
		    (close-input-port p)
		    (repl-load f)))))
    (let loop ((i args))
	(if (null? i)
	    #t
	    (let ((a (car i)))
		(cond
		    ((string=? a "-x")
			#f)
		    ((string=? a "-exit")
			(%exit 0))
		    ((string=? a "-e")
			(with-objects-from-port
			    (open-input-string (cadr i))
			    (lambda (o)
				(eval o envt)))
			(loop (cddr i)))
		    ((and (> (string-length a) 1)
			  (eq? (string-ref a 0) #\+))
			(dload envt (substring a 1))
			(loop (cdr i)))
		    ;; running RScheme in script mode
		    ;; causes the remaining arguments to be bound to *args*
		    ;; the file to be evaluated, and then
		    ;; exiting with code 0
		    ;; all error handlers are bound to a procedure which
		    ;; exits with code 1
		    ((and (string=? a "-script")
		          (> (length i) 1))
			(bind! envt (make <TopLevelVar> name: '*args*
							value: (cddr i)))
			(let ((kill (lambda 'script-death args 
				      (if *verbose*
					  (begin
					    (format #t "Died: ~s\n" args)
					    (ccbt)))
				      (%exit 1))))
			    (fluid-let ((*error* kill)
			    	        (*abort* kill)
					(*interrupt* kill)
					(*exception* kill))
				(repl-load-1* (cadr i) #t)
				(%exit 0))))
		    ((string=? a "--")
			(bind! envt (make <TopLevelVar> name: '*args*
							value: (cdr i)))
			#t)
		    (else
		        (repl-load-1 a)
			(loop (cdr i))))))))

(define *verbose* #t)
(define *license* #t)

(define ($start initial-loads)
    (if (and (pair? initial-loads)
	     (string=? (car initial-loads) "-q"))
	(begin
	    (set! *verbose* #f)
	    (set! initial-loads (cdr initial-loads))))
    (if (and (pair? initial-loads)
	     (string=? (car initial-loads) "-script"))
	(set! *verbose* #f))
    (if *verbose*
	(begin
	  (if *license*
	      (begin
		(display "Copyright (C) 1994 Donovan Kolbly, Paul Wilson\n")
		(display "RScheme comes with ABSOLUTELY NO WARRANTY.\n")
		(display
		 "type \",warranty\" for details; type \",help\" for help\n")))
	  (format #t "Compiled by: ~a\n" %version)
#|
	    ($if-benchmarks :all
		(display "(with all benchmarks)\n"))
	    ($if-benchmarks :conform-only
		(display "(with conform benchmark)\n"))
|#
	    ))
    (vector-set! exception-vector 0 repl-interrupt)
    (let ((envt (make-default-envt)))
	(fluid-let ((*top-level-envt* envt)
		    (*current-dir* (string->dir "."))
		    (*warn-unbound-var-ref* #f)
		    (*PS2* "> ")
		    (*HOME* (string->dir (or (getenv "HOME") "."))))
	    (fluid-let ((*load-quiet* #t))
	      (in-triv-context 
	       envt
	       (lambda ()
		 (process-args envt initial-loads))))
	    (read-eval-print envt "top"))))
