#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (make-envt-object envt)
    (lambda (msg . args)
	(case msg
	    ((lookup) (lookup envt (car args)))
	    (else (error/internal "Bad message (~s) to environment" msg)))))

(define (make-rewriter name proc type)
  (make <SpecialForm> 
	name: name 
	compiler: (case type
		    ((1) (lambda 'rewriter (sf form lex-envt dyn-envt mode)
			    (compile (proc form) 
				     lex-envt 
				     dyn-envt 
				     mode)))
		    ((2) (lambda 'extended-rewriter 
					    (sf form lex-envt dyn-envt mode)
			    (compile (proc form 
					   (lambda (expr)
					    (eval expr lex-envt)))
				     lex-envt 
				     dyn-envt 
				     mode)))
		    ((3) (lambda 'rewriter-in-envt 
					    (sf form lex-envt dyn-envt mode)
			    (compile (proc form 
					   (lambda (expr)
					      (eval expr lex-envt))
					   (make-envt-object lex-envt))
				     lex-envt 
				     dyn-envt 
				     mode)))
		    (else (abort 'make-rewriter 
				 "Bad rewriter: ~s type ~d" 
				 name
				 type)))))

(define (quasiquote-compiler sf form lxe dyne mode)
    (compile (%expand-quasiquote (cadr form)) lxe dyne mode))

(define *repl-envt* #f)

(define (snarf-compiler sf form lxe dyne mode)
    (if (not *repl-envt*)
	(set! *repl-envt* (make-repl-envt)))
    (for-each
	(lambda (name)
	    (let ((b (lookup *repl-envt* name)))
		(if b
		    (bind! (the-top-level lxe) b)
		    (warning "~a: Not bound in snooped envt" name))))
	(cdr form))
    #f)

(define (make-gensym)
    (let ((i 0)) 
	(lambda ()
	    (set! i (add1 i))
	    (string->symbol 
		(string-append "temp[" 
			       (number->string i) 
			       "]")))))

(define (define-gf-rewriter form)
    (let ((name (car (cdr form))))
	(list 'define name (list 'make-generic-fn (list 'quote name)))))

(define (the-ct-format-rewriter form)
    (ct-format-rewriter *default-ct-formatters* form))

(define (make-user-initial-forms)
    (list
        (make <TopLevelVar> name: 'load value: repl-load)
        (make <TopLevelVar> name: '%version value: $version)
	(make-rewriter 'let* let*-rewriter 1)
	(make-rewriter 'do do-rewriter 1)
	(make-rewriter '%define-getters %define-getters-rewriter 3)
	(make-rewriter '%define-setters %define-setters-rewriter 3)
	(make-rewriter 'define-class %define-class-rewriter 2)
	(make-rewriter 'make-class %make-class-rewriter 2)
	(make-rewriter 'define-method %define-method-rewriter 1)
	(make-rewriter 'define-generic-fn define-gf-rewriter 1)
	(make-rewriter 'define-generic-function define-gf-rewriter 1)
	(make-rewriter 'make %make-rewriter 2)
	(make-rewriter 'ct-format the-ct-format-rewriter 1)

	(make <SpecialForm> name: '%code compiler: compile/%code)

        (make <TopLevelVar> 
	    name: '$gensym 
	    value: $gensym)
	
        (make <TopLevelVar> 
	    name: '%list 
	    value: list)
	
        (make <TopLevelVar> 
	    name: '%initialize 
	    value: (lambda (thing) 
			(%initialize thing)))
	
        (make <TopLevelVar> 
	    name: 'make-method 
	    value: (lambda (specializers implementation) 
			(%make <Method> specializers implementation)))
	
        (make <TopLevelVar> 
	    name: '%define-method 
	    value: (lambda (generic-fn specializers code) 
			(add-method! generic-fn 
				     (make-method 
					    specializers 
					    (%make-closure '() code)))))
	
        (make <TopLevelVar> 
	    name: '%make-class* 
	    value: (lambda (name heap-type image-mode parents slots) 
			(%make-class* name 
				      heap-type 
				      image-mode 
				      parents 
				      slots)))
	
        (make <TopLevelVar> 
	    name: '%make-getter 
	    value: (lambda (name slot) 
			(%make-closure 
			    (%make <BindingEnvt> '() slot name)
			    $getter)))
	
        (make <TopLevelVar> 
	    name: '%make-unchecked-setter 
	    value: (lambda (name slot) 
			(%make-closure 
			    (%make <BindingEnvt> '() slot name)
			    $unchecked-setter)))
	
        (make <TopLevelVar> 
	    name: '%make-checked-setter 
	    value: (lambda (name slot type) 
			(%make-closure 
			    (%make <BindingEnvt> '() slot type name)
			    $checked-setter)))
	
	(make <SpecialForm> name: 'quasiquote compiler: quasiquote-compiler)
	(make <SpecialForm> name: 'snarf compiler: snarf-compiler)
	(make <SpecialForm> name: '%use compiler: use-compiler)
    ))

(define (repl-top-level-envt)
    (%top-level-environment))	; a VERY special form

(define (make-default-envt)
  (let ((envt (make-top-level-envt)))
    (multi-bind! envt (make-special-forms))
    (multi-bind! envt (make-user-initial-forms))
    (multi-bind! envt (make-definers))
    (bind-scheme-macros! envt)
    (bind-scheme-functions! envt)
    envt))

(define (make-repl-envt)
  (let ((envt (make-default-envt)))
    (multi-bind! envt (repl-top-level-envt))
    envt))

(define (tlv? x)
    (eq? (class-of x) <TopLevelVar>))

(define (warn-unbound-envt-vars envt)
    (table-for-each
	(table envt)
	(lambda (h k v)
	    (if (and (tlv? v)
	    	     (eq? (%gvec-read v 1) '#unbound))
		(warning "~s is unbound" k)))))

(define *compile-seeing-shadows* #f)

