#ifndef IGPS_HH
#define IGPS_HH

// This file contains the declaration of igps maintaining class.

#include "config.hh"
#include "langtype.h"
#include "scheduler.hh"

class igp_manager{
#ifdef GENERATIONAL
    friend double scheduling_manager::get_max_table_occupancy_ratio(int);
    friend void scheduling_manager::reset_at_gc_flip(void);
#endif
// a list of intergenerational pointer object.
// intergenerational pointers (or igp's) are pointers from one generation
// to some younger generation. igp_list contains the pointers to an object
// that contains an intergenerational pointer. add_igp adds a pointer to an
// intergenerational point to the list.
    
// Followings are the declaration of the intergenerational pointer management
// routines and data structures. Since this is a temporary implementation,
// we don't pay enough attention to its efficiency or realtimeness.

    // ipg_list[pointing][i] is an array of igps from the generation (pointing
    // + 1). (Since there is no igp object form the youngest generation, we
    // omit the entry.)
    pos_ptr_addr igp_list[NUMBER_OF_GENERATIONS - 1][2][NUM_IGP];  

    // old_igps[i] points to the complete igp list from the generation (i+1).
    // This is designed to point to igp_list[i][0] or igp_list[i][1];
    // new_igps[i] points to the igp list under construction and points to
    // one of the lists igp_list[i][0] and igp_list[i][1] which old_igps[i]
    // does not point to. At the end of the GC cycle of the pointing
    // generation, the values of od_gps[i] and new_igps[i] are exchanged,
    // in order to switch the role of the IGPS lists.
    
    pos_ptr_addr *old_igps[NUMBER_OF_GENERATIONS - 1];
    pos_ptr_addr *new_igps[NUMBER_OF_GENERATIONS - 1];

    // old_igp_num[i] is the number of entries in the igp_list pointed by
    // old_igps[i]. new_igp_num[i] is the number of entries in new_igps[i]
    // old_igp_next[pointing][pointed] indexes the next entry in
    // old_igps[pointing] for searching the generation pointed.
    // new_igp_next has the similar function.
    int old_igp_num[NUMBER_OF_GENERATIONS - 1];
    int new_igp_num[NUMBER_OF_GENERATIONS - 1];
    int old_igp_next[NUMBER_OF_GENERATIONS - 1][NUMBER_OF_GENERATIONS - 1];  
    int new_igp_next[NUMBER_OF_GENERATIONS - 1][NUMBER_OF_GENERATIONS - 1];	
  public:
    // add_igp() adds the argument to the new_igp_list. The first argument
    // should be the location of an IGP in a shaded object. (In the current
    // implementation, this is not true.) This function add the pointer as
    // an igp from the generation pointing to the generation pointed.
    void add_igp(pos_ptr_addr,int pointing, int pointed);

    // add_temporary_igp() adds the argument to the old_igp, which is thrown
    // away at the end of the GC cycle. The argument should be the location of
    // an IGP in a white object. (In the current implementation, this is not
    // true.)
    void add_temporary_igp(pos_ptr_addr, int pointing, int pointed);

    int scan_igp_list(int pointed);     // Return a IGP.

    // switch_igp_lists should be called at the end of a GC cycle of the
    // pointing generation.
    void switch_igp_lists(int pointing);

    // reset the indexes at the end of a GC cycle of the pointed generation,
    // to get ready for the next GC cycle.
    void reset_igp_lists(int pointed);
    igp_manager(void);
};
#endif // IGPS_HH
