/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
/* $Id: entry.c,v 1.2 94/11/01 16:40:53 donovan Exp $ */

#include <stdarg.h>
#include <setjmp.h>
#include "runtim.h"
#include "vinsns.h"
#ifdef ATRACE
#include "gcserver.h"
#include "atrace.h"
#endif
#include "intrs.h"
#include "osglue.h"


extern struct part_descr ctcprims_part;
extern struct function_descr finish_run_descr;

jmp_buf *restart_run = NULL;
static unsigned run( obj cont );
static jump_addr last_addr;

#ifdef GNU_VINSNS
static void *restart_run_addr;
static int restart_run_value;
#endif

#ifdef STEP_DUMP
unsigned insn_count = 0;
#endif
#ifdef TIMER_IS_MONOTONE_COUNTER
unsigned system_timeout = 0;
#endif

/* Defining STEP_DUMP and turning on do_step_dump
   causes a dump before the execution of
   every monotone, thereby slowing down executing slighly :-/
*/

#ifdef STEP_DUMP
int do_step_dump = 0;  /* most main()'s accept "-stepdump" to set this */
FILE *step_dump_file = NULL;
const char *step_dump_filename = NULL;
#endif /* STEP_DUMP */

/*  Manufacture a continuation which, when resumed,
    has the effect of calling the function with the
    given arguments
*/

static void vpush_call_continuation( obj closure, 
				     unsigned num_args, 
				     va_list the_args )
{
unsigned i;
jump_addr addr;

    arg_count_reg = num_args;
    addr = apply( closure );
    {
	PUSH_PARTCONT_ADDR( addr, arg_count_reg );
	
	for (i=0; i<arg_count_reg; i++)
	    SET_PARTCONT_REG( i, va_arg( the_args, obj ) );
    }
}

static void push_call_continuation( obj closure, unsigned num_args, ... )
{
va_list x;

    va_start( x, num_args );
    vpush_call_continuation( closure, num_args, x );
    va_end( x );
}

obj call_scheme( obj closure, unsigned num_args, ... )
{
obj t;
struct function_descr *fn;
va_list ap;
int n;

    switch_hw_regs_into_scheme();
    assert( CLOSURE_P(closure) );
    t = gvec_read( closure, SLOT(1) );
    fn = (struct function_descr *)VAL(gvec_read( t, SLOT(1) ));
#ifdef DEBUG_0
    printf( "Calling " );
    fflush( stdout );
    printf( "%s\n", fn->name );
#endif /* DEBUG_0 */

    /* manufacture a continuation which, when resumed,
	terminates the current run.  */
    
    envt_reg = NIL_OBJ;
    literals_reg = FALSE_OBJ;
    dynamic_state_reg = NIL_OBJ;
    continuation_reg = FALSE_OBJ;
    push_cont( finish_run_descr.monotones[0], 0 );
    
    va_start( ap, num_args );
    vpush_call_continuation( closure, num_args, ap );
    va_end( ap );

    n = run(continuation_reg);
#ifdef DEBUG_0
    for (i=0; i<n; i++)
    {
	printf( "Return value [%d] = ", i );
	fprinto( stdout, temp_space[i] );
	putchar( '\n' );
    }
#endif /* DEBUG_0 */
    switch_hw_regs_back_to_os();
    return temp_space[0];
}

#ifdef STEP_DUMP

static struct function_descr *get_function_descr( obj tmpl )
{
    return (struct function_descr *) 
    		OBJ_TO_RAW_PTR( gvec_read( tmpl, SLOT(1) ) );
}

void touch_step_dump_file( void )
{
    if (!step_dump_file)
	step_dump_file = os_fopen( step_dump_filename, "w" );
}

void step_dump( jump_addr f )
{
    touch_step_dump_file();
#ifdef TIMER_IS_MONOTONE_COUNTER
    fprintf( step_dump_file, "%u (timer %u) ==> ", 
    		insn_count, 
		system_timeout );
#else
    fprintf( step_dump_file, "%u ==> ", insn_count );
#endif
    if (EQ(literals_reg,FALSE_OBJ))
    {
        fprintf( step_dump_file, "*FINISH-RUN*\n" );
    }
    else
    {
    struct function_descr *fn = get_function_descr( literals_reg );
    unsigned i;
    
	i = 0;
	while (fn->monotones[i] != f)
	{
	    assert( fn->monotones[i] );
	    i++;
	}
        fprintf( step_dump_file, "%s", fn->name );
	if (i)
	    fprintf( step_dump_file, " [%d]", i );
        fprintf( step_dump_file, "  (%d args:", arg_count_reg );
	for (i=0; i<arg_count_reg; i++)
	{
	    fprintf( step_dump_file, " REG%d=", i );
	    fprinto( step_dump_file, reg_ref(i) );
	}
	fprintf( step_dump_file, ")\n" );
    }
    fflush( step_dump_file );
}

#endif /* STEP_DUMP */


static unsigned run( obj cont )
{
int i;
jump_addr f;
jmp_buf *prev_jmp_buf, current_jmp_buf;

#ifdef GNU_VINSNS
obj temp[1000];

    temp[0] = cont;
#endif

    temp_space[0] = cont;
    prev_jmp_buf = restart_run;
    restart_run = &current_jmp_buf;

    /* set up exception handler */
    while ((i = setjmp(current_jmp_buf)) == 1)
    { 
      /* nothing */
    }
    					/* this is where we start;
    					   on an exception, we come back
					   to here and reload the jump buffer
					*/
#ifdef GNU_VINSNS
    restart_run_value = 0;
restart:
    i = restart_run_value;
#endif

    if (i >= 2) /* finish_run was invoked */
    {
        restart_run = prev_jmp_buf;
	return i-2;
    }
    
    /* resume the continuation indicated by cross-over buffer */
    
    continuation_reg = temp_space[0];
    f = half_restore();
    arg_count_reg = restore_arb();
    
#ifndef GNU_VINSNS
    while (1)
    {
	/* eventually, this should be rolled
	   into the prologue of each monotone
	   for now, we probably don't allocate more than 1000 or so
	   words in any given monotone */
	safe_point( SLOT(1000) );
	if (interrupt_pending && interrupts_enabled)
	    f = dispatch_interrupt(f);
	
#ifdef STEP_DUMP
	if (do_step_dump)
	    step_dump(f);
#endif /* STEP_DUMP */
#ifdef ATRACE
	atrace_will_execute( f );
#endif
	last_addr = f;
	f = (jump_addr) f();
#ifdef STEP_DUMP
	insn_count++;
#endif

#ifdef TIMER_IS_MONOTONE_COUNTER
	if (system_timeout && (--system_timeout == 0))
	{
	    interrupt_pending = YES;
	    interrupt_flag[INTR_FLAG_TIMEOUT] = TRUE_OBJ;
	}
#endif /* TIMER_IS_MONOTONE_COUNTER */
    }
#else
#ifdef STEP_DUMP
    if (do_step_dump)
	step_dump(f);
    insn_count++;
#endif /* STEP_DUMP */
    restart_run_addr = &&restart;
    goto *f;
#endif
    return 0;	/* never reached */
}


volatile void scheme_error( const char *msg, unsigned num_args, ... )
{
va_list x;
int i;
obj the_args;

#if 0
    fprintf( stderr, "error! %s\n", msg );
    va_start( x, num_args );
    for (i=0; i<num_args; i++)
    {
	fprintf( stderr, "   arg[%d] ::= ", i );
	fprinto( stderr, va_arg( x, obj ) );
	fprintf( stderr, "\n" );
    }
    va_end( x );
#endif /* 0 */

    if (EQ(exception_handler,FALSE_OBJ))
    {
	fprintf( stderr, "** No exception handler in place **\n" );
	abort();
    }
    
    /* collect the args into a list */
    
    
    va_start( x, num_args );
    for (i=0; i<num_args; i++)
	temp_space[i] = va_arg( x, obj );
    va_end( x );

    the_args = NIL_OBJ;
    while (i>0)
	the_args = cons( temp_space[--i], the_args );

    /* invoke the exception handler */
    
    push_call_continuation( 
         exception_handler, 
	 6, ZERO, 
	    literals_reg, 
	    last_addr, 
	    envt_reg, 
	    make_string(msg), 
	    the_args );
    
    temp_space[0] = continuation_reg;
    longjmp( *restart_run, 1 );
}

/* entry point for finishing a call to run() */

#define FUNCTION "*FINISH-RUN*"

PROLOGUE(finish_run)

BEGIN_FWD(finish_run)
    FWD_MONOTONE(finish_run_0)
END_FWD(finish_run)

MONOTONE(finish_run_0)
{
unsigned i;

    if (!arg_count_reg)
	temp_space[0] = NOVALUE_OBJ;
    else
	for (i=0; i<arg_count_reg; i++)
	    temp_space[i] = reg_ref(i);

#ifdef GNU_VINSNS
    restart_run_value = arg_count_reg + 2;
    goto *restart_run_addr;
#else
    longjmp( *restart_run, arg_count_reg + 2 );
#endif
    return (jump_addr)NULL;		/* Never reached */
}

EPILOGUE(finish_run)
#undef FUNCTION

BEGIN_BACK(finish_run)
  BACK_MONOTONE(finish_run_0)
END_BACK(finish_run)

struct function_descr finish_run_descr = {
	JUMP_TABLE(finish_run),
	"*FINISH-RUN*",
	&ctcprims_part };

/* entry point for continuing after an interrupt */

#define FUNCTION "*CONTINUE*"

PROLOGUE(continue_intr)

BEGIN_FWD(continue_intr)
    FWD_MONOTONE(continue_intr_0)
END_FWD(continue_intr)

MONOTONE(continue_intr_0)
{
jump_addr f;

    RESTORE_CONT0();	/* personally, we don't have anything
    			   saved in our continuation */
    f = half_restore();			/* but get the interrupted
    					   context's fixed registers */
    arg_count_reg = restore_arb();	/* and all the other regs */
#ifdef GNU_VINSNS
    goto *f;
#else
    return f;
#endif
}

EPILOGUE(continue_intr)

BEGIN_BACK(continue_intr)
  BACK_MONOTONE(continue_intr_0)
END_BACK(continue_intr)

struct function_descr continue_intr_descr = {
	JUMP_TABLE(continue_intr),
	FUNCTION,
	&ctcprims_part };
#undef FUNCTION
