/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
/* $Id: init.c,v 1.6 1994/12/29 22:14:51 donovan Exp $ */

#include "runtim.h"
#include "heapi.h"
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include "hasht.h"
#include "intrs.h"
#include "osglue.h"

extern void init_dynamic_link( const char *argv0 );

/* #define DEBUG_0 */

#ifdef STEP_DUMP
extern int do_step_dump;
extern FILE *step_dump_file;
extern const char *step_dump_filename;

extern
#ifdef __cplusplus
  "C"
#endif
    int getpid( void );

void init_step_dump( void )
{
    if (!step_dump_filename)
	step_dump_filename = "stepdump.tra";
    step_dump_file = NULL;
}
#endif

extern struct module_descr ctcprims;
extern struct module_descr module_bltin;
extern struct module_descr module_scmio;
extern struct module_descr module_oslib;
extern struct module_descr module_repl;
extern struct module_descr module_usrlib;
extern struct module_descr module_tables;

/*
coming soon to an RScheme release near you...
   extern struct module_descr module_bci;
*/

static struct module_descr *(tab[]) = { 
    &ctcprims, 
    &module_bltin,
    &module_scmio,
    &module_oslib,
    &module_usrlib,
    &module_repl,
    &module_tables,
/*    &module_bci, */
    NULL };

static void run_vec( obj vec )
{
UINT_32 i, len = SIZEOF_PTR(vec);
obj x;

    for (i=0; i<len; i+=SLOT(1))
    {
	x = gvec_read( vec, i );
	if (VECTOR_P(x))
	{
	struct module_descr *md;
	int ix;
	
	    /* a TL-SET */
	    md = (struct module_descr *)
	    		OBJ_TO_RAW_PTR( gvec_read( x, SLOT(0) ) );
	    ix = fx2int(gvec_read( x, SLOT(1) ));
	    assert( ix >= 0 && ix < md->num_roots );
	    md->root_vector[ix] = gvec_read( x, SLOT(2) );
	}
	else if (CLOSURE_P(x))
	{
	    call_scheme( x, 0 );
	}
	else
	{
	    fprintf( stderr, "Not a <Vector> or <Closure>:\n" );
	    debug( x );
	}
    }
}

obj init_scheme( const char *executable,
		 const char *boot_image,
		 bool verbose )
{
obj start;

    switch_hw_regs_into_scheme();
    init_regs();
    init_dynamic_link( executable );
    init_linkage( tab );
    init_runtim(verbose);
    init_os();
#ifdef STEP_DUMP
    init_step_dump();
#endif /* STEP_DUMP */

    gc_safe_point( 1024*1024 );
    boot_list = load_image_file( boot_image, system_link_table );
    
    if (verbose)
	puts( string_text( gvec_read( boot_list, SLOT(0) ) ) );
    
    /* we have to switch back to OS regs because
       run_vec will call call_scheme(), which switches reg suites
       again and it doesn't behave in a stack-like fashion
    */
    switch_hw_regs_back_to_os();
    run_vec( gvec_read( boot_list, SLOT(1) ) );
    switch_hw_regs_into_scheme();
    init_interrupts();
    start = gvec_read( boot_list, SLOT(2) );
    boot_list = FALSE_OBJ;
    switch_hw_regs_back_to_os();
    return start;
}
