/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "intrs.h"
#include "smemory.h"
#include "vinsns.h"
#include "runtim.h"
#include <signal.h>

bool interrupt_pending;
bool interrupts_enabled;

#ifndef cast_to_signal_handler
#define cast_to_signal_handler(x) x
#endif


obj interrupt_flag[4] = { FALSE_OBJ, FALSE_OBJ, FALSE_OBJ, FALSE_OBJ };

int disable_signals( void )
{
#ifdef SIGSETMASK
    return sigsetmask( ~0 );
#else
    return 0;
#endif /* SIGSETMASK */
}

void reenable_signals( int info )
{
#ifdef SIGSETMASK
    sigsetmask( info );
#endif /* SIGSETMASK */
}

#ifdef __cplusplus
extern "C" {
#endif

/* static void set_interrupt_flag( int sig, int code, 
   struct sigcontext *scp, 
   char *addr )*/

static void set_interrupt_flag( int x )	/* signal handler */
{
    printf( "\n***Interrupt***\n" );
    if (truish(interrupt_flag[INTR_FLAG_USER_INTR]))
    {
	fprintf( stderr, "not handled timely enough... aborting\n" );
	abort();
    }
    interrupt_pending = YES;
    interrupt_flag[INTR_FLAG_USER_INTR] = TRUE_OBJ;
#ifdef NEED_SIGNAL_REINSTALL
    signal( SIGINT, cast_to_signal_handler(set_interrupt_flag) );
#endif
}

#ifdef __cplusplus
}
#endif

void init_interrupts( void )
{
    interrupt_pending = NO;
    interrupts_enabled = YES;
    interrupt_flag[INTR_FLAG_USER_INTR] = FALSE_OBJ;
    interrupt_flag[INTR_FLAG_TIMEOUT] = FALSE_OBJ;
    interrupt_flag[INTR_FLAG_DATA_READY] = FALSE_OBJ;
    interrupt_flag[INTR_FLAG_FINALIZE] = FALSE_OBJ;
    signal( SIGINT, cast_to_signal_handler(set_interrupt_flag) );
}


jump_addr dispatch_interrupt( jump_addr would_be_next )
{
int siginfo;
jump_addr f;

    save_cont( arg_count_reg, would_be_next );

    literals_reg = continue_intr_tmpl;
    f = OBJ_TO_JUMP_ADDR( gvec_read( literals_reg, SLOT(0) ) );
    save_cont( 0, f );

    siginfo = disable_signals();
    REG0 = interrupt_flag[0];	/* ^C */
    REG1 = interrupt_flag[1];	/* watchdog timer */
    REG2 = interrupt_flag[2];	/* data available */
    REG3 = interrupt_flag[3];	/* finalize some objects */
    interrupt_flag[0] = FALSE_OBJ;
    interrupt_flag[1] = FALSE_OBJ;
    interrupt_flag[2] = FALSE_OBJ;
    interrupt_flag[3] = FALSE_OBJ;
    interrupt_pending = NO;
    arg_count_reg = 4;
    reenable_signals( siginfo );

    return apply( interrupt_dispatch );
}
