/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
/*
    Scheme-level Interrupts
*/

#ifndef _H_INTRS
#define _H_INTRS

#include "obj.h"

#define INTR_FLAG_USER_INTR	(0)	/* ^C */
#define INTR_FLAG_TIMEOUT	(1)	/* watchdog timer */
#define INTR_FLAG_DATA_READY	(2)	/* input is possible */
#define INTR_FLAG_FINALIZE	(3)	/* need to finalize */

/*
    interrupt_pending is YES if any interrupt_flag is non-#f

    interrupts_enabled is YES if the runtime system should
    dispatch to the interrupt handler when it notices that 
    interrupt_pending is set.  Interrupts should not remain
    disabled for very long.
*/

extern bool interrupt_pending;
extern bool interrupts_enabled;

extern obj interrupt_flag[];	/* note: these are not part of the root
				   set, so only put IMMOB's here 
				   (eg, #t, #f) */

/*
    This function gets called to handle an interrupt.
    It should save the state of the currently executing
    thread and return the address of the scheme-level
    interrupt handler.
    (It gets called within finite time whenever 
    interrupt_pending is YES).
*/

jump_addr dispatch_interrupt( jump_addr would_be_next );

/*
    Initialize interrupts
    Install handlers for ^C, etc.
*/

void init_interrupts( void );

/*
    Arrange to be interrupted in a certain amount of time.
*/

void schedule_timer_interrupt( UINT_32 sec, UINT_32 usec );

#endif /* _H_INTRS */
