/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "runtim.h"
#include "gcclient.h"
#include <string.h>
#ifdef INLINES
#include "gcclient1.ci"
#include "gcclient2.ci"
#endif

struct module_descr **master_table = NULL;
static bool original_modules = YES;

static obj gate_test( obj a, obj b )
{
    return FX_ADD(a,b);
}

static obj finalize_gate( obj a )
{
#ifndef CONFIG_HH
extern void gc_register_finalizable_object( gc_obj_addr );

    if (OBJ_ISA_PTR(a))
      {
	gc_register_finalizable_object( PTR_TO_GCPTR(a) );
	return TRUE_OBJ;
      }
#endif
    return FALSE_OBJ;
}

static obj next_finalizing_gate( void )
{
#ifndef CONFIG_HH
gc_obj_addr p;
extern gc_obj_addr gc_next_finalizing_object( void );

    p = gc_next_finalizing_object();
    return p ? GCPTR_TO_PTR(p) : FALSE_OBJ;
#else
    return FALSE_OBJ;
#endif
}

static obj obj_addr_gate( obj a )
{
char temp[50];

    sprintf( temp, "%u", VAL(a) );
    return make_string(temp);
}

static obj convert_tags( obj a, obj new_tag )
{
    return OBJ((VAL(a) & ~3) | (fx2int(new_tag) & 3));
}

static obj check_tags( obj a, obj mask )
{
    return OBJ((VAL(a) << 2) & VAL(mask));
}

/*
extern "C" obj dgram_open( obj port );
extern "C" obj dgram_recv( obj local, obj net_addr_class );
extern "C" obj dgram_send( obj local, obj dest, obj data );
*/

struct call_gate_descr call_gates[] = {
    { "gate-test", (call_gate *)gate_test },
    { "finalize", (call_gate *)finalize_gate },
    { "next-finalizing", (call_gate *)next_finalizing_gate },
    { "&obj", (call_gate *)obj_addr_gate },
    { "convert-tags", (call_gate *)convert_tags },
    { "check-tags", (call_gate *)check_tags },
/*
    { "dgram-open", (call_gate *)dgram_open },
    { "dgram-recv", (call_gate *)dgram_recv },
    { "dgram-send", (call_gate *)dgram_send },
*/
    { NULL, NULL } };
    
#define IGNORE_FAULTS

void init_linkage( struct module_descr **initial_table )
{
    master_table = initial_table;
#ifdef GNU_VINSNS
    {
    struct module_descr **m;
    struct part_descr **p;
    struct function_descr **f;
    
	for (m=master_table; *m; m++)
	{
	    for (p=(*m)->parts; *p; p++)
	    {
		/*
		printf( "Fixing up %s,%s\n", (*m)->name, (*p)->name );
		*/
		for (f=(*p)->functions; *f; f++)
		{
		void **(*F)( void );
		
		    F = (void **(*)( void ))(*f)->monotones;
		    (*f)->monotones = (jump_addr *)F();
		}
	    }
	}
    }
#endif
}

void install_module( struct module_descr *new_module )
{
struct module_descr **s, **d, **new_table;
unsigned n;

    for (n=0, s=master_table; *s; s++, n++)
    {
      /* nothing */
    }

    new_table = (struct module_descr **)
    			malloc( sizeof( struct module_descr * ) * (n+2) );
    d = new_table;
    *d++ = new_module;
    s = master_table;
    while (*s)
	*d++ = *s++;
    *d++ = NULL;
    
    if (!original_modules)
	free( master_table );
    master_table = new_table;
    original_modules = NO;

    stable_root_reset();	/* reset the stable roots */
}

struct part_descr *find_part( struct module_descr *m, int tag )
{
struct part_descr **p;

    for (p=m->parts; *p; p++)
    {
	if ((*p)->tag == tag)
	{
	    /* we've found the part */
	    return *p;
	}
    }
    fprintf( stderr, 
		"Module %s has no part with tag %d\n", 
		m->name, tag );
#ifndef IGNORE_FAULTS
    abort();
#endif /* IGNORE_FAULTS */
    return NULL;
}

struct module_descr *find_module( const char *module )
{
struct module_descr **m;

    for (m=master_table; *m; m++)
    {
	if (strcmp((*m)->name,module) == 0)
	{
	    /* we've found the module */
	    return *m;
	}
    }
    fprintf( stderr, "Module %s not found\n", module );
#ifndef IGNORE_FAULTS
    abort();
#endif /* IGNORE_FAULTS */
    return NULL;
}
