/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#ifndef _H_OBJ
#define _H_OBJ

/*  core type system  */

/*
    WARNING:  Don't think that just because we have defines here
    	      that you can change the values and it will still work.
	      At least some of these values are frozen rock-solid.
*/


#include "platform.h"
#include <assert.h>

/*  declare the C types for the root classes  */

#ifdef GNU_VINSNS
typedef void *jump_addr;
#else
typedef void *(*jump_addr)( void );
#endif

/*  the macros VAL, IVAL, and OBJ convert back and forth between 
    obj's and [U]INT_32's
*/

#ifdef USE_CONSTRUCTOR_EXPRS

typedef struct _obj { UINT_32 v; } obj;

#define VAL(x)			((x).v)
#define IVAL(x)			((INT_32)(x).v)
#define OBJ(x)			((obj){x})

#else

typedef UINT_32 obj;

#define VAL(x)			(x)
#define IVAL(x)			((INT_32)(x))
#define OBJ(x)			(x)

#endif
#define BIG_ENDIAN_TO_HOST_OBJ(x)	OBJ(BIG_ENDIAN_TO_HOST_32(VAL(x)))
#define HOST_TO_BIG_ENDIAN_OBJ(x)	OBJ(HOST_TO_BIG_ENDIAN_32(VAL(x)))


#define EQ(x,y)			(VAL(x)==VAL(y)?YES:NO)
#define RB_NOT(x)		((x)?NO:YES)

/*  define the tag system  */
/*
   NOTE:  The only reason we are using `+' instead of `|' herein
          (since these are really bit-wise operations) is because
	  it's more likely that an optimizer will fold + than |
*/

#define PRIMARY_TAG_SIZE	(2)
#define SECONDARY_TAG_SIZE	(3)

#define COMBINED_TAG_SIZE	(PRIMARY_TAG_SIZE + SECONDARY_TAG_SIZE)
#define PRIMARY_TAG_MASK	((1<<PRIMARY_TAG_SIZE)-1)
#define COMBINED_TAG_MASK	((1<<(SECONDARY_TAG_SIZE+PRIMARY_TAG_SIZE))-1)
#define SECONDARY_TAG_MASK	(COMBINED_TAG_MASK & ~PRIMARY_TAG_MASK)

/*  define the tag values  */

/*  NOTE:  Because insns on MC680x0 are 16-bit aligned,
           the two even tags are immediate objects, so that
	   code pointers look like immobs instead of pointers.
*/

enum primary_tag {
    FIXNUM_TAG = 0,
    __unused_tag,		/* reserved for future use */
    IMMOB_TAG,
    POINTER_TAG
};

enum secondary_tag {
    BOOLEAN_TAG = 0,
    NIL_TAG,
    ASCII_CHAR_TAG,
    UNICODE_CHAR_TAG,
    UNIQUE_OBJ_TAG
};

enum unique_object_tag {
    NOVALUE_TAG,
    UNDEFINED_TAG,
    UNINITIALIZED_TAG,
    UNBOUND_TAG,
    REST_TAG,
    KEY_TAG };

#define TEST_PRIMARY_TAG(x,t)	  ((VAL(x) & PRIMARY_TAG_MASK) == t)

#define OBJ_ISA_FIXNUM(x)	  TEST_PRIMARY_TAG( x, FIXNUM_TAG )
#define OBJ_ISA_IMMOB(x)	  TEST_PRIMARY_TAG( x, IMMOB_TAG )
#define OBJ_ISA_PTR(x)	  	  TEST_PRIMARY_TAG( x, POINTER_TAG )

#define SECONDARY_TAG(x)          (enum secondary_tag)\
                                        ((VAL(x) & SECONDARY_TAG_MASK)\
                                                      >> PRIMARY_TAG_SIZE)


/************************* FIXNUMs *************************/

#define MAKE_FIXNUM(x)		OBJ(FIXNUM_TAG \
					+ ((UINT_32)(x) << PRIMARY_TAG_SIZE))

#define FIXNUM_TO_RAWINT(x)	(IVAL(x) >> PRIMARY_TAG_SIZE)
#define RAWINT_TO_FIXNUM(x)	MAKE_FIXNUM(x)
#define ZERO			MAKE_FIXNUM(0)
#define ADD1(x)			OBJ(IVAL(x)+(1<<PRIMARY_TAG_SIZE))
#define SUB1(x)			OBJ(IVAL(x)-(1<<PRIMARY_TAG_SIZE))
#define MUL2(x)			OBJ(IVAL(x)<<1)
#define DIV2(x)			OBJ((IVAL(x)>>1) & ~PRIMARY_TAG_MASK)

#define RAW_DIV4(x)		((x)/4)

#define FX_LT(x,y)		(IVAL(x)<IVAL(y))
#define FX_LE(x,y)		(IVAL(x)<=IVAL(y))
#define FX_GT(x,y)		(IVAL(x)>IVAL(y))
#define FX_GE(x,y)		(IVAL(x)>=IVAL(y))

#define FX_ADD(x,y)		OBJ(IVAL(x)+IVAL(y))
#define FX_SUB(x,y)		OBJ(IVAL(x)-IVAL(y))
#define FX_MUL(x,y)		OBJ(IVAL(x)*fx2int(y))
#define FX_DIV(x,y)		int2fx(fx2int(x)/fx2int(y))

#define FX_AND(x,y)		OBJ(VAL(x)&VAL(y))
#define FX_OR(x,y)		OBJ(VAL(x)|VAL(y))
#define FX_XOR(x,y)		OBJ((VAL(x)^VAL(y)) + FIXNUM_TAG)
#define FX_NOT(x)		OBJ(~VAL(x)-3)

#define RIBYTES_TO_FXWORDS(x)	OBJ(x)
#define FXWORDS_TO_RIBYTES(x)	VAL(x)

#define fx2int(x)		FIXNUM_TO_RAWINT(x)
#define int2fx(x)		RAWINT_TO_FIXNUM(x)

/************************* IMMOBs *************************/

/* TEST_SECONDARY_TAG implicitly tests the primary tag for IMMOB_TAG */

#define TEST_SECONDARY_TAG(x,t)	  ((VAL(x) & \
				     (SECONDARY_TAG_MASK|PRIMARY_TAG_MASK))\
					== ((t << PRIMARY_TAG_SIZE)\
					    |IMMOB_TAG))

#define OBJ_ISA_BOOLEAN(x)	  TEST_SECONDARY_TAG( x, BOOLEAN_TAG )
#define OBJ_ISA_NIL(x)	  	  TEST_SECONDARY_TAG( x, NIL_TAG )
#define OBJ_ISA_ASCII_CHAR(x)	  TEST_SECONDARY_TAG( x, ASCII_CHAR_TAG )
#define OBJ_ISA_UNICODE_CHAR(x)   TEST_SECONDARY_TAG( x, UNICODE_CHAR_TAG )
#define OBJ_ISA_UNDEFINED(x)	  TEST_SECONDARY_TAG( x, UNDEFINED_TAG )
#define OBJ_ISA_NOVALUE(x)	  TEST_SECONDARY_TAG( x, NOVALUE_TAG )
#define OBJ_ISA_UNINITIALIZED(x)  TEST_SECONDARY_TAG( x, UNINITIALIZED_TAG )
#define OBJ_ISA_UNBOUND(x) 	  TEST_SECONDARY_TAG( x, UNBOUND_TAG )

#define MAKE_IMMOB(t,x)		OBJ((IMMOB_TAG \
					+ ((UINT_32)(t) << PRIMARY_TAG_SIZE)) \
				  	+ ((UINT_32)(x) << COMBINED_TAG_SIZE))

#define TRUE_OBJ		(MAKE_IMMOB( BOOLEAN_TAG, 1 ))
#define FALSE_OBJ		(MAKE_IMMOB( BOOLEAN_TAG, 0 ))
#define NIL_OBJ			(MAKE_IMMOB( NIL_TAG, 0 ))

#define MAKE_UNIQ_OBJ(x)        (MAKE_IMMOB( UNIQUE_OBJ_TAG, x ))
#define UNDEFINED_OBJ		(MAKE_UNIQ_OBJ( UNDEFINED_TAG ))
#define NOVALUE_OBJ		(MAKE_UNIQ_OBJ( NOVALUE_TAG ))
#define UNINITIALIZED_OBJ	(MAKE_UNIQ_OBJ( UNINITIALIZED_TAG ))
#define UNBOUND_OBJ		(MAKE_UNIQ_OBJ( UNBOUND_TAG ))

#define MAKE_ASCII_CHAR(ch)	(MAKE_IMMOB( ASCII_CHAR_TAG, ch ))
#define MAKE_UNICODE_CHAR(ch)	(MAKE_IMMOB( UNICODE_CHAR_TAG, ch ))

#define GET_IMMEDIATE_VALUE(x)	(VAL(x) >> COMBINED_TAG_SIZE)


void debug( obj item );

#ifndef __TURBOC__
#define JUMP_ADDR_TO_OBJ(x)	OBJ((UINT_32)(x))
#define OBJ_TO_JUMP_ADDR(x)	((jump_addr)VAL(x))
#else
#define JUMP_ADDR_TO_OBJ(x)	OBJ((((INT_32)(x))-((INT_32)main))<<2)
#define OBJ_TO_JUMP_ADDR(x)	((jump_addr)((IVAL(x)>>2)+((INT_32)main)))
#endif

#define RAW_PTR_TO_OBJ(x)	OBJ((UINT_32)(x))
#define OBJ_TO_RAW_PTR(x)	((void *)VAL(x))

obj class_of( obj item );

#endif
