/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
/*
$Id: runtim.h,v 1.2 94/11/01 16:41:10 donovan Exp $
*/

#ifndef _H_RUNTIM
#define _H_RUNTIM

#include "obj.h"
#include "regs.h"
#include "linktype.h"
#include "smemory.h"
#include "allocns.h"
#include <stdio.h>

#include "bltin_r.h"
#include "tables_r.h"

obj init_scheme( const char *executable_path, 
		 const char *boot_image,
		 bool verbose );

obj call_scheme( obj closure, unsigned num_args, ... );
volatile void scheme_error( const char *msg, unsigned num_args, ... );


#define NUM_RUNTIM_ROOTS (30)

extern obj runtim_root_vector[NUM_RUNTIM_ROOTS];

#define system_link_table runtim_root_vector[0]
#define system_initial runtim_root_vector[1]
#define symbol_table runtim_root_vector[3]
#define address_table runtim_root_vector[5]
/* #define module_exports_table runtim_root_vector[6] */
#define backup_boots runtim_root_vector[8]

/***************** miscellaneous *****************/

CIH_DECL bool NULL_P( obj thing );
bool TEMPLATE_P( obj thing );
bool BYTE_VECTOR_P( obj thing );

bool LONGFLOAT_P( obj thing );
IEEE_64 extract_float( obj longfloat );
obj make_float( IEEE_64 longfloat );
IEEE_64 get_float( obj arg, const char *fn, const char *arg_name );

void fprinto( FILE *dest, obj item );
obj class_of( obj thing );
obj fluid_assq( obj key );
obj addr_to_name( obj tmpl, obj addr );

/***************** closures *****************/

CIH_DECL bool CLOSURE_P( obj thing );
CIH_DECL obj make_closure( obj envt, obj tmpl );

/***************** partial continuations *****************/

/***************** pairs *****************/

CIH_DECL bool PAIR_P( obj thing );
CIH_DECL obj cons( obj car, obj cdr );
CIH_DECL obj pair_car( obj pair );
CIH_DECL obj pair_cdr( obj pair );

/***************** vectors *****************/

bool VECTOR_P( obj thing );
obj make_empty_vector( UINT_32 length );

/***************** strings *****************/

bool STRING_P( obj thing );

const char *string_text( obj str );
UINT_32 string_length( obj str );
bool string_eq( obj str1, obj str2 );
bool string_ci_eq( obj str1, obj str2 );
int string_cmp( obj str1, obj str2 );
int string_ci_cmp( obj str1, obj str2 );
obj make_string( const char *text );
obj string_ref( obj str, UINT_32 index );
void string_set( obj str, UINT_32 index, obj ch );
obj string_hash( obj str );

/***************** symbols *****************/

bool SYMBOL_P( obj thing );
obj intern( obj str );
obj symbol_hash( obj symbol );
const char *symbol_text( obj symbol );
obj symbol_str( obj symbol );

obj lookup_symbol( const char *str ); /* str should be word aligned 
					 & zero padded */

/***************** top-level environments & vars *****************/

bool TLE_P( obj thing );
CIH_DECL bool TLV_P( obj thing );

obj tle_lookup_tlv( obj tle, obj name );

obj tlv_name( obj tlv );
obj tlv_value( obj tlv );
void tlv_set_value( obj tlv, obj value );

/***************** classes *****************/

bool CLASS_P( obj thing );
bool class_is_gvec( obj a_class );
obj class_name( obj a_class );
obj class_supers( obj a_class );
unsigned class_image_mode( obj a_class );

/* returns YES for indirect instances/subclasses as well */
bool obj_isa_instance_of( obj thing, obj a_class );
bool subclass_p( obj class1, obj class2 );

/** INITIALIZATION **/

void init_runtim( bool verbose );

#ifdef INLINES
#include "interim.ci"
#endif


#endif /* _H_RUNTIM */
