/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#ifndef _H_SMEMORY
#define _H_SMEMORY

#include "obj.h"
#include "gcserver.h"

typedef struct _POBHeader {
#ifdef ATRACE
    UINT_32	name;
    UINT_32	dummy;
#endif
    UINT_32	pob_size;
    obj		pob_class;
} POBHeader;

#define SLOT(i) ((i)*sizeof(obj))

#if 1 || defined(NDEBUG)
#define FRONT_PORCH	(0)	/* in WORDS */
#define BACK_PORCH	(0)
#else
#define FRONT_PORCH	(8)	/* in WORDS */
#define BACK_PORCH	(8)
#endif

/*  Low-level storage allocators.  Return uninitialized bits.  */

CIH_DECL obj alloc( UINT_32 bytes, obj obj_class );
CIH_DECL obj alloc1( obj obj_class );
CIH_DECL obj alloc2( obj obj_class );
CIH_DECL obj alloc3( obj obj_class );
CIH_DECL obj alloc4( obj obj_class );
CIH_DECL obj alloc5( obj obj_class );
CIH_DECL obj alloc6( obj obj_class );
CIH_DECL obj alloc7( obj obj_class );
CIH_DECL obj alloc8( obj obj_class );
CIH_DECL obj alloc9( obj obj_class );

/*  Allocate a gvec filled with FALSE_OBJ's  */

CIH_DECL obj gvec_alloc( UINT_32 num_slots, obj obj_class );

/*  Allocate a bvec, with last word set to zeros */

CIH_DECL obj bvec_alloc( UINT_32 num_bytes, obj bvec_class );

CIH_DECL UINT_32 SIZEOF_PTR( obj ptr );
CIH_DECL obj CLASSOF_PTR( obj ptr );
#define OBJ_ISA_PTR_OF_CLASS(x,c) ((OBJ_ISA_PTR(x) \
				    && EQ(CLASSOF_PTR(x),c)) \
				  ? YES : NO)

CIH_DECL void *PTR_TO_DATAPTR( obj ptr );

CIH_DECL POBHeader *PTR_TO_HDRPTR( obj ptr );
CIH_DECL obj HDRPTR_TO_PTR( POBHeader *hdr );

CIH_DECL gc_obj_addr PTR_TO_GCPTR( obj x );
CIH_DECL obj GCPTR_TO_PTR( gc_obj_addr x );

CIH_DECL void safe_point( UINT_32 headroom );


/******************** Memory Read Functions ********************/

CIH_DECL UINT_8 bvec_read( obj bvec, UINT_32 byte_offset );
CIH_DECL obj gvec_read( obj gvec, UINT_32 byte_offset );

/******************** Memory Write Functions ********************/

CIH_DECL void bvec_write( obj bvec, UINT_32 byte_offset, UINT_8 value );

CIH_DECL void gvec_write( obj gvec, UINT_32 byte_offset, obj value );
CIH_DECL void gvec_write_ptr( obj gvec, UINT_32 byte_offset, obj value );
CIH_DECL void gvec_write_non_ptr( obj gvec, UINT_32 byte_offset, obj value );

/* an object is FRESH from the time it is allocated to
   the next GC_SAFE_POINT.  This is extraordinarily useful
   if you are allocating white, in which case the write
   barrier becomes a NOP
*/

CIH_DECL void gvec_write_fresh( obj gvec, UINT_32 byte_offset, obj value );
CIH_DECL void gvec_write_fresh_ptr( obj gvec, UINT_32 byte_offset, obj value );
CIH_DECL void gvec_write_fresh_non_ptr( obj gvec, UINT_32 byte_offset, obj value );

/* lower-level allocation */

CIH_DECL void *malloc_aligned_32( UINT_32 bytes );
CIH_DECL void free_aligned_32( void *ptr );


#ifdef INLINES

#if !(defined(_CI_GCCLIENT1) || defined(_CI_GCCLIENT2))
#ifndef FOREIGN
#include "gcserver1.ci"
#include "gcserver2.ci"
#endif
#endif

#include "smemory.ci"
#include "readwrit.ci"
#endif

#endif /* _H_SMEMORY */
