/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "vinsns.h"

#ifdef STACK_CACHE

obj the_stack_cache[CACHE_SIZE];
obj *cache_iteration;

#ifndef INLINES
#include "stakcach.ci"
#endif /* INLINES */

void init_stack_cache( void )
{
}

/*  FLUSH_STACK_CACHE() -- Does the obvious thing, which is flush the stack
    cache.  In effect, it guarantees that the current continuation
    is a REAL object, flushing the cache sufficiently to ensure the
    truth of this.
*/

void flush_stack_cache( void )
{
UINT_32 cr = VAL(continuation_reg);
#define x  ((obj *)(cr-POINTER_TAG))
obj n, prev, next, first;
bool more;
UINT_32 i, size_in_bytes;

    if (!in_stack_cache(continuation_reg))
	return; /* nothing to do... already a real object */

#ifdef STEP_DUMP
    {
    extern FILE *step_dump_file;
    extern int do_step_dump;
    extern void touch_step_dump_file( void );
    
	if (do_step_dump && step_dump_file)
	{
	    touch_step_dump_file();
	    fprintf( step_dump_file, "** Flushing stack cache **\n" );
	}
    }
#endif /* STEP_DUMP */

    more = YES;
    prev = ZERO;
    first = ZERO;	/* suppress warning */
    do {
	next = x[3];
	if (in_stack_cache(next))
	{
	    size_in_bytes = VAL(next) - cr;
	}
	else
	{
	    size_in_bytes = cache_upper_limit - cr;
	    more = NO;
	}
	n = alloc( size_in_bytes, partcont_class );
	if (EQ(prev,ZERO))
	    first = n;
	else
	    gvec_write_fresh( prev, SLOT(3), n );
	gvec_write_fresh( n, SLOT(0), x[0] );
	gvec_write_fresh( n, SLOT(1), x[1] );
	gvec_write_fresh( n, SLOT(2), x[2] );
	for (i=SLOT(4); i<size_in_bytes; i+=SLOT(1))
	{
	    gvec_write_fresh( n, i, *(obj *)(i+(char *)x) );
	}
	prev = n;
	cr = VAL(next);
    } while (more);
    gvec_write_fresh( n, SLOT(3), next );

    continuation_reg = first;
#undef x
}



/*
    Restores a dynamically-determined number of
    argument-passing registers (ie, for use in the
    extremely rare cases when it cannot be determined
    statically -- nb, this only happens when compiling
    to closure-threaded code) and the continuation
    register itself
*/

unsigned restore_arb( void )
{
#define x ((obj *)(cr - POINTER_TAG))
UINT_32 cr = VAL(continuation_reg), size_in_bytes, next;
unsigned n;

    if (cr < cache_upper_limit && cr >= cache_lower_limit)
    {
	next = VAL(x[3]);
	if (next < cache_upper_limit && next >= cache_lower_limit)
	{
	    size_in_bytes = next - cr;
	}
	else
	{
	    size_in_bytes = cache_upper_limit - cr;
	}
    }
    else
	size_in_bytes = SIZEOF_PTR(continuation_reg);

    n  = (size_in_bytes / sizeof(obj)) - CONT_FIXED;
    restore_cont( n );
    return n;
#undef x
}

#endif /* STACK_CACHE */
