/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#ifndef _H_VINSNS_GNU
#define _H_VINSNS_GNU

/* trap_jump() intercepts a jump address at a safe point,
   possibly does some GC, and may return a different jump
   address (in case of an INTERRUPT)
*/

#ifdef STEP_DUMP
extern void step_dump( jump_addr f );
extern int do_step_dump;
extern unsigned insn_count;
#endif

static inline jump_addr trap_jump( jump_addr p )
{
#ifdef STEP_DUMP
    if (do_step_dump)
	step_dump(p);
    insn_count++;
#endif
    safe_point(4000);
    return p;
}

/*  Structuring things  */

#define PROLOGUE(fn_name)       static void **fn_name( void ) {
#define MONOTONE(name)		name:
#define EPILOGUE(fn_name)       }
				
#define BEGIN_FWD(fn_name)      static void *(label_entries[]) = {
#define FWD_MONOTONE(name)	&&name,
#define END_FWD(fn_name)        NULL }; return label_entries;

#define BEGIN_BACK(fn_name)   
#define BACK_MONOTONE(name)
#define END_BACK(fn_name)

#define JUMP_TABLE(fn_name)     (jump_addr *)fn_name

/*  The virtual machine instructions  */


#define STMT(body) do { body } while (0)

/* NOTE:  At least for now, we DO NOT trap on forward
   jumps.  This makes them EXTREMELY efficient.  It is
   also the reason for the distinction between fwd
   jumps and backward jumps.
*/

#ifdef STEP_DUMP
#define JUMP(num,label)	STMT(arg_count_reg = num; \
			     if (do_step_dump) step_dump(&&label); \
			     insn_count++; goto label;)
#else
#define JUMP(num,label)	STMT(arg_count_reg = num; goto label;)
#endif
#define BJUMP(num,label) STMT(arg_count_reg = num; \
			      goto *trap_jump(LABEL_TO_JUMP_ADDR(label));)

#define XJUMP(num,tmpl) STMT(arg_count_reg = num; \
			     literals_reg = tmpl; \
			     goto *trap_jump(OBJ_TO_JUMP_ADDR( \
					gvec_read( literals_reg, SLOT(0) ) ));)

#ifdef STEP_DUMP
#define APPLY(num,expr)	STMT(jump_addr f; arg_count_reg = num; \
			     f = apply(expr); \
			     if (do_step_dump) step_dump(f);\
			     insn_count++; goto *f;)
#define RETURN(num) STMT(jump_addr f; arg_count_reg = num; \
			     f = half_restore(); \
			     if (do_step_dump) step_dump(f);\
			     insn_count++; goto *f;)
#else
#define APPLY(num,expr)	STMT(arg_count_reg = num; goto *apply(expr);)
#define RETURN(num) STMT(arg_count_reg = num; goto *half_restore();)
#endif

#define LABEL_TO_JUMP_ADDR(label) &&label

#endif /* _H_VINSNS_GNU */
