#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define *default-ct-formatters*
    '((#\a . %display)
      (#\s . %write)
      (#\d . write-dec)
      (#\x . write-hex)
      (#\b . write-bin)
      (#\o . write-oct)
      (#\c . write-char)
      (#\> . write-tab4)
      (#\~ . #\~)
      (#\t . #\tab)
      (#\% . #\newline)
      (#\_ . #\space)))

(define (format:parse format-fns str)
    (let ((literal-buf '())
	  (result '()))

	(define (flush)
	    (if (pair? literal-buf)
		(begin
		    (push (list->string (reverse literal-buf)))
		    (set! literal-buf '()))))

	(define (add-ch ch)
	    (set! literal-buf (cons ch literal-buf)))
	
	(define (push x)
	    (set! result (cons x result)))

	(let loop ((str (string->list str)))
	    (if (pair? str)
		(if (eq? (car str) #\~)
		    (if (null? (cdr str))
			(add-ch (car str))
			(let ((cmd (assq (cadr str) format-fns)))
			    (if cmd
				(if (char? (cdr cmd))
				    (add-ch (cdr cmd))
				    (begin
					(flush)
					(push (cdr cmd))))
				(abort 'ct-format 
					    "Illegal format character: ~s" 
					    (cadr str)))
			    (loop (cddr str))))
		    (begin
			(add-ch (car str))
			(loop (cdr str))))))

	    (flush)
	    (reverse result)))

(define (format:wrap-args format-list args)
    (let loop  ((format-list format-list) 
    		(result '(begin)) 
		(args args))
	(if (null? format-list)
	    (if (null? args)
		(reverse result)
		(abort 'ct-format "Excess arguments: ~a" args))
	    (if (symbol? (car format-list))
		(if (null? args)
		    (abort 'ct-format "Missing arguments for: ~a" format-list)
		    (loop (cdr format-list) 
			  (cons (list (car format-list) (car args) '$$port)
				result)
			  (cdr args)))
		(loop (cdr format-list)
		      (cons (list '$$write-str '$$port (car format-list))
			    result)
		      args)))))

;; compile-time format

(define (ct-format-rewriter format-fns form)
    (let ((port-expr (cadr form))
	  (fmt-str (caddr form))
	  (fmt-args (cdddr form)))
	(type-check fmt-str string? <String>)
	(let ((port-init-expr (if port-expr
				    (if (eq? port-expr #t)
					'(current-output-port)
					port-expr)
				    '(open-output-string)))
	      (parsed-format (format:parse format-fns fmt-str)))
	    (list 'let*
		    (list (list '$$port port-init-expr)
		    	  '($$write-str 
				(find-implementation
				    %write-str
				    (list $$port ""))))
		    (format:wrap-args parsed-format fmt-args)
		    (if port-expr
			'(values)
			'(close-output-port $$port))))))
