#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;
;  higher-level i/o procedures...
;

(define (with-output-to-port port thunk)
    (%with-dynamic-state
	(cons (cons '*output-port* port)
	      %dynamic-state-reg)
	thunk))

(define (with-input-from-port port thunk)
    (%with-dynamic-state
	(cons (cons '*input-port* port)
	      %dynamic-state-reg)
	thunk))

; variants for accessing files...

(define (call-with-output-file string proc)
    (let ((file (open-output-file string)))
	(let ((result (proc file)))
	    (close-output-port file)
	    result)))

(define (with-output-to-file string thunk)
    (let ((file (open-output-file string)))
	(let ((result (with-output-to-port file thunk)))
	    (close-output-port file)
	    result)))

(define (call-with-input-file string proc)
    (let ((file (open-input-file string)))
	(let ((result (proc file)))
	    (close-input-port file)
	    result)))

(define (with-input-from-file string thunk)
    (let ((file (open-input-file string)))
	(let ((result (with-input-from-port file thunk)))
	    (close-input-port file)
	    result)))

(define (with-input-from-string string thunk)
    (let ((port (open-input-string string)))
	(let ((result (with-input-from-port port thunk)))
	    (close-input-port port)
	    result)))

; for-each for ports

(define (with-objects-from-port port proc)
    (let ((item (read port))
     	  (n 0))
	(if (eof-object? item)
	    n
	    (begin
		(proc item)
		(set! n (+ n 1))
		(with-objects-from-port port proc)))))

(define (with-objects-from-file file proc)
    (let ((port (open-input-file file)))
	(with-objects-from-port port proc)
	(close-input-port port)))

(export with-output-to-port with-input-from-port)
(export call-with-output-file with-output-to-file)
(export call-with-input-file with-input-from-file with-input-from-string)
(export with-objects-from-port with-objects-from-file)
