#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define-generic-fn get-thunk)
(define-generic-fn ready?-thunk)
(define-generic-fn num-left)

(define-generic-fn set-get-thunk!)
(define-generic-fn set-ready?-thunk!)
(define-generic-fn set-num-left!)

(define-class <Stream-InputPort> (<Object>)
    get-thunk
    ready?-thunk
    buffer
    index
    num-left)

(define (buffer-did-fill self item)
    (set-buffer! self item)
    (cond ((char? item)
    		(set-num-left! self 1))
	  ((string? item)
	  	(set-num-left! self (string-length item))
		(set-index! self 0))
	  (else
		(assert (%isa-ptr? item indirectly: <ByteVector>))
	  	;; assume it's a byte vector
		;; whose 1st byte represents the number of
		;; valid characters
		(set-num-left! self (%bvec-read item 0))
		(set-index! self 1))))
		
(define (fill-buffer! self)
    (let ((item ((get-thunk self))))
	(if item
	    (buffer-did-fill self item)
	    (set-get-thunk! self #f))))

(define (advance! self)
    (let ((n (sub1 (num-left self))))
	(if (eq? n 0)
	    (set-buffer! self #f)
	    (begin
		(set-num-left! self n)
		(set-index! self (add1 (index self)))))))


(define-method %char-ready? ((self <Stream-InputPort>))
    (or (buffer self)
    	(not (get-thunk self))
        ((ready?-thunk self))))

(define-method %peek-char ((self <Stream-InputPort>))
    (let loop ((b (buffer self)))
	(if b
	    (if (char? b)
		b
		(integer->char (%bvec-read b (index self))))
	    (if (get-thunk self)
		(begin
		    (fill-buffer! self)
		    (loop (buffer self)))
		*eof-object*))))

(define-method %read-char ((self <Stream-InputPort>))
    (let ((result (peek-char self)))
	(if (not (eof-object? result))
	    (advance! self))
	result))

(define-method %scan-token ((port <Stream-InputPort>))
    (abort 'scan-token "Not implemented for buffer ports (yet)"))

(define (open-buffered-input get-thunk ready?-thunk)
    (make <Stream-InputPort>
	get-thunk: get-thunk
	ready?-thunk: ready?-thunk
	buffer: #f
	index: #f
	num-left: 0))
	
(export open-buffered-input)
