/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "privio.h"
#include "bltin_p.h"
#include "scmio_p.h"
#include <stdio.h>

volatile void raise_exception( obj key, unsigned num_data, ... );

/*
    [4]  buffer
    [5]  index
    [6]  completions
    [7]  prompt
*/

static bool console_buffer_fill( obj port )
{
const char *result, *prompt;
obj t, completions;

    t = gvec_read( port, CIP_PROMPT );
    if (EQ(t,FALSE_OBJ))
	prompt = ">";
    else
	prompt = string_text( t );

    t = gvec_read( port, CIP_COMPLETIONS );
    if (EQ(t,FALSE_OBJ))
	completions = NIL_OBJ;
    else
	completions = t;

    result = read_console_line( completions, prompt );
    if (result)
    {
    obj str = make_string(result);
    
	gvec_write_ptr( port, CIP_BUFFER, str );
	return YES;
    }
    else
    {
	gvec_write_non_ptr( port, CIP_BUFFER, FALSE_OBJ );
	return NO;
    }
}


obj CIP_peek_char( obj console_inport )
{
char ch;
obj ix;
const char *cb;

    ix = gvec_read( console_inport, CIP_INDEX );
    if (EQ(ix, FALSE_OBJ))
    {
	if (!console_buffer_fill(console_inport))
	    return eof_object;
	gvec_write_non_ptr( console_inport, CIP_INDEX, ZERO );
	ix = ZERO;
    }
    cb = string_text( gvec_read( console_inport, CIP_BUFFER ) );
    ch = cb[fx2int(ix)];
    if (ch)
    {
	return MAKE_ASCII_CHAR( ch );
    }
    return MAKE_ASCII_CHAR( '\n' );
}

obj CIP_read_char( obj console_inport )
{
char ch;
obj ix;
const char *cb;

    ix = gvec_read( console_inport, CIP_INDEX );
    if (EQ(ix, FALSE_OBJ))
    {
	if (!console_buffer_fill(console_inport))
	    return eof_object;
	ix = ZERO;
    }
    cb = string_text( gvec_read( console_inport, CIP_BUFFER ) );
    ch = cb[fx2int(ix)];
    if (ch)
    {
	gvec_write_non_ptr( console_inport, CIP_INDEX, ADD1(ix) );
	return MAKE_ASCII_CHAR( ch );
    }
    gvec_write_non_ptr( console_inport, CIP_INDEX, FALSE_OBJ );
    return MAKE_ASCII_CHAR( '\n' );
}

bool CIP_char_ready_p( obj console_inport )
{
obj ix;

    ix = gvec_read( console_inport, CIP_INDEX );
    if (EQ(ix, FALSE_OBJ))
	return NO;
    return YES;
}
