/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "privio.h"
#include "bltin_p.h"
#include "scmio_p.h"
#include "osglue.h"

#include <stdio.h>

static inline obj path( obj file_port )
{
    return gvec_read( file_port, FIP_PATH );
}

static inline FILE *c_file( obj file_port )
{
FILE *f;

    f = (FILE *)OBJ_TO_RAW_PTR( gvec_read( file_port, FIP_CFILE ) );
    if (!f)
	scheme_error( "~s already closed", 1, path(file_port) );
    return f;
}


obj FIP_peek_char( obj file_inport )
{
obj ch;

    ch = gvec_read( file_inport, FIP_BUFFER );
    if (EQ(ch, FALSE_OBJ))
    {
        ch = MAKE_ASCII_CHAR( fgetc( c_file(file_inport) ) );
	if (feof(c_file(file_inport)))
	{
	    gvec_write_non_ptr( file_inport, FIP_BUFFER, NIL_OBJ );
	    return eof_object;
	}
	gvec_write_non_ptr( file_inport, FIP_BUFFER, ch );
    }
    else if (EQ(ch, NIL_OBJ))
    {
        ch = eof_object;
    }
    return ch;
}

obj FIP_read_char( obj file_inport )
{
obj ch;

    ch = gvec_read( file_inport, FIP_BUFFER );
    if (EQ(ch, FALSE_OBJ))
    {
        ch = MAKE_ASCII_CHAR( fgetc( c_file(file_inport) ) );
	if (feof(c_file(file_inport)))
	{
	    gvec_write_non_ptr( file_inport, FIP_BUFFER, NIL_OBJ );
	    return eof_object;
	}
    }
    else if (EQ(ch, NIL_OBJ))
    {
        ch = eof_object;
    }
    else
	gvec_write_non_ptr( file_inport, FIP_BUFFER, FALSE_OBJ );
    return ch;
}

/* R4RS: "If the port is at end of file then char-ready? returns #t" */

bool FIP_char_ready_p( obj file_port )
{
    return YES;	/* this may not be quite correct... If the port
    		   is attached to something that may hang on `getc',
		   like stdin, then sometimes (when?  how to tell?)
		   we should return NO */
}

void FP_close( obj fp )
{
    assert( !EQ(fp,ZERO) && !OBJ_ISA_PTR(fp) );
    fclose( (FILE *)OBJ_TO_RAW_PTR( fp ) );
}

void FP_flush( obj fp )
{
    assert( !EQ(fp,ZERO) && !OBJ_ISA_PTR(fp) );
    fflush( (FILE *)OBJ_TO_RAW_PTR( fp ) );
}

obj FIP_make( obj filename, obj fip_class )
{
FILE *f;

    assert( STRING_P(filename) );
    
    f = os_fopen( string_text(filename), "r" );
    if (!f)
	return FALSE_OBJ;

    return make3( fip_class, FALSE_OBJ, RAW_PTR_TO_OBJ(f), filename );
}
