#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(prologue {
#include "osglue.h"
})

(define-method %peek-char ((port <File-InputPort>))
    (%call ("FIP_peek_char" (<obj>) => <obj>) port))

(define-method %peek-char ((port <String-InputPort>))
    (%call ("SIP_peek_char" (<obj>) => <obj>) port))

(define-method %peek-char ((port <Console-InputPort>))
    (%call ("CIP_peek_char" (<obj>) => <obj>) port))


(define-method %read-char ((port <File-InputPort>))
    (%call ("FIP_read_char" (<obj>) => <obj>) port))

(define-method %read-char ((port <String-InputPort>))
    (%call ("SIP_read_char" (<obj>) => <obj>) port))

(define-method %read-char ((port <Console-InputPort>))
    (%call ("CIP_read_char" (<obj>) => <obj>) port))


(define-method %char-ready? ((port <File-InputPort>))
    (%call ("FIP_char_ready_p" (<obj>) => <raw-bool>) port))

(define-method %char-ready? ((port <String-InputPort>))
    (%call ("SIP_char_ready_p" (<obj>) => <raw-bool>) port))
    
(define-method %char-ready? ((port <Console-InputPort>))
    (%call ("CIP_char_ready_p" (<obj>) => <raw-bool>) port))

(%define-glue (%scan-token/FIP port)
{
    REG0 = inport_scan_token( port, FIP_peek_char, FIP_read_char );
    RETURN(1);
})

(define-method %scan-token ((port <File-InputPort>))
    (%scan-token/FIP port))

(%define-glue (%scan-token/SIP port)
{
    REG0 = inport_scan_token( port, SIP_peek_char, SIP_read_char );
    RETURN(1);
})

(define-method %scan-token ((port <String-InputPort>))
    (%scan-token/SIP port))

(%define-glue (%scan-token/CIP port)
{
    REG0 = inport_scan_token( port, CIP_peek_char, CIP_read_char );
    RETURN(1);
})

(define-method %scan-token ((port <Console-InputPort>))
    (%scan-token/CIP port))

(define-method close-input-port ((self <InputPort>))
    (values))	;; default is a NOP

(define-method %read-line ((port <InputPort>))
    (let ((line (make-seq)))
	(let loop ()
	    (let ((ch (%read-char port)))
		(if (eof-object? ch)
		    (if (seq-empty? line)
			*eof-object*
			(list->string (seq->list line)))
		    (if (eq? ch #\newline)
			(list->string (seq->list line))
			(begin
			    (seq-add! line ch)
			    (loop))))))))

(define *open-input-files* '())

(define (open-input-file* pathname)
    (type-check pathname string? <String>)
    (let ((p (%call ("FIP_make" (<obj> <obj>) => <obj>) 
    		pathname
		<File-InputPort>)))
	(if p
	    (set! *open-input-files*
		  (cons p *open-input-files*)))
	p))

(define-method close-input-port ((self <File-InputPort>))
    (if (eq? (c-file self) 0)
	(abort 'close-input-port "Port ~s already closed" self)
	(begin
	    (set! *open-input-files* (delq! self *open-input-files*))
	    (%call ("FP_close" (<obj>)) (c-file self))
	    (set-c-file! self 0)
	    (path self))))

(define (%for-each-input-file proc)
    (for-each proc *open-input-files*))


(%define-glue (file->string path)
{
FILE *f;
char *text;
size_t n;

    f = os_fopen( string_text(path), "r" );
    if (!f)
	REG0 = FALSE_OBJ;
    else
    {
	fseek( f, 0L, SEEK_END );
	n = ftell(f);
	fseek( f, 0L, SEEK_SET );
	
	REG0 = bvec_alloc( n+1, string_class );
	text = (char *)PTR_TO_DATAPTR(REG0);
	
	fread( text, 1, n, f );
	fclose( f );
    }
    RETURN(1);
})

(export file->string)
