#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-generic-fn %write-str)
(define-generic-fn %write-binary)
(define-generic-fn %write-bvec)

#|
    (define-generic-fn %write-int)
    (define-generic-fn %write-float)
    (define-generic-fn %write-char)
|#

(%define-glue (%write-str/FOP port str)
{
    assert( STRING_P(str) );
    FOP_write( port, string_text(str), string_length(str) );
    RETURN0();
})

(%define-glue (%write-str/SOP port str)
{
    assert( STRING_P(str) );
    SOP_write( port, string_text(str), string_length(str) );
    RETURN0();
})

(%define-glue (%write-str/BSOP port str)
{
    assert( STRING_P(str) );
    if (!BSOP_write( port, string_text(str), string_length(str) ))
    {
	APPLY( 1, gvec_read( port, BSOP_ESCAPE ) );
    }
    RETURN0();
})


(%define-glue (%write-bvec/FOP port bvec use_len)
{
    FOP_write( port, (const char *)PTR_TO_DATAPTR(bvec), fx2int(use_len) );
    RETURN0();
})

(%define-glue (%write-bvec/SOP port bvec use_len)
{
    SOP_write( port, (const char *)PTR_TO_DATAPTR(bvec), fx2int(use_len) );
    RETURN0();
})

(%define-glue (%write-bvec/BSOP port bvec use_len)
{
    if (!BSOP_write( port, (const char *)PTR_TO_DATAPTR(bvec),
    			   fx2int(use_len) ))
    {
	APPLY( 1, gvec_read( port, BSOP_ESCAPE ) );
    }
    RETURN0();
})


(%define-glue (%write-binary/FOP port val)
{
INT_32 i;

    assert (OBJ_ISA_FIXNUM(val));
    i = fx2int(val);
    FOP_write( port, (const char *)&i, sizeof(INT_32) );
    RETURN0();
})

(%define-glue (%write-binary/SOP port val)
{
INT_32 i;

    assert (OBJ_ISA_FIXNUM(val));
    i = fx2int(val);
    SOP_write( port, (const char *)&i, sizeof(INT_32) );
    RETURN0();
})

(%define-glue (%write-binary/BSOP port val)
{
INT_32 i;

    assert (OBJ_ISA_FIXNUM(val));
    i = fx2int(val);
    if (!BSOP_write( port, (const char *)&i, sizeof(INT_32) ))
    {
	APPLY( 1, gvec_read( port, BSOP_ESCAPE ) );
    }
    RETURN0();
})

(define-method %write-str ((port <File-OutputPort>) str)
    (%write-str/FOP port str))
    
(define-method %write-str ((port <String-OutputPort>) str)
    (%write-str/SOP port str))

(define-method %write-str ((port <BoundedString-OutputPort>) str)
    (%write-str/BSOP port str))


(define (%write-int port num)
    (%write-str port (number->string num)))
    
(define (%write-float port num)
    (%write-str port (number->string num)))
    
(define (%write-char port ch)
    (%write-str port (string ch)))


(define-method %write-binary ((port <File-OutputPort>) val)
    (%write-binary/FOP port val))

(define-method %write-binary ((port <String-OutputPort>) val)
    (%write-binary/SOP port val))

(define-method %write-binary ((port <BoundedString-OutputPort>) val)
    (%write-binary/BSOP port val))


(define-method %write-bvec ((port <File-OutputPort>) bvec use_len)
    (%write-bvec/FOP port bvec use_len))

(define-method %write-bvec ((port <String-OutputPort>) bvec use_len)
    (%write-bvec/SOP port bvec use_len))

(define-method %write-bvec ((port <BoundedString-OutputPort>) bvec use_len)
    (%write-bvec/BSOP port bvec use_len))

(%define-glue (%raw->string thing)
{
char temp[20];

    sprintf( temp, "%x", VAL(thing) );
    REG0 = make_string(temp);
    RETURN(1);
})

#|
(%define-glue (%number->string-w/radix value radix)
{
char temp[40], *p;
UINT_32 num = fx2int(value);
unsigned r = fx2int(radix);
short digit;

    p = &temp[40];
    *--p = 0;
    
    if (num)
    {
	while (num)
	{
	    digit = num % r;
	    num /= r;
	    *--p = ((digit < 10) ? '0' : ('a' - 10)) + digit;
	}
    }
    else
	*--p = '0';

    REG0 = make_string(p);
    RETURN(1);
})
|#

(define (%write-raw port thing)
    (%write-str port (%raw->string thing)))

(%define-glue (%COP-make consolename) literals: ((& <File-OutputPort>))
{
    REG0 = FOP_fmake( stdout, consolename, TLREF(0) );
    RETURN(1);
})

(define-syntax %FOP-make
    (syntax-form (filename)
	(%call ("FOP_make" (<obj> <obj>) => <obj>) filename 
						   <File-OutputPort>)))

(define-syntax %SOP-make
    (syntax-form ()
	(make <String-OutputPort>
	    buffer: (%bvec-alloc <ByteVector> 16)
	    index: 0
	    overflow: '())))
	    
(define-syntax %BSOP-make
    (syntax-form (using escape)
	(make <BoundedString-OutputPort> 
	    buffer: using 
	    index: 0 
	    escape: escape)))

(export %write-int %write-str %write-char %write-binary %write-bvec
	%write-raw %raw->string)


(define *open-output-files* '())

(define (open-output-file* pathname)
    (let ((p (%FOP-make pathname)))
	(if p
	    (set! *open-output-files*
		    (cons p *open-output-files*)))
	p))

(export open-output-file*)

(define-method close-output-port ((self <String-OutputPort>))
    (%call ("SOP_close" (<obj>) => <obj>) self))
    
(define-method close-output-port ((self <BoundedString-OutputPort>))
    (%call ("BSOP_close" (<obj>) => <obj>) self))

(define-method close-output-port ((self <File-OutputPort>))
    (if (eq? (c-file self) 0)
	#f
	(begin
	    (set! *open-output-files* (delq! self *open-output-files*))
	    (%call ("FP_close" (<obj>)) (c-file self))
	    (set-c-file! self 0)
	    (path self))))
;	(abort 'close-output-port "Port ~s already closed" self)

(define-method flush-output-port ((self <File-OutputPort>))
    (if (eq? (c-file self) 0)
	#f
	(begin
	    (%call ("FP_flush" (<obj>)) (c-file self))
	    (values))))
;	(abort 'flush-output-port "Port ~s is closed" self)

(define (%for-each-output-file proc)
    (for-each proc *open-output-files*))
