#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-syntax using-port
    (syntax-form (port)
	(if (null? port) 
	    (current-output-port) 
	    (car port))))

(define (newline . port)
    (%write-char (using-port port) #\newline ))

(define (write-char ch . port)
    (%write-char (using-port port) ch))


;; creating output ports

(define (open-output-file path)
    (let ((p (open-output-file* path)))
	(if p
	    p
	    (abort 'open-output-file "Could not open: ~s" path))))

(define (open-output-string)
    (%SOP-make))

;; recognizing them

(define (file-output-port? p)
    (%isa-ptr? p directly: <File-OutputPort>))

(define (output-port? p)
    (%isa-ptr? p indirectly: <OutputPort>))
    
;; higher-level access to bounded-string ports

(define (with-bounded-string-port size proc)
    (close-output-port
		(call-with-current-continuation
		    (lambda (exit)
			(let ((port (%BSOP-make (make-string size) exit)))
			    (proc port)
			    port)))))

(define (with-bounded-string-port* size proc finish-proc1 finish-proc2)
    (let ((p finish-proc2))
	(let ((result (close-output-port
			(call-with-current-continuation
			    (lambda (exit)
				(let ((port (%BSOP-make (make-string size) 
						        exit)))
				    (proc port)
				    (set! p finish-proc1)
				    port))))))
	    (p result))))

(define (object->bounded-string length thing)
    (with-bounded-string-port* 
    		length 
		(lambda (p) (write thing p))
		(lambda (str) str)
		(lambda (str) (string-append str "..."))))

(define (object->string thing)
    (let ((port (open-output-string)))
	(write thing port)
	(close-output-port port)))

;;

(export open-output-string open-output-file)

(export file-output-port? output-port?)

(export newline write-char)
(export object->string object->bounded-string)
(export with-bounded-string-port with-bounded-string-port*)
