#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (read:internal-error code info)
    (read:error "Internal[~d]: ~s" code info))

(define (read:error msg . info-list)
    (apply* 'read msg info-list abort))

(define (read:parse-vector scanner first last)
    (let ((item (read:parse-object* scanner)))
	(if (eof-object? item)
	    (read:error "missing <close-paren>")
	    (if (token? item)
		(case (token-type item)
		    ((<close-paren>) (list->vector first))
		    (else (read:internal-error 1 item)))
		(let ((i (cons item '())))
		    (if (null? first)
			(read:parse-vector scanner i i)
			(begin
			    (set-cdr! last i)
			    (read:parse-vector scanner first i)))))
	)))

(define (read:parse-list scanner first last)
    (let ((item (read:parse-object* scanner)))
	(if (eof-object? item)
	    (read:error "missing <close-paren>")
	    (if (token? item)
		(case (token-type item)
		    ((<close-paren>) first)
		    ((<dot>)
			(if (null? first)
			    (read:error "<dot> before any items")
			    (begin
				(set-cdr! last (read:parse-object* scanner))
				(let ((after (read:parse-object* scanner)))
				    (if (and (token? after)
				    	     (eq? (token-type after) 
					          '<close-paren>))
					first
					(read:error
					    "too many items after <dot>: ~s"
					    after))))))
		    (else (read:internal-error 2 item)))
		(let ((i (cons item '())))
		    (if (null? first)
			(read:parse-list scanner i i)
			(begin
			    (set-cdr! last i)
			    (read:parse-list scanner first i)))))
	)))

(define (read:parse-object* scanner)
    (let ((s (scanner #t)))
	(cond 
	    ((token? s)
		    (case (token-type s)
			((<open-paren>) 
			    (read:parse-list scanner '() '()))
			((<close-paren> <dot>) s)
			((<open-vector>)
			    (read:parse-vector scanner '() '()))
			((<number> <symbol> <literal>) 
			    (token-data s)) 
			((<quote>) 
			    (list 'quote 
			    	  (read:parse-object* scanner)))
			((<unquote>) 
			    (list 'unquote 
			    	  (read:parse-object* scanner)))
			((<unquote-splicing>) 
			    (list 'unquote-splicing
			    	  (read:parse-object* scanner)))
			((<quasiquote>) 
			    (list 'quasiquote 
			    	  (read:parse-object* scanner)))
			((<object-ref>)
			    (find-object-ref (token-data s)))
			(else
			    (read:error 
				"Strange token in input: ~s" 
				s))
		    ))
	    ((eof-object? s) s)
	    ((char? s) 
		(lexical-error (scanner #f) 
			       "(read) got invalid token: char ~s" 
			       s))
	    (#t (read:error 
		    "Unrecognized kind of token in input: ~s" 
		    s)))))

#|
(define (lexical-error port)
    (let ((ch (read-char port)))
	(let loop ((i (peek-char port)))
	    (if (or (eq? i #\newline) (eof-object? i))
		(read:error "Lexical error at: ~s" ch)
		(begin
		    (read-char port)
		    (loop (peek-char port)))))))
|#

(define (read:parse-object scanner)
    (let ((result (read:parse-object* scanner)))
	(if (token? result)
	    (case (token-type result)
		((<close-paren>) 
		    (read:error "Unmatched <close-paren>"))
		((<dot>)
		    (read:error "Misplaced <dot>"))
		(else
		    (read:internal-error 3 result)))
	    result)))

(define (read:make-console-scanner port)
    (let ((alt-prompt (fluid-ref *PS2* "> ")))
	(lambda (flag?)
	    (if flag?
		; with #t, we scan the next token and return it
		(let ((result (scan-token port)))
		    (set-input-port-prompt! port alt-prompt)
		    result)
		; with #f, we return the port that we're scanning
		port))))

(define (read . args)
    (let ((port (if (null? args) (current-input-port) (car args))))
	(type-check port input-port? <InputPort>)
	(if (console-input-port? port)
	    (let ((old-prompt (input-port-prompt port))
	    	  (result (read:parse-object 
	    			(read:make-console-scanner port))))
		(set-input-port-prompt! port old-prompt)
		result)
	    (read:parse-object (lambda (flag) 
				(if flag
				    (scan-token port)
				    port))))))

#|
	(call-with-current-continuation
	    (lambda (exit-read)
		(%with-dynamic-state
		    (cons (cons '*EOI* (lambda (key . data)
					    (exit-read *eof-object*)))
			    %dynamic-state-reg)
		    (lambda ()
			(read:parse-object 
				      (if (console-input-port? port)
					  (read:make-console-scanner port)
					  (lambda () 
					     (scan-token port))))))))					     
|#
