#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-class <User-OutputPort> (<OutputPort>)
    (write-more '#uninit <Closure>)
    (close-proc #f))

(define-method %write-str ((self <User-OutputPort>) (str <String>))
    ((write-more self) str))

(define-method close-output-port ((self <User-OutputPort>))
    (if (close-proc self)
	((close-proc self))))

(define-class <User-InputPort> (<InputPort>)
    (buffer '())
    (index 0)
    (read-more '#uninit <Closure>))

(define (user-fill! (self <User-InputPort>))
    (let* ((reader (read-more self))
    	   (more (if reader (reader) #f)))
	(if more
	    (begin
		(if (string? more)
		    (set-buffer! self (string->list more))
		    (abort 'user-fill! "read-more of ~s returned: ~s"
		    	   self
			   more))
		#t)
	    (begin
		(set-read-more! self #f)
		#f))))

(define-method %read-char ((self <User-InputPort>))
    (let loop ()
	(if (null? (buffer self))
	    (begin
		(if (user-fill! self)
		    (loop)
		    *eof-object*))
	    (let ((b (buffer self)))
		(set-buffer! self (cdr b))
		(car b)))))

(define-method %peek-char ((self <User-InputPort>))
    (let loop ()
	(if (null? (buffer self))
	    (begin
		(if (user-fill! self)
		    (loop)
		    *eof-object*)
		(loop))
	    (car (buffer self)))))

(define-method %char-ready? ((self <User-InputPort>))
    (pair? (buffer self)))

(export <User-InputPort> <User-OutputPort>)
