#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
#|
    $Id: writers.scm,v 1.3 1994/12/11 00:31:22 donovan Exp $

    This file defines the writer (and display, when they're different)
    methods for the builtin classes.
    
    Pointers to these functions wind up being stored in the class
    objects.
|#

;----------------- <Pair> -----------------

(define (reader-macro-print pair)
    (let ((m (assq (%car pair)
		   '((quote . "'")
		     (quasiquote . "`")
		     (unquote . ",")
		     (unquote-splicing . ",@")))))
	(if m
	    (if (pair? (%cdr pair))
		(if (null? (%cdr (%cdr pair)))
		    (%cdr m)
		    #f)
		#f)
	    #f)))

(define (write/pair pair port)
    (let ((m (reader-macro-print pair)))
	(if m
	    (begin
		(%write-str port m)
		(%write (%car (%cdr pair)) port))
	    (begin
		(%write-char port #\()
		(%write (%car pair) port)
		(let loop ((x (%cdr pair)))
		    (if (pair? x)
			(begin
			    (%write-char port #\space)
			    (%write (%car x) port)
			    (loop (%cdr x)))
			(if (%not (null? x))
			    (begin
				(%write-str port " . ")
				(%write x port)))))
		(%write-char port #\))))))

(define (display/pair pair port)
    (let ((m (reader-macro-print pair)))
	(if m
	    (begin
		(%write-str port m)
		(%display (%car (%cdr pair)) port))
	    (begin
		(%write-char port #\()
		(%display (%car pair) port)
		(let loop ((x (%cdr pair)))
		    (if (pair? x)
			(begin
			    (%write-char port #\space)
			    (%display (%car x) port)
			    (loop (%cdr x)))
			(if (%not (null? x))
			    (begin
				(%write-str port " . ")
				(%display x port)))))
		(%write-char port #\))))))

;----------------- <String> -----------------

(prologue {
#include <ctype.h>
})

(%define-glue (string->printable str str_index)
{
    /* stores it's results directly into REG0 and REG1 */
    printablize_string( str, str_index );
    RETURN(2);
})

(define (write/string str port)
    (%write-char port #\")
    (let loop ((i 0))
	(bind ((str-part next-i (string->printable str i)))
	    (%write-str port str-part)
	    (if next-i
		(loop next-i))))
    (%write-char port #\"))

(define (display/string str port)
    (%write-str port str))

;----------------- <C-Text> -----------------

(define (write/c-text text port)
    (%write-char port #\{)
    (%write-str port (c-text->string text))
    (%write-char port #\}))

(define (display/c-text text port)
    (%write-str port (c-text->string text)))


;----------------- <FixNum> -----------------

;----------------- <Long-Float> -----------------

;----------------- <Vector> -----------------

(define (write/vector vec port)
    (%write-str port "#(")
    (let ((len (vector-length vec)))
	(if (> len 0)
	    (begin
		(%write (vector-ref vec 0) port)
		(let loop ((i 1))
		    (if (< i len)
			(begin
			    (%write-char port #\space)
			    (%write (vector-ref vec i) port)
			    (loop (add1 i))))))))
    (%write-char port #\)))

(define (display/vector vec port)
    (%write-str port "#(")
    (let ((len (vector-length vec)))
	(if (> len 0)
	    (begin
		(%display (vector-ref vec 0) port)
		(let loop ((i 1))
		    (if (< i len)
			(begin
			    (%write-char port #\space)
			    (%display (vector-ref vec i) port)
			    (loop (add1 i))))))))
    (%write-char port #\)))


;----------------- <Symbol> -----------------

(define (write/symbol sym port)
    (%write-str port (symbol->string sym)))

;----------------- <Token> -----------------

(define (write/token tok port)
    (%write-str port "#[<Token> ")
    (let ((type (token-type tok)))
	(%write-str port (symbol->string type))
	(if (memq type '(<literal> <symbol> <number> <object-ref>))
	    (begin
		(%write-str port " data:")
		(%write (token-data tok) port)))
	(%write-str port "]")))

;----------------- <ASCII-Char> -----------------

(define $ascii-char-names '#("nul" "soh" "stx" "etx" 
			     "eot" "enq" "ack" "bel" 
			     "bs" "tab" "newline" "vt" 
			     "np" "cr" "so" "si" 
			     "dle" "dc1" "dc2" "dc3" 
			     "dc4" "nak" "syn" "etb" 
			     "can" "em" "sub" "esc" 
			     "fs" "gs" "rs" "us" "space"))

(define (write/ascii-char ch port)
    (%write-str port "#\\")
    (let ((i (char->integer ch)))
	(if (>= i 128)
	    (begin
		(set! i (- i 128))
		(%write-str port "M-")))
	(if (< i 33)
	    (%write-str port (vector-ref $ascii-char-names i))
	    (if (eq? i 127)
		(%write-str port "del")
		(%write-char port (integer->char i))))))

    
;----------------- <Boolean> -----------------

(define (write/boolean bool port)
    (%write-str port (if bool "#t" "#f")))

;----------------- <<Class>> -----------------

(define (write/class class port)
    (%write-str port "#'")
    (%write-str port (symbol->string (class-name class))))

;----------------- <Closure> -----------------

(define-macro (addr->name tmpl addr)
    (%call ("addr_to_name" (<obj> <obj>) => <obj>) 
	   tmpl 
	   addr))

(define (closure-compute-name closure tmpl)
    (cond 
	((eq? (%gvec-read tmpl 0) (%gvec-read %ctc-template 0))
	    (let ((name (%gvec-read (%gvec-read tmpl 2) 3)))
		(if name
		    (object->string name)
		    "anon")))
	((generic-function? closure)
	    (symbol->string (generic-function-name closure)))
	((eq? tmpl $getter)
	    (let ((envt (%gvec-read closure 0)))
		(format #f "(~a get[~d])"
			(%gvec-read envt 2)
			(%gvec-read envt 1))))
	((eq? tmpl $unchecked-setter)
	    (let ((envt (%gvec-read closure 0)))
		(format #f "(~a set[~d])"
			(%gvec-read envt 2)
			(%gvec-read envt 1))))
	((eq? tmpl $checked-setter)
	    (let ((envt (%gvec-read closure 0)))
		(format #f "(~a set[~d] ~s)"
			(%gvec-read envt 3)
			(%gvec-read envt 1)
			(%gvec-read envt 2))))
	(else (addr->name tmpl (%gvec-read tmpl 0)))))

(define *closure-names* #f)

(define (closure-name closure)
    (if (not *closure-names*)
	(set! *closure-names* (make-object-table)))
    (let ((cached-name (table-lookup *closure-names* closure)))
	(if cached-name
	    cached-name
	    (let ((new-name (closure-compute-name closure 
						  (%gvec-read closure 1))))
		(table-insert! *closure-names*
			       closure
			       new-name)
		new-name))))

(define (write/closure closure port)
    (if (generic-function? closure)
	(%write-str port "#[<GF> ")
	(%write-str port "#[<closure> "))
    (%write-str port (closure-name closure))
    (%write-str port "]"))

;----------------- <Partial-Continuation> -----------------

(define (write/part-cont part-cont port)
    (%write-str port "#[<PartialCont> at ")
    (%write-str port (addr->name (%gvec-read part-cont 1) 
			         (%gvec-read part-cont 4)))
    (%write-str port "]"))


;----------------- <Template> -----------------
;----------------- <ByteVector> -----------------

(define (write/byte-vector bvec port)
    (let ((len (%bvec-length bvec)))
	(%write-str port "#[")
	(display (class-name (class-of bvec)) port)
	(%write-str port " ")
	(%write-int port len)
	(%write-str port " bytes:")
	(let loop ((i 0))
	    (if (eq? i len)
		(%write-str port "]")
		(begin
		    (write-hex-w/pad (%bvec-read bvec i) port 2)
		    (loop (add1 i)))))))
	    
;----------------- <table> -----------------

(define (write/table table port)
    (%write-str port "#[<table>")
    (let ((first #t))
	(table-for-each
	    table
	    (lambda (hash key value)
		(%write-str port (if first " " ", "))
		(set! first #f)
		(%write key port)
		(%write-str port "->")
		(%write value port)))
	(%write-str port "]")))

(define (display/table table port)
    (%write-str port "#[<table>")
    (let ((first #t))
	(table-for-each
	    table
	    (lambda (hash key value)
		(%write-str port (if first " " ", "))
		(set! first #f)
		(%display key port)
		(%write-str port "->")
		(%display value port)))
	(%write-str port "]")))

;----------------- <InputPort> -----------------
;----------------- <OutputPort> -----------------
;----------------- <File-Output-Port> -----------------

(define (write/file-output-port p port)
    (%write-str port "#[<File-Output-Port> ")
    (write/string (%gvec-read p 3) port)
    (%write-char port #\]))

;----------------- <Bounded-String-Output-Port2> -----------------

(define (write/bsp p port)
    (%write-str port "#[<Bounded-String-Port> ")
    (%write (%gvec-read p 3) port)
    (%write-char port #\/)
    (%write (string-length (%gvec-read p 0)) port)
    (%write-char port #\]))

;----------------- <String-Output-Port> -----------------

(define (write/sp p port)
    (%write-str port "#[<String-Output-Port> ")
    (%write (let loop ((i (%gvec-read p 1)) (overflow (%gvec-read p 2)))
    		(if (null? overflow)
		    i
		    (loop (+ i (%bvec-length (car overflow)))
		          (cdr overflow))))
	    port)
    (%write-char port #\]))

;----------------- <Top-Level-Var> -----------------

(define (write/tlv tlv port)
    (%write-str port "#[tlv ")
    (%write-str port (symbol->string (%gvec-read tlv 0)))
    (if (not (eq? (%gvec-read tlv 1) '#unbound))
	(begin
	    (%write-char port #\=)
	    (%write (%gvec-read tlv 1) port)))
    (%write-char port #\]))

;----------------- <Top-Level-Envt> -----------------

;----------------- <UniqueObject> -----------------

(%define-glue (immob->value immob)
{
    REG0 = int2fx( GET_IMMEDIATE_VALUE(immob) );
    RETURN(1);
})

(define (write/unique-object thing port)
    (let ((i (immob->value thing))
    	  (v '#("#none" "#undef" "#uninit" "#unbound" "#rest" "#key")))
	(if (< i (vector-length v))
	    (%write-str port (vector-ref v i))
	    (begin
		(%write-str port "#unique?")
		(%write-int port i)))))

;----------------- default: -----------------

(define (write/default thing port)
    (case thing
	((#nil) (%write-str port "()"))
	((#undef) (%write-str port "#undef"))
	((#uninit) (%write-str port "#uninit"))
	((#none) (%write-str port "#none"))
	((#unbound) (%write-str port "#unbound"))
	(else   (%write-str port "#[")
		(%write-str port 
			    (symbol->string (%gvec-read (class-of thing) 0)))
		(%write-str port "]"))))

(define (write/float num port)
    (%write-float port num))
    	
(define (display/ascii-char ch port)
    (%write-char port ch))


;--- set 'em up ---

;; note that we don't actually store the implementation
;; in the class itself; that will happen lazily as instances
;; of the class are displayed or written

(fast-gfe-set-methods!
    (gf->gfe write)
    (list
	(make-method (list <String>) write/string)
	(make-method (list <FixNum>) write-dec)
	(make-method (list <Pair>) write/pair)
	(make-method (list <ASCII-Char>) write/ascii-char)
	(make-method (list <double-float>) write/float)
	(make-method (list <Symbol>) write/symbol)
	(make-method (list <Boolean>) write/boolean)
	
	(make-method (list <C-Text>) write/c-text)
	(make-method (list <Token>) write/token)
	(make-method (list <<Class>>) write/class)
	(make-method (list <Vector>) write/vector)
	(make-method (list <Closure>) write/closure)
	(make-method (list <PartialContinuation>) write/part-cont)
	(make-method (list <ByteVector>) write/byte-vector)
	(make-method (list <table>) write/table)
	(make-method (list <TopLevelVar>) write/tlv)
	(make-method (list <UniqueObject>) write/unique-object)
	
	(make-method (list <Object>) write/default)))  ;; must be last

(fast-gfe-set-methods!
    (gf->gfe display)
    (list
	(make-method (list <String>) display/string)
	(make-method (list <FixNum>) write-dec)
	(make-method (list <Pair>) display/pair)
	(make-method (list <ASCII-Char>) display/ascii-char)
	(make-method (list <double-float>) write/float)
	(make-method (list <Symbol>) write/symbol)
	(make-method (list <Boolean>) write/boolean)

	(make-method (list <C-Text>) display/c-text)
	(make-method (list <Token>) write/token)
	(make-method (list <<Class>>) write/class)
	(make-method (list <Vector>) display/vector)
	(make-method (list <Closure>) write/closure)
	(make-method (list <PartialContinuation>) write/part-cont)
	(make-method (list <ByteVector>) write/byte-vector)
	(make-method (list <table>) display/table)
	(make-method (list <TopLevelVar>) write/tlv)
	(make-method (list <UniqueObject>) write/unique-object)

	(make-method (list <Object>) write/default)))  ;; must be last

(export addr->name closure-name)

