/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "privio.h"
#include "bltin_p.h"
#include "scmio_p.h"
#include <stdio.h>

/*
   $Id: op_str.c,v 1.2 94/10/29 13:57:45 donovan Exp $
*/

#define SOP_BLOCK_SIZE (64-16)


obj SOP_close( obj port )
{
int len;
obj dst, overflow;
char *endptr;
const char *src;

    len = fx2int( gvec_read( port, SOP_INDEX ) );
    overflow = gvec_read( port, SOP_OVERFLOW );
    
    while (!EQ(overflow,NIL_OBJ))
    {
	len += SIZEOF_PTR(pair_car(overflow));
	overflow = pair_cdr(overflow);
    }

    dst = bvec_alloc( len+1, string_class );
    endptr = ((char *)string_text(dst)) + len;
    *endptr = 0;

    src = (const char *)PTR_TO_DATAPTR( gvec_read( port, SOP_BUFFER ) );
    len = fx2int( gvec_read( port, SOP_INDEX ) );
    overflow = gvec_read( port, SOP_OVERFLOW );
    
    while (1)
    {
	endptr -= len;
	memcpy( endptr, src, len );
	if (EQ(overflow,NIL_OBJ))
	    break;

	src = (const char *)PTR_TO_DATAPTR(pair_car(overflow));
	len = SIZEOF_PTR(pair_car(overflow));
	overflow = pair_cdr(overflow);
    }
    return dst;
}

void SOP_write( obj port, const char *src, UINT_32 len )
{
obj buf, fxpos;
char *ptr;
UINT_32 n, max, pos;

    buf = gvec_read( port, SOP_BUFFER );
    fxpos = gvec_read( port, SOP_INDEX );

    assert( BYTE_VECTOR_P(buf) );
    assert( OBJ_ISA_FIXNUM(fxpos) );

    max = SIZEOF_PTR( buf );
    pos = fx2int(fxpos);
    
    ptr = (char *)PTR_TO_DATAPTR(buf);

    while (pos + len > max)
    {
	n = max - pos;
	memcpy( ptr + pos, src, n );
	src += n;
	len -= n;
	gvec_write( port, 
		    SOP_OVERFLOW, 
		    cons( buf, gvec_read( port, SOP_OVERFLOW ) ) );
	max = SOP_BLOCK_SIZE;
	buf = alloc( max, byte_vector_class );
	gvec_write( port, SOP_BUFFER, buf );
	pos = 0;
	ptr = (char *)PTR_TO_DATAPTR(buf);
    }

    memcpy( ptr + pos, src, len );
    pos += len;

    gvec_write_non_ptr( port, SOP_INDEX, int2fx(pos) );
}

jump_addr do_write_str( jump_addr (*fn)( obj, const char *src, UINT_32 len ),
	            	obj port,
			obj datum )
{
    if (STRING_P(datum))
	return fn( port, string_text(datum), string_length(datum) );
    else
    {
	assert( OBJ_ISA_PTR(datum) && EQ(CLASSOF_PTR(datum),c_text_class) );
	return fn( port, 
		   (const char *)PTR_TO_DATAPTR(datum), 
		   SIZEOF_PTR(datum) );
    }
}

/* stores it's results directly into REG0 and REG1 */

void printablize_string( obj str, obj str_index )
{
#define TEMP_LENGTH (50)
char ec, *d, *limit, temp[TEMP_LENGTH+6];
const char *s0, *s;

    s0 = s = string_text(str) + fx2int(str_index);
    d = temp;
    limit = d + TEMP_LENGTH;
    while (*s && d < limit)
    {
	switch (*s)
	{
	    case '\"':  ec = '\"';
			goto use_escaped;
	    case '\\':  ec = '\\';
			goto use_escaped;
	    case '\n':  ec = 'n';
			goto use_escaped;
	    case '\t':  ec = 't';
	    use_escaped:
			*d++ = '\\';
			*d++ = ec;
			break;
	    default:    
			if (isprint(*s))
			    *d++ = *s;
			else
			{
			    sprintf( d, "\\%03o", *(UINT_8 *)s );
			    d += 4;
			}
			break;
	}
	s++;
    }
    *d++ = 0;
    if (*s)
    {
	REG0 = make_string( temp );
	REG1 = FX_ADD(str_index,int2fx(s-s0));
    }
    else if (EQ(str_index,ZERO) && ((d - temp) == SIZEOF_PTR(str)))
    {
	REG0 = str;
	REG1 = FALSE_OBJ;
    }
    else
    {
	REG0 = make_string( temp );
	REG1 = FALSE_OBJ;
    }
}
