#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;  The old gcd and lcm algorithms
;;  which work only for non-negative fixnums

; from Algorithms, 2nd. Ed.
;      Sedgewick
;      p. 8

(define (raw-gcd u v)
  (if (binary<? v u)
      (raw-gcd v u)
      (let loop ((u u) (v v))
	(if (eq? u 0)
	    v
	    (loop (remainder v u) u)))))

(define (raw-lcm x y)
  (let ((gcd (raw-gcd x y)))
    (if (eq? gcd 0)
	0
	(quotient (binary* x y) gcd))))

;; The generalized gcd and lcm algorithms
;; Date: Sat, 04 Jun 1994 21:13:15 -0500
;; From: Jun Sawada <sawada@cs.utexas.edu>

(define (gcd . args) 
  (cond ((not (pair? args)) 0)
	((not (pair? (cdr args)))
	 (if (number? (car args))
	     (abs (car args))
	     (abort 'gcd "num ~a is not a number." (car args))))
	(else
	 (let* ((return-inexact? #f)
		(normals
		 (map (lambda (x)
			(if (number? x)
			    (if (exact? x)
				(abs x)
				(let ((y (inexact->exact x)))
				  (set! return-inexact? #t)
				  (if (binary=? (binary- y x) 0)
				      (abs y)
				      (abort 'gcd "num ~a is not an integer." x))))
				
			    (abort 'gcd "num ~a is not a number." x)))
		      args)))
	   (let loop ((first (car normals))
		      (rest (cdr normals)))
	     (if (pair? rest)
		 (loop (raw-gcd first (car rest)) (cdr rest))
		 (if return-inexact?
		     (exact->inexact first)
		     first)))))))

(define (lcm . args) 
  (cond ((not (pair? args)) 1)
	((not (pair? (cdr args)))
	 (if (number? (car args))
	     (abs (car args))
	     (abort 'lcm "num ~a is not a number." (car args))))
	(else
	 (let* ((return-inexact? #f)
		(normals
		 (map (lambda (x)
			(if (number? x)
			    (if (exact? x)
				(abs x)
				(let ((y (inexact->exact x)))
				  (set! return-inexact? #t)
				  (if (binary=? (binary- y x) 0)
				      (abs y)
				      (abort 'lcm "num ~a is not an integer." x))))
				
			    (abort 'lcm "num ~a is not a number." x)))
		      args)))
	   (let loop ((first (car normals))
		      (rest (cdr normals)))
	     (if (pair? rest)
		 (loop (raw-lcm first (car rest)) (cdr rest))
		 (if return-inexact?
		     (exact->inexact first)
		     first)))))))

(export gcd lcm)
