#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define (map1 proc list)
    (if (pair? list)
	(cons (proc (car list)) (map1 proc (cdr list)))
	(if (null? list)
	    '()
	    (abort 'map "Not a list at: ~s" list))))

;; unsafe-map-c[ad]r are unsafe with regard to the list
;; and with regard to the contents of the list.
;; that is, they require that `lst' be a proper list,
;; and each of the elements of `lst' to be a pair.

(define (unsafe-map-car list)
    (if (null? list)
	'()
	(cons (car (car list)) (unsafe-map-car (cdr list)))))

(define (unsafe-map-cdr list)
    (if (null? list)
	'()
	(cons (cdr (car list)) (unsafe-map-cdr (cdr list)))))

(define (all-pairs? lst)
    (let loop ((l lst))
	(if (null? l)
	    #t
	    (if (pair? (car l))
		(loop (cdr l))
		#f))))

(define (unsafe-car x)
    (car x))

(define (unsafe-cdr x)
    (cdr x))

(define (mapn proc lists)
    (if (all-pairs? lists)
	(cons (apply* (unsafe-map-car lists) proc)
	      (mapn proc (unsafe-map-cdr lists)))
	'()))


(define (map proc list1 . more-lists)
    (if (null? more-lists)
	(map1 proc list1)
	(mapn proc (cons list1 more-lists))))

(define (for-each1 proc list)
    (if (pair? list)
	(if (null? (cdr list))
	    (proc (car list))
	    (begin
		(proc (car list))
		(for-each1 proc (cdr list))))
	(values)))

;; for-eachn requires that lists be non-empty

(define (for-eachn proc lists)
    (if (all-pairs? lists)
	(if (null? (cdr (car lists)))
	    (apply* (unsafe-map-car lists) proc)
	    (begin
		(apply* (unsafe-map-car lists) proc)
		(for-eachn proc (unsafe-map-cdr lists))))
	(abort 'for-each "Not all lists at: ~s" lists)))

(define (for-each proc list1 . more-lists)
    (if (null? more-lists)
	(for-each1 proc list1)
	(if (null? list1)
	    (values)
	    (for-eachn proc (cons list1 more-lists)))))

(define (any? predicate list)
    (let loop ((x list))
	(if (pair? x)
	    (if (predicate (car x))
		#t
		(loop (cdr x)))
	    #f)))

(define (all? predicate list)
    (let loop ((x list))
	(if (pair? x)
	    (if (predicate (car x))
		(loop (cdr x))
		#f)
	    #t)))

(define-shadow map map)
(define-shadow for-each for-each)

(export for-each map any? all?)

#|
these versions assume `lst' is a list but don't assume
the elements are pairs

(define (unsafe-map-car lst)
    (if (null? lst)
	'()
	(let ((item (car lst)))
	    (if (pair? item)
		(cons (car item) (unsafe-map-car (cdr lst)))
		(abort 'unsafe-map-car 
		       "List element not a pair at: ~s" 
		       lst)))))

(define (unsafe-map-car lst)
    (if (null? lst)
	'()
	(let ((item (car lst)))
	    (if (pair? item)
		(cons (car item) (unsafe-map-car (cdr lst)))
		(abort 'unsafe-map-car 
		       "List element not a pair at: ~s" 
		       lst)))))
|#
