#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(prologue {
#include "str2num.h"
})
(include-h "str2num.h")
(include-c "str2num.c")

(define (number? x)
    (or (fixnum? x) (double-float? x)))

(define complex? number?)
(define real? number?)
(define rational? number?)

(define (integer? x)
    (or (fixnum? x)
	(and (double-float? x)
	     (binary=? x (round x)))))	

(define exact? fixnum?)
(define inexact? double-float?)

	
(define (_+ . args)
    (let loop ((x args) (sum 0))
	(if (null? x)
	    sum
	    (loop (cdr x) (binary+ sum (car x))))))

(define (* . args)
    (let loop ((x args) (prod 1))
	(if (null? x)
	    prod
	    (loop (cdr x) (binary* prod (car x))))))


(define (_- arg1 . args)
    (if (null? args)
	(unary- arg1)
	(let loop ((x (cdr args)) 
		   (diff (binary- arg1 (car args))))
	    (if (null? x)
		diff
		(loop (cdr x) (binary- diff (car x)))))))

(define (/ arg1 . args)
    (if (null? args)
	(unary/ arg1)
	(let loop ((x (cdr args))
		   (quotient (binary/ arg1 (car args))))
	    (if (null? x)
		quotient
		(loop (cdr x) (binary/ quotient (car x)))))))


(define-syntax %transitively
    (syntax-form (binary? arg1 arg2 rest)
	(if (null? rest)
	    (binary? arg1 arg2)
	    (let loop ((args (cons* arg1 arg2 rest)))
		(if (null? (cdr (cdr args)))
		    (binary? (car args) (car (cdr args)))
		    (if (binary? (car args) (car (cdr args)))
			(loop (cdr args))
			#f))))))

(define (= arg1 arg2 . args)
    (%transitively binary=? arg1 arg2 args))

(define (_> arg1 arg2 . args)
    (%transitively binary>? arg1 arg2 args))

(define (_< arg1 arg2 . args)
    (%transitively binary<? arg1 arg2 args))

(define (_>= arg1 arg2 . args)
    (%transitively binary>=? arg1 arg2 args))

(define (_<= arg1 arg2 . args)
    (%transitively binary<=? arg1 arg2 args))


(define (zero? x)
    (binary=? x 0))

(define (negative? x)
    (binary<? x 0))

(define (positive? x)
    (binary>? x 0))


(define (even? n)
    (let ((parity (bitwise-and n 1)))
	(if (eq? parity 0) #t #f)))

(define (odd? n)
    (let ((parity (bitwise-and n 1)))
	(if (eq? parity 1) #t #f)))

#| there is a bug in these definitions...
   R4RS p.21:
	Note:  If any argument is inexact, then the result will
	also be inexact...
|#

(define (min arg1 . args)
    (let loop ((x args) (smallest arg1))
	(if (pair? x)
	    (loop (cdr x)
		  (if (binary<? (car x) smallest)
		      (car x)
		      smallest))
	    smallest)))

(define (max arg1 . args)
    (let loop ((x args) (largest arg1))
	(if (pair? x)
	    (loop (cdr x)
		  (if (binary>? (car x) largest)
		      (car x)
		      largest))
	    largest)))


(define (abs x)
    (if (binary<? x 0)
	(unary- x)
	x))

(define (string->number str . rest)
    (type-check str string? <String>)
    (if (null? rest)
	(%call ("string_to_number" (<raw-str> <raw-int>) => <obj>) 
		str
		10)
	(let ((radix (car rest)))
	    (type-check radix fixnum? <FixNum>)
	    (%call ("string_to_number" (<raw-str> <raw-int>) => <obj>) 
		str
		radix))))

(%define-glue (double-float->string num)
{
char temp[64];

    REG0 = make_string( double_float_to_string( temp, num ) );
    RETURN(1);
})

(%define-glue (fixnum->string num radix)
{
char temp[32];

    REG0 = make_string( fixnum_to_string( &temp[32], num, fx2int(radix) ) );
    RETURN(1);
})

(define (number->string num . rest)
    (cond 
	((fixnum? num)
	    (if (null? rest)
		(fixnum->string num 10)
		(let ((radix (car rest)))
		    (type-check radix fixnum? <FixNum>)
		    (if (< radix 2)
			(abort 'number->string "radix ~d < 2" radix)
			(if (> radix 36)
			    (abort 'number->string "radix ~d > 36" radix)
			    (fixnum->string num radix))))))
	((double-float? num)
	    (if (null? rest)
		(double-float->string num)
		(let ((radix (car rest)))
		    (type-check radix fixnum? <FixNum>)
		    (if (eq? radix 10)
			(double-float->string num)
			(abort 'number->string "radix ~d != 10" radix)))))
	(else
	    (abort 'number->string "num ~s not a number" num))))

(export abs string->number number->string)
(export max min odd? even?)

(export number? complex? real? rational? integer? exact? inexact?)
(export zero? negative? positive?)

(export (_+ +)
	* 
	(_- -)
	/
	=
	(_< <)
	(_<= <=)
	(_> >)
	(_>= >=))
	
