#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

; a sort facility, using heapsort
;
; adapted from the HeapSort algorithm
; presented in:
;	Algorithms, 2nd ed.
;	Robert Sedgewick
;	Addison Wesley

(define (sort:down-heap heap N compare-fn k0)
    (let ((v (vector-ref heap k0)))
	(let loop ((k k0))
	    (if (< k (quotient N 2))
		(let ((j (add1 (mul2 k))))
		    (if (< (add1 j) N)
			(if (compare-fn (vector-ref heap j)
					(vector-ref heap (add1 j)))
			    (set! j (add1 j))))
		    (if (not (compare-fn v (vector-ref heap j)))
			(vector-set! heap k v)
			(begin
			    (vector-set! heap k (vector-ref heap j))
			    (loop j))))
		(vector-set! heap k v)))))

(define (sort:heap-sort heap N compare-fn)
    (let loop1 ((k (div2 (add1 N))))
	(if (> k 0)
	    (begin
		(sort:down-heap heap N compare-fn (sub1 k))
		(loop1 (sub1 k)))))

    (let loop2 ((result '()) (n (sub1 N)))
	(if (eq? n -1)
	    result
	    (let ((top (vector-ref heap 0)))
		(vector-set! heap 0 (vector-ref heap n))
		(sort:down-heap heap n compare-fn 0)
		(loop2 (cons top result) (sub1 n)))
	)))

(define (sort the-list compare-fn)
    (let ((heap (list->vector the-list)))
	(sort:heap-sort heap (vector-length heap) compare-fn)))

(export sort)
