/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "str2num.h"
#include "runtim.h"
#include <string.h>

unsigned char digit_values[256] = {
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
     0,  1,  2,  3,   4,  5,  6,  7,   8,  9, 99, 99,  99, 99, 99, 99,
    99, 10, 11, 12,  13, 14, 15, 16,  17, 18, 19, 20,  21, 22, 23, 24,
    25, 26, 27, 28,  29, 30, 31, 32,  33, 34, 35, 99,  99, 99, 99, 99,
    99, 10, 11, 12,  13, 14, 15, 16,  17, 18, 19, 20,  21, 22, 23, 24,
    25, 26, 27, 28,  29, 30, 31, 32,  33, 34, 35, 99,  99, 99, 99, 99,

    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,
    99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99,  99, 99, 99, 99 };

static inline int digit_value( unsigned char ch )
{
    return digit_values[ch];
}

static char value_digits[37] = "0123456789abcdefghijklmnopqrstuvwxyz";

/* fills up buffer right-to-left!  buffer starts out pointing
   to the END of the buffer! */

char *fixnum_to_string( char *buffer, obj value, unsigned radix )
{
int neg, v;

    *--buffer = 0;
    v = fx2int(value);
    if (v < 0)
    {
	neg = 1;
	v = -v;
    }
    else
	neg = 0;
    if (v == 0)
	*--buffer = '0';
    else
    {
	while (v > 0)
	{
	    *--buffer = value_digits[v % radix];
	    v /= radix;
	}
    }
    if (neg)
	*--buffer = '-';
    return buffer;
}

char *double_float_to_string( char *buffer, obj value )
{
    sprintf( buffer, "%g", extract_float(value) );
    if (!strchr(buffer,'.'))
        strcat( buffer, ".0" );
    return buffer;
}

obj string_to_number( const char *str, unsigned radix )
{
int i, neg;

    if (*str == '#')
    {
	str++;
	switch (*str++)
	{
	  case 'b': radix = 2; break;
	  case 'o': radix = 8; break;
	  case 'd': radix = 10; break;
	  case 'x': radix = 16; break;
	  default:  return FALSE_OBJ;
	}
    }
    if (*str == '-')
    {
	str++;
	neg = 1;
    }
    else if (*str == '+')
    {
	str++;
	neg = 0;
    }
    else
	neg = 0;

    if (strchr(str,'.'))
    {
    double x, r = radix, v = 0.0;
    unsigned num_digits = 0;

	while (*str != '.')
	{
	    i = digit_value( *str++ );
	    if (i >= radix)
		return FALSE_OBJ;
	    v = v * radix + i;
	    num_digits++;
	}
	str++;
	r = 1.0 / r;
	x = r;
	while (*str)
	{
	    i = digit_value( *str++ );
	    if (i >= radix)
		return FALSE_OBJ;
	    v += i * x;
	    x *= r;
	    num_digits++;
	}
        if (num_digits == 0)
	  return FALSE_OBJ;  /* no digits -- "." is invalid */

	return make_float( neg ? -v : v );
    }
    else
    {
    unsigned v = 0, preq, prem;

        /* compute the maximum value & digit that is
	   allowed BEFORE a new digit is added.
	   For example, if the limit is 1024 (max 1023)
	   and we're in base 10, the preq is 102 and prem is 4
	   is if the value so far is 102 and we see a 4, we know
	   that's too much, but if we see a 3, that's OK */
	   
	preq = (1<<29) / radix;
	prem = (1<<29) % radix;
	/* printf( "preq = %d, prem = %d\n", preq, prem ); */

        if (!*str)
	    return FALSE_OBJ;  /* no digits! */

	while (*str)
	{
	    i = digit_value( *str++ );
	    if (i >= radix)
		return FALSE_OBJ;
	    /* printf( "  %c: v = %d, i = %d\n", str[-1], v, i ); */
	    if ((v > preq)
	        || ((v == preq) 
	            && ((i > prem) || ((i == prem) && !neg))))
	    {
		return FALSE_OBJ;
	    }
	    v = v * radix + i;
	}
	return int2fx( neg ? -v : v );
    }
}
