#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; vectors-as-lists operations
;; $Id: vectops.scm,v 1.3 94/10/30 10:35:42 donovan Exp $

(define (vector-for-each1/i proc (vec <vector>))
    (if (eq? (vector-length vec) 0)
	(values)
	(let ((max (sub1 (vector-length vec))))
	    (let loop (((i <fixnum>) 0))
		(if (eq? i max)
		    (proc i (vector-ref vec i))
		    (begin
			(proc i (vector-ref vec i))
			(loop (add1 i))))))))

(define (vector-for-each2/i proc (v1 <vector>) (v2 <vector>))
    (let ((max-index (sub1 (min (vector-length v1)
    				(vector-length v2)))))
	(if (eq? max-index -1)
	    (values)
	    (let loop (((i <fixnum>) 0))
		(if (eq? i max-index)
		    (proc i (vector-ref v1 i) (vector-ref v2 i))
		    (begin
			(proc i (vector-ref v1 i) (vector-ref v2 i))
			(loop (add1 i))))))))

(define (vectors-min-len vs)
    (apply* (map vector-length vs) min))
    
(define (vector-for-eachn/i proc vs)
    (assert (all? vector? vs))
    (let ((max-index (sub1 (vectors-min-len vs))))
	(if (eq? max-index -1)
	    (values)
	    (let loop (((i <fixnum>) 0))
		(if (eq? i max-index)
		    (apply* i (map (lambda ((v <vector>))
		    			(vector-ref v i))
				    vs)
			      proc)
		    (begin
			(apply* i (map (lambda ((v <vector>))
		    			(vector-ref v i))
				    vs)
			      proc)
			(loop (add1 i))))))))

(define (vector-for-each/i proc v1 more-vs)
    (assert (and (vector? v1) (all? vector? more-vs)))
    (if (null? more-vs)
	(vector-for-each1/i proc v1)
	(if (null? (cdr more-vs))
	    (vector-for-each2/i proc v1 (car more-vs))
	    (vector-for-eachn/i proc (cons v1 more-vs)))))

(define (vector-for-each proc v1 . more-vs)
    (vector-for-each/i (if (null? more-vs)
    			   (lambda (i elem)
				(proc elem))
			    (if (null? (cdr more-vs))
				;; optimize the 2-vector case
				(lambda (i e1 e2)
				    (proc e1 e2))
				(lambda (i . elems)
				    (apply* elems proc))))
			v1
			more-vs))

(define (vector-map proc v1 . more-vs)
    (assert (and (vector? v1) (all? vector? more-vs)))
    (let ((result (make-vector (vectors-min-len (cons v1 more-vs)))))
        (vector-for-each/i
	    (if (null? more-vs)
		(lambda (i elem)
		    (vector-set! result i (proc elem)))
		(if (null? (cdr more-vs))
		    ;; optimize the 2-vector case
		    (lambda (i e1 e2)
			(vector-set! result i (proc e1 e2)))
		    (lambda (i . elems)
			(vector-set! result 
					i 
					(apply* elems proc)))))
	    v1
	    more-vs)
	result))

(%define-glue (vector-bool-op vec1 vec2 op newv)
{
char c;
int b, opcode = fx2int(op);
UINT_32 i, len;
obj dest;

    COUNT_ARGS(4);
    if(!(VECTOR_P(vec1) && VECTOR_P(vec2) 
    	 && OBJ_ISA_FIXNUM(op) && OBJ_ISA_BOOLEAN(newv)))
    {
	scheme_error( "vector-bool-op: bad arg", 0 );
    }
    len = SIZEOF_PTR(vec1);
    if (SIZEOF_PTR(vec2) < len)
	len = SIZEOF_PTR(vec2);
	
    if (truish(newv))
	dest = alloc( len, vector_class );
    else
	dest = vec1;
    for (i=0; i<len; i+=SLOT(1))
    {
	if (EQ(gvec_read(vec1,i),FALSE_OBJ))
	{
	    if (EQ(gvec_read(vec2,i),FALSE_OBJ))
		b = opcode & 1;
	    else
		b = opcode & 2;
	}
	else
	{
	    if (EQ(gvec_read(vec2,i),FALSE_OBJ))
		b = opcode & 4;
	    else
		b = opcode & 8;
	}
	gvec_write_non_ptr( dest, i, b ? TRUE_OBJ : FALSE_OBJ );
    }
    REG0 = dest;
    RETURN(1);
})

(define (vector-or vec1 vec2)
    (vector-bool-op vec1 vec2 #b1110 #t))
    
(define (vector-and! vec1 vec2)
    (vector-bool-op vec1 vec2 #b1000 #f))
    
(define (vector-andn! vec1 vec2)
    (vector-bool-op vec1 vec2 #b0001 #f))

(export vector-bool-op vector-or vector-and! vector-andn!)

(export vector-map vector-for-each)
