#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
#|
(define $log2-tree '#(128 #(8 #(2 #(1 0 1)
				  #(4 2 3))
			      #(32 #(16 4 5)
			           #(64 6 7)))
			  #(2048 #(512 #(256 8 9)
			               #(1024 10 11))
				 #(8192 #(4096 12 13)
				        #(16384 14 15)))))
				   

(define (binary-search tree (value <fixnum>))
    (if (fixnum? tree)
	tree
	(let (((t <vector>) tree))
	    (if (<= value (vector-ref t 0))
		(binary-search (vector-ref t 1) value)
		(binary-search (vector-ref t 2) value)))))
|#

(define (log2 (n <fixnum>))
    (let loop (((i <fixnum>) 1) (j 0))
	(if (>= i n)
	    j
	    (loop (bit-shift-left i 1) (add1 j)))))

(define (capacity->directory-size cap)
    (if cap (log2 cap) 4))

(define (make-table eq-proc hash-proc . cap)
    (let ((initial-size (capacity->directory-size 
			    (if (null? cap) 
				#f 
				(car cap)))))
	(cond
	    ((and (eq? eq-proc string=?)
		  (eq? hash-proc string->hash))
		(make <string-table> directory-bits: initial-size))
	    ((and (eq? eq-proc string-ci=?)
		  (eq? hash-proc string-ci->hash))
		(make <string-ci-table> directory-bits: initial-size))
	    ((and (eq? eq-proc eq?)
		  (eq? hash-proc symbol->hash))
		(make <symbol-table> directory-bits: initial-size))
	    ((and (eq? eq-proc eq?)
		  (eq? hash-proc identity))
	        (make <integer-table> directory-bits: initial-size))
	    ((and (eq? eq-proc eq?)
		  (eq? hash-proc integer->hash))
	        (make <hash-integer-table> directory-bits: initial-size))
	    ((eq? eq-proc eq?)
		(make <eq-table> directory-bits: initial-size
				 table-hash-function: hash-proc))
	    (else
		(make <generic-table> directory-bits: initial-size
				      table-hash-function: hash-proc
				      table-equal-function: eq-proc)))))

(define (table? thing)
    (instance? thing <table>))
    
(export make-table table?)
