#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

(define (table-replace! table key proc new-value)
    (let* ((old-value (table-lookup table key))
    	   (new-value (proc old-value new-value)))
	(if (not (eq? old-value new-value))
	    (table-insert! table key new-value))
	new-value))

(define (seq-adder old new)
    (if (not old)
	(set! old (make-seq)))
    (seq-add! old new)
    old)

(define (table-add-to-seq! table key new-value)
    (table-replace! table key seq-adder new-value))

(define-method initialize ((self <table>))
    (set-directory! self (make-vector (expt 2 (directory-bits self))))
    self)

(define (table-join a b a-only b-only a-and-b)
  (if (not (and (eq? (table-equal-function a)
		     (table-equal-function b))
		(eq? (table-hash-function a)
		     (table-hash-function b))))
      (abort 'table-merge "incommensurate tables given: ~s and ~s" a b)
      (let ((join (make-table eq? identity))
	    (eq-fn (table-equal-function a)))
	(table-for-each
	 a
	 (lambda (h k v)
	   ;(format #t "join: ~s ~s ~s\n" h k v)
	   (let ((e (table-lookup join h)))
	     (if (not e)
		 (begin
		   (set! e (make-seq))
		   (table-insert! join h e)))
	     (seq-add! e (cons k v)))))
	(table-for-each
	 b
	 (lambda (h k v)
	   ;(format #t "in-b: ~s ~s ~s\n" h k v)
	   (let ((e (table-lookup join h)))
	     (if e
		 (let loop ((as (seq->list e)))
		   (if (null? as)
		       (if b-only (b-only k v))
		       (if (and (car as)
				(eq-fn (caar as) k))
			   (begin
			     (if a-and-b
				 (a-and-b (caar as) (cdar as) k v))
			     (set-car! as #f))
			   (loop (cdr as)))))
		 (if b-only (b-only k v))))))
	(if a-only
	    (table-for-each
	     join
	     (lambda (h k v)
	       ;(format #t "in-join: ~s ~s ~s\n" h k (seq->list v))
	       (for-each
		(lambda (ae)
		  (if ae
		      (a-only (car ae) (cdr ae))))
		(seq->list v))))))))

(export table-replace! table-add-to-seq! table-join)
