#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(define-alias $bucket-size 8)

(define-syntax (bucket-hash bucket ix)
    (%gvec-read bucket (+ ix 2)))
    
(define-syntax (bucket-key bucket ix)
    (%gvec-read bucket (+ ix (+ 2 $bucket-size))))
    
(define-syntax (bucket-value bucket ix)
    (%gvec-read bucket (+ ix (+ 2 (* 2 $bucket-size)))))

(define (bucket-for-each (bucket <table-bucket>) proc)
    (let buckets-loop (((b <table-bucket>) bucket))
	(let loop (((i <fixnum>) 0))
	    (if (< i $bucket-size)
		(begin
		    (if (bucket-hash b i)
			(proc (bucket-hash b i)
			      (bucket-key b i)
			      (bucket-value b i)))
		    (loop (add1 i)))))
	(if (bucket-overflow b)
	    (buckets-loop (bucket-overflow b)))))

;; counts the # entries in a bucket

(define (bucket-count b)
    (let loop (((i <fixnum>) 0) ((n <fixnum>) 0))
	(if (< i $bucket-size)
	    (loop (add1 i)
	          (if (bucket-hash b i)
		      (add1 n)
		      n))
	    n)))

(define-method table-for-each ((self <table>) (proc <closure>))
    (for-each
	(lambda (bucket)
	    (bucket-for-each bucket proc))
	(%call ("hashtable_chains" (<obj>) => <obj>) self)))

(define-method table-keys->list ((self <table>))
    (%call ("hashtable_keys_to_list" (<obj>) => <obj>) self))

(define-method table-values->list ((self <table>))
    (%call ("hashtable_values_to_list" (<obj>) => <obj>) self))

;; returns a pair of the length of a chain in # buckets 
;; and the # entries along the chain

(define (chain-length chain)
    (let loop ((i 0) (n 0) (b chain))
	(if b
	    (loop (add1 i) 
		  (+ n (bucket-count b))
		  (bucket-overflow b))
	    (cons i n))))

;; returns a vector:
;;   [0] is a vector of the size of each bucket chain
;;   [1] is the total # of (unique) chains
;;   [2] is a list of the lengths of the unique chains

(define (table-stats (self <table>))
    (let ((chains (%call ("hashtable_chains" (<obj>) => <obj>) self)))
	(vector
	    (vector-map chain-length (directory self))
	    (length chains)
	    (map chain-length chains))))

(export table-for-each table-keys->list table-values->list)
(export table-stats)
