#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
(prologue {
#include "trs.h"
#include "osglue.h"
#include "intrs.h"
})

(include-h "trs.h")

(%define-glue (%call-with-lightweight-continuation the_proc)
{
obj proc;

    flush_stack_cache();
    proc = the_proc;
    REG0 = continuation_reg;
    APPLY( 1, proc );
})

(%define-glue (%jump-with-timeout state timeout)
{
    os_set_timer( fx2int( timeout ) );
    continuation_reg = state;
    RETURN0();
})

(%define-glue (%save-and-suspend-current-thread) 
	      literals: ((& %dispatch-next-thread)
			 (& *ready-queue*)
			 (& *current-thread*))
{
obj ready_queue = TLREF(1);
obj next;

    flush_stack_cache();
    gvec_write( TLREF(2), THREAD_STATE, continuation_reg );

    interrupts_enabled = YES;
    os_halt_timer(); /* turn off the timer */
    next = pair_cdr( ready_queue );
    gvec_write_ptr( ready_queue, SLOT(0), pair_car( next ) );
    gvec_write_ptr( ready_queue, SLOT(1), pair_cdr( next ) );
    APPLY(0,TLREF(0));
})

(%define-glue (%suspend-current-thread) literals: ((& %dispatch-next-thread)
						   (& *ready-queue*))
{
obj ready_queue = TLREF(1);
obj next;

    interrupts_enabled = YES;
    os_halt_timer(); /* turn off the timer */
    next = pair_cdr( ready_queue );
    gvec_write_ptr( ready_queue, SLOT(0), pair_car( next ) );
    gvec_write_ptr( ready_queue, SLOT(1), pair_cdr( next ) );
    APPLY(0,TLREF(0));
})

(%define-glue (%disable-interrupts)
{
    interrupts_enabled = NO;
    RETURN0();
})

(%define-glue (%reenable-interrupts)
{
    interrupts_enabled = YES;
    RETURN0();
})

#|
   Oops.. This is completely wrong
   (it causes the proc to be started with 
    0 args because jump-with-timeout ends
    that way (RETURN0()))
    
(%define-glue (%lightweight-continue-apply proc)
{
jump_addr addr;
unsigned i, reg_space = arg_count_reg - 1;

    COUNT_ARGS_AT_LEAST(1);
    addr = apply( proc );
    REG0 = alloc( SLOT( CONT_FIXED + reg_space ), partcont_class );

    gvec_write_fresh( REG0, SLOT(0), envt_reg );
    gvec_write_fresh( REG0, SLOT(1), literals_reg );
    gvec_write_fresh( REG0, SLOT(2), dynamic_state_reg );
    gvec_write_fresh_non_ptr( REG0, SLOT(3), FALSE_OBJ );
    gvec_write_fresh_non_ptr( REG0, SLOT(4), JUMP_ADDR_TO_OBJ(addr) );

    for (i=1; i<arg_count_reg; i++)
	gvec_write_fresh( REG0, SLOT(i+1), reg_ref(i) );
    RETURN(1);
}
("do_continue" {
unsigned n;

    n = restore_arb();
    RETURN(n);
})
|#
    



;   inserts the unblocked thread immediately after the
;   current thread in the ready queue

(%define-glue (%mark-thread-ready thread) literals: ((& *ready-queue*))
{
obj entry;
obj ready_queue = TLREF(0);

    entry = cons( thread, pair_cdr( ready_queue ) );
    gvec_write( ready_queue, SLOT(1), entry );
    RETURN(1);
})

#|
(%define-glue (print item) literals: ((& *current-thread*))
{
obj name = gvec_read( TLREF(0), SLOT(2) );
int i;

    printf( "%s ==> ", string_text( pair_cdr(name) ) );
    for (i=fx2int(pair_car(name)); i>0; i--)
	printf( "          " );
    fprinto( stdout, item );
    putchar( '\n' );
    RETURN0();
})
|#

;; it doesn't matter what this procedure returns --
;; it is of INTERRUPT type, so the interrupted thread's
;; state is completely saved

(%define-glue (%timeout) literals: ((& %dispatch-next-thread)
				    (& *ready-queue*))
{
obj ready_queue = TLREF(1);
obj current_thread;

    if (EQ(ready_queue,NIL_OBJ))
	RETURN0();

    os_halt_timer();
    current_thread = pair_car( ready_queue );
    
    flush_stack_cache();
    gvec_write( current_thread, THREAD_STATE, continuation_reg );
    TLSET(1,pair_cdr(ready_queue));
    APPLY( 0, TLREF(0) );
})
