#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
#|
    front    ==> the front of the queue (oldest item)
    back     ==> the end of the queue (newest item)
    blocked  ==> a list of threads blocked on the queue
|#

(define-class <Queue> (<Object>) :no-ivar-gfs
    (front '())
    (back '())
    (blocked '()))

(define-macro (front q)
    (%gvec-read q 0))


(define-method write ((self <Queue>) port)
    (format port "#[<Queue> ~s]" (front self)))

(define-method display ((self <Queue>) port)
    (format port "#[<Queue> ~s]" (front self)))

(define (queue-empty? q)
    (null? (front q)))
    
(%define-glue (dequeue! queue) literals: ((& %suspend-current-thread)
					  (& *current-thread*)
					  <Queue>
					  '(dequeue!)
					  'queue
					  (list <Queue>))
("dequeue_loop" {
obj front, current_thread = TLREF(1);

    COUNT_ARGS(1);
    if (!OBJ_ISA_PTR(queue) || !EQ(CLASSOF_PTR(queue),LITERAL(2)))
    {
	failed_type_check( LITERAL(3), LITERAL(4), queue, LITERAL(5) );
    }
    front = gvec_read( queue, QUEUE_FRONT );
    if (EQ(front,NIL_OBJ))
    {
	/* install us on the queue's BLOCKED list */
	gvec_write_ptr( queue,
		        QUEUE_BLOCKED,
		        cons( current_thread, 
			      gvec_read( queue, QUEUE_BLOCKED ) ) );
	SAVE_CONT1(dequeue_prep_try_again);
	flush_stack_cache();
	gvec_write( current_thread, THREAD_STATE, continuation_reg );
	
	/* remove us from the ready queue */

	os_halt_timer();	/* turn off the timer
				   (otherwise, we could get a timeout
				   interrupt before the new thread
				   even starts to run!) */
	APPLY(0,TLREF(0));	/* suspend the current thread */
    }
    else
    {
    obj next = pair_cdr(front);
    
	gvec_write( queue, QUEUE_FRONT, next );
	if (EQ(next,NIL_OBJ))
	    gvec_write( queue, QUEUE_BACK, NIL_OBJ );
	REG0 = pair_car(front);
	RETURN(1);
    }
})
("dequeue_prep_try_again" {
    RESTORE_CONT1();
    BJUMP(1,dequeue_loop);
}))

(%define-glue (enqueue! queue item) literals: ((& %threads-did-unblock)
						<Queue>
						'(dequeue!)
						'queue
						(list <Queue>))
{
obj did_unblock, back;

    COUNT_ARGS(2);
    if (!OBJ_ISA_PTR(queue) || !EQ(CLASSOF_PTR(queue),LITERAL(1)))
    {
	failed_type_check( LITERAL(2), LITERAL(3), queue, LITERAL(4) );
    }
    item = cons( item, NIL_OBJ );
    back = gvec_read( queue, QUEUE_BACK );
    if (EQ(back,NIL_OBJ))
    {
        /* the queue is empty -- so unblock any waiting threads */
	did_unblock = gvec_read( queue, QUEUE_BLOCKED );
	gvec_write_non_ptr( queue, QUEUE_BLOCKED, NIL_OBJ );
	gvec_write_ptr( queue, QUEUE_FRONT, item );
	gvec_write_ptr( queue, QUEUE_BACK, item );
	if (EQ(did_unblock,NIL_OBJ))
	    RETURN0();
	else
	{
	    REG0 = queue;
	    REG1 = did_unblock;
	    APPLY( 2, TLREF(0) );
	}
    }
    else
    {
	/* the queue is not empty -- nobody should be blocked on us */
	gvec_write_ptr( gvec_read( queue, QUEUE_BACK ), SLOT(1), item );
	gvec_write_ptr( queue, QUEUE_BACK, item );
	RETURN0();
    }
})

