/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "runtim.h"
#include <string.h>
#include <ctype.h>
#ifdef MACH_ATRACE
extern "C" {
#include "atrace-client.h"
}
#endif

#ifndef INSTALL_LIB_DIR
#define INSTALL_LIB_DIR "/usr/local/lib"
#endif

const char *find_system_image( const char *arg0 );

int main( int argc, const char **argv )
{
obj start, args;
const char *system_file
#ifdef __NeXT__
			= "sect:IMAGE:system";
#else
			= NULL;
#endif
int i;
bool verbose = YES;

    i = 1;
    if (argc >= 3 && (strcmp( argv[1], "-image" ) == 0))
    {
	system_file = argv[2];
	i += 2;
    }
    else if (argc >= 3 && (strcmp( argv[1], "-qimage" ) == 0))
    {
	system_file = argv[2];
	i += 2;
	verbose = NO;
    }
#ifndef __NeXT__
    else
        system_file = find_system_image( argv[0] );
#endif

#ifdef STEP_DUMP
    if (i < argc && (strcmp(argv[i],"-stepdump") == 0))
    {
    extern int do_step_dump;
    
	do_step_dump = 1;
	i++;
    }
#endif
    if (i < argc && ((strcmp(argv[i],"-q") == 0)
    		     || (strcmp(argv[i],"-script") == 0)))
    {
	verbose = NO;
	/* we leave "-q" in argv[] to be seen by the scheme program */
    }

    start = init_scheme( argv[0], system_file, verbose );
#ifdef MACH_ATRACE
     atrace_init( (void *)0x2000, (void *)0xffffff, "rscheme.1" );
#endif
    args = NIL_OBJ;
    while (argc > i)
    {
        args = cons( make_string( argv[--argc] ), args );
    }
    call_scheme( start, 1, args );
    return 0;
}

/* try to find the system image if it
   wasn't specified on the command line
*/

extern bool os_file_exists_p( const char *path );

const char *find_system_image( const char *arg0 )
{
static char temp[1024];
char last = arg0[strlen(arg0)-1];

  /* part 1.  If argv[0] is of the form bar/baz/foo
              then strip "foo" and look for bar/baz/system.img
  */
	       
  if (strchr(arg0,'/'))
    {
      strcpy( temp, arg0 );
      strrchr( temp, '/' )[1] = 0;
      strcat( temp, "system.img" );
    }
  else
    strcpy( temp, "system.img" );

  /* part 2.  If argv[0] ends in a digit X, look for system.imgX 
              instead of system.img
  */

  if (isdigit(last))
    {
      char t2[2];
      
      t2[0] = last;
      t2[1] = 0;
      strcat( temp, t2 );
    }

  if (os_file_exists_p(temp))
    return temp;

  /* attempt 2.  look in ~/lib/rs/0.5/system.img  */

  sprintf(temp, 
	  "%s/lib/rs/%s/system.img",
	  getenv("HOME"),
	  RS_VERSION);
  if (os_file_exists_p(temp))
    return temp;

  /* attempt 3.  look in $INSTALL_LIB_DIR/rs/0.5/system.img */

  strcpy( temp, INSTALL_LIB_DIR "/rs/" RS_VERSION "/system.img" );
  if (os_file_exists_p(temp))
    return temp;

#ifdef DEFAULT_IMAGE
  /* attempt 4.  look in $DEFAULT_IMAGE */

  strcpy( temp, DEFAULT_IMAGE );
  if (os_file_exists_p(temp))
    return temp;
#endif

  fprintf( stderr, "could not find system image file\n" );
  exit(1);
  return NULL;
}
