/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "runtim.h"
#include "heapi.h"
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include "hasht.h"

main(...)
{
obj start = init_scheme();
    call_scheme( gvec_read( x, SLOT(2) ), 1, args );
}


extern void init_dynamic_link( const char *argv0 );
extern void init_interrupts( void );

/* #define DEBUG_0 */

unsigned numFlushedObjects = 0;

#ifdef DEBUG_0
void sync_with_browser( const char *msg );
#endif

#ifdef STEP_DUMP
extern int do_step_dump;
extern FILE *step_dump_file;
extern const char *step_dump_filename;
#endif

void run_vec( obj vec );

#ifdef PRE_2H
#define module_oslib module_unxlib
#endif

extern struct module_descr ctcprims;
extern struct module_descr module_bltin;
extern struct module_descr module_scmio;
extern struct module_descr module_oslib;
extern struct module_descr module_repl;

static struct module_descr *(tab[]) = { 
    &ctcprims, 
    &module_bltin,
    &module_scmio,
    &module_oslib,
    &module_repl,
    /* &module_conform, */
    NULL };

#ifdef GUI_VIBRANT
INT_16 Nlm_Main( void )
#else
int main( int argc, const char **argv )
#endif
{
obj x, args;
const char *system_file = getenv("ES_SYSTEM_IMAGE");
#ifndef GUI_VIBRANT
int i = 1;
#endif

#ifdef STEP_DUMP
char sdf_buf[128];
extern int getpid( void );

    step_dump_filename = sdf_buf;
    sprintf( (char *)step_dump_filename,
    	     "/tmp/steptrace.dump.%05d", getpid() );
#ifndef GUI_VIBRANT
    if (i<(argc-1) && (strcmp(argv[i],"-dump") == 0))
    {
	step_dump_filename = argv[i+1];
	i += 2;
    }
    step_dump_file = NULL;
#endif /* GUI_VIBRANT */
#endif /* STEP_DUMP */

    if (!system_file) 
    {
#ifdef __NeXT__
	system_file = "sect:IMAGE:system";
#else
	system_file = "system.img";
#endif /* DEBUG_0 */
    }

#ifndef GUI_VIBRANT
    if (i<(argc-1) && (strcmp(argv[i],"-image") == 0))
    {
	system_file = argv[i+1];
	i += 2;
    }
#endif /* GUI_VIBRANT */

#ifdef GUI_VIBRANT
    init_dynamic_link( NULL );
#else
    init_dynamic_link( argv[0] );
#endif
    init_linkage( tab );
    init_runtim();

    gc_safe_point( 1024*1024 );
    boot_list = 
	x = 
	backup_boots = load_image_file( system_file, system_link_table );

#ifdef DEBUG_0
    sync_with_browser( "loaded image" );
#endif /* DEBUG_0 */
    
    puts( string_text( gvec_read( x, SLOT(0) ) ) );
    run_vec( gvec_read( x, SLOT(1) ) );

#ifdef DEBUG_0
    sync_with_browser( "ran inits" );
#endif /* DEBUG_0 */

#ifndef GUI_VIBRANT
    args = NIL_OBJ;
    while (argc > 0)
    {
        args = cons( make_string( argv[--argc] ), args );
    }
#else
    args = cons( make_string( "vibrant.app" ), NIL_OBJ );
#endif /* GUI_VIBRANT */
    init_interrupts();

#if 1
    call_scheme( gvec_read( x, SLOT(2) ), 1, args );
#else
    x = call_scheme( gvec_read( x, SLOT(2) ), 1, args );
    debug(x);
#endif

#ifdef ATRACE
    atrace_flush(1);
    printf( "%u flushes\n", numFlushedObjects );
#endif
    return 0;
}

void run_vec( obj vec )
{
UINT_32 i, len = SIZEOF_PTR(vec);
obj x;

    for (i=0; i<len; i+=SLOT(1))
    {
	x = gvec_read( vec, i );
	if (VECTOR_P(x))
	{
	struct module_descr *md;
	int ix;
	
	    /* a TL-SET */
	    md = (struct module_descr *)
	    		OBJ_TO_RAW_PTR( gvec_read( x, SLOT(0) ) );
	    ix = fx2int(gvec_read( x, SLOT(1) ));
	    assert( ix >= 0 && ix < md->num_roots );
	    md->root_vector[ix] = gvec_read( x, SLOT(2) );
	}
	else if (CLOSURE_P(x))
	{
	    call_scheme( x, 0 );
	}
	else
	{
	    fprintf( stderr, "Not a <Vector> or <Closure>:\n" );
	    debug( x );
	}
    }
}

#ifdef DEBUG_0
void sync_with_browser( const char *msg )
{ 
char temp[100];
int getpid( void );

    printf( "-m %d %d   at `%s' continue?", getpid(), (int)master_table, msg );
    gets(temp);
}
#endif /* DEBUG_0 */
