#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
; dynamic load ONLY (ie, no compilation)

(define *dload-defs* #f)

(define ($init-dload)
    (let-syntax (((add-system-def name thing) 
		    (table-insert! *dload-defs* name thing)))
	(set! *dload-defs* (make-table string=? string->hash))
	(add-system-def "<<Class>>" <<Class>>)
	(add-system-def "<String>" <String>)
	(add-system-def "<Vector>" <Vector>)
	(add-system-def "<Symbol>" <Symbol>)
	(add-system-def "<Pair>" <Pair>)
	(add-system-def "<Closure>" <Closure>)
	(add-system-def "<Template>" <Template>)
	(add-system-def "<TopLevelVar>" <TopLevelVar>)
	(add-system-def "<TopLevelEnvt>" <TopLevelEnvt>)
	(add-system-def "<BindingEnvt>" <BindingEnvt>)))

;(%use dynamic-linking pathname hacker)

(run-named $init-dload)

(define (base-path->dl-file base-path)
    (make <file-name> filename: (last-directory base-path) 
    		      directory: base-path 
		      extension: "o"))

(define (base-path->image-file base-path)
    (make <file-name> filename: (last-directory base-path) 
    		      directory: base-path 
		      extension: "img"))

(define (base-path->module-name base-path)
    (let* ((l (last-directory base-path))
    	   (x (- (string-length l) 2)))
	(if (and (> x 0)
		 (string=? (substring l x) "_m"))
	    (string->symbol (substring l 0 x))
	    (string->symbol l))))

(define (dynamic-load base-path)
    (if (string? base-path)
	(dynamic-load (string->dir base-path))
    (dynamic-load-module
		    (pathname->string (base-path->dl-file base-path))
		    (pathname->string (base-path->image-file base-path))
		    (base-path->module-name base-path))))

;; use DYNAMIC-LOAD-MODULE when neither the code nor
;; the heap image has been loaded.  In this case, after
;; we load the object code, loading the heap image automatically
;; swizzles code pointers in the heap


(define (dynamic-load-module path image-path name)
    (let ((dlf (open-dynamic-load path)))
	(if dlf
	    (let ((md (dynamic-load-lookup 
			    dlf 
			    (string-append "_module_" 
					   (symbol->string name)))))
		(install-module md)
		(close-dynamic-load dlf)
		(let ((img (%load-image image-path *dload-defs*)))
		    (format #t "    Loading: ~a\n" (vector-ref img 0))
		    img))
	    (abort 'dynamic-load-module "could not load ~s" path))))

(export *dload-defs* dynamic-load dynamic-load-module)
