#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;$Id: print.scm,v 1.3 94/11/01 16:50:12 donovan Exp $

(define-generic-function print)

(define (print.all-slots c)
    (apply append 
	(map %class-slots (reverse (%linear-ancestors c)))))

(define (print.print1 name index item)
    (let ((s (if name
		 (format #f "    [~d] ~s = " index name)
		 (format #f "    [~d] = " index))))
	(display s)
	(display (object->bounded-string (- 76 (string-length s))
					    item))
	(newline)))

(define (print.print* name index)
    (if name
	(format #t "    [~d]? ~s\n" index name)
	(format #t "    [~d]?\n" index)))

(define-method print ((l <pair>))
    (print-summary l)
    (let loop ((i 0) (x l))
	(if (pair? x)
	    (begin
		(print.print1 #f i (car x))
		(loop (add1 i) (cdr x)))
	    (begin
		(if (not (null? x))
		    (print.print1 'rest i x))
		l))))

(define-method print ((l <null>))
    (print-summary l)
    (display "    the empty list\n")
    '())

(define-method print ((v <vector>))
    (print-summary v)
    (let loop ((i 0))
	(if (< i (vector-length v))
	    (begin
		(print.print1 #f i (vector-ref v i))
		(loop (add1 i)))
	    v)))

(define (print-summary x)
    (display (object->bounded-string 70 x))
    (newline))
    
(define (print-gvec x)
    (let ((slots (print.all-slots (class-of x)))
	  (limit 0))
	    
	(for-each
	    (lambda (name index)
		(if (> index limit)
		    (set! limit index))
		(if (< index (%gvec-length x))
		    (print.print1 name index (%gvec-read x index))
		    (print.print* name index)))
	    (map %slot-name slots)
	    (map %slot-position slots))
	    
	(let loop ((index (add1 limit)))
	    (if (< index (%gvec-length x))
		(begin
		    (print.print1 #f index (%gvec-read x index))
		    (loop (add1 index)))))))

;; XXXX: XXXX XXXX XXXX XXXX  XXXX XXXX XXXX XXXX  | ................

(define $post-blk '#("" " " "" " " "" " " "" "  "
		     "" " " "" " " "" " " "" "  | "))

(define (print-bvec-line bvec index len ixfw)
    (let ((s (number->string index 16)))
	(format #t "~a~a: " (make-string (- ixfw (string-length s)) #\0)
			    s)
	(let loop ((i index) (n len))
	    (if (eq? n 0)
		(let loop ((i i) (n len))
		    (if (< n 16)
			(begin
			    (display "  ")
			    (display (vector-ref $post-blk (bitwise-and i 15)))
			    (loop (+ i 1) (+ n 1)))))
		(let ((byte (%bvec-read bvec i)))
		    (display (substring (number->string (+ byte 256) 16) 1))
		    (display (vector-ref $post-blk (bitwise-and i 15)))
		    (loop (+ i 1) (- n 1)))))
	(let loop ((i index) (n len))
	    (if (not (eq? n 0))
		(let ((byte (%bvec-read bvec i)))
		    (if (and (>= byte 32) (< byte 127))
			(write-char (integer->char byte))
			(write-char #\.))
		    (loop (+ i 1) (- n 1)))))
	(newline)))	    
    
(define (print-bvec x)
    (format #t "bvec ~s: ~d bytes\n" 
           (class-name (class-of x))
    	   (%bvec-length x))
    (let* ((n (%bvec-length x))
	  (ixfw (string-length (number->string n 16))))
	(if (< ixfw 4)
	    (set! ixfw 4))
	(let loop ((i 0))
	    (if (< (+ i 16) n)
		(begin
		    (print-bvec-line x i 16 ixfw)
		    (loop (+ i 16)))
		(print-bvec-line x i (- n i) ixfw)))))

(define-method print ((x <object>))
    (print-summary x)
    (cond
	((gvec? x)
	    (print-gvec x))
	((bvec? x)
	    (print-bvec x)))	
    x)

(define-method print ((self <table>))
    (print-summary self)
    
    ;; print the bucket occupation levels
    (let (((stats <vector>) (table-stats self)))
	(format #t "implementation: ~s, " (class-name (class-of self)))
	(format #t "directory: ~d bits\n" (directory-bits self))
	(format #t "~d chains, sizes: ~s\noccupancy: "
		(vector-ref stats 1)
		(map cdr (vector-ref stats 2)))
	(let (((n <fixnum>) 0))
	    (vector-for-each
		(lambda (e)
		    (if (eq? n 0)
			(display (cdr e))
			(if (eq? n 32)
			    (begin
				(format #t "\n           ~d" (cdr e))
				(set! n 0))
			    (format #t " ~d" (cdr e))))
		    (set! n (add1 n)))
		(vector-ref stats 0))
	    (newline)))
    ;; print the contents
    (table-for-each
	self
	(lambda (h k v)
	    (format #t "   ~s => ~s\n" k v)))
    self)

(export print print-gvec print-bvec)
