#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;  this is a library of debugging functionality, which
;;  goes behind the scenes to install debugging code directly
;;  into closures.
;;
;;  to use it, call debug!* with a name and a closure,
;;  or debug! (a special form) on a top-level variable name
;;  (preceded by an optional tl-environment)

(define *untrace* "*the-untrace-token*")

(define (closure-set! dest source)
    (%gvec-write dest 0 (%gvec-read source 0))
    (%gvec-write dest 1 (%gvec-read source 1)))

(define (make-trace-proc name proc orig)
    (lambda args
	(if (and (pair? args)
	         (eq? (car args) *untrace*))
	    (begin
		(format #t "Untracing: ~a\n" name)
		(closure-set! orig proc)
		orig)
	    (begin
		(format #t "Entering: ~a\n" name)
		(format #t "      ==> ~s\n" args)
		(bind ((#rest result (apply proc args)))
		    (format #t "Exiting: ~s\n" name)
		    (format #t "   with: ~s\n" result)
		    (list->values result))))))

(define-syntax (debug! proc)
    (closure-set! 
	proc 
	(make-trace-proc (mquote proc)
			 (clone proc)
			 proc)))
