#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; PostScript-generating output port
;; with tabbing facilities
;
;  created --dmk 94.08.18

(define-generic-function %write-char)
(define-generic-function %write-str)
(define-generic-function flush-output)

#|
(define (load-afm file)
    (call-with-input-file
	(string-append "/NextLibrary/Fonts/afm/" file ".afm")
	(lambda (port)
	    (let ((char-widths (make-vector 256)))
		(let loop ()
		    (let ((line (read-line port)))
			(if (eof-object? line)
			    char-widths
			    (if (and (> (string-length line) 2)
				     (string=? (substring line 0 2) "C "))
				(
|#

#|
	#!/usr/local/bin/perl
	#
	#  convert AFM file to DRAWEXPR-suitable format
	
	#print "converting $ARGV[0]\n";
	
	open( AFM, $ARGV[0] ) || die( "Could not open input\n" );
	
	printf "( ";
	while (<AFM>)
	{
	    if (m/^C ([0-9]+).*WX ([0-9]+)/)
	    {
		print "($1 . $2) ";
	    }
	}
	printf ")\n";
|#

(define (load-afm file)
    (let ((char-widths (make-vector 256)))
	(for-each
	    (lambda (fw)
		(vector-set! char-widths (car fw) (cdr fw)))
	    (call-with-input-file
		(string-append file ".fw")
		read))
	char-widths))
		
(define-class <font> (<object>)
    ps-name
    afm-name
    size
    char-widths
    default-line-spacing)

(define-method initialize ((self <font>))
    (if (not (afm-name self))
	(set-afm-name! self 
		       (format #f "~a.~d" (ps-name self) (size self))))
    (set-char-widths! self (load-afm (afm-name self)))
    (if (not (default-line-spacing self))
	(set-default-line-spacing! self (+ (size self) 4)))
    self)

(define *default-font* (make <font> ps-name: "Screen-Helvetica" size: 12))

(define (char-width (font <font>) char)
    (vector-ref (char-widths font) (char->integer char)))
    
(define (ps-moveto ps-port x y)
    (format ps-port "~d ~d moveto " x y))

(define (ps-show ps-port str)
    (format ps-port "(")
    (for-each (lambda (ch)
    		(if (or (eq? ch #\()
			(eq? ch #\)))
		    (write-char #\\ ps-port))
		(write-char ch ps-port))
	      (string->list str))
    (format ps-port ") show\n"))

(define (ps-set-font ps-port (font <font>))
    (format ps-port "/~a findfont [~d 0 0 -~d 0 0] makefont setfont\n" 
		    (ps-name font) 
		    (size font)
		    (size font)))
    

;; a "drawing" port

(define-class <drawing-port> (<object>)
    (underlying-port #f)
    (x-position 0)
    (y-position 0)
    (left-margin 0)
    (right-margin 100)
    (line-spacing #f)
    (buffer #f)
    (buffers-x 0)
    (font-selection #f))

(define-method initialize ((self <drawing-port>))
    (if (not (font-selection self))
	(set-font-selection! self *default-font*))
    (if (not (underlying-port self))
	(set-underlying-port! self (current-output-port)))
    (if (not (line-spacing self))
	(set-line-spacing! self (default-line-spacing (font-selection self))))
    (display "% hello\n" (underlying-port self))
    (display "50 150 translate 1 -1 scale\n" (underlying-port self))
    (ps-set-font (underlying-port self) (font-selection self))
    self)

(define-method flush-output ((self <drawing-port>))
    (if (buffer self)
	(begin
	    (ps-moveto (underlying-port self) 
		       (buffers-x self) 
		       (y-position self))
	    (ps-show (underlying-port self)
		     (list->string (seq->list (buffer self))))
	    (set-buffers-x! self (x-position self))
	    (set-buffer! self #f)))
    (values))

(define-method %write-char ((self <drawing-port>) (ch <ascii-char>))
    (if (eq? ch #\newline)
	(begin
	    (flush-output self)
	    (set-x-position! self (left-margin self))
	    (set-buffers-x! self (left-margin self))
	    (set-y-position! self (+ (y-position self) (line-spacing self))))
	(begin
	    (if (not (buffer self))
		(set-buffer! self (make-seq)))
	    (seq-add! (buffer self) ch)
	    (set-x-position! self (+ (x-position self)
				     (char-width (font-selection self) ch))))))

(define-method %write-str ((self <drawing-port>) (str <string>))
    (for-each
	(lambda (c)
	    (%write-char self c))
	(string->list str)))

(define-method close-output-port ((self <drawing-port>))
    (flush-output self)
    (format (underlying-port self) "% thank you\n"))
    
;(define-generic-function draw-string)
;(define-generic-function indented)

(define (draw-string (self <drawing-port>) (format-str <string>) . args)
    (%write-str self (apply format #f format-str args)))

(define (indent-to-tab (self <drawing-port>) tab-stob thunk)
  (flush-output self)
  (let ((old-margin (left-margin self)))
    (set-left-margin! self tab-stob)
    (set-x-position! self (left-margin self))
    (set-buffers-x! self (left-margin self))
    (thunk)
    (set-left-margin! self old-margin)
    (if (not (buffer self))
	(begin
	    (set-buffers-x! self old-margin)
	    (set-x-position! self old-margin)))))

(define (indented (self <drawing-port>) amount thunk)
    (indent-to-tab self (+ (left-margin self) amount) thunk))

(define (indented-here (self <drawing-port>) thunk)
    (indent-to-tab self (x-position self) thunk))

#|
(define (indented-here (self <drawing-port>) thunk)
  (flush-output self)
  (let ((old-margin (left-margin self)))
    (set-left-margin! self (x-position self))
    (set-x-position! self (left-margin self))
    (set-buffers-x! self (left-margin self))
    (thunk)
    (set-left-margin! self old-margin)))
|#    


(define (test)
    (let ((p (make <drawing-port>)))
	(draw-string p "foo = ~s" '(1 2 3))
	(draw-string p " (~a)\n" "blah")
	(indented p 5
	    (lambda ()
		(draw-string p "bar = ~s [" '(1 2 3))
		(indented-here p 
		    (lambda ()
			(draw-string p "1 2 3\n4 5 6\n7 8 9]\n")))))
	(draw-string p "(lambda (x y)\n")
	(indented p $body-indent
	    (lambda ()
		(draw-string p "(let (")
		(indented-here p
		    (lambda ()
			 (draw-string p "(quux (foo 1 2))\n")   
			 (draw-string p "(bleen (blech 9 2)))\n")))
		(indented p $body-indent
		    (lambda ()
			(draw-string p "(body1 ...)")
			(indent-to-tab p $comment-tab
			    (lambda ()
				(draw-string p "; this is a comment\n")))
			(draw-string p "(body2 ...)))")
			(indent-to-tab p $comment-tab
			    (lambda ()
				(draw-string p "; this is more comments\n")))))))	
	(close-output-port p)))

(define $body-indent 10)
(define $comment-tab 100)

#|
% hello
50 150 translate 1 -1 scale
/Screen-Helvetica findfont [12 0 0 -12 0 0] makefont setfont
0 0 moveto (foo = \(1 2 3\) \(blah\)) show
5 16 moveto (bar = \(1 2 3\) [) show
77 16 moveto (1 2 3) show
77 32 moveto (4 5 6) show
77 48 moveto (7 8 9]) show
0 64 moveto (\(lambda \(x y\)) show
10 80 moveto (\(let \() show
34 80 moveto (\(quux \(foo 1 2\)\)) show
34 96 moveto (\(bleen \(blech 9 2\)\)\)) show
20 112 moveto (\(body1 ...\)) show
100 112 moveto (; this is a comment) show
20 128 moveto (\(body2 ...\)\)\)) show
100 128 moveto (; this is more comments) show
% thank you
|#
