#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; list a directory hierarchy reversly by date

(%use os pathname)
;(load "../unxlib/unixutil.scm")

(define (safe-stat path)
    (call-with-current-continuation
	(lambda (exit)
	    (fluid-let ((*exception* (lambda args (exit #f))))
		(stat path)))))

(define-class <seq> (<Object>) :no-ivar-gfs
    (seq-first '(1) <Pair>)
    (seq-last '(1) <Pair>))

(define-method initialize ((self <seq>))
    (let ((i (cons 0 '())))
	(%gvec-write self 0 i)
	(%gvec-write self 1 i))
    self)

(define-syntax (seq-add! seq item)
    (let ((s seq)
    	  (i (cons item '())))
	(set-cdr! (%gvec-read s 1) i)
	(%gvec-write s 1 i)
	(values)))

(define-syntax (seq-as-list seq)
    (cdr (%gvec-read seq 0)))

(define (dir-list path seq)
    (format #t "Checking ~a\n" path)
    (let ((subs '()))
	(for-each
	    (lambda (d)
		(let* ((f (string-append path "/" d))
		       (s (safe-stat f)))
		    (if s
			(if (stat-directory? s)
			    (set! subs (cons f subs))
			    (seq-add! seq (cons f s))))))
	    (cddr (scandir path)))
	(for-each (lambda (i) (dir-list i seq)) subs)))

(define (sort-by-date l)
    (sort l
    	  (lambda (a b)
	    (time>? (vector-ref (cdr a) 1)
	            (vector-ref (cdr b) 1)))))

(define (print-dir-list l)
    (let ((w (apply max (map (lambda (s) (string-length (car s))) l))))
	(for-each 
	    (lambda (i) 
		(format #t "~a~a ~a\n" 
			(car i)
			(make-string (- w (string-length (car i))) #\space)
			(vector-ref (cdr i) 1))) 
	    l)))

;(msdos-check ".")
;(%exit 0)
