#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; a program that creates a new, bootable system image
;; which will start & call thunk
(%use hacker tables dynamic-linking)

(define (make-system-defs)
    (let ((sys '()))
	(let-syntax ((define-system 
			(syntax-form (name thing)
			    (set! sys (cons (cons name thing) sys)))))
	    (define-system "<<Class>>" <<Class>>)
	    (define-system "<Symbol>" <Symbol>)
	    (define-system "<Closure>" <Closure>)
	    (define-system "<BindingEnvt>" <BindingEnvt>)
	    (define-system "<Vector>" <Vector>)
	    (define-system "<String>" <String>)
	    (define-system "<Pair>" <Pair>)
	    (define-system "<TopLevelVar>" <TopLevelVar>)
	    (define-system "<TopLevelEnvt>" <TopLevelEnvt>)
	    (define-system "<Template>" <Template>)
	    (define-system "<LongFloat>" <double-float>)
	    (define-system "<table>" <table>)
	    (define-system "<string-table>" <string-table>)
	    (define-system "<string-ci-table>" <string-ci-table>)
	    (define-system "<object-table>" <object-table>)
	    (define-system "<eq-table>" <eq-table>)
;	    (define-system "<symbol-table>" <symbol-table>)
	    (define-system "<table-bucket>" <table-bucket>)
	    (list->vector sys))))

(define *system-defs* #f)

(define (get-system-defs)
    (if (not *system-defs*)
	(set! *system-defs* (make-system-defs)))
    *system-defs*)

(define *module-reverse* (make-table eq? integer->hash))

(define (make-module-descriptor m)
    (make <PartDescr> 
	module-name: m
	part-tag: 0))

(for-each
    (lambda (m)
	(table-insert! *module-reverse* 
		       (name->module m) 
		       (make-module-descriptor m)))
    '("bltin" "scmio" "tables"))

(define *fixed-its* 
  (vector-map
   (lambda (it)
     (if (vector? it)
	 (vector (table-lookup *module-reverse*
			       (vector-ref it 0))
		 (vector-ref it 1)
		 (vector-ref it 2))
	 it))
   *initialization-thunks*))
  
;; the procedure proc should take one argument,
;; which will be the list of command-line arguments

(define (save-image path version-string proc)
    (%save-image path
    		 (vector version-string
    			 *fixed-its*
			 proc)
		(get-system-defs)))
